/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		QTSSModuleUtils.cpp

	Contains:	Implements utility routines defined in QTSSModuleUtils.h.
					
	$Log: QTSSModuleUtils.cpp,v $
	Revision 1.2  2000/01/27 02:09:00  rtucker
	Merge changes from v81
	
	Revision 1.12  2000/01/19 20:35:29  serenyi
	Moved to new NEW macro
	
	Revision 1.11  1999/12/21 01:43:18  lecroy
	*** empty log message ***
	
	Revision 1.10  1999/11/15 01:24:28  serenyi
	Added in 4 char debugging codes
	
	Revision 1.9  1999/11/09 12:09:12  murata
	Added GetValueTerm
	// This function is the same as GetValue but returns inBuffer.Ptr as
	// the function result and 0 terminates the buffer.
	
	Revision 1.8  1999/11/08 23:02:07  serenyi
	Updated to latest version of QTSS API
	
	Revision 1.7  1999/11/04 22:35:31  serenyi
	Added extra string arg to LogError
	
	Revision 1.6  1999/10/22 00:28:16  serenyi
	Got rid of Asserts in GetPref functions
	
	Revision 1.5  1999/10/21 22:07:22  lecroy
	made access & fileprivs modules dynamic, added AFTER_BUILD scripts to strip them and move them around, renamed QTSSPasswd to qtpasswd, probably some other stuff too...
	
	Revision 1.4  1999/10/21 21:52:02  serenyi
	Got rid of bogus assert condition
	
	Revision 1.3  1999/10/16 00:37:30  murata
	Added typed accessors for prefs dictionary.
	
	Revision 1.2  1999/10/07 08:57:47  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.1.1.1  1999/09/22 23:01:11  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:30  rtucker
	initial checkin from rrv14
	
	

	Created: Tue, Mar 2, 1999 @ 4:23 PM
*/

#include "QTSSModuleUtils.h"
#include "QTSS_Private.h"

#include "StrPtrLen.h"
#include "OSArrayObjectDeleter.h"
#include "OSMemory.h"
#include "MyAssert.h"
#include "StringFormatter.h"

QTSS_TextMessagesObject 	QTSSModuleUtils::sMessages = NULL;
QTSS_PrefsObject 			QTSSModuleUtils::sPrefs = NULL;
QTSS_StreamRef 			QTSSModuleUtils::sErrorLog = NULL;


void	QTSSModuleUtils::Initialize(QTSS_TextMessagesObject inMessages,
									QTSS_PrefsObject inPrefs,
									QTSS_StreamRef inErrorLog)
{
	sMessages = inMessages;
	sPrefs = inPrefs;
	sErrorLog = inErrorLog;
}

void 	QTSSModuleUtils::LogError(	QTSS_ErrorVerbosity inVerbosity,
									QTSS_AttributeID inTextMessage,
									UInt32 /*inErrNumber*/,
									char* inArgument,
									char* inArg2)
{
	static char* sEmptyArg = "";
	
	// Retrieve the specified text message from the text messages dictionary.
	
	StrPtrLen theMessage;
	(void)QTSS_GetValuePtr(sMessages, inTextMessage, 0, (void**)&theMessage.Ptr, &theMessage.Len);
	if ((theMessage.Ptr == NULL) || (theMessage.Len == 0))
		(void)QTSS_GetValuePtr(sMessages, qtssMsgNoMessage, 0, (void**)&theMessage.Ptr, &theMessage.Len);

	if ((theMessage.Ptr == NULL) || (theMessage.Len == 0))
		return;
	
	// ::sprintf and ::strlen will crash if inArgument is NULL
	if (inArgument == NULL)
		inArgument = sEmptyArg;
	if (inArg2 == NULL)
		inArg2 = sEmptyArg;
	
	// Create a new string, and put the argument into the new string.
	
	UInt32 theMessageLen = theMessage.Len + ::strlen(inArgument) + ::strlen(inArg2);

	OSCharArrayDeleter theLogString(NEW char[theMessageLen + 1]);
	::sprintf(theLogString.GetObject(), theMessage.Ptr, inArgument, inArg2);
	Assert(theMessageLen >= ::strlen(theLogString.GetObject()));
	
	(void)QTSS_Write(sErrorLog, theLogString.GetObject(), ::strlen(theLogString.GetObject()),
						NULL, inVerbosity);
}


char* QTSSModuleUtils::GetFullPath(	QTSS_RTSPRequestObject inRequest,
									QTSS_AttributeID whichFileType,
									UInt32* outLen,
									StrPtrLen* suffix)
{
	Assert(outLen != NULL);
	
	// Get the proper file path attribute. This may return an error if
	// the file type is qtssFilePathTrunc attr, because there may be no path
	// once its truncated. That's ok. In that case, we just won't append a path.
	StrPtrLen theFilePath;
	(void)QTSS_GetValuePtr(inRequest, whichFileType, 0, (void**)&theFilePath.Ptr, &theFilePath.Len);

	StrPtrLen theRootDir;
	QTSS_Error theErr = QTSS_GetValuePtr(inRequest, qtssRTSPReqRootDir, 0, (void**)&theRootDir.Ptr, &theRootDir.Len);
	Assert(theErr == QTSS_NoErr);

	//construct a full path out of the root dir path for this request,
	//and the url path.
	*outLen = theFilePath.Len + theRootDir.Len + 2;
	if (suffix != NULL)
		*outLen += suffix->Len;
	
	char* theFullPath = NEW char[*outLen];
	
	//write all the pieces of the path into this new buffer.
	StringFormatter thePathFormatter(theFullPath, *outLen);
	thePathFormatter.Put(theRootDir);
	thePathFormatter.Put(theFilePath);
	if (suffix != NULL)
		thePathFormatter.Put(*suffix);
	thePathFormatter.PutTerminator();

	*outLen = *outLen - 2;
	return theFullPath;
}

QTSS_Error	QTSSModuleUtils::SendErrorResponse(	QTSS_RTSPRequestObject inRequest,
														QTSS_RTSPStatusCode inStatusCode,
														QTSS_AttributeID inTextMessage,
														StrPtrLen* inStringArg)
{			
	static Bool16 sFalse = false;
	
	//set RTSP headers necessary for this error response message
	(void)QTSS_SetValue(inRequest, qtssRTSPReqStatusCode, 0, &inStatusCode, sizeof(inStatusCode));
	(void)QTSS_SetValue(inRequest, qtssRTSPReqRespKeepAlive, 0, &sFalse, sizeof(sFalse));
	
	//send the response header. In all situations where errors could happen, we
	//don't really care, cause there's nothing we can do anyway!
	(void)QTSS_SendRTSPHeaders(inRequest);
	
	// Retrieve the specified message out of the text messages dictionary.
	StrPtrLen theMessage;
	(void)QTSS_GetValuePtr(sMessages, inTextMessage, 0, (void**)&theMessage.Ptr, &theMessage.Len);

	if ((theMessage.Ptr == NULL) || (theMessage.Len == 0))
	{
		// If we couldn't find the specified message, get the default
		// "No Message" message, and return that to the client instead.
		
		(void)QTSS_GetValuePtr(sMessages, qtssMsgNoMessage, 0, (void**)&theMessage.Ptr, &theMessage.Len);
	}
	Assert(theMessage.Ptr != NULL);
	Assert(theMessage.Len > 0);
	
	// Look for a %s in the string, and if one exists, replace it with the
	// argument passed into this function.
	
	//we can safely assume that message is in fact NULL terminated
	char* stringLocation = ::strstr(theMessage.Ptr, "%s");
	if (stringLocation != NULL)
	{
		//write first chunk
		(void)QTSS_Write(inRequest, theMessage.Ptr, stringLocation - theMessage.Ptr, NULL, 0);
		if (inStringArg != NULL && inStringArg->Len > 0)
		{
			//write string arg if it exists
			(void)QTSS_Write(inRequest, inStringArg->Ptr, inStringArg->Len, NULL, 0);
			stringLocation += 2;
		}
		//write last chunk
		(void)QTSS_Write(inRequest, stringLocation, (theMessage.Ptr + theMessage.Len) - stringLocation, NULL, 0);
	}
	else
		(void)QTSS_Write(inRequest, theMessage.Ptr, theMessage.Len, NULL, 0);
	return QTSS_RequestFailed;
}

void	QTSSModuleUtils::SendDescribeResponse(QTSS_RTSPRequestObject inRequest,
													QTSS_ClientSessionObject inSession,
													iovec* describeData,
													UInt32 inNumVectors,
													UInt32 inTotalLength)
{
	//write content size header
	char buf[32];
	::sprintf(buf, "%ld", inTotalLength);
	(void)QTSS_AppendRTSPHeader(inRequest, qtssContentLengthHeader, &buf[0], ::strlen(&buf[0]));

	(void)QTSS_SendStandardRTSPResponse(inRequest, inSession, 0);
	(void)QTSS_WriteV(inRequest, describeData, inNumVectors, inTotalLength, NULL);
}

void	QTSSModuleUtils::GetValue(QTSS_Object inDictionary,
										QTSS_AttributeID inAttrID,
										StrPtrLen* inBuffer,
										UInt32 inExtraSpace)
{
	Assert(inBuffer->Ptr == NULL);
	Assert(inBuffer->Len == 0);
	
	QTSS_Error theErr = QTSS_GetValue(inDictionary, inAttrID, 0, NULL, &inBuffer->Len);
	if (inBuffer->Len > 0)
	{
		inBuffer->Ptr = NEW char[inBuffer->Len + inExtraSpace + 1];
		theErr = QTSS_GetValue(inDictionary, inAttrID, 0, inBuffer->Ptr, &inBuffer->Len);
		if (theErr != QTSS_NoErr)
		{
			delete [] inBuffer->Ptr;
			inBuffer->Ptr = NULL;
			inBuffer->Len = 0;
		}
		else
		{
			Assert(inBuffer->Len > 0);
			#if !DEBUG
				inBuffer->Ptr[inBuffer->Len] = 0;
			#endif
			
		}
	}
}


char *QTSSModuleUtils::GetValueTerm	(	QTSS_Object inDictionary,
							QTSS_AttributeID inAttrID,
							StrPtrLen* inBuffer,
							UInt32 inExtraSpace)
{
		QTSSModuleUtils::GetValue(  inDictionary,
									inAttrID,
									inBuffer,
									inExtraSpace);
								
		if (inBuffer->Len > 0)
		{
			inBuffer->Ptr[inBuffer->Len] = 0;
		}
		
		return inBuffer->Ptr;
}


UInt32 QTSSModuleUtils::GetUInt32StringPref(QTSS_AttributeID inAttrID, const UInt32 inDefaultValue)
{
	UInt32 retVal = inDefaultValue;
	StrPtrLen theValue;
	QTSSModuleUtils::GetValue(sPrefs, inAttrID, &theValue, 1);
	if (theValue.Ptr != NULL)
	{
		Assert(theValue.Len > 0);
		theValue.Ptr[theValue.Len] = '\0';
		retVal = ::strtol(theValue.Ptr, NULL, 10);
	}
	delete [] theValue.Ptr;
	
	return retVal;
}

SInt16 QTSSModuleUtils::GetSInt16Pref(QTSS_AttributeID inAttrID, const SInt16 inDefaultValue)
{
	SInt16 retVal = inDefaultValue;
	StrPtrLen theValue((char *) &retVal, sizeof(retVal) );	
	(void)QTSS_GetValue(sPrefs, inAttrID, 0, theValue.Ptr, &theValue.Len);
	return retVal;
}

UInt16 QTSSModuleUtils::GetUInt16Pref(QTSS_AttributeID inAttrID, const UInt16 inDefaultValue)
{
	return (UInt16) QTSSModuleUtils::GetSInt16Pref( inAttrID, (SInt16) inDefaultValue);
}

SInt32 QTSSModuleUtils::GetSInt32Pref(QTSS_AttributeID inAttrID, const SInt32 inDefaultValue)
{
	SInt32 retVal = inDefaultValue;
	StrPtrLen theValue((char *) &retVal, sizeof(retVal) );	
	(void)QTSS_GetValue(sPrefs, inAttrID, 0, theValue.Ptr, &theValue.Len);
	
	return retVal;
}

UInt32 QTSSModuleUtils::GetUInt32Pref(QTSS_AttributeID inAttrID, const UInt32 inDefaultValue)
{
	return (UInt32) QTSSModuleUtils::GetSInt32Pref( inAttrID, (SInt32) inDefaultValue);
}



Bool16 QTSSModuleUtils::GetBoolStringPref(QTSS_AttributeID inAttrID, const Bool16 inDefaultValue)
{
	static StrPtrLen sEnabledString = "enabled";
	static StrPtrLen sDisabledString = "disabled";

	Bool16 retVal = inDefaultValue;
	StrPtrLen theValue;
	QTSSModuleUtils::GetValue(sPrefs, inAttrID, &theValue);
	if (theValue.Ptr != NULL)
	{ 
		Assert((theValue.Len == sEnabledString.Len) || (theValue.Len == sDisabledString.Len));
		if (theValue.Equal(sEnabledString))
			retVal = true;
		else
			retVal = false;
	}
	delete [] theValue.Ptr;
	return retVal;
}

Bool16 QTSSModuleUtils::GetBool16Pref(QTSS_AttributeID inAttrID, const Bool16 inDefaultValue)
{

	Bool16 retVal = inDefaultValue;
	StrPtrLen theValue((char *) &retVal, sizeof(retVal) );	
	(void)QTSS_GetValue(sPrefs, inAttrID, 0, theValue.Ptr, &theValue.Len);
	return retVal;
}

char* QTSSModuleUtils::GetStringPref(QTSS_AttributeID inAttrID, const char* inDefaultValue)
{
	StrPtrLen theValue;
	QTSSModuleUtils::GetValue(sPrefs, inAttrID, &theValue, 1);
	if (theValue.Ptr != NULL)
	{
		theValue.Ptr[theValue.Len] = '\0';// make sure what we return is NULL terminated
		return theValue.Ptr;
	}
	
	// If we couldn't sucessfully retrieve the preference, return the default,
	// and make sure to copy it because caller assumes it has to free the
	// memory returned from this function
	char* retVal = NEW char[::strlen(inDefaultValue) + 2];
	::strcpy(retVal, inDefaultValue);
	return retVal;
}
