/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		QTSSModuleUtils.h

	Contains:	Utility routines for modules to use.
					
	$Log: QTSSModuleUtils.h,v $
	Revision 1.1.1.1  2000/01/17 23:35:54  wsanchez
	Import of StreamingServer-2.0 (Apple)
	
	Revision 1.8  1999/12/21 01:43:20  lecroy
	*** empty log message ***
	
	Revision 1.7  1999/11/09 12:09:13  murata
	Added GetValueTerm
	// This function is the same as GetValue but returns inBuffer.Ptr as
	// the function result and 0 terminates the buffer.
	
	Revision 1.6  1999/11/08 23:02:07  serenyi
	Updated to latest version of QTSS API
	
	Revision 1.5  1999/11/04 22:35:32  serenyi
	Added extra string arg to LogError
	
	Revision 1.4  1999/10/21 22:07:22  lecroy
	made access & fileprivs modules dynamic, added AFTER_BUILD scripts to strip them and move them around, renamed QTSSPasswd to qtpasswd, probably some other stuff too...
	
	Revision 1.3  1999/10/16 00:37:30  murata
	Added typed accessors for prefs dictionary.
	
	Revision 1.2  1999/10/07 08:57:48  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.1.1.1  1999/09/22 23:01:11  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:30  rtucker
	initial checkin from rrv14
	
	

	Created: Tue, Mar 2, 1999 @ 4:23 PM
*/


#ifndef _QTSS_MODULE_UTILS_H_
#define _QTSS_MODULE_UTILS_H_

#include <stdlib.h>
#include "QTSS.h"
#include "StrPtrLen.h"

class QTSSModuleUtils
{
	public:
	
		static void		Initialize(QTSS_TextMessagesObject inMessages,
									QTSS_PrefsObject inPrefs,
									QTSS_StreamRef inErrorLog);
	
		// Using a message out of the text messages dictionary is a common
		// way to log errors to the error log. Here is a function to
		// make that process very easy.
		
		static void 	LogError(	QTSS_ErrorVerbosity inVerbosity,
									QTSS_AttributeID inTextMessage,
									UInt32 inErrNumber,
									char* inArgument = NULL,
									char* inArg2 = NULL);
									
		// This function constructs a C-string of the full path to the file being requested.
		// You may opt to append an optional suffix, or pass in NULL. You are responsible
		// for disposing this memory

		static char* GetFullPath(	QTSS_RTSPRequestObject inRequest,
									QTSS_AttributeID whichFileType,
									UInt32* outLen,
									StrPtrLen* suffix = NULL);

		// This function sends an error to the RTSP client. You must provide a
		// status code for the error, and a text message ID to describe the error.
		//
		// It always returns QTSS_RequestFailed.

		static QTSS_Error	SendErrorResponse(	QTSS_RTSPRequestObject inRequest,
														QTSS_RTSPStatusCode inStatusCode,
														QTSS_AttributeID inTextMessage,
														StrPtrLen* inStringArg = NULL);

		//Modules most certainly don't NEED to use this function, but it is awfully handy
		//if they want to take advantage of it. Using the SDP data provided in the iovec,
		//this function sends a standard describe response.
		//NOTE: THE FIRST ENTRY OF THE IOVEC MUST BE EMPTY!!!!
		static void	SendDescribeResponse(QTSS_RTSPRequestObject inRequest,
													QTSS_ClientSessionObject inSession,
													iovec* describeData,
													UInt32 inNumVectors,
													UInt32 inTotalLength);
													
		// This function retrieves a dictionary attribute & copies it into
		// the provided StrPtrLen. Caller must deallocate the newly allocated memory.
		// Function will allocate inExtraSpace bytes on top of what is necessary
		// to actually store the attribute. If inBuffer.Ptr is non-NULL on output,
		// inBuffer.Len is guarenteed to be greater than 0.
		static void GetValue(QTSS_Object inDictionary,
									QTSS_AttributeID inAttrID,
									StrPtrLen* inBuffer,
									UInt32 inExtraSpace = 0);
									
									
		// This function is the same as GetValue but returns inBuffer.Ptr as 
		// the function result and 0 terminates the buffer. 
		// This call like GetValue allocates memory.
		static char *GetValueTerm(QTSS_Object inDictionary,
									QTSS_AttributeID inAttrID,
									StrPtrLen* inBuffer,
									UInt32 inExtraSpace = 0);
									
		// This function returns a UInt32 value from the prefs dictionary.
		// If the value isn't found, inDefaultValue is used instead
		static UInt32 GetUInt32StringPref(QTSS_AttributeID inAttrID, const UInt32 inDefaultValue);
		static Bool16 GetBoolStringPref(QTSS_AttributeID inAttrID, const Bool16 inDefaultValue);

		static SInt16 GetSInt16Pref(QTSS_AttributeID inAttrID, const SInt16 inDefaultValue);
		static UInt16 GetUInt16Pref(QTSS_AttributeID inAttrID, const UInt16 inDefaultValue);
		
		static Bool16 GetBool16Pref(QTSS_AttributeID inAttrID, const Bool16 inDefaultValue);
		static UInt32 GetUInt32Pref(QTSS_AttributeID inAttrID, const UInt32 inDefaultValue);
		static SInt32 GetSInt32Pref(QTSS_AttributeID inAttrID, const SInt32 inDefaultValue);
		
		// Returns a newly allocated buffer always. If inAttrID exists in the
		// prefs dictionary, it will return a buffer containing the attribute value,
		// NULL terminated. Otherwise, this function returns a newly allocated buffer
		// containing inDefaultValue
		static char* GetStringPref(QTSS_AttributeID inAttrID, const char* inDefaultValue);


	private:
	
		static QTSS_TextMessagesObject sMessages;
		static QTSS_PrefsObject 		sPrefs;
		static QTSS_StreamRef 		sErrorLog;
};

#endif //_QTSS_MODULE_UTILS_H_
