/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		QTSSAccessModule.cpp

	Contains:	Implementation of QTSSAccessModule. 
					
	
	$Log: QTSSAccessModule.cpp,v $
	Revision 1.2  2000/01/27 02:09:01  rtucker
	Merge changes from v81
	
	Revision 1.8  2000/01/19 20:35:23  serenyi
	Moved to new NEW macro
	
	Revision 1.7  1999/12/21 01:43:22  lecroy
	*** empty log message ***
	
	Revision 1.6  1999/11/22 23:30:55  murata
	Fixed setting the realm for accepted and denied requests.
	
	Revision 1.5  1999/11/09 12:13:28  murata
	Fixed memory leak and buffer over-run in AccessChecker and
	changed QTSSAccessModule.cpp to use OSCharArrayDeleter to clean up
	the code.
	
	Revision 1.4  1999/11/08 23:02:04  serenyi
	Updated to latest version of QTSS API
	
	Revision 1.3  1999/11/04 22:34:44  serenyi
	Eliminated duplicate LogError function, replaced with QTSSModuleUtils::LogError
	
	Revision 1.2  1999/11/04 02:30:13  murata
	Added check of config file name for / ' and " and uses default if found and logs the error.
	
	Revision 1.1  1999/10/21 22:10:02  lecroy
	added access and fileprovs dynamic module projects
	
	Revision 1.7  1999/10/19 20:54:15  lecroy
	fixed realm, zero-termed user/passwd
	
	Revision 1.6  1999/10/18 20:21:20  serenyi
	changed "dictType" to "objectType"
	
	Revision 1.5  1999/10/16 00:38:22  murata
	now using typed accessors for prefs dictionary.
	
	Revision 1.4  1999/10/13 03:45:39  murata
	just some #include changes for MacOS compiles.
	
	Revision 1.3  1999/10/09 01:09:33  lecroy
	Added QTSSAccessModule
	
	Revision 1.2  1999/10/07 15:51:37  lecroy
	added AuthUserFile, AuthGroupFile, AuthName support
	
	Revision 1.1  1999/10/05 20:04:47  lecroy
	first added
	
	Revision 1.1.1.1  1999/09/22 23:01:10  lecroy
	importing
	
	Revision 1.8  1999/09/17 23:54:28  lecroy
	use OSThread::GetErrNo() instead of my dumb macro
	
	Revision 1.7  1999/09/17 21:40:48  lecroy
	fixed up so that filenotfound will return filenotfound  still needs to be cleaned up
	
	Revision 1.6  1999/09/16 00:23:43  lecroy
	added initialize role
	
	Revision 1.5  1999/09/15 22:28:03  lecroy
	added "file_authorization" config value to enable/disable QTSSAuthorizationModule
	functionality
	
	Revision 1.4  1999/09/15 17:00:18  serenyi
	Added __MW_
	
	Revision 1.3  1999/09/14 22:16:00  john
	removed filtering on Describe. All requests must be authenticated.
	
	Revision 1.2  1999/09/14 02:31:47  murata
	Fixed bug authentication was failing on broadcasts (.sdp files) added check of .sdp
	
	Revision 1.1  1999/09/13 21:43:00  murata
	added auth
	
	Revision 1.2  1999/02/19 23:08:25  ds
	Created
	

*/

#include "QTSSAccessModule.h"


#include "OSArrayObjectDeleter.h"
#include "QTSS_Private.h"
#include "StrPtrLen.h"
#include "OSMemory.h"
#include "MyAssert.h"
#include "StringFormatter.h"
#include "StrPtrLen.h"
#include "StringParser.h"
#include "QTSSModuleUtils.h"
#include "base64.h"
#include "OS.h"
#include "AccessChecker.h"


#ifndef __MW_
	#include <pwd.h>
	#include <unistd.h>
	#include <fcntl.h>
	#include <sys/stat.h>
	#include <sys/errno.h>
#endif

#pragma mark _QTSS_AUTHENTICATE_MODULE_

// ATTRIBUTES

// STATIC DATA


#define MODPREFIX_ "modAccess_"

static StrPtrLen	sSDPSuffix(".sdp");
static OSMutex*		sUserMutex 				= NULL;

const char* 			sAccessModuleEnabledName		= MODPREFIX_"enabled";
const Bool16			sDefaultAuthenticationEnabled	= true;
static QTSS_AttributeID	sAuthenticationEnabledPref 		= qtssIllegalAttrID;
static Bool16			sAuthenticationEnabled			= true;

const char* sUsersFilePathPrefName		= MODPREFIX_"usersfilepath";
const char* sDefaultUsersFilePath = "/etc/qtusers";
static QTSS_AttributeID	sUsersFilePathPrefID 		= qtssIllegalAttrID;
static char* sUsersFilePath = NULL;

const char* sGroupsFilePathPrefName		= MODPREFIX_"groupsfilepath";
const char* sDefaultGroupsFilePath = "/etc/qtgroups";
static QTSS_AttributeID	sGroupsFilePathPrefID 		= qtssIllegalAttrID;
static char* sGroupsFilePath = NULL;

const char* sAccessFileNamePrefName		= MODPREFIX_"qtaccessfilename";
const char* sDefaultAccessFileName = "qtaccess";
static QTSS_AttributeID sAccessFileNamePrefID 		= qtssIllegalAttrID;
static char* sAccessFileName = NULL;

static QTSS_AttributeID	sBadNameMessageAttrID				= qtssIllegalAttrID;
const int kBuffLen = 512;

static QTSS_StreamRef 			sErrorLogStream = NULL;
static QTSS_TextMessagesObject  sMessages = NULL;

// FUNCTION PROTOTYPES

static QTSS_Error QTSSAccessModuleDispatch(QTSS_Role inRole, QTSS_RoleParamPtr inParams);
static QTSS_Error Register();
static QTSS_Error Initialize(QTSS_Initialize_Params* inParams);
static QTSS_Error Shutdown();
static QTSS_Error RereadPrefs();
static QTSS_Error AuthenticateRTSPRequest(QTSS_StandardRTSP_Params* inParams);
bool QTSSAccess(QTSS_StandardRTSP_Params* inParams, const char* pathBuff, const char* movieRootDir, char* ioRealmName);
char* GetCheckedFileName();


// FUNCTION IMPLEMENTATIONS
#pragma mark __QTSS_AUTHENTICATE_MODULE__

QTSS_Error QTSSAccessModule_Main(void* inPrivateArgs)
{
	return _stublibrary_main(inPrivateArgs, QTSSAccessModuleDispatch);
}


QTSS_Error 	QTSSAccessModuleDispatch(QTSS_Role inRole, QTSS_RoleParamPtr inParams)
{
	switch (inRole)
	{
		case QTSS_Register_Role:
			return Register();
		break;
		
		case QTSS_Initialize_Role:
			return Initialize(&inParams->initParams);
		break;
		
		case QTSS_RereadPrefs_Role:
			return RereadPrefs();
		break;
			
		case QTSS_RTSPAuthorize_Role:
			if (sAuthenticationEnabled)
				return AuthenticateRTSPRequest(&inParams->rtspRequestParams);
		break;
			
		case QTSS_Shutdown_Role:
			return Shutdown();
		break;
	}
	
	return QTSS_NoErr;
}

QTSS_Error Register()
{
	// Do role & attribute setup
	(void)QTSS_AddRole(QTSS_Initialize_Role);
	(void)QTSS_AddRole(QTSS_RereadPrefs_Role);
	(void)QTSS_AddRole(QTSS_RTSPAuthorize_Role);
		
	// Add AuthenticateName and Password attributes
	QTSS_Error theErr = QTSS_NoErr;
	static char*		sBadAccessFileName	= "QTSSAccessModuleBadAccessFileName";
	
	theErr = QTSS_AddAttribute(qtssTextMessagesObjectType, sBadAccessFileName, NULL);Assert(theErr == QTSS_NoErr);
	theErr = QTSS_IDForAttr(qtssTextMessagesObjectType, sBadAccessFileName, &sBadNameMessageAttrID);Assert(theErr == QTSS_NoErr);
	
	theErr = QTSS_AddAttribute(qtssPrefsObjectType,sAccessModuleEnabledName, NULL);Assert(theErr == QTSS_NoErr);
	theErr = QTSS_IDForAttr(qtssPrefsObjectType, sAccessModuleEnabledName, &sAuthenticationEnabledPref);Assert(theErr == QTSS_NoErr);
	
	theErr = QTSS_AddAttribute(qtssPrefsObjectType,sUsersFilePathPrefName, NULL);Assert(theErr == QTSS_NoErr);
	theErr = QTSS_IDForAttr(qtssPrefsObjectType, sUsersFilePathPrefName, &sUsersFilePathPrefID);Assert(theErr == QTSS_NoErr);
	
	theErr = QTSS_AddAttribute(qtssPrefsObjectType,sGroupsFilePathPrefName, NULL);Assert(theErr == QTSS_NoErr);
	theErr = QTSS_IDForAttr(qtssPrefsObjectType, sGroupsFilePathPrefName, &sGroupsFilePathPrefID);Assert(theErr == QTSS_NoErr);
	
	theErr = QTSS_AddAttribute(qtssPrefsObjectType,sAccessFileNamePrefName, NULL);Assert(theErr == QTSS_NoErr);
	theErr = QTSS_IDForAttr(qtssPrefsObjectType, sAccessFileNamePrefName, &sAccessFileNamePrefID);Assert(theErr == QTSS_NoErr);

	
	return QTSS_NoErr;
}


QTSS_Error Initialize(QTSS_Initialize_Params* inParams)
{
	// Setup module utils
	QTSSModuleUtils::Initialize(inParams->inMessages, inParams->inPrefs, inParams->inErrorLogStream);
	sErrorLogStream = inParams->inErrorLogStream;
	sMessages = inParams->inMessages;
	sUserMutex = NEW OSMutex();
	RereadPrefs();

	return QTSS_NoErr;
}

QTSS_Error Shutdown()
{
	return QTSS_NoErr;
}

char* GetCheckedFileName()
{
	char		*result = NULL;
	static char *badChars = "/'\"";
	char		theBadCharMessage[] = "' '";
	char		*theBadChar = NULL;
	result = QTSSModuleUtils::GetStringPref(sAccessFileNamePrefID, sDefaultAccessFileName);Assert(result != NULL);
	StrPtrLen searchStr(result);
	
	theBadChar = strpbrk(searchStr.Ptr, badChars);
	if ( theBadChar!= NULL) 
	{
		theBadCharMessage[1] = theBadChar[0];
		QTSSModuleUtils::LogError(qtssWarningVerbosity,sBadNameMessageAttrID, 0, theBadCharMessage, (char*)sDefaultAccessFileName);
				
		delete[] result;
		result = NEW char[::strlen(sDefaultAccessFileName) + 2];
		::strcpy(result, (char*) sDefaultAccessFileName);	
	}
	return result;
}

QTSS_Error RereadPrefs()
{
	OSMutexLocker locker(sUserMutex);
	
	sAuthenticationEnabled = QTSSModuleUtils::GetBoolStringPref(sAuthenticationEnabledPref, sDefaultAuthenticationEnabled);
	
	delete[] sUsersFilePath;
	sUsersFilePath = QTSSModuleUtils::GetStringPref(sUsersFilePathPrefID, sDefaultUsersFilePath);Assert(sUsersFilePath != NULL);
	
	delete[] sGroupsFilePath;
	sGroupsFilePath = QTSSModuleUtils::GetStringPref(sGroupsFilePathPrefID, sDefaultGroupsFilePath);Assert(sGroupsFilePath != NULL);
	
	delete[] sAccessFileName;
	sAccessFileName = GetCheckedFileName();
	
	return QTSS_NoErr;
}

bool QTSSAccess(QTSS_StandardRTSP_Params* inParams, 
				const char* pathBuff, 
				const char* movieRootDir,
				char* ioRealmName)
{
	QTSS_RTSPRequestObject	theRTSPRequest = inParams->inRTSPRequest;		
	

	AccessChecker accessChecker(movieRootDir, sAccessFileName, sUsersFilePath, sGroupsFilePath);
	
	//If there are no access files, then allow world access
	if ( !accessChecker.GetAccessFile(pathBuff) ) 
		return true;

	if ( accessChecker.GetRealmHeaderPtr() != NULL )
		::strcpy(ioRealmName, accessChecker.GetRealmHeaderPtr());
		
	StrPtrLen	nameBuffStr;
	OSCharArrayDeleter nameBuffPtr(QTSSModuleUtils::GetValueTerm(theRTSPRequest,qtssRTSPReqUserName, &nameBuffStr, 0));
	if (0 == nameBuffStr.Len) return false;
	
	StrPtrLen	passwordBuffStr;
	OSCharArrayDeleter passwordBuffPtr(QTSSModuleUtils::GetValueTerm(theRTSPRequest,qtssRTSPReqUserPassword, &passwordBuffStr, 0));
	if (0 == passwordBuffStr.Len) return false;
				
	if ( accessChecker.CheckAccess(nameBuffStr.Ptr, passwordBuffStr.Ptr) )
		return true;

	return false;
}


QTSS_Error AuthenticateRTSPRequest(QTSS_StandardRTSP_Params* inParams)
{
	QTSS_Error 				theErr = QTSS_NoErr;
	QTSS_RTSPRequestObject	theRTSPRequest = inParams->inRTSPRequest;
	
	OSMutexLocker locker(sUserMutex);

	if  ( (NULL == inParams) || (NULL == inParams->inRTSPRequest) )
		return QTSS_RequestFailed;
	
	//get the local file path
	StrPtrLen	pathBuffStr;
	OSCharArrayDeleter pathBuffPtr(QTSSModuleUtils::GetValueTerm(theRTSPRequest,qtssRTSPReqLocalPath, &pathBuffStr, 0));
	if (0 == pathBuffStr.Len) return QTSS_RequestFailed;
	

	//get the root movie directory
	StrPtrLen	movieRootDirStr;
	OSCharArrayDeleter movieRootDirPtr(QTSSModuleUtils::GetValueTerm(theRTSPRequest,qtssRTSPReqRootDir, &movieRootDirStr, 0));
	if (0 == movieRootDirStr.Len) return QTSS_RequestFailed;
	

	//check if this user is allowed to see this movie
	char realmName[kBuffLen] = {};
	StrPtrLen	realmNameStr(realmName,kBuffLen -1);
	Bool16 allowRequest = ::QTSSAccess(inParams, pathBuffStr.Ptr, movieRootDirStr.Ptr, realmName);


	if ( realmName[0] != '\0' ) 	//set the realm if we have one
	{
		(void) QTSS_SetValue(theRTSPRequest,qtssRTSPReqURLRealm, 0, realmName, strlen(realmName) );
	}
	

	if (allowRequest)
	{
		return QTSS_NoErr;	//everything's kosher - let the request continue
	}
	else	//request denied
	{
		// We are denying the request so pass false back to the server.
		theErr = QTSS_SetValue(theRTSPRequest,qtssRTSPReqUserAllowed, 0, &allowRequest, sizeof(allowRequest));
		if (theErr != QTSS_NoErr) 
			return QTSS_RequestFailed; // Bail on the request. The Server will handle the error
		
	}

	return QTSS_NoErr;
}







