/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		QTSSRawFileModule.cpp

	Contains:	Implementation of Raw File module

	$Log: QTSSRawFileModule.cpp,v $
	Revision 1.2  2000/01/27 02:09:02  rtucker
	Merge changes from v81
	
	Revision 1.7  2000/01/19 20:35:19  serenyi
	Moved to new NEW macro
	
	Revision 1.6  1999/12/21 01:43:24  lecroy
	*** empty log message ***
	
	Revision 1.5  1999/11/08 02:27:48  lecroy
	long long printf macro & sgi mods
	
	Revision 1.4  1999/11/02 04:24:43  serenyi
	Added ability to return file with extra ".raw"
	
	Revision 1.3  1999/10/18 20:20:12  serenyi
	changed "dictType" to "objectType"
	
	Revision 1.2  1999/10/07 09:01:02  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.1.1.1  1999/09/22 23:01:13  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:33  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:08:28  ds
	Created
	

*/

#include "QTSSRawFileModule.h"
#include "OSFileSource.h"
#include "OSHeaders.h"
#include "StrPtrLen.h"
#include "OSArrayObjectDeleter.h"
#include "QTSSModuleUtils.h"
#include "OSMemory.h"
#include "ev.h"

#define RAWFILE_FILE_ASYNC 1
#define RAW_FILE_DEBUGGING 0


// ATTRIBUTES IDs

static QTSS_AttributeID	sStateAttr 					= qtssIllegalAttrID;
static QTSS_AttributeID	sEventContextAttr 			= qtssIllegalAttrID;
static QTSS_AttributeID	sFileAttr 					= qtssIllegalAttrID;
static QTSS_AttributeID	sFileBufferAttr 			= qtssIllegalAttrID;
static QTSS_AttributeID	sReadOffsetAttr 			= qtssIllegalAttrID;
static QTSS_AttributeID	sWriteOffsetAttr 			= qtssIllegalAttrID;

// STATIC DATA

static StrPtrLen	sRawSuffix(".raw");

static const UInt32 kReadingBufferState = 0;
static const UInt32 kWritingBufferState = 1;

// FUNCTIONS

static QTSS_Error QTSSRawFileModuleDispatch(QTSS_Role inRole, QTSS_RoleParamPtr inParams);
static QTSS_Error 	Register(QTSS_Register_Params* inParams);
static QTSS_Error Preprocess(QTSS_StandardRTSP_Params* inParams);



QTSS_Error QTSSRawFileModule_Main(void* inPrivateArgs)
{
	return _stublibrary_main(inPrivateArgs, QTSSRawFileModuleDispatch);
}


QTSS_Error QTSSRawFileModuleDispatch(QTSS_Role inRole, QTSS_RoleParamPtr inParams)
{
	switch (inRole)
	{
		case QTSS_Register_Role:
			return Register(&inParams->regParams);
		case QTSS_RTSPPreProcessor_Role:
			return Preprocess(&inParams->rtspPreProcessorParams);
	}
	return QTSS_NoErr;
}

QTSS_Error Register(QTSS_Register_Params* inParams)
{
	(void)QTSS_AddRole(QTSS_RTSPPreProcessor_Role);

	static char*		sStateName 			= 	"QTSSRawFileModuleState";
	static char*		sEventContextName 	= 	"QTSSRawFileModuleContext";
	static char*		sFileName		 	= 	"QTSSRawFileModuleFile";
	static char*		sFileBufferName		= 	"QTSSRawFileModuleFileBuffer";
	static char*		sReadOffsetName		= 	"QTSSRawFileModuleReadOffset";
	static char*		sWriteOffsetName	= 	"QTSSRawFileModuleWriteOffset";

	(void)QTSS_AddAttribute(qtssRTSPRequestObjectType, sStateName, NULL);
	(void)QTSS_IDForAttr(qtssRTSPRequestObjectType, sStateName, &sStateAttr);
	
	(void)QTSS_AddAttribute(qtssRTSPRequestObjectType, sEventContextName, NULL);
	(void)QTSS_IDForAttr(qtssRTSPRequestObjectType, sEventContextName, &sEventContextAttr);

	(void)QTSS_AddAttribute(qtssRTSPRequestObjectType, sFileName, NULL);
	(void)QTSS_IDForAttr(qtssRTSPRequestObjectType, sFileName, &sFileAttr);

	(void)QTSS_AddAttribute(qtssRTSPRequestObjectType, sFileBufferName, NULL);
	(void)QTSS_IDForAttr(qtssRTSPRequestObjectType, sFileBufferName, &sFileBufferAttr);

	(void)QTSS_AddAttribute(qtssRTSPRequestObjectType, sReadOffsetName, NULL);
	(void)QTSS_IDForAttr(qtssRTSPRequestObjectType, sReadOffsetName, &sReadOffsetAttr);

	(void)QTSS_AddAttribute(qtssRTSPRequestObjectType, sWriteOffsetName, NULL);
	(void)QTSS_IDForAttr(qtssRTSPRequestObjectType, sWriteOffsetName, &sWriteOffsetAttr);

	// Tell the server our name!
	static char* sModuleName = "QTSSRawFileModule";
	::strcpy(inParams->outModuleName, sModuleName);

	return QTSS_NoErr;
}


QTSS_Error Preprocess(QTSS_StandardRTSP_Params* inParams)
{
	static UInt32 sFileBufSize = 32768;
	static UInt32 sInitialState = kReadingBufferState;
	static UInt32 sZero = 0;
	
	UInt32 theLen = 0;
	UInt32* theStateP = NULL;

	OSFileSource* theFile = NULL;
	
	(void)QTSS_GetValuePtr(inParams->inRTSPRequest, sStateAttr, 0, (void**)&theStateP, &theLen);
	if ((theStateP == NULL) || (theLen != sizeof(UInt32)))
	{
		// Initial state. We haven't started sending the file yet, so
		// check to see if this is our request, and if it is, set everything up.
		
		// Only operate if this is a DESCRIBE
		QTSS_RTSPMethod* theMethod = NULL;
		UInt32 theLen = 0;
		if ((QTSS_GetValuePtr(inParams->inRTSPRequest, qtssRTSPReqMethod, 0,
				(void**)&theMethod, &theLen) != QTSS_NoErr) || (theLen != sizeof(QTSS_RTSPMethod)))
		{
			Assert(0);
			return QTSS_RequestFailed;
		}

		if (*theMethod != qtssDescribeMethod)
			return QTSS_RequestFailed;
			
		char* theFilePath = NULL;
		
		// Check to see if this is a raw file request
		(void)QTSS_GetValuePtr(inParams->inRTSPRequest, qtssRTSPReqLocalPath, 0, (void**)&theFilePath, &theLen);

		// Copy the full path, and append a ".raw"
		OSCharArrayDeleter rawPath(NEW char[theLen + sRawSuffix.Len + 4]);
		::memcpy(rawPath.GetObject(), theFilePath, theLen);
		::strcpy(rawPath.GetObject() + theLen, sRawSuffix.Ptr);
		
		theFile = NEW OSFileSource(rawPath.GetObject());
		
		// If the file doesn't exist, and if this is a path with a '.raw' at the end,
		// check to see if the path without the extra .raw exists
		if (theFile->GetLength() == 0)
		{
			delete theFile;
			theFile = NULL;
			rawPath.GetObject()[theLen] = '\0';
			
			if (theLen > sRawSuffix.Len)
			{
				StrPtrLen comparer((theFilePath + theLen) - sRawSuffix.Len, sRawSuffix.Len);
				if (comparer.Equal(sRawSuffix))
					theFile = NEW OSFileSource(rawPath.GetObject());
			}
		}

		// If the file doesn't exist, we should probably return a 404 not found.
		if ((theFile == NULL) || (theFile->GetLength() == 0))
		{
			delete theFile;
			return QTSS_RequestFailed;
		}

		// We have a real file. Setup all the dictionary values we need
		(void)QTSS_SetValue(inParams->inRTSPRequest, sFileAttr, 0, &theFile, sizeof(theFile));

		// Create a buffer to store data.
		char* theFileBuffer = NEW char[sFileBufSize];
		(void)QTSS_SetValue(inParams->inRTSPRequest, sFileBufferAttr, 0, &theFileBuffer, sizeof(theFileBuffer));

		// Store our initial state
		(void)QTSS_SetValue(inParams->inRTSPRequest, sStateAttr, 0, &sInitialState, sizeof(sInitialState));
		theStateP = &sInitialState; // so we can proceed normally

		(void)QTSS_SetValue(inParams->inRTSPRequest, sReadOffsetAttr, 0, &sZero, sizeof(sZero));
		(void)QTSS_SetValue(inParams->inRTSPRequest, sWriteOffsetAttr, 0, &sZero, sizeof(sZero));
	
#if RAWFILE_FILE_ASYNC
		// The event context is responsible for closing the fd.
		theFile->DontCloseFD();
		
		// Create an event context for the file, so we can do async I/O
		QTSS_EventContextRef theRef = NULL;
		(void)QTSS_CreateEventContext(theFile->GetFD(), &theRef, theFile->GetOSFileEventContextData());
		Assert(theRef != NULL);
		(void)QTSS_SetValue(inParams->inRTSPRequest, sEventContextAttr, 0, &theRef, sizeof(theRef));
#endif
	}
	
	// Get our attributes
	char** theFileBufferP = NULL;
	(void)QTSS_GetValuePtr(inParams->inRTSPRequest, sFileBufferAttr, 0, (void**)&theFileBufferP, &theLen);
	Assert(theFileBufferP != NULL);
	Assert(theLen == sizeof(char*));
	
	OSFileSource** theFileP = NULL;
	(void)QTSS_GetValuePtr(inParams->inRTSPRequest, sFileAttr, 0, (void**)&theFileP, &theLen);
	Assert(theFileP != NULL);
	Assert(theLen == sizeof(OSFileSource*));

	QTSS_EventContextRef* theRefP = NULL;
#if RAWFILE_FILE_ASYNC
	(void)QTSS_GetValuePtr(inParams->inRTSPRequest, sEventContextAttr, 0, (void**)&theRefP, &theLen);
	Assert(theRefP != NULL);
	Assert(theLen == sizeof(QTSS_EventContextRef*));
#endif

	QTSS_EventContextRef* theSocketRefP = NULL;
	(void)QTSS_GetValuePtr(inParams->inRTSPSession, qtssRTSPSesEventCntxt, 0, (void**)&theSocketRefP, &theLen);
	Assert(theSocketRefP != NULL);
	Assert(theLen == sizeof(QTSS_EventContextRef*));
	
	UInt32* theReadOffsetP = NULL;
	(void)QTSS_GetValuePtr(inParams->inRTSPRequest, sReadOffsetAttr, 0, (void**)&theReadOffsetP, &theLen);
	Assert(theReadOffsetP != NULL);
	Assert(theLen == sizeof(QTSS_EventContextRef*));

	UInt32* theWriteOffsetP = NULL;
	(void)QTSS_GetValuePtr(inParams->inRTSPRequest, sWriteOffsetAttr, 0, (void**)&theWriteOffsetP, &theLen);
	Assert(theWriteOffsetP != NULL);
	Assert(theLen == sizeof(QTSS_EventContextRef*));
	
	UInt32 theReadOffset = *theReadOffsetP;
	UInt32 theWriteOffset = *theWriteOffsetP;
	UInt32 theState = *theStateP;

	Bool16 isBlocked = false;
	
	while (!isBlocked)
	{
		// If we have less than the full buffer size left to go in the file,
		// down adjust our buffer size to be the amount of data remaining in the file
		UInt32 theBufferSize = sFileBufSize;
		if (((*theFileP)->GetLength() - (*theFileP)->GetCurOffset()) < sFileBufSize)
			theBufferSize = (*theFileP)->GetLength() - (*theFileP)->GetCurOffset();

		switch (theState)
		{
			case kReadingBufferState:
			{
				// Read as much data as possible out of the file
				UInt32 theRecvLen = 0;
				(*theFileP)->Read(	(*theFileBufferP) + theReadOffset,
									theBufferSize - theReadOffset,
									&theRecvLen);
				theReadOffset += theRecvLen;
#if RAW_FILE_DEBUGGING
				printf("Got %lu bytes back from file read. Now at: %"_64BITARG_"u\n", theRecvLen, (*theFileP)->GetCurOffset());
#endif
				if (theReadOffset < theBufferSize)
				{
#if RAW_FILE_DEBUGGING
					printf("Flow controlled on file. Waiting for read event\n");
#endif
					isBlocked = true;
					break;
				}
				
				theReadOffset = 0;
				Assert(theWriteOffset == 0);
				theState = kWritingBufferState;
			}
			case kWritingBufferState:
			{
				UInt32 theWrittenLen = 0;
				
				// for debugging purposes, construct an IOVec out of this data
				iovec theVec[5];
				UInt32 units = (theBufferSize - theWriteOffset) /4;
				UInt32 offset = theWriteOffset;
				theVec[1].iov_base = (*theFileBufferP) + offset;
				theVec[1].iov_len = units;
				offset += units;
				theVec[2].iov_base = (*theFileBufferP) + offset;
				theVec[2].iov_len = units;
				offset += units;
				theVec[3].iov_base = (*theFileBufferP) + offset;
				theVec[3].iov_len = units;
				offset += units;
				theVec[4].iov_base = (*theFileBufferP) + offset;
				theVec[4].iov_len = theBufferSize - offset;
				
				(void)QTSS_WriteV(	inParams->inRTSPSession,
									theVec, 5,
									theBufferSize - theWriteOffset,
									&theWrittenLen);
				theWriteOffset += theWrittenLen;
#if RAW_FILE_DEBUGGING
				printf("Got %lu bytes back from socket write.\n", theWrittenLen);
#endif
				if (theWriteOffset < theBufferSize)
				{
#if RAW_FILE_DEBUGGING
					printf("Flow controlled on socket. Waiting for write event.\n");
#endif
					isBlocked = true;
					break;
				}
				
				// Check to see if we're done. If we are, delete stuff and return
				if ((*theFileP)->GetCurOffset() == (*theFileP)->GetLength())
				{
#if RAW_FILE_DEBUGGING
					printf("File transfer complete\n");
#endif
					// If we've gotten here, we're done sending the file!
					// Delete all our resources
					delete [] *theFileBufferP;
					delete *theFileP;
					(void)QTSS_DestroyEventContext(*theRefP);
	
					return QTSS_NoErr;
				}
				
				theWriteOffset = 0;
				Assert(theReadOffset == 0);
				theState = kReadingBufferState;
			}
		}
	}
	
	Assert(isBlocked);
	
	// We've reached a blocking condition for some reason.
	// Save our state, request an event, and return.
	(void)QTSS_SetValue(inParams->inRTSPRequest, sReadOffsetAttr, 0, &theReadOffset, sizeof(theReadOffset));
	(void)QTSS_SetValue(inParams->inRTSPRequest, sWriteOffsetAttr, 0, &theWriteOffset, sizeof(theWriteOffset));
	(void)QTSS_SetValue(inParams->inRTSPRequest, sStateAttr, 0, &theState, sizeof(theState));

	// If we're reading, wait for the file to become readable
	if (theState == kReadingBufferState)
	{
		Assert(theRefP != NULL);
		(void)QTSS_RequestEvent(*theRefP, EV_RE);
	}
	// If we're writing, wait for the socket to become writable
	else
		(void)QTSS_RequestEvent(*theSocketRefP, EV_WR);
	return QTSS_NoErr;
}