/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		QTSSReflectorModule.cpp

	Contains:	Implementation of QTSSReflectorModule class. 
					
	
	$Log: QTSSReflectorModule.cpp,v $
	Revision 1.4  2000/02/11 01:08:32  serenyi
	Modules support 304 Not Modified and append cache control header
	
	Revision 1.3  2000/01/28 08:33:51  serenyi
	Fixed preliminary reflector bugs
	
	Revision 1.2  2000/01/27 02:07:45  rtucker
	Merge changes from v81
	
	Revision 1.21  2000/01/26 21:43:32  serenyi
	Fix 2399850 and now allow reflector to act as an RTSP client
	
	Revision 1.20  2000/01/21 23:52:34  rtucker
	
	logging changes
	
	Revision 1.19  2000/01/19 20:34:30  serenyi
	Moved to new NEW macro
	
	Revision 1.18  1999/12/21 01:43:25  lecroy
	*** empty log message ***
	
	Revision 1.17  1999/12/06 19:40:36  rtucker
	removed warning on second DECSCRIBE
	
	Revision 1.16  1999/12/04 09:29:06  serenyi
	Fix #2419499. Got rid of #DEBUG around code that wasn't just for debugging
	
	Revision 1.15  1999/11/15 01:24:32  serenyi
	Added in 4 char debugging codes
	
	Revision 1.14  1999/11/08 23:02:01  serenyi
	Updated to latest version of QTSS API
	
	Revision 1.13  1999/11/04 22:33:58  serenyi
	Added some error feedback to the relay parsing stuff
	
	Revision 1.12  1999/10/28 01:45:24  serenyi
	Fix 2404454 ReflectorSession & module were still assuming trackIDs were always 1-based
	
	Revision 1.11  1999/10/26 00:15:23  serenyi
	Make sure to check that there is no output for this session on a DESCRIBE
	
	Revision 1.10  1999/10/22 23:01:55  serenyi
	Made it so the reflector never reflects any more data after a TEARDOWN has been issued.
	
	Revision 1.9  1999/10/21 18:30:05  rtucker
	fixed "double send" bug in relfector
	
	Revision 1.8  1999/10/19 22:17:29  serenyi
	ReflectorSession now keeps track of its bitrate, each client puts that bitrate into the RTPSession
	
	Revision 1.7  1999/10/18 20:20:10  serenyi
	changed "dictType" to "objectType"
	
	Revision 1.6  1999/10/16 00:38:39  murata
	now using typed accessors for prefs dictionary.
	
	Revision 1.5  1999/10/14 01:17:27  rtucker
	updates for thinning
	
	Revision 1.4  1999/10/11 18:44:17  serenyi
	Use new RCFSourceInfo class for parsing inline relay configuratiion info
	
	Revision 1.3  1999/10/07 09:01:49  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.2  1999/09/30 00:23:01  serenyi
	Changed ipAddr field of SourceInfo to be fDestIPAddr
	
	Revision 1.1.1.1  1999/09/22 23:01:07  lecroy
	importing
	
	Revision 1.3  1999/09/17 22:56:40  serenyi
	Fixed a memory leak: sdp data & local sdp data wasn't being destroyed
	
	Revision 1.2  1999/09/15 17:01:31  serenyi
	Made relay, reflector work
	
	Revision 1.1.1.1  1999/09/13 16:16:34  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:08:25  ds
	Created
	

*/

#include "QTSSReflectorModule.h"
#include "QTSSModuleUtils.h"
#include "ReflectorSession.h"
#include "OSArrayObjectDeleter.h"
#include "QTSS_Private.h"

#include "OSMemory.h"
#include "OSRef.h"
#include "OSFileSource.h"
#include "IdleTask.h"
#include "Task.h"
#include "OS.h"
#include "Socket.h"
#include "SocketUtils.h"
#include "FilePrefsSource.h"
#include "OSArrayObjectDeleter.h"

//ReflectorOutput objects
#include "RTPSessionOutput.h"

//SourceInfo objects
#include "SDPSourceInfo.h"


#if DEBUG
#define REFLECTOR_MODULE_DEBUGGING 0
#else
#define REFLECTOR_MODULE_DEBUGGING 0
#endif

// ATTRIBUTES
static QTSS_AttributeID			sOutputAttr		 			=	qtssIllegalAttrID;
static QTSS_AttributeID			sExpectedDigitFilenameErr	=	qtssIllegalAttrID;
static QTSS_AttributeID			sReflectorBadTrackIDErr		=	qtssIllegalAttrID;

// STATIC DATA

// ref to the prefs dictionary object
static QTSS_Object 		sServerPrefs;
static OSRefTable* 		sSessionMap = NULL;
static const StrPtrLen 		kCacheControlHeader("no-cache");

// Important strings
static StrPtrLen	sSDPSuffix(".sdp");

// FUNCTION PROTOTYPES

static QTSS_Error QTSSReflectorModuleDispatch(QTSS_Role inRole, QTSS_RoleParamPtr inParams);
static QTSS_Error 	Register(QTSS_Register_Params* inParams);
static QTSS_Error Initialize(QTSS_Initialize_Params* inParams);
static QTSS_Error Shutdown();
static QTSS_Error ProcessRTSPRequest(QTSS_StandardRTSP_Params* inParams);
static QTSS_Error DoDescribe(QTSS_StandardRTSP_Params* inParams);
ReflectorSession* FindOrCreateSession(StrPtrLen* inPath, QTSS_RTSPRequestObject inRequest);
static QTSS_Error DoSetup(QTSS_StandardRTSP_Params* inParams, ReflectorSession* inSession);
static QTSS_Error DoPlay(QTSS_StandardRTSP_Params* inParams, ReflectorSession* inSession);
static QTSS_Error DestroySession(QTSS_ClientSessionClosing_Params* inParams);
static void RemoveOutput(ReflectorOutput* inOutput, ReflectorSession* inSession);



// FUNCTION IMPLEMENTATIONS
#pragma mark __QTSS_REFLECTOR_MODULE__

QTSS_Error QTSSReflectorModule_Main(void* inPrivateArgs)
{
	return _stublibrary_main(inPrivateArgs, QTSSReflectorModuleDispatch);
}


QTSS_Error 	QTSSReflectorModuleDispatch(QTSS_Role inRole, QTSS_RoleParamPtr inParams)
{
	switch (inRole)
	{
		case QTSS_Register_Role:
			return Register(&inParams->regParams);
		case QTSS_Initialize_Role:
			return Initialize(&inParams->initParams);
		case QTSS_RTSPPreProcessor_Role:
			return ProcessRTSPRequest(&inParams->rtspRequestParams);
		case QTSS_ClientSessionClosing_Role:
			return DestroySession(&inParams->clientSessionClosingParams);
		case QTSS_Shutdown_Role:
			return Shutdown();
	}
	return QTSS_NoErr;
}


QTSS_Error Register(QTSS_Register_Params* inParams)
{
	// Do role & attribute setup
	(void)QTSS_AddRole(QTSS_Initialize_Role);
	(void)QTSS_AddRole(QTSS_Shutdown_Role);
	(void)QTSS_AddRole(QTSS_RTSPPreProcessor_Role);
	(void)QTSS_AddRole(QTSS_ClientSessionClosing_Role);
	
	// Add text messages attributes
	static char*		sExpectedDigitFilenameName		= "QTSSReflectorModuleExpectedDigitFilename";
	static char*		sReflectorBadTrackIDErrName		= "QTSSReflectorModuleBadTrackID";

	(void)QTSS_AddAttribute(qtssTextMessagesObjectType, sExpectedDigitFilenameName, NULL);
	(void)QTSS_IDForAttr(qtssTextMessagesObjectType, sExpectedDigitFilenameName, &sExpectedDigitFilenameErr);

	(void)QTSS_AddAttribute(qtssTextMessagesObjectType, sReflectorBadTrackIDErrName, NULL);
	(void)QTSS_IDForAttr(qtssTextMessagesObjectType, sReflectorBadTrackIDErrName, &sReflectorBadTrackIDErr);
	
	// Add an RTP session attribute for tracking ReflectorSession objects
	static char*		sOutputName	= "QTSSReflectorModuleOutput";

	(void)QTSS_AddAttribute(qtssClientSessionObjectType, sOutputName, NULL);
	(void)QTSS_IDForAttr(qtssClientSessionObjectType, sOutputName, &sOutputAttr);
	
	// Reflector session needs to setup some parameters too.
	ReflectorStream::Register();
	// RTPSessionOutput needs to do the same
	RTPSessionOutput::Register();

	// Tell the server our name!
	static char* sModuleName = "QTSSReflectorModule";
	::strcpy(inParams->outModuleName, sModuleName);

	return QTSS_NoErr;
}


QTSS_Error Initialize(QTSS_Initialize_Params* inParams)
{
	// Setup module utils
	QTSSModuleUtils::Initialize(inParams->inMessages, inParams->inPrefs, inParams->inErrorLogStream);
	sSessionMap = new OSRefTable();
	sServerPrefs = inParams->inPrefs;

#if QTSS_REFLECTOR_EXTERNAL_MODULE
	// The reflector is dependent on a number of objects in the Common Utilities
	// library that get setup by the server if the reflector is internal to the
	// server.
	//
	// So, if the reflector is being built as a code fragment, it must initialize
	// those pieces itself
#if !MACOSXEVENTQUEUE
	::select_startevents();//initialize the select() implementation of the event queue
#endif
	OS::Initialize();
	Socket::Initialize();
	SocketUtils::Initialize();

	const UInt32 kNumReflectorThreads = 8;
	TaskThreadPool::AddThreads(kNumReflectorThreads);
	IdleTask::Initialize();
	Socket::StartThread();
#endif
	
	// Call helper class initializers
	ReflectorStream::Initialize();
	ReflectorSession::Initialize();
	
	return QTSS_NoErr;
}

QTSS_Error Shutdown()
{
#if QTSS_REFLECTOR_EXTERNAL_MODULE
	TaskThreadPool::RemoveThreads();
#endif
	return QTSS_NoErr;
}

QTSS_Error ProcessRTSPRequest(QTSS_StandardRTSP_Params* inParams)
{
	QTSS_RTSPMethod* theMethod = NULL;
	UInt32 theLen = 0;
	if ((QTSS_GetValuePtr(inParams->inRTSPRequest, qtssRTSPReqMethod, 0,
			(void**)&theMethod, &theLen) != QTSS_NoErr) || (theLen != sizeof(QTSS_RTSPMethod)))
	{
		Assert(0);
		return QTSS_RequestFailed;
	}

	if (*theMethod == qtssDescribeMethod)
		return DoDescribe(inParams);
		
	RTPSessionOutput** theOutput = NULL;
	QTSS_Error theErr = QTSS_GetValuePtr(inParams->inClientSession, sOutputAttr, 0, (void**)&theOutput, &theLen);
	if ((theErr != QTSS_NoErr) || (theLen != sizeof(RTPSessionOutput*)))
		return QTSS_RequestFailed;
	
	switch (*theMethod)
	{
		case qtssSetupMethod:
			return DoSetup(inParams, (*theOutput)->GetReflectorSession());
		case qtssPlayMethod:
			return DoPlay(inParams, (*theOutput)->GetReflectorSession());
		case qtssTeardownMethod:
			// Tell the server that this session should be killed, and send a TEARDOWN response
			(void)QTSS_Teardown(inParams->inClientSession);
			(void)QTSS_SendStandardRTSPResponse(inParams->inRTSPRequest, inParams->inClientSession, 0);
			break;
		case qtssPauseMethod:
			(void)QTSS_Pause(inParams->inClientSession);
			(void)QTSS_SendStandardRTSPResponse(inParams->inRTSPRequest, inParams->inClientSession, 0);
			break;
		default:
			break;
	}			
	return QTSS_NoErr;
}


QTSS_Error DoDescribe(QTSS_StandardRTSP_Params* inParams)
{
	// Check and see if the full path to this file matches an existing ReflectorSession
	StrPtrLen thePathPtr;
	OSCharArrayDeleter sdpPath(QTSSModuleUtils::GetFullPath(	inParams->inRTSPRequest,
																qtssRTSPReqFilePath,
																&thePathPtr.Len, &sSDPSuffix));
	
	thePathPtr.Ptr = sdpPath.GetObject();

	// If the actual file path has a .sdp in it, first look for the URL without the extra .sdp
	if (thePathPtr.Len > (sSDPSuffix.Len * 2))
	{
		// Check and see if there is a .sdp in the file path.
		// If there is, truncate off our extra ".sdp", cuz it isn't needed
		StrPtrLen endOfPath(&sdpPath.GetObject()[thePathPtr.Len - (sSDPSuffix.Len * 2)], sSDPSuffix.Len);
		if (endOfPath.Equal(sSDPSuffix))
		{
			sdpPath.GetObject()[thePathPtr.Len - sSDPSuffix.Len] = '\0';
			thePathPtr.Len -= sSDPSuffix.Len;
		}
	}
	
	ReflectorSession* theSession = FindOrCreateSession(&thePathPtr, inParams->inRTSPRequest);
	if (theSession == NULL)
		return QTSS_RequestFailed;

	RTPSessionOutput** theOutput = NULL;
	UInt32 theLen = 0;
	QTSS_Error theErr = QTSS_GetValuePtr(inParams->inClientSession, sOutputAttr, 0, (void**)&theOutput, &theLen);
	//WarnV(theErr != QTSS_NoErr, "Found sOutputAttr on DESCRIBE"); // Can this ever happen?
	// yes the client will sometimes ask twice for a DESCRIBE

	// Only add a new RTPSessionOutput if there isn't already one for this RTP session
	if (theErr != QTSS_NoErr)
	{
		//ok, we've found or setup the proper reflector session, create an RTPSessionOutput object,
		//and add it to the session's list of outputs
		RTPSessionOutput* theNewOutput = NEW RTPSessionOutput(inParams->inClientSession, theSession, sServerPrefs );
		theSession->AddOutput(theNewOutput);

		// And vice-versa, store this reflector session in the RTP session.
		(void)QTSS_SetValue(inParams->inClientSession, sOutputAttr, 0, &theNewOutput, sizeof(theNewOutput));
	}
		
	// Finally, send the DESCRIBE response
	
	//above function has signalled that this request belongs to us, so let's respond
	iovec theDescribeVec[2] = { 0 };
	
	Assert(theSession->GetLocalSDP()->Ptr != NULL);
	theDescribeVec[1].iov_base = theSession->GetLocalSDP()->Ptr;
	theDescribeVec[1].iov_len = theSession->GetLocalSDP()->Len;
	(void)QTSS_AppendRTSPHeader(inParams->inRTSPRequest, qtssCacheControlHeader,
								kCacheControlHeader.Ptr, kCacheControlHeader.Len);
	QTSSModuleUtils::SendDescribeResponse(inParams->inRTSPRequest, inParams->inClientSession,
											&theDescribeVec[0], 2, theSession->GetLocalSDP()->Len);	
	return QTSS_NoErr;
}

ReflectorSession* FindOrCreateSession(StrPtrLen* inPath, QTSS_RTSPRequestObject inRequest)
{
	// This function assumes that inPath is NULL terminated
	
	// Ok, look for a reflector session matching this full path as the ID
	OSMutexLocker locker(sSessionMap->GetMutex());
	OSRef* theSessionRef = sSessionMap->Resolve(inPath);
	ReflectorSession* theSession = NULL;

		if (theSessionRef == NULL)
		{
			//If this URL doesn't already have a reflector session, we must make a new
		//one. The first step is to create an SDPSourceInfo object.
		
		StrPtrLen theFileData;
		SDPSourceInfo* theInfo = NULL;
		
		OSFileSource::ReadEntireFile(inPath->Ptr, &theFileData);
		if (theFileData.Len > 0)
			theInfo = NEW SDPSourceInfo(theFileData.Ptr, theFileData.Len);
		else
			return NULL;
			
		if (!theInfo->IsReflectable())
		{
			delete theInfo;
			return NULL;
		}
			
			//create a reflector session, and bind the sockets
		theSession = NEW ReflectorSession(inPath);
		QTSS_Error theErr = theSession->SetupReflectorSession(theInfo, inRequest);
			if (theErr != QTSS_NoErr)
			{
				delete theSession;
				return NULL;
			}
			
			//put the session's ID into the session map.
			theErr = sSessionMap->Register(theSession->GetRef());
			Assert(theErr == QTSS_NoErr);

			//unless we do this, the refcount won't increment (and we'll delete the session prematurely
		OSRef* debug = sSessionMap->Resolve(inPath);
			Assert(debug == theSession->GetRef());
		}
		else
			theSession = (ReflectorSession*)theSessionRef->GetObject();
			
		Assert(theSession != NULL);
	return theSession;
}


QTSS_Error DoSetup(QTSS_StandardRTSP_Params* inParams, ReflectorSession* inSession)
{
	//unless there is a digit at the end of this path (representing trackID), don't
	//even bother with the request
	StrPtrLen theDigit;
	(void)QTSS_GetValuePtr(inParams->inRTSPRequest, qtssRTSPReqFileDigit, 0, (void**)&theDigit.Ptr, &theDigit.Len);
	if (theDigit.Len == 0)
		return QTSSModuleUtils::SendErrorResponse(inParams->inRTSPRequest, qtssClientBadRequest,
													sExpectedDigitFilenameErr);
	UInt32 theTrackID = ::strtol(theDigit.Ptr, NULL, 10);
	QTSS_Error theErr = QTSS_NoErr;
	
	// Get info about this trackID
	SourceInfo::StreamInfo* theStreamInfo = inSession->GetSourceInfo()->GetStreamInfoByTrackID(theTrackID);
	// If theStreamInfo is NULL, we don't have a legit track, so return an error
	if (theStreamInfo == NULL)
		return QTSSModuleUtils::SendErrorResponse(inParams->inRTSPRequest, qtssClientBadRequest,
													sReflectorBadTrackIDErr);
													
	StrPtrLen* thePayloadName = &theStreamInfo->fPayloadName;
	QTSS_RTPPayloadType thePayloadType = theStreamInfo->fPayloadType;
	
	QTSS_RTPStreamObject newStream = NULL;
	{
		// Ok, this is completely crazy but I can't think of a better way to do this that's
		// safe so we'll do it this way for now. Because the ReflectorStreams use this session's
		// stream queue, we need to make sure that each ReflectorStream is not reflecting to this
		// session while we call QTSS_AddRTPStream. One brutal way to do this is to grab each
		// ReflectorStream's mutex, which will stop every reflector stream from running.
		
		for (UInt32 x = 0; x < inSession->GetNumStreams(); x++)
			inSession->GetStreamByIndex(x)->GetMutex()->Lock();
			
		theErr = QTSS_AddRTPStream(inParams->inClientSession, inParams->inRTSPRequest, &newStream, 0);

		for (UInt32 x = 0; x < inSession->GetNumStreams(); x++)
			inSession->GetStreamByIndex(x)->GetMutex()->Unlock();
			
		if (theErr != QTSS_NoErr)
			return theErr;
	}
	
	// Set up dictionary items for this stream
	theErr = QTSS_SetValue(newStream, qtssRTPStrPayloadName, 0, thePayloadName->Ptr, thePayloadName->Len);
	Assert(theErr == QTSS_NoErr);
	theErr = QTSS_SetValue(newStream, qtssRTPStrPayloadType, 0, &thePayloadType, sizeof(thePayloadType));
	Assert(theErr == QTSS_NoErr);
	theErr = QTSS_SetValue(newStream, qtssRTPStrTrackID, 0, &theTrackID, sizeof(theTrackID));
	Assert(theErr == QTSS_NoErr);

	// Set the number of quality levels.
	static UInt32 sNumQualityLevels = ReflectorSession::kNumQualityLevels;
	theErr = QTSS_SetValue(newStream, qtssRTPStrNumQualityLevels, 0, &sNumQualityLevels, sizeof(sNumQualityLevels));
	Assert(theErr == QTSS_NoErr);
	
	//send the setup response
	(void)QTSS_AppendRTSPHeader(inParams->inRTSPRequest, qtssCacheControlHeader,
								kCacheControlHeader.Ptr, kCacheControlHeader.Len);
	(void)QTSS_SendStandardRTSPResponse(inParams->inRTSPRequest, newStream, 0);
	return QTSS_NoErr;
}


QTSS_Error DoPlay(QTSS_StandardRTSP_Params* inParams, ReflectorSession* inSession)
{
	// Tell the session what the bitrate of this reflection is. This is nice for logging,
	// it also allows the server to scale the TCP buffer size appropriately if we are
	// interleaving the data over TCP. This must be set before calling QTSS_Play so the
	// server can use it from within QTSS_Play
	UInt32 bitsPerSecond =	inSession->GetBitRate();
	(void)QTSS_SetValue(inParams->inClientSession, qtssCliSesMovieAverageBitRate, 0, &bitsPerSecond, sizeof(bitsPerSecond));

	//Server shouldn't send RTCP (reflector does it), server shouldn't write rtp header either
	QTSS_Error theErr = QTSS_Play(inParams->inClientSession, inParams->inRTSPRequest, 0);
	if (theErr != QTSS_NoErr)
		return theErr;
	
	//and send a standard play response
	(void)QTSS_SendStandardRTSPResponse(inParams->inRTSPRequest, inParams->inClientSession, 0);
	return QTSS_NoErr;
}

QTSS_Error DestroySession(QTSS_ClientSessionClosing_Params* inParams)
{
	RTPSessionOutput** theOutput = NULL;
	UInt32 theLen = 0;
	QTSS_Error theErr = QTSS_GetValuePtr(inParams->inClientSession, sOutputAttr, 0, (void**)&theOutput, &theLen);
	if ((theErr != QTSS_NoErr) || (theLen != sizeof(RTPSessionOutput*)) || (theOutput == NULL))
		return QTSS_RequestFailed;
	
	RemoveOutput(*theOutput, (*theOutput)->GetReflectorSession());
	return QTSS_NoErr;
}

void RemoveOutput(ReflectorOutput* inOutput, ReflectorSession* inSession)
{
	// This function removes the output from the ReflectorSession, then
	// checks to see if the session should go away. If it should, this deletes it
	inSession->RemoveOutput(inOutput);

	//check if the ReflectorSession should be deleted
	//(it should if its ref count has dropped to 0)
	OSMutexLocker locker (sSessionMap->GetMutex());
	//decrement the ref count
	sSessionMap->Release(inSession->GetRef());
	if (inSession->GetRef()->GetRefCount() == 0)
	{
		sSessionMap->UnRegister(inSession->GetRef());
		delete inSession;
	}	
	delete inOutput;
}

