/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		RTSPReflectorOutput.cpp

	Contains:	Implementation of object in .h file
					
	$Log: RTPSessionOutput.cpp,v $
	Revision 1.1.1.1  2000/01/17 23:35:56  wsanchez
	Import of StreamingServer-2.0 (Apple)
	
	Revision 1.12  1999/12/21 01:43:26  lecroy
	*** empty log message ***
	
	Revision 1.11  1999/11/08 23:02:01  serenyi
	Updated to latest version of QTSS API
	
	Revision 1.10  1999/10/27 17:52:45  serenyi
	ntegrated Matthias's MacOS / GUSI changes
	
	Revision 1.9  1999/10/26 00:15:23  serenyi
	Make sure to check that there is no output for this session on a DESCRIBE
	
	Revision 1.8  1999/10/22 23:01:56  serenyi
	Made it so the reflector never reflects any more data after a TEARDOWN has been issued.
	
	Revision 1.7  1999/10/21 18:30:05  rtucker
	fixed "double send" bug in relfector
	
	Revision 1.6  1999/10/19 21:42:55  rtucker
	changed local default of video delay tolerance
	
	Revision 1.5  1999/10/18 20:20:10  serenyi
	changed "dictType" to "objectType"
	
	Revision 1.4  1999/10/16 00:57:27  rtucker
	fix reflector bugs from thinning
	
	Revision 1.3  1999/10/14 01:17:27  rtucker
	updates for thinning
	
	Revision 1.2  1999/10/07 09:01:49  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.1.1.1  1999/09/22 23:01:08  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:34  rtucker
	initial checkin from rrv14
	
	

*/


#include "RTPSessionOutput.h"


#include <sys/errno.h>


#if DEBUG 
#define RTP_SESSION_DEBUGGING 0
#else
#define RTP_SESSION_DEBUGGING 0
#endif

// ATTRIBUTES

static QTSS_AttributeID 	sNextSeqNumAttr				= qtssIllegalAttrID;
static QTSS_AttributeID 	sSeqNumOffsetAttr			= qtssIllegalAttrID;

RTPSessionOutput::RTPSessionOutput(QTSS_ClientSessionObject inClientSession, ReflectorSession* inReflectorSession, QTSS_Object serverPrefs)
				: fClientSession(inClientSession), fReflectorSession(inReflectorSession)
{
	
	// get server prefs for tolerance in delivering late packets
	// important for HTTP thinning.
	UInt32 	valueSize =  sizeof(SInt32);
	
	fMaxVideoDelayTolerance = 500;
	
	QTSS_Error getErr = QTSS_GetValue( serverPrefs, qtssPrefsTCPVideoDelayToleranceInMSec, 0,
			(void*)&fMaxVideoDelayTolerance, &valueSize);
			
	Assert( getErr == QTSS_NoErr );
	Assert( valueSize == sizeof(SInt32) );

	fMaxAudioDelayTolerance = 2500;
	
	getErr = QTSS_GetValue( serverPrefs, qtssPrefsTCPAudioDelayToleranceInMSec, 0,
			(void*)&fMaxAudioDelayTolerance, &valueSize);

	Assert( getErr == QTSS_NoErr );
	Assert( valueSize == sizeof(SInt32) );
	
	#if RTP_SESSION_DEBUGGING
	printf( "RTPSessionOutput:ctor inReflectorSession->GetNumStreams: %li\n", (long) inReflectorSession->GetNumStreams() );
	#endif	
	// create a bookmark for each stream we'll reflect
	this->InititializeBookmarks( inReflectorSession->GetNumStreams() );
}

void RTPSessionOutput::Register()
{
	// Add some attributes to QTSS_RTPStream dictionary for thinning
	static char*		sNextSeqNum 			= "qtssNextSeqNum";
	static char*		sSeqNumOffset			= "qtssSeqNumOffset";

	(void)QTSS_AddAttribute(qtssRTPStreamObjectType, sNextSeqNum, NULL);
	(void)QTSS_IDForAttr(qtssRTPStreamObjectType, sNextSeqNum, &sNextSeqNumAttr);

	(void)QTSS_AddAttribute(qtssRTPStreamObjectType, sSeqNumOffset, NULL);
	(void)QTSS_IDForAttr(qtssRTPStreamObjectType, sSeqNumOffset, &sSeqNumOffsetAttr);
}


QTSS_Error	RTPSessionOutput::WritePacket(StrPtrLen* inPacket, UInt32 inStreamID, UInt32 inFlags, SInt64 packetLatenessInMSec )
{
	QTSS_RTPSessionState* 	theState = NULL;
	UInt32 					theLen = 0;
	QTSS_Error	 			writeErr = QTSS_NoErr;
	
	(void)QTSS_GetValuePtr(fClientSession, qtssCliSesState, 0, (void**)&theState, &theLen);
	if (*theState != qtssPlayingState)
		return QTSS_NoErr;
		
	
	//TODO -- currently the code bails out when it gets a write err on the first stream
	// that wants a given packet.  It -is- possible that  a client could have multiple
	// streams for the same track.  In that case, they might see duplicate packets on 
	// a retrasnmission caused by one stream or the other.

	//make sure all RTP streams with this ID see this packet
	QTSS_RTPStreamObject* theStream = NULL;
						
	for (int z = 0; QTSS_GetValuePtr(fClientSession, qtssCliSesStreamObjects, z, (void**)&theStream, &theLen) == QTSS_NoErr; z++)
	{
		UInt32* theTrackID = NULL;
		QTSS_RTPPayloadType* thePayloadType = NULL;
		(void)QTSS_GetValuePtr(*theStream, qtssRTPStrTrackID, 0, (void**)&theTrackID, &theLen);
		if ((theTrackID != NULL) && (*theTrackID == inStreamID))
		{
			(void)QTSS_GetValuePtr(*theStream, qtssRTPStrPayloadType, 0, (void**)&thePayloadType, &theLen);
			//We may have thinned this client in the past. If that's the case,
			//the sequence number of this packet needs to be adjusted, but only
			//for this client, so we need to make sure we can restore the old sequence
			//number after we mess around with it.
			UInt16 theSeqNumber = this->GetPacketSeqNumber(inPacket);
			
			if ((thePayloadType == NULL) ||
				(*thePayloadType != qtssVideoPayloadType) ||
				(inFlags == qtssWriteFlagsIsRTCP) ||
				(!this->PacketShouldBeThinned(*theStream, inPacket)))
			{
				// now that we know which payload type needs to be delivered
				// apply the lateness threshold to it.  
				if (*thePayloadType == qtssVideoPayloadType && packetLatenessInMSec > fMaxVideoDelayTolerance )
				{	
					#if RTP_SESSION_DEBUGGING > 1
					printf( "video thinned  @ WritePacket packetLatenessInMSec: %li\n", (long)packetLatenessInMSec );
					#endif
					
					break;
				}
				else if ( packetLatenessInMSec > fMaxAudioDelayTolerance )
				{	
					#if RTP_SESSION_DEBUGGING > 1
					printf( "audio thinned  @ WritePacket\n" );
					#endif
					break;
				}
					
				writeErr = QTSS_Write(*theStream, inPacket->Ptr, inPacket->Len, NULL, inFlags);
				
				#if RTP_SESSION_DEBUGGING
				if ( writeErr != QTSS_WouldBlock )
					::printf( "rtp::WritePacket: sess 0x%lx payload %li, sequence %li\n", (long)this, (long)*thePayloadType, (long)theSeqNumber );
				#endif
			}

			//Reset the old sequence number
			this->SetPacketSeqNumber(inPacket, theSeqNumber);

		}

		if ( writeErr != QTSS_NoErr )
			break;
	}
	
	return writeErr;
}


UInt16 RTPSessionOutput::GetPacketSeqNumber(StrPtrLen* inPacket)
{
	if (inPacket->Len < 4)
		return 0;
	
	//The RTP seq number is the second short of the packet
	UInt16* seqNumPtr = (UInt16*)inPacket->Ptr;
	return ntohs(seqNumPtr[1]);
}

void RTPSessionOutput::SetPacketSeqNumber(StrPtrLen* inPacket, UInt16 inSeqNumber)
{
	if (inPacket->Len < 4)
		return;

	//The RTP seq number is the second short of the packet
	UInt16* seqNumPtr = (UInt16*)inPacket->Ptr;
	seqNumPtr[1] = htons(inSeqNumber);
}


Bool16 RTPSessionOutput::PacketShouldBeThinned(QTSS_RTPStreamObject inStream, StrPtrLen* inPacket)
{
	static UInt16 sZero = 0;
	
	//This function determines whether the packet should be dropped.
	//It also adjusts the sequence number if necessary

	if (inPacket->Len < 4)
		return false;
	
	UInt16 curSeqNum = this->GetPacketSeqNumber(inPacket);
	UInt32* curQualityLevel = NULL;
	UInt16* nextSeqNum = NULL;
	UInt16* theSeqNumOffset = NULL;
	
	UInt32 theLen = 0;
	(void)QTSS_GetValuePtr(inStream, qtssRTPStrQualityLevel, 0, (void**)&curQualityLevel, &theLen);
	if ((curQualityLevel == NULL) || (theLen != sizeof(UInt32)))
		return false;
	(void)QTSS_GetValuePtr(inStream, sNextSeqNumAttr, 0, (void**)&nextSeqNum, &theLen);
	if ((nextSeqNum == NULL) || (theLen != sizeof(UInt16)))
	{
		nextSeqNum = &sZero;
		(void)QTSS_SetValue(inStream, sNextSeqNumAttr, 0, nextSeqNum, sizeof(UInt16));
	}
	(void)QTSS_GetValuePtr(inStream, sSeqNumOffsetAttr, 0, (void**)&theSeqNumOffset, &theLen);
	if ((theSeqNumOffset == NULL) || (theLen != sizeof(UInt16)))
	{
		theSeqNumOffset = &sZero;
		(void)QTSS_SetValue(inStream, sSeqNumOffsetAttr, 0, theSeqNumOffset, sizeof(UInt16));
	}
	UInt16 newSeqNumOffset = *theSeqNumOffset;
	
	//Check to see if we need to drop to audio only
	if ((*curQualityLevel == ReflectorSession::kAudioOnlyQuality) &&
		(*nextSeqNum == 0))
	{
#if REFLECTOR_THINNING_DEBUGGING || RTP_SESSION_DEBUGGING
		printf(" *** Reflector Dropping to audio only *** \n");
#endif
		//All we need to do in this case is mark the sequence number of the first dropped packet
		(void)QTSS_SetValue(inStream, sNextSeqNumAttr, 0, &curSeqNum, sizeof(UInt16));		
	}
	
	//Check to see if we can reinstate video
	if ((*curQualityLevel == ReflectorSession::kNormalQuality) && (*nextSeqNum != 0))
	{
		//Compute the offset amount for each subsequent sequence number. This offset will
		//alter the sequence numbers so that they increment normally (providing the illusion to the
		//client that there are no missing packets)
		newSeqNumOffset = (*theSeqNumOffset) + (curSeqNum - (*nextSeqNum));
		(void)QTSS_SetValue(inStream, sSeqNumOffsetAttr, 0, &newSeqNumOffset, sizeof(UInt16));
		(void)QTSS_SetValue(inStream, sNextSeqNumAttr, 0, &sZero, sizeof(UInt16));
#if REFLECTOR_THINNING_DEBUGGING || RTP_SESSION_DEBUGGING
		printf("Reflector Reinstating video to probe the client. Offset = %li\n",  (long)newSeqNumOffset );
#endif
	}
	
	//tell the caller whether to drop this packet or not.
	if (*curQualityLevel == ReflectorSession::kAudioOnlyQuality)
		return true;
	else
	{
		//Adjust the sequence number of the current packet based on the offset, if any
		curSeqNum -= newSeqNumOffset;
		this->SetPacketSeqNumber(inPacket, curSeqNum);
		return false;
	}
}

