/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		ReflectorSession.h

	Contains:	This object supports reflecting an RTP multicast stream to N
				RTPStreams. It spaces out the packet send times in order to
				maximize the randomness of the sending pattern and smooth
				the stream.
					
	$Log: ReflectorSession.h,v $
	Revision 1.2  2000/01/27 02:07:45  rtucker
	Merge changes from v81
	
	Revision 1.11  2000/01/26 21:43:32  serenyi
	Fix 2399850 and now allow reflector to act as an RTSP client
	

*/

#ifndef __REFLECTOR_SESSION__
#define __REFLECTOR_SESSION__

#include "QTSS.h"
#include "OSRef.h"
#include "StrPtrLen.h"
#include "ResizeableStringFormatter.h"
#include "MyAssert.h"

#include "ReflectorStream.h"
#include "SourceInfo.h"



class ReflectorSession
{
	public:
	
		// Public interface to generic RTP packet forwarding engine
		
		//
		// Initialize
		//
		// Call initialize before calling any other function in this class
		static void Initialize();
			
		// Create one of these ReflectorSessions per source broadcast. For mapping purposes,
		// the object can be constructred using an optional source ID.
		//
		// Caller may also provide a SourceInfo object, though it is not needed and
		// will also need to be provided to SetupReflectorSession when that is called.
		ReflectorSession(StrPtrLen* inSourceID, SourceInfo* inInfo = NULL);
		~ReflectorSession();
		
		//
		// MODIFIERS
		
		// Call this to initialize and setup the source sockets. Once this function
		// completes sucessfully, Outputs may be added through the function calls below.
		//
		// The SourceInfo object passed in here will be owned by the ReflectorSession. Do not
		// delete it.
		
		enum
		{
			kMarkSetup = 0,		//After SetupReflectorSession is called, IsSetup returns true
			kDontMarkSetup = 1	//After SetupReflectorSession is called, IsSetup returns false
		};
		
		QTSS_Error		SetupReflectorSession(SourceInfo* inInfo, QTSS_RTSPRequestObject inRequest,
												UInt32 inFlags = kMarkSetup);
												
		// Packets get forwarded by attaching ReflectorOutput objects to a ReflectorSession.

		void	AddOutput(ReflectorOutput* inOutput);
		void 	RemoveOutput(ReflectorOutput* inOutput);
		void	ManuallyMarkSetup()	{ fIsSetup = true; }
		
		// For the Relay's status, a ReflectorSession can format an informative bit of
		// HTML to describe the source. This must be called after the ReflectorSession
		// is all setup.
		
		void	FormatHTML(StrPtrLen* inURL);

		//
		// ACCESSORS
		
		OSRef*			GetRef()			{ return &fRef; }
		OSQueueElem*	GetQueueElem()		{ return &fQueueElem; }
		UInt32			GetNumOutputs()		{ return fNumOutputs; }
		UInt32			GetNumStreams() 	{ return fSourceInfo->GetNumStreams(); }
		StrPtrLen*		GetSourceInfoHTML() { return &fSourceInfoHTML; }
		SourceInfo*		GetSourceInfo()		{ return fSourceInfo; }
		StrPtrLen*		GetLocalSDP()		{ return &fLocalSDP; }
		Bool16			IsSetup()			{ return fIsSetup; }
		ReflectorStream*GetStreamByIndex(UInt32 inIndex) { return fStreamArray[inIndex]; }
		
		// A ReflectorSession keeps track of the aggregate bit rate each
		// stream is reflecting (RTP only). Initially, this will return 0
		// until enough time passes to compute an accurate average.
		UInt32			GetBitRate();
		
		// Returns true if this SourceInfo structure is equivalent to this
		// ReflectorSession.
		Bool16 Equal(SourceInfo* inInfo);
	
		//Reflector quality levels:
		enum
		{
			kMaxHTMLSize = 128,
			kAudioOnlyQuality = 1,		//UInt32
			kNormalQuality = 0,			//UInt32
			kNumQualityLevels = 2		//UInt32
		};

	private:
	
		// Is this session setup?
		Bool16		fIsSetup;

		// For storage in the session map		
		OSRef  		fRef;
		StrPtrLen 	fSourceID;
		OSQueueElem	fQueueElem; // Relay uses this.
		
		unsigned int		fNumOutputs;
		
		ReflectorStream** 	fStreamArray;
		
		char		fHTMLBuf[kMaxHTMLSize];
		StrPtrLen	fSourceInfoHTML;
		ResizeableStringFormatter fFormatter;
		
		// The reflector session needs to hang onto the source info object
		// for it's entire lifetime. Right now, this is used for reflector-as-client.
		SourceInfo*	fSourceInfo;
		StrPtrLen fLocalSDP;
};

#endif

