/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		QTSSWebStatsModule.cpp

	Contains:	Implements web stats module

	$Log: QTSSWebStatsModule.cpp,v $
	Revision 1.1.1.1  2000/01/17 23:35:57  wsanchez
	Import of StreamingServer-2.0 (Apple)
	
	Revision 1.14  2000/01/08 00:22:59  rtucker
	added error handling for cases when DNS is not present
	
	Revision 1.13  1999/12/21 01:43:29  lecroy
	*** empty log message ***
	
	Revision 1.12  1999/12/16 02:21:49  murata
	Fix display of api version number.
	
	Revision 1.11  1999/11/17 20:44:37  serenyi
	Wasn't outputting "is out of file descriptors" and "num udp sockets" properly
	
	Revision 1.10  1999/11/08 23:01:55  serenyi
	Updated to latest version of QTSS API
	
	Revision 1.9  1999/11/08 02:27:48  lecroy
	long long printf macro & sgi mods
	
	Revision 1.8  1999/10/29 00:47:28  serenyi
	Prints out all server ports, not just one
	
	Revision 1.7  1999/10/27 17:52:48  serenyi
	ntegrated Matthias's MacOS / GUSI changes
	
	Revision 1.6  1999/10/18 20:19:04  serenyi
	Changed "dictType" to "objectType"
	
	Revision 1.5  1999/10/16 00:38:55  murata
	now using typed accessors for prefs dictionary.
	
	Revision 1.4  1999/10/14 17:59:29  lecroy
	Added qtssAPIVersion attribute
	
	Revision 1.3  1999/10/07 09:02:00  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.2  1999/09/30 00:23:23  serenyi
	Added http / rtsp session separation
	
	Revision 1.1.1.1  1999/09/22 23:01:04  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:36  rtucker
	initial checkin from rrv14
	
	
	Revision 1.1  1999/02/26 05:14:57  cl
	Added HTTP stats (QTSSWebStatsModule)
	

	

*/

#include <time.h>

#include <unistd.h>		/* for getopt() et al */
#include <stdio.h>		/* for printf */
#include <stdlib.h>		/* for getloadavg & other useful stuff */
#if __MacOSX__
#include <mach/mach.h>	/* for mach cpu state struct and functions */
#endif

#include "QTSSWebStatsModule.h"
#include "OSArrayObjectDeleter.h"
#include "StringParser.h"
#include "StrPtrLen.h"
#include "QTSSModuleUtils.h"

// STATIC DATA

static char* sResponseHeader = "HTTP/1.0 200 OK\r\nServer: QTSS/2.0\r\nConnection: Close\r\nContent-Type: text/html\r\n\r\n";

static QTSS_PrefsObject			sPrefs = NULL;
static QTSS_ServerObject			sServer = NULL;
static Bool16						sFalse = false;
static time_t					sStartupTime = 0;

// ATTRIBUTES
static QTSS_AttributeID			sWebStatsURLID = qtssIllegalAttrID;

static QTSS_AttributeID			sAccessLogEnabledPref 			= qtssIllegalAttrID;
static QTSS_AttributeID			sAccessLogDirPref 				= qtssIllegalAttrID;
static QTSS_AttributeID			sAccessLogNamePref 				= qtssIllegalAttrID;
static QTSS_AttributeID			sMaxAccessLogBytesPref 			= qtssIllegalAttrID;
static QTSS_AttributeID			sAccessRollIntervalPref 		= qtssIllegalAttrID;
static QTSS_AttributeID			sReflectorDelayPref 			= qtssIllegalAttrID;
static QTSS_AttributeID			sReflectorBucketSizePref 		= qtssIllegalAttrID;
static QTSS_AttributeID			sSvrControlUpdateIntervalPref 	= qtssIllegalAttrID;


//**************************************************
class CPUStats {
public:
	static void GetStats( double *dpLoad, long *lpTicks);
#if __MacOSX__
	static void GetMachineInfo(machine_info* ioMachineInfo) 
	{	if (!CPUStats::sInited)
			CPUStats::Init();
		*ioMachineInfo = sMachineInfo;
	}
#endif
	static void Init();
private:
	static int	_CPUCount ;
	static int	_CPUSlots [16] ;
#if __MacOSX__
	static long _laLastTicks [CPU_STATE_MAX] ;
	static struct machine_info sMachineInfo;
#endif
	static Bool16 sInited;

};
Bool16 CPUStats::sInited = false;
int	CPUStats::_CPUCount = 1;
int	CPUStats::_CPUSlots [16] = {-1};
#if __MacOSX__
long CPUStats::_laLastTicks[CPU_STATE_MAX];
struct machine_info CPUStats::sMachineInfo  = {};
#endif


static QTSS_Error 	QTSSWebStatsModuleDispatch(QTSS_Role inRole, QTSS_RoleParamPtr inParams);
static QTSS_Error 	Register(QTSS_Register_Params* inParams);
static QTSS_Error Initialize(QTSS_Initialize_Params* inParams);
static QTSS_Error FilterRequest(QTSS_Filter_Params* inParams);
static void SendStats(QTSS_StreamRef inStream, UInt32  refreshInterval, Bool16 displayHelp, StrPtrLen* fieldList);


// FUNCTION IMPLEMENTATIONS
#pragma mark __QTSS_WEB_STATS_MODULE__

QTSS_Error QTSSWebStatsModule_Main(void* inPrivateArgs)
{
	return _stublibrary_main(inPrivateArgs, QTSSWebStatsModuleDispatch);
}


QTSS_Error 	QTSSWebStatsModuleDispatch(QTSS_Role inRole, QTSS_RoleParamPtr inParams)
{
	switch (inRole)
	{
		case QTSS_Register_Role:
			return Register(&inParams->regParams);
		case QTSS_Initialize_Role:
			return Initialize(&inParams->initParams);
		case QTSS_RTSPFilter_Role:
			return FilterRequest(&inParams->rtspFilterParams);
	}
	return QTSS_NoErr;
}


QTSS_Error Register(QTSS_Register_Params* inParams)
{
	static char*		sWebStatsURLName = "web_stats_url";

	// Do role & attribute setup
	(void)QTSS_AddRole(QTSS_Initialize_Role);
	(void)QTSS_AddRole(QTSS_RTSPFilter_Role);
	
	(void)QTSS_AddAttribute(qtssPrefsObjectType, sWebStatsURLName, NULL);
	(void)QTSS_IDForAttr(qtssPrefsObjectType, sWebStatsURLName, &sWebStatsURLID);

	// Tell the server our name!
	static char* sModuleName = "QTSSWebStatsModule";
	::strcpy(inParams->outModuleName, sModuleName);

	return QTSS_NoErr;
}

QTSS_Error Initialize(QTSS_Initialize_Params* inParams)
{
	// Setup module utils
	QTSSModuleUtils::Initialize(inParams->inMessages, inParams->inPrefs, inParams->inErrorLogStream);

	sPrefs = inParams->inPrefs;
	sServer = inParams->inServer;
	sStartupTime = ::time(NULL);
	
	// Get IDs of attributes setup by other modules
	static char*		sAccessLogEnabledName 		= 	"request_logging";
	static char*		sAccessLogNameName 			= 	"request_logfile_name";
	static char*		sAccessLogDirName	 		= 	"request_logfile_dir";
	static char*		sMaxAccessLogBytesName		= 	"request_logfile_size";
	static char*		sAccessRollIntervalName 	= 	"request_logfile_interval";

	static char*		sReflectorDelayName 			= "reflector_delay";
	static char*		sReflectorBucketSizeName		= "reflector_bucket_size";
	
	static char*		sSvrControlUpdateIntervalName = "history_update_interval";

	(void)QTSS_IDForAttr(qtssPrefsObjectType, sAccessLogEnabledName, &sAccessLogEnabledPref);
	(void)QTSS_IDForAttr(qtssPrefsObjectType, sAccessLogDirName, &sAccessLogDirPref);
	(void)QTSS_IDForAttr(qtssPrefsObjectType, sAccessLogNameName, &sAccessLogNamePref);
	(void)QTSS_IDForAttr(qtssPrefsObjectType, sMaxAccessLogBytesName, &sMaxAccessLogBytesPref);
	(void)QTSS_IDForAttr(qtssPrefsObjectType, sAccessRollIntervalName, &sAccessRollIntervalPref);

	(void)QTSS_IDForAttr(qtssPrefsObjectType, sReflectorDelayName, &sReflectorDelayPref);
	(void)QTSS_IDForAttr(qtssPrefsObjectType, sReflectorBucketSizeName, &sReflectorBucketSizePref);

	(void)QTSS_IDForAttr(qtssPrefsObjectType, sSvrControlUpdateIntervalName, &sSvrControlUpdateIntervalPref);

	return QTSS_NoErr;
}

QTSS_Error FilterRequest(QTSS_Filter_Params* inParams)
{
	UInt8 sParamStopMask[] =
	{
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //0-9      //stop unless a '\t', ' ', or '&'
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //10-19  
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //20-29
		0, 0, 0, 0, 0, 0, 0, 0, 1, 0, //30-39   
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //40-49
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //50-59
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //60-69
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //70-79
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //80-89
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //90-99
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //100-109
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //110-119
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //120-129
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //130-139
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //140-149
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //150-159
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //160-169
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //170-179
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //180-189
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //190-199
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //200-209
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //210-219
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //220-229
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //230-239
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //240-249
		0, 0, 0, 0, 0, 1 			 //250-255
	};

	//check to see if we should handle this request. Invokation is triggered
	//by a "GET /" request
	
	QTSS_RTSPRequestObject theRequest = inParams->inRTSPRequest;
	
	StrPtrLen theFullRequest;
	(void)QTSS_GetValuePtr(theRequest, qtssRTSPReqFullRequest, 0, (void**)&theFullRequest.Ptr, &theFullRequest.Len);
	StringParser fullRequest(&theFullRequest);

	char theWebStatsURLBuf[256];
	theWebStatsURLBuf[0] = '\0';
	StrPtrLen theWebStatsURL(&theWebStatsURLBuf[0], 256);
	
	StrPtrLen	strPtr;
	StrPtrLen	paramName;
	StrPtrLen	fieldsList;
	
	fullRequest.ConsumeWord(&strPtr);
	if ( strPtr.Equal(StrPtrLen("GET")) )	//it's a "Get" request
	{
		fullRequest.ConsumeWhitespace();
		if ( fullRequest.Expect('/') )
		{
			UInt32  refreshInterval = 0;
			Bool16  displayHelp = false;
			
			QTSS_Error theErr = QTSS_GetValue(sPrefs, sWebStatsURLID, 0, theWebStatsURL.Ptr, &theWebStatsURL.Len);			
			if (theErr != QTSS_NoErr)
				theWebStatsURL.Len = 0;

			// If there isn't any web stats URL, we can just return at this point
			if (theWebStatsURL.Len == 0)
				return QTSS_NoErr;
				
			fullRequest.ConsumeWord(&strPtr);
			if ( strPtr.Len != 0 && strPtr.Equal(theWebStatsURL) )	//it's a "stats" request
			{
				if ( fullRequest.Expect('?') )
				{
					do {
						fullRequest.ConsumeWord(&paramName);
						
						if( paramName.Len != 0)
						{
							
							if ( paramName.Equal(StrPtrLen("refresh",strlen("refresh"))) )
							{
								if (fullRequest.Expect('='))
									refreshInterval  = fullRequest.ConsumeInteger(NULL);
							}
							else if ( paramName.Equal(StrPtrLen("help",strlen("help"))) )
							{
								displayHelp = true;
							}
							else if ( paramName.Equal(StrPtrLen("fields",strlen("fields"))) )
							{
								if (fullRequest.Expect('='))
									fullRequest.ConsumeUntil(&fieldsList, (UInt8*)sParamStopMask);
								
							}
						}
					} while ( paramName.Len != 0 && fullRequest.Expect('&') );
				}
				
				// Before sending a response, set keep alive to off for this connection
				(void)QTSS_SetValue(theRequest, qtssRTSPReqRespKeepAlive, 0, &sFalse, sizeof(sFalse));
				SendStats(inParams->inRTSPRequest, refreshInterval, displayHelp, (fieldsList.Len != 0) ? &fieldsList : NULL);
			}
		}
	}
	return QTSS_NoErr;
}


void SendStats(QTSS_StreamRef inStream, UInt32  refreshInterval, Bool16 displayHelp, StrPtrLen* fieldList)
{
	struct FieldIndex {
		char*	fieldName;
		int fieldIndex;
	};
	
	const FieldIndex kFieldIndexes[] = {
									{"title", 1},
									{"dnsname", 2},
									{"curtime", 3},
									{"linebreak", 4},
									{"serververs", 5},
									{"serverbornon", 6},
									{"serverstatus", 7},
									{"linebreak", 8},
									{"noop", 9},
									{"kernelinfo", 10},
									{"cpuload", 11},
									{"noop", 12},
									{"linebreak", 13},
									{"currtp", 14},
									{"currtsp", 15},
									{"currtsphttp", 16},
									{"curthru", 17},
									{"curpkts", 18},
									{"totbytes", 19},
									{"totconns", 20},
									{"linebreak", 21},
									{"connlimit", 22},
									{"thrulimit", 23},
									{"moviedir", 24},
									{"rtspip", 25},
									{"rtspport", 26},
									{"rtsptimeout", 27},
									{"rtptimeout", 28},
									{"secstobuffer", 29},
									{"linebreak", 30},
									{"accesslog", 31},
									{"accesslogdir",32},
									{"accesslogname", 33},
									{"accessrollsize", 34},
									{"accessrollinterval", 35},
									{"linebreak", 36},
									{"errorlog", 37},
									{"errorlogdir", 38},
									{"errorlogname", 39},
									{"errorrollsize", 40},
									{"errorrollinterval", 41},
									{"errorloglevel", 42},
									{"linebreak", 43},
									{"assertbreak", 44},
									{"autostart", 45},
									{"totbytesupdateinterval", 46},
									{"reflectordelay", 47},
									{"reflectorbucketsize", 48},
									{"historyinterval", 49},
									{"outoffiledesc", 50},
									{"numudpsockets", 51},
									{"apiversion", 52}
	};
	const int kMaxFieldNum = 52;

	
	char buffer[1024];

	(void)QTSS_Write(inStream, sResponseHeader, ::strlen(sResponseHeader), NULL, 0);

	if (refreshInterval > 0)
	{
		sprintf(buffer, "<META HTTP-EQUIV=Refresh CONTENT=%lu>\n", refreshInterval);
		(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);
	}
		
	sprintf(buffer, "<body text=\"#000000\" bgcolor=\"#C0C0C0\" link=\"#0000FF\" vlink=\"#551A8B\" alink=\"#0000FF\">\n");
	(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);
	(void)QTSS_Write(inStream, "<HTML><BODY>\n", ::strlen("<HTML><BODY>\n"), NULL, 0);
	
	if (displayHelp)
	{
		static StrPtrLen sHelpLine1("<P><b>QuickTime Streaming Server Statistics Help</b></P>\n");
		static StrPtrLen sHelpLine2("<P>Example:</P>\n");
		static StrPtrLen sHelpLine3("<BLOCKQUOTE><P>http://server/statsURL?help&amp;refresh=15&amp;fields=curtime,cpuload </P>\n");

		static StrPtrLen sHelpLine4("\"?\" means that there are options being attached to the stats request.<BR>\n");
		static StrPtrLen sHelpLine5("\"&amp;\" separates multiple stats options<BR>\n<BR>\n");

		static StrPtrLen sHelpLine6("<P>The three possible parameters to stats are:</P>\n");
		static StrPtrLen sHelpLine7("<P>\"help\"  -- shows the help information you're reading right now.</P>\n");
		static StrPtrLen sHelpLine8("<P>\"refresh=&#91;n&#93;\"  -- tells the browser to automatically update the page every &#91;n&#93; seconds.</P>\n");
		static StrPtrLen sHelpLine9("<P>\"fields=&#91;fieldList&#93;\"  -- show only the fields specified in &#91;fieldList&#93;</P>\n");
		static StrPtrLen sHelpLine10("<BLOCKQUOTE>The following fields are available for use with the \"fields\" option:</P><BLOCKQUOTE><DL>\n");
		
		(void)QTSS_Write(inStream, sHelpLine1.Ptr, sHelpLine1.Len, NULL, 0);
		(void)QTSS_Write(inStream, sHelpLine2.Ptr, sHelpLine2.Len, NULL, 0);
		(void)QTSS_Write(inStream, sHelpLine3.Ptr, sHelpLine3.Len, NULL, 0);

		(void)QTSS_Write(inStream, sHelpLine4.Ptr, sHelpLine4.Len, NULL, 0);
		(void)QTSS_Write(inStream, sHelpLine5.Ptr, sHelpLine5.Len, NULL, 0);

		(void)QTSS_Write(inStream, sHelpLine6.Ptr, sHelpLine6.Len, NULL, 0);
		(void)QTSS_Write(inStream, sHelpLine7.Ptr, sHelpLine7.Len, NULL, 0);
		(void)QTSS_Write(inStream, sHelpLine8.Ptr, sHelpLine8.Len, NULL, 0);
		(void)QTSS_Write(inStream, sHelpLine9.Ptr, sHelpLine9.Len, NULL, 0);
		(void)QTSS_Write(inStream, sHelpLine10.Ptr, sHelpLine10.Len, NULL, 0);		
	
		for (short i = 0; i < kMaxFieldNum; i++)
		{
				sprintf(buffer, "<DT><I>%s</I></DT>\n", kFieldIndexes[i].fieldName);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
		}
		
		static StrPtrLen sHelpLine11("</DL></BLOCKQUOTE></BLOCKQUOTE></BLOCKQUOTE><BR><P><HR>");
		(void)QTSS_Write(inStream, sHelpLine11.Ptr, sHelpLine11.Len, NULL, 0);		
	}

	StringParser fieldNamesParser(fieldList);
	StrPtrLen fieldName;
	int fieldNum = 0;
	do
	{
	
			
		if (fieldList != NULL)
		{
			fieldNum = 0;
			
			fieldNamesParser.ConsumeWord(&fieldName);
			
			for (short i = 0; i < kMaxFieldNum; i++)
			{
				if ( fieldName.Equal(StrPtrLen(kFieldIndexes[i].fieldName, ::strlen(kFieldIndexes[i].fieldName))) )
				{
					fieldNum = kFieldIndexes[i].fieldIndex;
					break;
				}
			}
		}
		else
		{
			fieldNum++;
			if ( fieldNum > kMaxFieldNum )
				fieldNum = 0;
		}

		UInt32 theLen = 0;
		
		switch (fieldNum)
		{
			case 1:
			{
#if __MacOSX__
					//sprintf(buffer, "<TITLE>QuickTime Streaming Server Stats</TITLE><BR>\n");
					static StrPtrLen sStatsLine1("<TITLE>QuickTime Streaming Server Stats</TITLE><BR>\n");
					(void)QTSS_Write(inStream, sStatsLine1.Ptr, sStatsLine1.Len, NULL, 0);		
#else
					//sprintf(buffer, "<TITLE>Streaming Server Stats</TITLE><BR>\n");
					static StrPtrLen sStatsLine1("<TITLE>Streaming Server Stats</TITLE><BR>\n");
					(void)QTSS_Write(inStream, sStatsLine1.Ptr, sStatsLine1.Len, NULL, 0);		
#endif
					
#if __MacOSX__
					//sprintf(buffer, "<center><h1>QuickTime Streaming Server Statistics</h1></center>\n");
					static StrPtrLen sStatsLine2("<center><h1>QuickTime Streaming Server Statistics</h1></center>\n");
					(void)QTSS_Write(inStream, sStatsLine2.Ptr, sStatsLine2.Len, NULL, 0);		
#else
					//sprintf(buffer, "<center><h1>Streaming Server Statistics</h1></center>\n");
					static StrPtrLen sStatsLine2("<center><h1>Streaming Server Statistics</h1></center>\n");
					(void)QTSS_Write(inStream, sStatsLine2.Ptr, sStatsLine2.Len, NULL, 0);		
#endif
			}
			break;
		
			case 2:
			{
				StrPtrLen theDNS;
				(void)QTSS_GetValuePtr(sServer, qtssSvrDefaultDNSName, 0, (void**)&theDNS.Ptr, &theDNS.Len);
				
				if ( theDNS.Ptr == NULL )
				{	// no DNS, try for the IP address only.
					(void)QTSS_GetValuePtr(sServer, qtssSvrDefaultIPAddr, 0, (void**)&theDNS.Ptr, &theDNS.Len);
					
				}
				
				if ( theDNS.Ptr != NULL )
				{
					sprintf(buffer, "<b>DNS Name (default): </b> %s<BR>\n", theDNS.Ptr);
					(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
				}
			}
			break;
			
			case 3:
			{
				time_t curTime = ::time(NULL); 
				sprintf(buffer, "<b>Current Time: </b> %s<BR>\n", ctime(&curTime));
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
			
			case 4:
			{
				(void)QTSS_Write(inStream, "<P><HR>", ::strlen("<P><HR>"), NULL, 0);		
			}
			break;
	
			case 5:
			{
				StrPtrLen theVersion;
				(void)QTSS_GetValuePtr(sServer, qtssSvrServerVersion, 0, (void**)&theVersion.Ptr, &theVersion.Len);
				Assert(theVersion.Ptr != NULL);
				sprintf(buffer, "<b>Server Version: </b> %s<BR>\n", theVersion.Ptr);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			case 6:
			{
				StrPtrLen theBuildDate;
				(void)QTSS_GetValuePtr(sServer, qtssSvrServerBuildDate, 0, (void**)&theBuildDate.Ptr, &theBuildDate.Len);
				Assert(theBuildDate.Ptr != NULL);
				sprintf(buffer, "<b>Server Build Date: </b> %s<BR>\n", theBuildDate.Ptr);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			case 7:
			{
			
				const char* states[]	= {	"Starting Up",
											"Running",
											"Refusing Connections",
											"Fatal Error",
											"Shutting Down"
										  };
				QTSS_ServerState theState = qtssRunningState;
				theLen = sizeof(theState);
				(void)QTSS_GetValue(sServer, qtssSvrState, 0, &theState, &theLen);

				if (theState == qtssRunningState)
					sprintf(buffer, "<b>Status: </b> %s since %s<BR>", states[theState], ctime(&sStartupTime));
				else
					sprintf(buffer, "<b>Status: </b> %s<BR>", states[theState]);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
			
			case 8:
			{
				(void)QTSS_Write(inStream, "<P><HR>", ::strlen("<P><HR>"), NULL, 0);		
			}
			break;
	
			case 9:
			{
				
				//NOOP
			}
			break;
			
		//**********************************
		//misc machines stats
			case 10:
			{
#if __MacOSX__
				machine_info machineInfo = {};


				CPUStats::GetMachineInfo(&machineInfo);
			
//				sprintf(buffer, "<b>Kernel Info: </b> Mach kernel is v%d.%d, built for %d CPU[s]<BR>\n", 
//								machineInfo.major_version, machineInfo.minor_version, machineInfo.max_cpus);
				sprintf(buffer, "<b>Kernel Info: </b> Mach kernel built for %d CPU[s]<BR>\n", machineInfo.max_cpus);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
#else
				static StrPtrLen sKernelInfoStr("<b>Kernel info not available</b><BR>");
				(void)QTSS_Write(inStream, sKernelInfoStr.Ptr, sKernelInfoStr.Len, NULL, 0);		
#endif
			}
			break;
	
			case 11:
			{
#if __MacOSX__
				double cpuLoad [3] ;
			    long cpuTicks[CPU_STATE_MAX] ;
				
				CPUStats::GetStats(cpuLoad, cpuTicks) ;
				sprintf(buffer, "<b>CPU Stats: </b> %ld%% idle, %ld%% user, %ld%% system, load: %.2f, %.2f, %.2f<BR>", 
								cpuTicks[CPU_STATE_IDLE], cpuTicks[CPU_STATE_USER], cpuTicks[CPU_STATE_SYSTEM], cpuLoad[0], cpuLoad[1], cpuLoad[2]);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
#else
				static StrPtrLen sKernelInfoStr2("<b>CPU stats not available</b><BR>");
				(void)QTSS_Write(inStream, sKernelInfoStr2.Ptr, sKernelInfoStr2.Len, NULL, 0);		
#endif
			}
			break;
	
			case 12:
			{		
			/*	
				struct vm_statistics vmStats = {};
				if (vm_statistics (current_task (), &vmStats) != KERN_SUCCESS)
			        memset (&stats, '\0', sizeof (vmStats)) ;
			*/
				(void)QTSS_Write(inStream, "<P><HR>", ::strlen("<P><HR>"), NULL, 0);		
			}
			break;
			
			case 13:
			{
				(void)QTSS_Write(inStream, "<P><HR>", ::strlen("<P><HR>"), NULL, 0);		
			}
			break;
				
			//**********************************



	
			case 14:
			{
				UInt32 curConn = 0;
				theLen = sizeof(curConn);
				(void)QTSS_GetValue(sServer, qtssRTPSvrCurConn, 0, &curConn, &theLen);

				sprintf(buffer, "<b>Current RTP Connections: </b> %lu<BR>\n", curConn);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			case 15:
			{
				UInt32 curRTSPConn = 0;
				theLen = sizeof(curRTSPConn);
				(void)QTSS_GetValue(sServer, qtssRTSPCurrentSessionCount, 0, &curRTSPConn, &theLen);

				sprintf(buffer, "<b>Current RTSP Connections: </b> %lu<BR>\n", curRTSPConn);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			case 16:
			{
				UInt32 curRTSPHTTPConn = 0;
				theLen = sizeof(curRTSPHTTPConn);
				(void)QTSS_GetValue(sServer, qtssRTSPHTTPCurrentSessionCount, 0, &curRTSPHTTPConn, &theLen);

				sprintf(buffer, "<b>Current RTSP over HTTP Connections: </b> %lu<BR>\n", curRTSPHTTPConn);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			case 17:
			{
				UInt32 curBandwidth = 0;
				theLen = sizeof(curBandwidth);
				(void)QTSS_GetValue(sServer, qtssRTPSvrCurBandwidth, 0, &curBandwidth, &theLen);

				sprintf(buffer, "<b>Current Throughput: </b> %lu kbits<BR>\n", curBandwidth/1024);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			case 18:
			{
				UInt32 curPackets = 0;
				theLen = sizeof(curPackets);
				(void)QTSS_GetValue(sServer, qtssRTPSvrCurPackets, 0, &curPackets, &theLen);

				sprintf(buffer, "<b>Current Packets Per Second: </b> %lu <BR>\n", curPackets);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			case 19:
			{
				UInt64 totBytes = 0;
				theLen = sizeof(totBytes);
				(void)QTSS_GetValue(sServer, qtssRTPSvrTotalBytes, 0, &totBytes, &theLen);

				sprintf(buffer, "<b>Total Bytes Served: </b> %"_64BITARG_"u<BR>\n", totBytes);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			case 20:
			{
				UInt32 totConn = 0;
				theLen = sizeof(totConn);
				(void)QTSS_GetValue(sServer, qtssRTPSvrTotalConn, 0, &totConn, &theLen);

				sprintf(buffer, "<b>Total Connections: </b> %lu<BR>", totConn);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
			
			case 21:
			{
				(void)QTSS_Write(inStream, "<P><HR>", ::strlen("<P><HR>"), NULL, 0);		
			}
			break;
	
			//**************************************
			case 22:
			{
				SInt32 theMaxConns = QTSSModuleUtils::GetSInt32Pref(qtssPrefsMaximumConnections, 0);
				sprintf(buffer, "<b>Maximum Connections: </b> %ld<BR>\n", theMaxConns);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);	
			}
			break;
	
			case 23:
			{
				SInt32 theMaxBandwidth = QTSSModuleUtils::GetSInt32Pref(qtssPrefsMaximumBandwidth, 0);
				sprintf(buffer, "<b>Maximum Throughput: </b> %ld<BR>\n",theMaxBandwidth);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);	
			}
			break;
	
			case 24:
			{
				OSCharArrayDeleter theMovieFolderLoc(QTSSModuleUtils::GetStringPref(qtssPrefsMovieFolder, "?"));
				sprintf(buffer, "<b>Movie Folder Path: </b> %s<BR>\n", theMovieFolderLoc.GetObject());
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			case 25:
			{
				OSCharArrayDeleter theIPAddr(QTSSModuleUtils::GetStringPref(qtssPrefsRTSPIPAddr, "INADDR_ANY"));
				sprintf(buffer, "<b>RTSP IP Address: </b> %s<BR>\n", theIPAddr.GetObject());
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			case 26:
			{
				static StrPtrLen sRTSPPortsStart("<b>RTSP Ports: </b> ");
				(void)QTSS_Write(inStream, sRTSPPortsStart.Ptr, sRTSPPortsStart.Len, NULL, 0);		

				StrPtrLen thePort;
				for (UInt32 theIndex = 0; true; theIndex++)
				{
					QTSS_Error theErr = QTSS_GetValuePtr(sServer, qtssSvrRTSPPorts, theIndex, (void**)&thePort.Ptr, &thePort.Len);
					if (theErr != QTSS_NoErr)
						break;

					Assert(thePort.Ptr != NULL);
					char temp[20];
					sprintf(temp, "%u ", *(UInt16*)thePort.Ptr);
					(void)QTSS_Write(inStream, temp, ::strlen(temp), NULL, 0);		
				}
				
				static StrPtrLen sRTSPPortsEnd("<BR>\n");
				(void)QTSS_Write(inStream, sRTSPPortsEnd.Ptr, sRTSPPortsEnd.Len, NULL, 0);		
			}
			break;
	
			case 27:
			{
				UInt32 theRTSPTimeout = QTSSModuleUtils::GetUInt32Pref(qtssPrefsRTSPTimeout, 0);
				sprintf(buffer, "<b>RTSP Timeout: </b> %lu<BR>\n", theRTSPTimeout);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);	
			}
			break;
	
			case 28:
			{
				UInt32 theRTPTimeout = QTSSModuleUtils::GetUInt32Pref(qtssPrefsRTPTimeout, 0);
				sprintf(buffer, "<b>RTP Timeout: </b> %lu<BR>\n", theRTPTimeout);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);	
			}
			break;
	
			case 29:
			{
				(void)QTSS_Write(inStream, "<P><HR>", ::strlen("<P><HR>"), NULL, 0);		
			}
			break;
	
			case 30:
			{
				(void)QTSS_Write(inStream, "<P><HR>", ::strlen("<P><HR>"), NULL, 0);		
			}
			break;
	
			case 31:
			{
				OSCharArrayDeleter theAccessLogEnabled(QTSSModuleUtils::GetStringPref(sAccessLogEnabledPref, "disabled"));
				sprintf(buffer, "<b>Access Logging: </b> %s<BR>\n", theAccessLogEnabled.GetObject());
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			case 32:
			{
				OSCharArrayDeleter theAccessLogDir(QTSSModuleUtils::GetStringPref(sAccessLogDirPref, "?"));
				sprintf(buffer, "<b>Access Log Directory: </b> %s<BR>\n", theAccessLogDir.GetObject());
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			case 33:
			{
				OSCharArrayDeleter theAccessLogName(QTSSModuleUtils::GetStringPref(sAccessLogNamePref, "?"));
				sprintf(buffer, "<b>Access Log Name: </b> %s<BR>\n", theAccessLogName.GetObject());
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			case 34:
			{
				OSCharArrayDeleter theMaxAccessLogBytes(QTSSModuleUtils::GetStringPref(sMaxAccessLogBytesPref, "?"));
				sprintf(buffer, "<b>Access Log Roll Size: </b> %s<BR>\n", theMaxAccessLogBytes.GetObject());
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			case 35:
			{
				OSCharArrayDeleter theAccessRollInterval(QTSSModuleUtils::GetStringPref(sAccessRollIntervalPref, "?"));
				sprintf(buffer, "<b>Access Log Roll Interval (days): </b> %s<BR>", theAccessRollInterval.GetObject());
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
			
			case 36:
			{
				(void)QTSS_Write(inStream, "<P><HR>", ::strlen("<P><HR>"), NULL, 0);		
			}
			break;
	
			case 37:
			{
				Bool16 theErrLogEnabled = QTSSModuleUtils::GetBool16Pref(qtssPrefsErrorLogEnabled, 0);
				sprintf(buffer, "<b>Error Logging: </b> %s<BR>\n", theErrLogEnabled ? "enabled" : "disabled");
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);	
			}
			break;
	
			case 38:
			{
				OSCharArrayDeleter theErrorLogDir(QTSSModuleUtils::GetStringPref(qtssPrefsErrorLogDir, "?"));
				sprintf(buffer, "<b>Error Log Directory: </b> %s<BR>\n", theErrorLogDir.GetObject());
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			case 39:
			{
				OSCharArrayDeleter theErrorName(QTSSModuleUtils::GetStringPref(qtssPrefsErrorLogName, "?"));
				sprintf(buffer, "<b>Error Log Name: </b> %s<BR>\n", theErrorName.GetObject());
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			case 40:
			{
				UInt32 theErrorRollSize = QTSSModuleUtils::GetUInt32Pref(qtssPrefsMaxErrorLogSize, 0);
				sprintf(buffer, "<b>Error Log Roll Size: </b> %lu<BR>\n", theErrorRollSize);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);	
			}
			break;
	
			case 41:
			{
				UInt32 theErrorRollInterval = QTSSModuleUtils::GetUInt32Pref(qtssPrefsErrorRollInterval, 0);
				sprintf(buffer, "<b>Error Log Roll Interval (days): </b> %lu<BR>\n",theErrorRollInterval);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);	
			}
			break;
	
			case 42:
			{
				UInt32 theErrorLogVerbosity = QTSSModuleUtils::GetUInt32Pref(qtssPrefsErrorLogVerbosity, 0);
				sprintf(buffer, "<b>Error Log Verbosity: </b> %lu<BR>", theErrorLogVerbosity);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);	
			}
			break;
			
			case 43:
			{
				(void)QTSS_Write(inStream, "<P><HR>", ::strlen("<P><HR>"), NULL, 0);		
			}
			break;
	
			case 44:
			{
				Bool16 theBreakOnAssert = QTSSModuleUtils::GetBool16Pref(qtssPrefsBreakOnAssert, 0);
				sprintf(buffer, "<b>Break On Assert: </b> %s<BR>\n", theBreakOnAssert ? "enabled" : "disabled");
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);	
			}
			break;
	
			case 45:
			{
				Bool16 theAutoRestart = QTSSModuleUtils::GetBool16Pref(qtssPrefsAutoRestart, 0);
				sprintf(buffer, "<b>AutoStart: </b> %s<BR>\n", theAutoRestart ? "enabled" : "disabled");
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);	
			}
			break;
	
			case 46:
			{
				UInt32 totBytesInterval = QTSSModuleUtils::GetUInt32Pref(qtssPrefsTotalBytesUpdate, 0);
				sprintf(buffer, "<b>Total Bytes Update Interval: </b> %lu<BR>\n", totBytesInterval);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);	
			}
			break;
	
			case 47:
			{
				OSCharArrayDeleter theRefDelay(QTSSModuleUtils::GetStringPref(sReflectorDelayPref, "?"));
				sprintf(buffer, "<b>Reflector Delay Time: </b> %s<BR>\n", theRefDelay.GetObject());
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			case 48:
			{
				OSCharArrayDeleter theRefBucketSize(QTSSModuleUtils::GetStringPref(sReflectorBucketSizePref, "?"));
				sprintf(buffer, "<b>Reflector Bucket Size: </b> %s<BR>\n", theRefBucketSize.GetObject());
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			case 49:
			{
				OSCharArrayDeleter theHistoryUpdate(QTSSModuleUtils::GetStringPref(sSvrControlUpdateIntervalPref, "?"));
				sprintf(buffer, "<b>History Update Interval: </b> %s<BR>\n", theHistoryUpdate.GetObject());
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			case 50:
			{
				Bool16 isOutOfDescriptors = false;
				theLen = sizeof(isOutOfDescriptors);
				(void)QTSS_GetValue(sServer, qtssSvrIsOutOfDescriptors, 0, &isOutOfDescriptors, &theLen);

				sprintf(buffer, "<b>Out of file descriptors: </b> %d<BR>\n", isOutOfDescriptors);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			case 51:
			{
				UInt32 totUDPSock = 0;
				theLen = sizeof(totUDPSock);
				(void)QTSS_GetValue(sServer, qtssRTPSvrNumUDPSockets, 0, &totUDPSock, &theLen);

				sprintf(buffer, "<b>Number of UDP sockets: </b> %lu<BR>\n", totUDPSock);
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			case 52:
			{
				UInt32 apiVersion =  QTSSModuleUtils::GetUInt32Pref(qtssServerAPIVersion, 0);
				UInt32 size = sizeof(UInt32);
				(void)QTSS_GetValue(sServer, qtssServerAPIVersion, 0, &apiVersion, &size);
				sprintf(buffer, "<b>API version: </b> %d.%d<BR>\n", (int)( (UInt32) (apiVersion & (UInt32) 0xFFFF0000L) >> 16), (int)(apiVersion &(UInt32) 0x0000FFFFL));
				(void)QTSS_Write(inStream, buffer, ::strlen(buffer), NULL, 0);		
			}
			break;
	
			default:		
				break;
	
				
		} //switch fieldNum
			
			
		if (fieldList != NULL && !fieldNamesParser.Expect(','))
			fieldNum = 0;
			
	} while (fieldNum != 0);
	
	
	(void)QTSS_Write(inStream, "</BODY></HTML>\n", ::strlen("</BODY></HTML>\n"), NULL, 0);		
}






void CPUStats::Init()
{
#if __MacOSX__
    if (xxx_host_info (current_task (), &sMachineInfo) == KERN_SUCCESS) {
        register int		i, j = 0 ;
        struct machine_slot	slot ;

        _CPUCount = 0 ;
        for (i = sMachineInfo.max_cpus ; i-- ; )
            if ((xxx_slot_info (current_task (), i, &slot)
                 == KERN_SUCCESS) && (slot.is_cpu == 1)) {
                _CPUSlots[j++] = i ;
                _CPUCount++ ;
            }
        _CPUSlots[j] = -1 ;
		
	}
#endif
	
	CPUStats::sInited = true;
}

//courtesy of CJ
void CPUStats::GetStats ( double *dpLoad, long *lpTicks)
{
#if __MacOSX__
    register int	i ;
    register long	lTicks ;
    long			laSave [CPU_STATE_MAX];

	if (!CPUStats::sInited)
		CPUStats::Init();

    // Get the curent load average.
    if (-1 == getloadavg (dpLoad, 3))
    	dpLoad[0] = dpLoad[1] = dpLoad[2] = 0.0 ;

    // Return if mach says it doesn't have any CPUs.
    if (_CPUSlots[0] == -1) {
        lpTicks[0] = lpTicks[1] = lpTicks[2] = 0 ;
        return ;
    }

    // Accumulate all CPU ticks into one entry.
    laSave[0] = laSave[1] = laSave[2] = 0 ;
    for (i = _CPUCount ; i-- ; ) {
        struct machine_slot	slot ;

        // Get the percentages of CPU usage for each CPU.
        if ((xxx_slot_info (current_task (), _CPUSlots[i], &slot)
             == KERN_SUCCESS) && (slot.is_cpu == 1)) {
            laSave[CPU_STATE_USER] += slot.cpu_ticks[CPU_STATE_USER] ;
            laSave[CPU_STATE_SYSTEM] += slot.cpu_ticks[CPU_STATE_SYSTEM] ;
            laSave[CPU_STATE_IDLE] += slot.cpu_ticks[CPU_STATE_IDLE] ;
        }
    }

    // Get the difference between samples.
    for (lTicks = 0, i = CPU_STATE_MAX ; i-- ; ) {
        lpTicks[i] = laSave[i] - _laLastTicks[i] ;
        _laLastTicks[i] = laSave[i] ;
        lTicks += lpTicks[i] ;
        lpTicks[i] *= 100 ;
    }

    // Scale it to a percentage.
    if (lTicks) {
        lpTicks[CPU_STATE_USER] /= lTicks ;
        lpTicks[CPU_STATE_SYSTEM] /= lTicks ;
        lpTicks[CPU_STATE_IDLE] /= lTicks ;
    }
#endif
}
