/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */

#ifndef QTSS_H
#define QTSS_H

#ifdef __cplusplus
extern "C" {
#endif

#include "OSHeaders.h"
#include "QTSSRTSPProtocol.h"

#include <sys/uio.h>

#define QTSS_API_VERSION 				0x00010001
#define QTSS_MAX_MODULE_NAME_LENGTH		64
#define QTSS_MAX_SESSION_ID_LENGTH		32
#define QTSS_MAX_ATTRIBUTE_NAME_SIZE 	64


//*******************************
// ENUMERATED TYPES

/**********************************/
// Error Codes

enum
{
	QTSS_NoErr 				= 0,
	QTSS_RequestFailed 		= -1,
	QTSS_Unimplemented 		= -2,
	QTSS_RequestArrived 	= -3,
	QTSS_OutOfState 		= -4,
	QTSS_NotAModule 		= -5,
	QTSS_WrongVersion 		= -6,
	QTSS_IllegalService 	= -7,
	QTSS_BadIndex			= -8,
	QTSS_ValueNotFound		= -9,
	QTSS_BadArgument		= -10,
	QTSS_ReadOnly			= -11,
	QTSS_NotPreemptiveSafe	= -12,
	QTSS_NotEnoughSpace		= -13,
	QTSS_WouldBlock			= -14,
	QTSS_NotConnected		= -15
};
typedef SInt32 QTSS_Error;

// QTSS_AddStreamFlags used in the QTSS_AddStream Callback function
enum
{
	qtssASFlagsAllowDestination 	= 0x00000001,
	qtssASFlagsForceInterleave		= 0x00000002
};
typedef UInt32 QTSS_AddStreamFlags;

// QTSS_PlayFlags used in the QTSS_Play Callback function.
enum 
{
	qtssPlayFlagsSendRTCP			= 0x00000001	
};
typedef UInt32 QTSS_PlayFlags;

// Flags for QTSS_Write when writing to a QTSS_ClientSessionObject.
enum 
{
	qtssWriteFlagsNoFlags			= 0x00000000,
	qtssWriteFlagsIsRTP			 	= 0x00000001,
	qtssWriteFlagsIsRTCP			= 0x00000002	
};
typedef UInt32 QTSS_WriteFlags;

// Flags for QTSS_SendStandardRTSPResponse
enum
{
	qtssPlayRespWriteTrackInfo		= 0x00000001
};

/**********************************/
// RTP SESSION STATES
//
// Is this session playing, paused, or what?
enum
{
	qtssPausedState 		= 0,
	qtssPlayingState 		= 1
};
typedef UInt32 QTSS_RTPSessionState;

//*********************************/
// CLIENT SESSION CLOSING REASON
//
// Why is this Client going away?
enum
{
	qtssCliSesCloseClientTeardown	= 0, // Client or Module has issued a TEARDOWN
	qtssCliSesCloseTimeout			= 1 // Server is timing this session out
};
typedef UInt32 QTSS_CliSesClosingReason;

// CLIENT SESSION TEARDOWN REASON
//
//  An attribute in the QTSS_ClientSessionObject 
//
//  When calling QTSS_Teardown, a module should specify the QTSS_CliSesTeardownReason in the QTSS_ClientSessionObject 
//  if the tear down was not a client request.
//  
enum
{
	qtssCliSesTearDownClientRequest				= 0,
	qtssCliSesTearDownUnsupportedMedia 			= 1,
	qtssCliSesTearDownServerShutdown 			= 2,
	qtssCliSesTearDownServerInternalErr 		= 3
	
};
typedef UInt32  QTSS_CliSesTeardownReason;


/**********************************/
// RTSP SESSION TYPES
//
// Is this a normal RTSP session or an RTSP / HTTP session?
enum
{
	qtssRTSPSession			= 0,
	qtssRTSPHTTPSession		= 1,
	qtssRTSPHTTPInputSession= 2 //The input half of an RTSPHTTP session. These session types are usually very short lived.
};
typedef UInt32 QTSS_RTSPSessionType;

/**********************************/
// PAYLOAD TYPES
//
// When a module adds an RTP stream to a client session, it must specify
// the stream's payload type. This is so that other modules can find out
// this information in a generalized fashion. Here are the currently
// defined payload types
enum
{
	qtssUnknownPayloadType 	= 0,
	qtssVideoPayloadType 	= 1,
	qtssAudioPayloadType 	= 2
};
typedef UInt32 QTSS_RTPPayloadType;

/**********************************/
// QTSS API OBJECT TYPES
enum
{
	qtssRTPStreamDictType 			= FOUR_CHARS_TO_INT('r', 's', 't', 'd'), //rstd
	qtssRTPSessionDictType	 		= FOUR_CHARS_TO_INT('r', 's', 'e', 'd'), //rsed
	qtssRTSPSessionDictType 		= FOUR_CHARS_TO_INT('s', 's', 'e', 'd'), //ssed
	qtssRTSPRequestDictType 		= FOUR_CHARS_TO_INT('s', 'r', 'q', 'd'), //srqd
	qtssRTSPHeaderDictType	 		= FOUR_CHARS_TO_INT('s', 'h', 'd', 'd'), //shdd
	qtssServerDictType	 			= FOUR_CHARS_TO_INT('s', 's', 'v', 'd'), //ssvd
	qtssPrefsDictType 				= FOUR_CHARS_TO_INT('p', 'r', 'f', 'd'), //prfd
	qtssTextMessagesDictType 		= FOUR_CHARS_TO_INT('t', 'x', 't', 'd')  //txtd
};
enum
{
	qtssRTPStreamObjectType 		= FOUR_CHARS_TO_INT('r', 's', 't', 'o'), //rsto
	qtssClientSessionObjectType	 	= FOUR_CHARS_TO_INT('c', 's', 'e', 'o'), //cseo
	qtssRTSPSessionObjectType 		= FOUR_CHARS_TO_INT('s', 's', 'e', 'o'), //sseo
	qtssRTSPRequestObjectType 		= FOUR_CHARS_TO_INT('s', 'r', 'q', 'o'), //srqo
	qtssRTSPHeaderObjectType	 	= FOUR_CHARS_TO_INT('s', 'h', 'd', 'o'), //shdo
	qtssServerObjectType	 		= FOUR_CHARS_TO_INT('s', 'e', 'r', 'o'), //sero
	qtssPrefsObjectType 			= FOUR_CHARS_TO_INT('p', 'r', 'f', 'o'), //prfo
	qtssTextMessagesObjectType 		= FOUR_CHARS_TO_INT('t', 'x', 't', 'o')  //txto
};
typedef UInt32 QTSS_ObjectType;

/**********************************/
// ERROR LOG VERBOSITIES
//
// This provides some information to the module on the priority or
// type of this error message.
//
// When modules write to the error log stream (see below),
// the verbosity is qtssMessageVerbosity.
enum
{
	qtssFatalVerbosity				= 0,
	qtssWarningVerbosity			= 1,
	qtssMessageVerbosity			= 2,
	qtssAssertVerbosity				= 3,
	qtssDebugVerbosity				= 4,
	
	qtssIllegalVerbosity			= 5
};
typedef UInt32 QTSS_ErrorVerbosity;


/**********************************/
// SERVER STATES
//
//  An attribute in the QTSS_ServerObject returns the server state
//  as a QTSS_ServerState. Modules may also set the server state.
//
//  Setting the server state to qtssFatalErrorState, or qtssShuttingDownState
//  will cause the server to quit.
//
//  Setting the state to qtssRefusingConnectionsState will cause the server
//  to start refusing new connections.
enum
{
	qtssStartingUpState 			= 0,
	qtssRunningState 				= 1,
	qtssRefusingConnectionsState 	= 2,
	qtssFatalErrorState 			= 3,//a fatal error has occurred, not shutting down yet
	qtssShuttingDownState			= 4
};
typedef UInt32 QTSS_ServerState;

/**********************************/
// ILLEGAL ATTRIBUTE ID
enum
{
	qtssIllegalAttrID 				= -1,
	qtssIllegalServiceID 			= -1
};

//*********************************/
// QTSS DON'T CALL SENDPACKETS AGAIN
// If this time is specified as the next packet time when returning
// from QTSS_SendPackets_Role, the module won't get called again in
// that role until another QTSS_Play is issued
enum
{
	qtssDontCallSendPacketsAgain 	= -1
};

/**********************************/
//BUILT IN SERVER ATTRIBUTES

//The server maintains many attributes internally, and makes these available to plug-ins.
//Each value is a standard attribute, with a name and everything. Plug-ins may resolve the id's of
//these values by name if they'd like, but in the initialize role they will receive a struct of
//all the ids of all the internally maintained server parameters. This enumerated type block defines the indexes
//in that array for the id's.

enum
{
	//QTSS_RTPStreamObject parameters. All of these are preemptive safe.
	
	qtssRTPStrTrackID				= 0,	//r/w		//UInt32			//Unique ID identifying each stream. This will default to 0 unless set explicitly by a module.
	qtssRTPStrSSRC					= 1,	//read		//UInt32			//SSRC (Synchronization Source) generated by the server. Guarenteed to be unique amongst all streams in the session.
																			//This SSRC will be included in all RTCP Sender Reports generated by the server. See The RTP / RTCP RFC for more info on SSRCs.
	qtssRTPStrPayloadName			= 2,	//r/w		//char array		//Payload name of the media on this stream. This will be empty unless set explicitly by a module.
	qtssRTPStrPayloadType			= 3,	//r/w		//QTSS_RTPPayloadType	//Payload type of the media on this stream. This will default to qtssUnknownPayloadType unless set explicitly by a module.
	qtssRTPStrFirstSeqNumber		= 4,	//r/w		//SInt16			//Sequence number of the first RTP packet generated for this stream after the last PLAY request was issued. If known, this must be set by a module before calling QTSS_Play. It is used by the server to generate a proper RTSP PLAY response.
	qtssRTPStrFirstTimestamp		= 5,	//r/w		//SInt32			//RTP timestamp of the first RTP packet generated for this stream after the last PLAY request was issued. If known, this must be set by a module before calling QTSS_Play. It is used by the server to generate a proper RTSP PLAY response.
	qtssRTPStrTimescale				= 6,	//r/w		//SInt32			//Timescale for the track. If known, this must be set before calling QTSS_Play.
	qtssRTPStrQualityLevel			= 7,	//r/w		//UInt32			//Private
	qtssRTPStrNumQualityLevels		= 8,	//r/w		//UInt32			//Private
	qtssRTPStrBufferDelayInSecs		= 9,	//r/w		//Float32			//Size of the client's buffer. Server always sets this to 3 seconds, it is up to a module to determine what the buffer size is and set this attribute accordingly.

	// All of these parameters come out of the last RTCP packet received on this stream.
	// If the corresponding field in the last RTCP packet was blank, the attribute value will be 0.
	
	qtssRTPStrFractionLostPackets	= 10,	//read		//UInt32			// Fraction lost packets so far on this stream.
	qtssRTPStrTotalLostPackets		= 11,	//read		//UInt32			// Total lost packets so far on this stream.
	qtssRTPStrJitter				= 12,	//read		//UInt32			// Cumulative jitter on this stream.
	qtssRTPStrRecvBitRate			= 13,	//read		//UInt32			// Average bit rate received by the client in bits / sec.
	qtssRTPStrAvgLateMilliseconds	= 14,	//read		//UInt16			// Average msec packets received late.
	qtssRTPStrPercentPacketsLost	= 15,	//read		//UInt16			// Percent packets lost on this stream, as a fixed %.
	qtssRTPStrAvgBufDelayInMsec		= 16,	//read		//UInt16			// Average buffer delay in milliseconds
	qtssRTPStrGettingBetter			= 17,	//read		//UInt16			// Non-zero if the client is reporting that the stream is getting better.
	qtssRTPStrGettingWorse			= 18,	//read		//UInt16			// Non-zero if the client is reporting that the stream is getting worse.
	qtssRTPStrNumEyes				= 19,	//read		//UInt32			// Number of clients connected to this stream.
	qtssRTPStrNumEyesActive			= 20,	//read		//UInt32			// Number of clients playing this stream.
	qtssRTPStrNumEyesPaused			= 21,	//read		//UInt32			// Number of clients connected but currently paused.
	qtssRTPStrTotPacketsRecv		= 22,	//read		//UInt32			// Total packets received by the client
	qtssRTPStrTotPacketsDropped		= 23,	//read		//UInt16			// Total packets dropped by the client.
	qtssRTPStrTotPacketsLost		= 24,	//read		//UInt16			// Total packets lost.
	qtssRTPStrClientBufFill			= 25,	//read		//UInt16			// How much the client buffer is filled in 10ths of a second.
	qtssRTPStrFrameRate				= 26,	//read		//UInt16			// Current frame rate, in frames per second.
	qtssRTPStrExpFrameRate			= 27,	//read		//UInt16			// Expected frame rate, in frames per second.
	qtssRTPStrAudioDryCount			= 28,	//read		//UInt16			// Number of times the audio has run dry.
	// Address & network related parameters
	qtssRTPStrIsTCP					= 29,	//read		//Bool16			//Is this RTP stream being sent over TCP? If false, it is being sent over UDP.
	qtssRTPStrStreamRef				= 30,	//read		//QTSS_StreamRef	//A QTSS_StreamRef used for sending RTP or RTCP packets to the client. Use the QTSS_WriteFlags to specify whether each packet is an RTP or RTCP packet.
	// Does anyone need to src / dst IP add / port pair?

	// measured by the writing module
	// how late was the most recently wrtten packet in ms 
	// <0 packet delivered early, >0 packet delivered late
	// modules should report this value to the dictionary before 
	// calling qtss_write with a RTP packet
	qtssRTPStrCurrentPacketDelay  	= 31,	// r/w		//SInt32			
	
	qtssRTPStrNumParams 			= 32

};
typedef UInt32 QTSS_RTPStreamAttributes;

enum
{
	//QTSS_ClientSessionObject parameters. All of these are preemptive safe
	
	qtssCliSesStreamObjects			= 0,	//read		//QTSS_RTPStreamObject//Iterated attribute. All the QTSS_RTPStreamRefs belonging to this session.
	qtssCliSesCreateTimeInMsec		= 1,	//read		//SInt64		//Time in milliseconds the session was created.
	qtssCliSesFirstPlayTimeInMsec	= 2,	//read		//SInt64		//Time in milliseconds the first QTSS_Play call was issued.
	qtssCliSesPlayTimeInMsec		= 3,	//read		//SInt64		//Time in milliseconds the most recent play was issued.
	qtssCliSesAdjustedPlayTimeInMsec= 4,	//read		//SInt64		//Time in milliseconds the most recent play was issued, adjusted forward to delay sending packets until the play response is issued.
	qtssCliSesRTPBytesSent			= 5,	//read		//UInt32		//Number of RTP bytes sent so far on this session.
	qtssCliSesRTPPacketsSent		= 6,	//read		//UInt32		//Number of RTP packets sent so far on this session.
	qtssCliSesState					= 7,	//read		//QTSS_RTPSessionState // State of this session: is it paused or playing currently?
	qtssCliSesPresentationURL		= 8,	//read		//char array	//Presentation URL for this session. This URL is the "base" URL for the session. RTSP requests to this URL are assumed to affect all streams on the session.
	qtssCliSesFirstUserAgent		= 9,	//read		//char array	//Private
	qtssCliSesMovieDurationInSecs 	= 10,	//r/w		//Float64		//Duration of the movie for this session in seconds. This will default to 0 unless set by a module.
	qtssCliSesMovieSizeInBytes		= 11,	//r/w		//UInt64		//Movie size in bytes. This will default to 0 unless explictly set by a module
	qtssCliSesMovieAverageBitRate	= 12,	//r/w		//UInt32 		//average bits per second based on total RTP bits/movie duration. This will default to 0 unless explictly set by a module.
	qtssCliSesLastRTSPSession		= 13,	//read		//QTSS_RTSPSessionObject //Private
	qtssCliSesFullURL				= 14,	//read		//char array	//full Presentation URL for this session. Same as qtssCliSesPresentationURL, but includes rtsp://domain.com prefix
	qtssCliSesHostName				= 15,	//read		//char array	//requestes host name for s session. Just the "domain.com" portion from qtssCliSesFullURL above

	qtssCliRTSPSessRemoteAddrStr	= 16,		//read		//char array		//IP address addr of client, in dotted-decimal format.
	qtssCliRTSPSessLocalDNS			= 17,		//read		//char array		//DNS name of local IP address for this RTSP connection.
	qtssCliRTSPSessLocalAddrStr		= 18,		//read		//char array		//Ditto, in dotted-decimal format.

	qtssCliRTSPSesUserName 			= 19,		//read		//char array		// from the most recent (last) request.
	qtssCliRTSPSesUserPassword 		= 20,		//read		//char array		// from the most recent (last) request.
	qtssCliRTSPSesURLRealm 			= 21,		//read		//char array		// from the most recent (last) request.
	
	qtssCliRTSPReqRealStatusCode 	= 22,		//read		//UInt32			//Same as qtssRTSPReqRTSPReqRealStatusCode, the status from the most recent (last) request.
	qtssCliTeardownReason			= 23,		//r/w		//QTSS_CliSesTeardownReason // Must be set by a module that calls QTSS_Teardown if it is not a client requested disconnect.
	qtssCliSesNumParams 			= 24		//read		//char array		// Private

};
typedef UInt32 QTSS_ClientSessionAttributes;

enum
{
	//QTSS_RTSPSessionObject parameters
	
	//Valid in any role that receives a QTSS_RTSPSessionObject
	qtssRTSPSesID	 		= 0,		//read		//UInt32		//This is a unique ID for each session since the server started up.
	qtssRTSPSesLocalAddr	= 1,		//read		//UInt32		//Local IP address for this RTSP connection
	qtssRTSPSesLocalAddrStr	= 2,		//read		//char array			//Ditto, in dotted-decimal format.
	qtssRTSPSesLocalDNS		= 3,		//read		//char array			//DNS name of local IP address for this RTSP connection.
	qtssRTSPSesRemoteAddr	= 4,		//read		//UInt32		//IP address of client.
	qtssRTSPSesRemoteAddrStr= 5,		//read		//char array			//IP address addr of client, in dotted-decimal format.
	qtssRTSPSesEventCntxt	= 6,		//read		//QTSS_EventContextRef //An event context for the RTSP connection to the client. This should primarily be used to wait for EV_WR events if flow-controlled when responding to a client. 
	qtssRTSPSesType			= 7,		//read		//QTSS_RTSPSessionType //Is this a normal RTSP session, or is it a HTTP tunnelled RTSP session?
	qtssRTSPSesStreamRef	= 8,		//read		//QTSS_StreamRef	// A QTSS_StreamRef used for sending data to the RTSP client.

	qtssRTSPSesLastUserName 		= 9,//read		//char array		// Private
	qtssRTSPSesLastUserPassword 	= 10,//read		//char array		// Private
	qtssRTSPSesLastURLRealm 		= 11,//read		//char array		// Private
	
	qtssRTSPSesNumParams 	= 12
};
typedef UInt32 QTSS_RTSPSessionAttributes;

enum 
{
	//All text names are identical to the enumerated type names

	//QTSS_RTSPRequestObject parameters. All of these are pre-emptive safe parameters

	//Available in every role that receives the QTSS_RTSPRequestObject
	
	qtssRTSPReqFullRequest	 		= 0,	//read		//char array		//The full request sent by the client
	
	//Available in every method that receives the QTSS_RTSPRequestObject except for the QTSS_FilterMethod
	
	qtssRTSPReqMethodStr 			= 1,	//read		//char array		//RTSP Method of this request.
	qtssRTSPReqFilePath 			= 2,	//read		//char array		//URI for this request, converted to a local file system path.
	qtssRTSPReqURI 					= 3,	//read		//char array		//URI for this request
	qtssRTSPReqFilePathTrunc 		= 4, 	//read		//char array		//Same as qtssFilePathParam, without the last element of the path
	qtssRTSPReqFileName 			= 5, 	//read		//char array		//Everything after the last path separator in the file system path
	qtssRTSPReqFileDigit 			= 6,	//read		//char array		//If the URI ends with one or more digits, this points to those.
	qtssRTSPReqAbsoluteURL 			= 7,	//read		//char array		//The full URL, starting from "rtsp://"
	qtssRTSPReqTruncAbsoluteURL 	= 8, 	//read		//char array		//Absolute URL without last element of path
	qtssRTSPReqMethod 				= 9,	//read		//QTSS_RTSPMethod	//Method as QTSS_RTSPMethod
	qtssRTSPReqStatusCode			= 10,	//r/w		//QTSS_RTSPStatusCode	//The current status code for the request as QTSS_RTSPStatusCode. By default, it is always qtssSuccessOK. If a module sets this attribute, and calls QTSS_SendRTSPHeaders, the status code of the header generated by the server will reflect this value.
	qtssRTSPReqStartTime			= 11,	//read		//Float64	//Start time specified in Range: header of PLAY request.
	qtssRTSPReqStopTime				= 12,	//read		//Float64	//Stop time specified in Range: header of PLAY request.
	qtssRTSPReqRespKeepAlive		= 13,	//r/w		//Bool16		//Will (should) the server keep the connection alive. Set this to false if the connection should be terminated after completion of this request.
	qtssRTSPReqRootDir				= 14,	//r/w		//char array	//Root directory to use for this request. The default value for this parameter is the server's media folder path. Modules may set this attribute from the QTSS_RTSPRoute_Role.
	qtssRTSPReqRealStatusCode		= 15,	//read		//UInt32	//Same as qtssRTSPReqStatusCode, but translated from QTSS_RTSPStatusCode into an actual RTSP status code.
	qtssRTSPReqStreamRef			= 16,	//read		//QTSS_StreamRef //A QTSS_StreamRef for sending data to the RTSP client. This stream ref, unlike the one provided as an attribute in the QTSS_RTSPSessionObject, will never return EWOULDBLOCK in response to a QTSS_Write or a QTSS_WriteV call.

	qtssRTSPReqUserName				= 17,	//read		//char array//decoded Authentication information when provided by the RTSP request. See RTSPSessLastUserName.
	qtssRTSPReqUserPassword			= 18,	//read		//char array //decoded Authentication information when provided by the RTSP request. See RTSPSessLastUserPassword.
	qtssRTSPReqUserAllowed			= 19,	//r/w		//Bool16	//Default is true, set to false if request is denied. Missing or bad movie files should allow the server to handle the situation and return true.
	qtssRTSPReqURLRealm				= 20,	//r/w		//char array //The authorization entity for the client to display "Please enter password for -realm- at server name. The default realm is QT Streaming Server.
	qtssRTSPReqLocalPath			= 21,	//read		//char array //The full local path to the file. This Attribute is set after the Routing Role has run and before any other role is called. 
	
	qtssRTSPReqIfModSinceDate		= 22,	//read		//time_t	// If the RTSP request contains an If-Modified-Since header, this is the if-modified date, converted to a time_t

	qtssRTSPReqNumParams 			= 23
	
};
typedef UInt32 QTSS_RTSPRequestAttributes;

enum
{
	//QTSS_ServerObject parameters
	
	// These parameters ARE pre-emptive safe.
	
	qtssServerAPIVersion 			= 0,	//read	//UInt32			//The API version supported by this server (format 0xMMMMmmmm, where M=major version, m=minor version)
	qtssSvrDefaultDNSName 			= 1,	//read	//char array		//The "default" DNS name of the server
	qtssSvrDefaultIPAddr 			= 2,	//read	//UInt32			//The "default" IP address of the server
	qtssSvrServerName				= 3,	//read	//char array		//Name of the server
	qtssSvrServerVersion			= 4,	//read	//char array		//Version of the server
	qtssSvrServerBuildDate			= 5,	//read	//char array		//When was the server built?
	qtssSvrRTSPPorts			 	= 6,	//read	//UInt16			//Indexed parameter: all the ports the server is listening on
	qtssSvrRTSPServerHeader		 	= 7,	//read	//char array		//Server: header that the server uses to respond to RTSP clients

	// These parameters are NOT pre-emptive safe, they cannot be accessed
	// via. QTSS_GetValuePtr.
	
	qtssSvrState					= 8,	//r/w	//QTSS_ServerState	//The current state of the server. If a module sets the server state, the server will respond in the appropriate fashion. Setting to qtssRefusingConnectionsState causes the server to refuse connections, setting to qtssFatalErrorState or qtssShuttingDownState causes the server to quit.
	qtssSvrIsOutOfDescriptors		= 9,	//read	//Bool16			//true if the server has run out of file descriptors, false otherwise
	qtssRTSPCurrentSessionCount 	= 10,	//read	//UInt32			//Current number of connected clients over standard RTSP
	qtssRTSPHTTPCurrentSessionCount = 11,	//read	//UInt32			//Current number of connected clients over RTSP / HTTP

	qtssRTPSvrNumUDPSockets			= 12,	//read		//UInt32	//Number of UDP sockets currently being used by the server
	qtssRTPSvrCurConn 				= 13,	//read		//UInt32	//Number of clients currently connected to the RTP server
	qtssRTPSvrTotalConn				= 14,	//read		//UInt32	//Total number of RTP clients since startup
	qtssRTPSvrCurBandwidth			= 15,	//read		//UInt32	//Current bandwidth being output by the RTP server in bits per second
	qtssRTPSvrTotalBytes 			= 16,	//read		//UInt64	//Total number of bytes served since startup
	qtssRTPSvrAvgBandwidth 			= 17,	//read		//UInt32	//Average bandwidth being output by the RTP server in bits per second
	qtssRTPSvrCurPackets 			= 18,	//read		//UInt32	//Current packets per second being output by the RTP server
	qtssRTPSvrTotalPackets			= 19,	//read		//UInt64	//Total number of bytes served since startup
	
	qtssSvrNumParams				= 20
};
typedef UInt32 QTSS_ServerAttributes;

enum
{
	//QTSS_PrefsObject parameters

	// Valid in all methods. None of these are pre-emptive safe, so the version
	// of QTSS_GetAttribute that copies data must be used.
	
	// All of these parameters are read-only. 
	
	qtssPrefsRTSPTimeout 			= 0,	//"rtsp_timeout"				//UInt32	//RTSP timeout in seconds sent to the client.
	qtssPrefsRealRTSPTimeout 		= 1,	//"real_rtsp_timeout"			//UInt32	//Amount of time in seconds the server will wait before disconnecting idle RTSP clients. 0 means no timeout
	qtssPrefsRTPTimeout 			= 2,	//"rtp_timeout"					//UInt32	//Amount of time in seconds the server will wait before disconnecting idle RTP clients. 0 means no timeout
	qtssPrefsMaximumConnections 	= 3,	//"maximum_connections"			//SInt32	//Maximum # of concurrent RTP connections allowed by the server. -1 means unlimited.
	qtssPrefsMaximumBandwidth 		= 4,	//"maximum_bandwidth"			//SInt32	//Maximum amt of bandwidth the server is allowed to serve in K bits. -1 means unlimited.
	qtssPrefsMovieFolder 			= 5,	//"movie_folder"				//char array	//Path to the root movie folder
	qtssPrefsRTSPIPAddr 			= 6,	//"bind_ip_addr"				//char array	//IP address the server should accept RTSP connections on. 0.0.0.0 means all addresses on the machine.
	qtssPrefsBreakOnAssert			= 7,	//"break_on_assert"				//Bool16		//If true, the server will break in the debugger when an assert fails.
	qtssPrefsAutoRestart			= 8,	//"auto_restart"				//Bool16		//If true, the server will automatically restart itself if it crashes.
	qtssPrefsTotalBytesUpdate 		= 9,	//"total_bytes_update"			//UInt32	//Interval in seconds between updates of the server's total bytes and current bandwidth statistics
	qtssPrefsAvgBandwidthUpdate 	= 10,	//"average_bandwidth_update"	//UInt32	//Interval in seconds between computations of the server's average bandwidth
	qtssPrefsSafePlayDuration 		= 11,	//"safe_play_duration"			//UInt32	//Hard to explain... see streamingserver.conf
	qtssPrefsModuleFolder 			= 12,	//"module_folder"				//char array	//Path to the module folder

	// There is a compiled-in error log module that loads before all the other modules
	// (so it can log errors from the get-go). It uses these prefs.
	
	qtssPrefsErrorLogName 			= 13,	//"error_logfile_name"			//char array		//Name of error log file
	qtssPrefsErrorLogDir 			= 14,	//"error_logfile_dir"			//char array		//Path to error log file directory
	qtssPrefsErrorRollInterval 		= 15,	//"error_logfile_interval"		//UInt32	//Interval in days between error logfile rolls
	qtssPrefsMaxErrorLogSize 		= 16,	//"error_logfile_size"			//UInt32	//Max size in bytes of the error log
	qtssPrefsErrorLogVerbosity 		= 17,	//"error_logfile_verbosity"		//UInt32	//Max verbosity level of messages the error logger will log
	qtssPrefsScreenLogging 			= 18,	//"screen_logging"				//Bool16		//Should the error logger echo messages to the screen?
	qtssPrefsErrorLogEnabled 		= 19,	//"error_logging"				//Bool16		//Is error logging enabled?

	qtssPrefsTCPMinThinDelayToleranceInMSec = 20, 	//"tcp_min_thin_delay_tolerance"	//SInt32	// max ms delay for min (bframe) thinning bandwidth
	qtssPrefsTCPMaxThinDelayToleranceInMSec = 21, 	//"tcp_max_thin_delay_tolerance"	//SInt32	// max ms delay for max (keyframe) thinning bandwidth
	qtssPrefsTCPVideoDelayToleranceInMSec 	= 22, 	//"tcp_max_video_delay_tolerance"	//SInt32	// max ms delay before dropping video 
	qtssPrefsTCPAudioDelayToleranceInMSec 	= 23, 	//"tcp_max_audio_delay_tolerance"	//SInt32	// max ms delay before dropping audio 
	
	qtssPrefsMinTCPBufferSizeInBytes		= 24,	//"min_tcp_buffer_size"	//UInt32	// When streaming over TCP, this is the minimum size the TCP socket send buffer can be set to
	qtssPrefsMaxTCPBufferSizeInBytes		= 25,	//"max_tcp_buffer_size"	//UInt32	// When streaming over TCP, this is the maximum size the TCP socket send buffer can be set to
	qtssPrefsTCPSecondsToBuffer				= 26,	//"tcp_seconds_to_buffer" //Float32	// When streaming over TCP, the size of the TCP send buffer is scaled based on the bitrate of the movie. It will fit all the data that gets sent in this amount of time.
	
	qtssPrefsDoReportHTTPConnectionAddress 	= 27,	//"do_report_http_connection_ip_address"	//Bool16	// when behind a round robin DNS, the client needs to be told the specific ip address of the maching handling its request. this pref tells the server to repot its IP address in the reply to the HTTP GET request when tunneling RTSP through HTTP

	qtssPrefsDefaultAuthorizationRealm 		= 28, 	// "default_authorization_realm" //char array	//
	
	qtssPrefsPidFileName 					= 29,	//"pid_file_name"		//char array		//Name of process ID (pid) file
	qtssPrefsRunUserName 					= 30,	//"set_user_name"		//char array		//Run under this user's account
	qtssPrefsRunGroupName					= 31,	//"set_group_name"		//char array		//Run under this group's account
	
	qtssPrefsNumParams 						= 32
};

typedef UInt32 QTSS_PrefsAttributes;

enum
{
	//QTSS_TextMessagesObject parameters
	
	// All of these parameters are read-only, char*'s, and preemptive-safe.
	
	qtssMsgNoMessage				= 0,	//"NoMessage"
	qtssMsgNoURLInRequest			= 1,
	qtssMsgBadRTSPMethod			= 2,
	qtssMsgNoRTSPVersion			= 3,
	qtssMsgNoRTSPInURL				= 4,
	qtssMsgURLTooLong				= 5,
	qtssMsgURLInBadFormat			= 6,
	qtssMsgNoColonAfterHeader		= 7,
	qtssMsgNoEOLAfterHeader			= 8,
	qtssMsgRequestTooLong			= 9,
	qtssMsgNoModuleFolder			= 10,
	qtssMsgCouldntListen			= 11,
	qtssMsgInitFailed				= 12,
	qtssMsgNotConfiguredForIP		= 13,
	qtssMsgDefaultRTSPAddrUnavail	= 14,
	qtssMsgBadModule				= 15,
	qtssMsgRegFailed				= 16,
	qtssMsgRefusingConnections		= 17,
	qtssMsgTooManyClients			= 18,
	qtssMsgTooMuchThruput			= 19,
	qtssMsgNoSessionID				= 20,
	qtssMsgFileNameTooLong			= 21,
	qtssMsgNoClientPortInTransport	= 22,
	qtssMsgRTPPortMustBeEven		= 23,
	qtssMsgRTCPPortMustBeOneBigger	= 24,
	qtssMsgOutOfPorts				= 25,
	qtssMsgNoModuleForRequest		= 26,
	qtssMsgAltDestNotAllowed		= 27,
	qtssMsgCantSetupMulticast		= 28,
	qtssListenPortInUse				= 29,
	qtssListenPortAccessDenied		= 30,
	qtssListenPortError				= 31,
	qtssMsgBadBase64				= 32,
	qtssMsgSomePortsFailed			= 33,
	qtssMsgNoPortsSucceeded			= 34,
	qtssMsgCannotCreatePidFile		= 35,
	qtssMsgCannotSetRunUser 		= 36,
	qtssMsgCannotSetRunGroup		= 37,
	
	qtssMsgNumParams 				= 38
	
};
typedef UInt32 QTSS_TextMessagesAttributes;


/********************************************************************/
// QTSS API ROLES
//
// Each role represents a unique situation in which a module may be
// invoked. Modules must specify which roles they want to be invoked for. 

enum
{
	//Global
	QTSS_Register_Role =			 FOUR_CHARS_TO_INT('r', 'e', 'g', ' '), //reg  //All modules get this once at startup
	QTSS_Initialize_Role = 			 FOUR_CHARS_TO_INT('i', 'n', 'i', 't'),	//init //Gets called once, later on in the startup process
	QTSS_Shutdown_Role = 			 FOUR_CHARS_TO_INT('s', 'h', 'u', 't'),	//shut //Gets called once at shutdown
	
	QTSS_ErrorLog_Role =			 FOUR_CHARS_TO_INT('e', 'l', 'o', 'g'), //elog //This gets called when the server wants to log an error.
	QTSS_RereadPrefs_Role =			 FOUR_CHARS_TO_INT('p', 'r', 'e', 'f'), //pref //This gets called when the server rereads preferences.
	
	//RTSP-specific
	QTSS_RTSPFilter_Role = 			 FOUR_CHARS_TO_INT('f', 'i', 'l', 't'),	//filt //Filter all RTSP requests before the server parses them
	QTSS_RTSPRoute_Role =			 FOUR_CHARS_TO_INT('r', 'o', 'u', 't'), //rout //Route all RTSP requests to the correct root folder.
	QTSS_RTSPAuthorize_Role =		 FOUR_CHARS_TO_INT('a', 'u', 't', 'h'), //auth //Authorize RTSP requests to proceed
	QTSS_RTSPPreProcessor_Role = 	 FOUR_CHARS_TO_INT('p', 'r', 'e', 'p'),	//prep //Pre-process all RTSP requests before the server responds.
										//Modules may opt to "steal" the request and return a client response.
	QTSS_RTSPRequest_Role = 		 FOUR_CHARS_TO_INT('r', 'e', 'q', 'u'),	//requ //Process an RTSP request & send client response
	QTSS_RTSPPostProcessor_Role = 	 FOUR_CHARS_TO_INT('p', 'o', 's', 't'),	//post //Post-process all RTSP requests
	QTSS_RTSPSessionClosing_Role =	 FOUR_CHARS_TO_INT('s', 'e', 's', 'c'),	//sesc //RTSP session is going away

	//RTP-specific
	QTSS_RTPSendPackets_Role =			 FOUR_CHARS_TO_INT('s', 'e', 'n', 'd'),	//send //Send RTP packets to the client
	QTSS_ClientSessionClosing_Role =	 FOUR_CHARS_TO_INT('d', 'e', 's', 's'),	//dess //Client session is going away
	
	//RTCP-specific
	QTSS_RTCPProcess_Role =			 FOUR_CHARS_TO_INT('r', 't', 'c', 'p')	//rtcp //Process all RTCP packets sent to the server
};
typedef UInt32 QTSS_Role;


//***********************************************/
// TYPEDEFS

typedef void* 			QTSS_StreamRef;
typedef void*			QTSS_EventContextRef;
typedef void*		 	QTSS_Object;
typedef void* 			QTSS_ServiceFunctionArgsPtr;
typedef SInt32 			QTSS_AttributeID;
typedef SInt32 			QTSS_ServiceID;

typedef QTSS_Object 			QTSS_RTPStreamObject;
typedef QTSS_Object 			QTSS_RTSPSessionObject;
typedef QTSS_Object 			QTSS_RTSPRequestObject;
typedef QTSS_Object 			QTSS_RTSPHeaderObject;
typedef QTSS_Object 			QTSS_ClientSessionObject;
typedef QTSS_Object 			QTSS_ServerObject;
typedef QTSS_Object 			QTSS_PrefsObject;
typedef QTSS_Object 			QTSS_TextMessagesObject;

//***********************************************/
// ROLE PARAMETER BLOCKS
//
// Each role has a unique set of parameters that get passed
// to the module.

typedef struct
{
	char outModuleName[QTSS_MAX_MODULE_NAME_LENGTH];
} QTSS_Register_Params;

typedef struct
{
	QTSS_ServerObject 			inServer;			// Global dictionaries
	QTSS_PrefsObject			inPrefs;
	QTSS_TextMessagesObject		inMessages;
	QTSS_StreamRef				inErrorLogStream;	// Writing to this stream causes modules to
													// be invoked in the QTSS_ErrorLog_Role
} QTSS_Initialize_Params;

typedef struct
{
	QTSS_ErrorVerbosity 		inVerbosity;
	char*						inBuffer;
	
} QTSS_ErrorLog_Params;

typedef struct 
{
	QTSS_RTSPSessionObject		inRTSPSession;
	QTSS_RTSPRequestObject		inRTSPRequest;
	QTSS_RTSPHeaderObject		inRTSPHeaders;
	QTSS_ClientSessionObject	inClientSession;

} QTSS_StandardRTSP_Params;

typedef struct 
{
	QTSS_RTSPSessionObject		inRTSPSession;
	QTSS_RTSPRequestObject		inRTSPRequest;
	char**						outNewRequest;

} QTSS_Filter_Params;

typedef struct
{
	QTSS_RTSPSessionObject		inRTSPSession;
} QTSS_RTSPSession_Params;

typedef struct
{
	QTSS_ClientSessionObject		inClientSession;
	SInt64							inCurrentTime;
	SInt64							outNextPacketTime;
} QTSS_RTPSendPackets_Params;

typedef struct
{
	QTSS_ClientSessionObject		inClientSession;
	QTSS_CliSesClosingReason		inReason;
} QTSS_ClientSessionClosing_Params;

typedef struct
{
	QTSS_ClientSessionObject	inClientSession;
	QTSS_RTPStreamObject		inRTPStream;
	void*						inRTCPPacketData;
	UInt32						inRTCPPacketDataLen;
} QTSS_RTCPProcess_Params;

typedef union
{
	QTSS_Register_Params				regParams;
	QTSS_Initialize_Params 				initParams;
	QTSS_ErrorLog_Params				errorParams;

	QTSS_Filter_Params 					rtspFilterParams;
	QTSS_StandardRTSP_Params 			rtspRouteParams;
	QTSS_StandardRTSP_Params 			rtspAuthParams;
	QTSS_StandardRTSP_Params 			rtspPreProcessorParams;
	QTSS_StandardRTSP_Params 			rtspRequestParams;
	QTSS_StandardRTSP_Params 			rtspPostProcessorParams;
	QTSS_RTSPSession_Params 			rtspSessionClosingParams;

	QTSS_RTPSendPackets_Params 			rtpSendPacketsParams;
	QTSS_ClientSessionClosing_Params 	clientSessionClosingParams;
	QTSS_RTCPProcess_Params 			rtcpProcessParams;
	
} QTSS_RoleParams, *QTSS_RoleParamPtr;

/********************************************************************/
// ENTRYPOINTS & FUNCTION TYPEDEFS

// MAIN ENTRYPOINT FOR MODULES
//
// Every QTSS API must implement two functions: a main entrypoint, and a dispatch
// function. The main entrypoint gets called by the server at startup to do some
// initialization. Your main entrypoint must follow the convention established below
//
// QTSS_Error mymodule_main(void* inPrivateArgs)
// {
//		return _stublibrary_main(inPrivateArgs, MyDispatchFunction);
// }
//
//

typedef QTSS_Error (*QTSS_MainEntryPointPtr)(void* inPrivateArgs);
typedef QTSS_Error (*QTSS_DispatchFuncPtr)(QTSS_Role inRole, QTSS_RoleParamPtr inParamBlock);

// STUB LIBRARY MAIN
QTSS_Error _stublibrary_main(void* inPrivateArgs, QTSS_DispatchFuncPtr inDispatchFunc);



/********************************************************************/
// QTSS API CALLBACK FUNCTIONS


/********************************************************************/
//  QTSS_New
//	QTSS_Delete
//
//	These should be used for all dynamic memory allocation done from
//	within modules. The memoryIdentifier is used for debugging:
//	the server can track this memory to make memory leak debugging easier.
void* 	QTSS_New(FourCharCode inMemoryIdentifier, UInt32 inSize);
void	QTSS_Delete(void* inMemory);

/********************************************************************/
//  QTSS_Milliseconds
//
//	The server maintains a millisecond timer internally. The current
//	value of that timer can be obtained from this function. This value
//	is not valid between server executions.
//
//  All millisecond values used in QTSS API use this timer, unless otherwise noted
SInt64	QTSS_Milliseconds();


/********************************************************************/
//  QTSS_MilliSecsTo1970Secs
//
//	Convert milliseconds from the QTSS_Milliseconds call to 
//	second's since 1970
//
time_t	QTSS_MilliSecsTo1970Secs(SInt64 inQTSS_MilliSeconds);

/********************************************************************/
//  QTSS_AddRole
//
//  Only available from QTSS_Initialize role. Call this for all the roles you
//  would like your module to operate on.
//
//  Returns: 	QTSS_NoErr
//				QTSS_OutOfState: If this function isn't being called from the Register role
//				QTSS_RequestFailed: 	If module is registering for the QTSS_RTSPRequest_Role
//										and there already is such a module.
//				QTSS_BadArgument:	Registering for a nonexistent role.
QTSS_Error QTSS_AddRole(QTSS_Role inRole);


/*****************************************/
//  ATTRIBUTE / OBJECT CALLBACKS
//


/********************************************************************/
//  QTSS_AttrFunctionPtr
//
//	Modules have the option of registering an attribute in any predefined object type, and having
//	the server call a module-implemented function whenever someone calls QTSS_GetAttribute
//	on that attribute. Any attribute with a param retrieval function
//  must NOT be indexed. The function must conform to the following typedef.
//
//	On input, the QTSS_Object for the specified parameter will be passed in.
//
//	The function has 2 options for returning the attribute value to the caller:
//
//	1) 	Return the attribute body as the return value, and the attribute length in the
//		second parameter. This will return the attribute to the caller, but not update
//		the object, so each subsequent time that paramter is requested, the function
//		will be invoked again.
//
//	2)	Use QTSS_SetAttribute to set the attribute in the object. Once the attribute
//		is set in the object, your function will not be called again for this object.
typedef void* (*QTSS_AttrFunctionPtr)(QTSS_Object, UInt32* );




/********************************************************************/
//  QTSS_AddAttribute
//
//  Adds a new attribute to a predefined object type (note that this is
//  different from the object itself, a object type refers
//  to all objects of a certain type).
//
//  This may only be called from the QTSS_Register role.
//
//  Returns: 	QTSS_NoErr
//				QTSS_OutOfState: If this function isn't being called from the Register role
//				QTSS_BadArgument:	Adding an attribute to a nonexistent object type, attribute
//						name too long, or NULL arguments.
QTSS_Error QTSS_AddAttribute(QTSS_ObjectType inObjectType, const char* inAttributeName, QTSS_AttrFunctionPtr inFunctionPtr);


/********************************************************************/
//  QTSS_IDForAttr
//
//  Given an attribute name, this returns its accompanying attribute ID.
//  The ID can in turn be used to retrieve the attribute value from
//  a object.
//
//  Returns:	QTSS_NoErr
//				QTSS_BadArgument: Bad argument
QTSS_Error QTSS_IDForAttr(QTSS_ObjectType inObjectType, const char* inAttributeName, QTSS_AttributeID* outID);


/********************************************************************/
//  QTSS_GetValuePtr
//
//  NOT TO BE USED WITH NON-PREEMPTIVE-SAFE attributes
//
//  Returns:	QTSS_NoErr
//				QTSS_BadArgument: Bad argument
//				QTSS_NotPreemptiveSafe: Attempt to get a non-preemptive safe attribute
//				QTSS_BadIndex: Attempt to get non-existent index.
//
QTSS_Error QTSS_GetValuePtr (QTSS_Object inObject, QTSS_AttributeID inID, UInt32 inIndex, void** outBuffer, UInt32* outLen);

/********************************************************************/
//  QTSS_GetValue
//
//  Copies the data into provided buffer. If E2BIG is returned, outLen is still set.
//
//  Returns:	QTSS_NoErr
//				QTSS_BadArgument: Bad argument
//				QTSS_NotEnoughSpace: Value is too big for buffer provided.
//				QTSS_BadIndex: Attempt to get non-existent index.
//
QTSS_Error QTSS_GetValue (QTSS_Object inObject, QTSS_AttributeID inID, UInt32 inIndex, void* ioBuffer, UInt32* ioLen);

/********************************************************************/
//  QTSS_SetValue
//
//  Returns:	QTSS_NoErr
//				QTSS_BadArgument: Bad argument
//				QTSS_ReadOnly: Attribute is read only.
//				QTSS_BadIndex: Attempt to set non-0 index of attribute with a param retrieval function.
//
QTSS_Error QTSS_SetValue (QTSS_Object inObject, QTSS_AttributeID inID, UInt32 inIndex, const void* inBuffer,  UInt32 inLen);


/*****************************************/
//  STREAM CALLBACKS
//
//  The QTSS API provides QTSS_StreamRefs as a generalized stream abstraction. Mostly,
//  QTSS_StreamRefs are used for communicating with the client. For instance,
//  in the QTSS_RTSPRequest_Role, modules receive a QTSS_StreamRef which can be
//  used for reading RTSP data from the client, and sending RTSP response data to the client.
//
//  Additionally, QTSS_StreamRefs are generalized enough to be used in many other situations.
//  For instance, modules receive a QTSS_StreamRef for the error log. When modules want
//  to report errors, they can use these same routines, passing in the error log StreamRef.

/********************************************************************/
//  QTSS_Write
//
//	Writes data to a stream.
//
//  Returns: 	QTSS_NoErr
//				QTSS_WouldBlock: The stream cannot accept any data at this time.
//				QTSS_NotConnected: The stream receiver is no longer connected.
//				QTSS_BadArgument:	NULL argument.
QTSS_Error	QTSS_Write(QTSS_StreamRef inRef, const void* inBuffer, UInt32 inLen, UInt32* outLenWritten, QTSS_WriteFlags inFlags);

/********************************************************************/
//  QTSS_WriteV
//
//	Works similar to the POSIX WriteV, and takes a POSIX iovec.
//  THE FIRST ENTRY OF THE IOVEC MUST BE BLANK!!!
//
//  Returns: 	QTSS_NoErr
//				QTSS_WouldBlock: The stream cannot accept any data at this time.
//				QTSS_NotConnected: The stream receiver is no longer connected.
//				QTSS_BadArgument:	NULL argument.
QTSS_Error	QTSS_WriteV(QTSS_StreamRef inRef, iovec* inVec, UInt32 inNumVectors, UInt32 inTotalLength, UInt32* outLenWritten);

/********************************************************************/
//  QTSS_Flush
//
//  Some QTSS_StreamRefs (QTSS_RequestRef, for example) buffers data before sending it
//  out. Calling this forces the stream to write the data immediately.
//
//  Returns: 	QTSS_NoErr
//				QTSS_WouldBlock: Stream cannot be completely flushed at this time.
//				QTSS_NotConnected: The stream receiver is no longer connected.
//				QTSS_BadArgument:	NULL argument.
QTSS_Error	QTSS_Flush(QTSS_StreamRef inRef);

/*****************************************/
//  SERVICES
//
//  Oftentimes modules have functionality that they want accessable from other
//  modules. An example of this might be a logging module that allows other
//  modules to write messages to the log.
//
//  Modules can use the following callbacks to register and invoke "services".
//  Adding & finding services works much like adding & finding attributes in
//  an object. A service has a name. In order to invoke a service, the calling
//  module must know the name of the service and resolve that name into an ID.
//
//  Each service has a parameter block format that is specific to that service.
//  Modules that are exporting services should carefully document the services they
//  export, and modules calling services should take care to fail gracefully
//  if the service isn't present or returns an error.

typedef QTSS_Error (*QTSS_ServiceFunctionPtr)(QTSS_ServiceFunctionArgsPtr);

/********************************************************************/
//  QTSS_AddService
//
//  This function registers a service with the specified name, and
//  associates it with the specified function pointer.
//  QTSS_AddService may only be called from the QTSS_Register role
//
//  Returns: 	QTSS_NoErr
//				QTSS_OutOfState: If this function isn't being called from the Register role
//				QTSS_BadArgument:	Service name too long, or NULL arguments.
QTSS_Error QTSS_AddService(const char* inServiceName, QTSS_ServiceFunctionPtr inFunctionPtr);


/********************************************************************/
//  QTSS_IDForService
//
//  Much like QTSS_IDForAttr, this resolves a service name into its
//  corresponding QTSS_ServiceID. The QTSS_ServiceID can then be used to
//  invoke the service.
//
//  Returns:	QTSS_NoErr
//				QTSS_BadArgument: Bad argument
QTSS_Error QTSS_IDForService(const char* inTag, QTSS_ServiceID* outID);

/********************************************************************/
//  QTSS_DoService
//
//  Invokes the service. Return value from this function comes from the service
//  function itself, unless the QTSS_IllegalService errorcode is returned,
//  which is returned when the QTSS_ServiceID is bad.
QTSS_Error QTSS_DoService(QTSS_ServiceID inID, QTSS_ServiceFunctionArgsPtr inArgs);

/********************************************************************/
//  BUILT-IN SERVICES
//
//  The server registers some built-in services when it starts up.
//  Here are macros for their names & descriptions of what they do

// Rereads the preferences, also causes the QTSS_RereadPrefs_Role to be invoked
#define QTSS_REREAD_PREFS_SERVICE	"RereadPreferences"



/*****************************************/
//  RTSP HEADER CALLBACKS
//
//  As a convience to modules that want to send RTSP responses, the server
//  has internal utilities for formatting a proper RTSP response. When a module
//  calls QTSS_SendRTSPHeaders, the server sends a proper RTSP status line, using
//  the request's current status code, and also sends the proper CSeq header,
//  session ID header, and connection header.
//
//  Any other headers can be appended by calling QTSS_AppendRTSPHeader. They will be
//  sent along with everything else when QTSS_SendRTSPHeaders is called.
//
//  Returns:	QTSS_NoErr
//				QTSS_BadArgument: Bad argument
QTSS_Error QTSS_SendRTSPHeaders(QTSS_RTSPRequestObject inRef);
//
//  Returns:	QTSS_NoErr
//				QTSS_BadArgument: Bad argument
QTSS_Error QTSS_AppendRTSPHeader(QTSS_RTSPRequestObject inRef, QTSS_RTSPHeader inHeader, const char* inValue, UInt32 inValueLen);


/*****************************************/
//  QTSS_SendStandardRTSPResponse
//
//  This function is also provided as an optional convienence to modules who are sending
//  "typical" RTSP responses to clients. The function uses the QTSS_RTSPRequestObject and
//  the QTSS_Object as inputs, where the object may either be a QTSS_ClientSessionObject
//  or a QTSS_RTPStreamObject, depending on the method. The response is written to the
//  stream provided.
//
//  Below is a description of what is returned for each method this function supports:
//
//  DESCRIBE:
//
//   Writes status line, CSeq, SessionID, Connection headers as determined by the request.
//   Writes a Content-Base header with the Content-Base being the URL provided.
//   Writes a Content-Type header of "application/sdp"
//   QTSS_Object must be a QTSS_ClientSessionObject.
//
//  SETUP:
//
//   Writes status line, CSeq, SessionID, Connection headers as determined by the request.
//	 Writes a Transport header with the client & server ports (if connection is over UDP).
//   QTSS_Object must be a QTSS_RTPStreamObject.
//
//  PLAY:
//
//   Writes status line, CSeq, SessionID, Connection headers as determined by the request.
//   QTSS_Object must be a QTSS_ClientSessionObject.
//
//   Specify whether you want the server to append the seq#, timestamp, & ssrc info to
//   the RTP-Info header via. the qtssPlayRespWriteTrackInfo flag.
//
//  PAUSE:
//
//   Writes status line, CSeq, SessionID, Connection headers as determined by the request.
//   QTSS_Object must be a QTSS_ClientSessionObject.
//
//  TEARDOWN:
//
//   Writes status line, CSeq, SessionID, Connection headers as determined by the request.
//   QTSS_Object must be a QTSS_ClientSessionObject.
//
//  Returns:	QTSS_NoErr
//				QTSS_BadArgument: Bad argument
QTSS_Error	QTSS_SendStandardRTSPResponse(QTSS_RTSPRequestObject inRTSPRequest, QTSS_Object inRTPInfo, UInt32 inFlags);


/*****************************************/
//  CLIENT SESSION CALLBACKS
//
//  QTSS API Modules have the option of generating and sending RTP packets. Only
//  one module currently can generate packets for a particular session. In order
//  to do this, call QTSS_AddRTPStream. This must be done in response to a RTSP
//  request, and typically is done in response to a SETUP request from the client.
//
//  After one or more streams have been added to the session, the module that "owns"
//  the packet sending for that session can call QTSS_Play to start the streams playing.
//  After calling QTSS_Play, the module will get invoked in the QTSS_SendPackets_Role.
//  Calling QTSS_Pause stops playing.
//
//  The "owning" module may call QTSS_Teardown at any time. Doing this closes the
//  session and will cause the QTSS_SessionClosing_Role to be invoked for this session. 
//
//  Returns:	QTSS_NoErr
//				QTSS_BadArgument: Bad argument
//				QTSS_RequestFailed: QTSS_RTPStreamObject couldn't be created.
QTSS_Error	QTSS_AddRTPStream(QTSS_ClientSessionObject inClientSession, QTSS_RTSPRequestObject inRTSPRequest, QTSS_RTPStreamObject* outStream, QTSS_AddStreamFlags inFlags);
//
//  Returns:	QTSS_NoErr
//				QTSS_BadArgument: Bad argument
//				QTSS_RequestFailed: No streams added to this session.
QTSS_Error	QTSS_Play(QTSS_ClientSessionObject inClientSession, QTSS_RTSPRequestObject inRTSPRequest, QTSS_PlayFlags inPlayFlags);
//
//  Returns:	QTSS_NoErr
//				QTSS_BadArgument: Bad argument
QTSS_Error	QTSS_Pause(QTSS_ClientSessionObject inClientSession);
//
//  Returns:	QTSS_NoErr
//				QTSS_BadArgument: Bad argument
QTSS_Error	QTSS_Teardown(QTSS_ClientSessionObject inClientSession);


/*****************************************/
//  ASYNC I/O CALLBACKS
//
//  QTSS modules must be kind in how they use the CPU. The server doesn't
//  prevent a poorly implemented QTSS module from hogging the processing
//  capability of the server, at the expense of other modules and other clients.
//
//  It is therefore imperitive that a module use non-blocking, or async, I/O.
//  If a module were to block, say, waiting to read file data off disk, this stall
//  would affect the entire server.
//
//  This problem is resolved in QTSS API in a number of ways.
//
//  Firstly, all QTSS_StreamRefs provided to modules are non-blocking, or async.
//  Modules should be prepared to receive EWOULDBLOCK errors in response to
//  QTSS_Read, QTSS_Write, & QTSS_WriteV calls, with certain noted exceptions
//  in the case of responding to RTSP requests.
//
//  Modules that open their own file descriptors for network or file I/O can
//  create separate threads for handling I/O. In this case, these descriptors
//  can remain blocking, as long as they always block on the private module threads.
//
//  In most cases, however, creating a separate thread for I/O is not viable for the
//  kind of work the module would like to do. For instance, a module may wish
//  to respond to a RTSP DESCRIBE request, but can't immediately because constructing
//  the response would require I/O that would block.
//
//  The problem is once the module returns from the QTSS_RTSPProcess_Role, the
//  server will mistakenly consider the request handled, and move on. It won't
//  know that the module has more work to do before it finishes processing the DESCRIBE.
//
//  In this case, the module needs to tell the server to delay processing of the
//  DESCRIBE request until the file descriptor's blocking condition is lifted.
//  The module can do this by using the provided "event" callback routines.
//
//  Each file descriptor may have an "event context" associated with it. If a module
//  would like to receive events on file descriptor, it must create an event context
//  for that descriptor. Certain dictionaries (QTSS_RTSPRequestRefs & QTSS_RTPStreamRefs),
//  have pre-constructed event contexts available as attributes.
//
//  Once a module has constructed the event context, it should call QTSS_RegisterForEvent,
//  pass in the context and the events it would like to receive.
//
//  Once it has done this, the module should return control back to the server. The
//  server will re-invoke the module in the same role with the same dictionaries
//  in exactly the same state once the event conditions have been satisfied.

//  Always pass in NULL for the inPrivate argument
//
//  Returns:	QTSS_NoErr
//				QTSS_BadArgument: Bad argument
QTSS_Error	QTSS_CreateEventContext(int inFileDesc, QTSS_EventContextRef* outContext, void* inPrivate); 
//
//  Returns:	QTSS_NoErr
//				QTSS_BadArgument: Bad argument
QTSS_Error	QTSS_DestroyEventContext(QTSS_EventContextRef inContext);

//
//  Returns:	QTSS_NoErr
//				QTSS_BadArgument: Bad argument
//				QTSS_OutOfState: if this callback is made from a role that doesn't allow async I/O events
//				QTSS_RequestFailed: Not currently possible to request an event. 
QTSS_Error  QTSS_RequestEvent(QTSS_EventContextRef inContext, int inEventMask);
QTSS_Error  QTSS_SetIdleTimer(SInt64 inIdleMsec);

#ifdef __cplusplus
}
#endif

#endif
