/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		EventContext.cpp

	Contains:	Impelments object in .h file
					
	$Log: EventContext.cpp,v $
	Revision 1.2  2000/01/28 08:32:40  serenyi
	Made it so #DEBUG doesn't turn on printfs
	
	Revision 1.1.1.1  2000/01/17 23:35:57  wsanchez
	Import of StreamingServer-2.0 (Apple)
	
	Revision 1.8  1999/12/23 18:45:07  rtucker
	mods to support platforms with non-preemptive threading packages
	
	Revision 1.7  1999/12/21 01:43:30  lecroy
	*** empty log message ***
	
	Revision 1.6  1999/12/13 22:12:30  serenyi
	Part of fixing 2415738. In RequestEvent, fWatchEventCalled *must* be set before calling watchevent.
	
	Revision 1.5  1999/11/02 04:01:03  serenyi
	Cosmetic change
	
	Revision 1.4  1999/10/27 17:53:45  serenyi
	Integrated Matthias's MacOS / GUSI changes
	
	Revision 1.3  1999/10/23 03:19:33  murata
	Changes to allow debugging codewarrior compiles on MacOS for PlaylistBroadcaster
	
	Revision 1.2  1999/09/30 00:24:40  serenyi
	Snarf bug fixes?
	
	Revision 1.1.1.1  1999/09/22 23:01:16  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:37  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:13:18  ds
	Created
	
	
*/

#include "EventContext.h"
#include "OSThread.h"
#include "atomic.h"

#include <fcntl.h>
#include <unistd.h>
#include <sys/errno.h>
#if MACOSXEVENTQUEUE
#include "tempcalls.h" //includes MacOS X prototypes of event queue functions
#endif

#define EVENT_CONTEXT_DEBUG 0

#if EVENT_CONTEXT_DEBUG
#include "OS.h"
#endif

unsigned int EventContext::sUniqueID = 1;


EventContext::EventContext(EventThread* inThread)
:	fFileDesc(kInvalidFileDesc),
 	fUniqueID(0),
	fUniqueIDStr((char*)&fUniqueID, sizeof(fUniqueID)),
	fEventThread(inThread),
	fWatchEventCalled(false),
	fEventBits(0)
{}

void EventContext::Cleanup()
{
	int err = 0;
	
	if (fFileDesc != kInvalidFileDesc)
	{
		//if this object is registered in the table, unregister it now
		if (fUniqueID > 0)
		{
			fEventThread->fRefTable.UnRegister(&fRef);

#if !MACOSXEVENTQUEUE
			select_removeevent(fFileDesc);//The eventqueue / select shim requires this
#else
			//On Linux (possibly other UNIX implementations) you MUST NOT close the fd before
			//removing the fd from the select mask, and having the select function wake up
			//to register this fact. If you close the fd first, bad things may happen, like
			//the socket not getting unbound from the port & IP addr.
			//
			//So, what we do is have the select thread itself call close. This is triggered
			//by calling removeevent.
			err = ::close(fFileDesc);
#endif		
		}
		else
			err = ::close(fFileDesc);
	}

	fFileDesc = kInvalidFileDesc;
	fUniqueID = 0;
	
	AssertV(err == 0, OSThread::GetErrno());//we don't really care if there was an error, but it's nice to know
}


void EventContext::SnarfEventContext( EventContext &fromContext )
{  
	//+ show that we called watchevent
	// copy the unique id
	// set our fUniqueIDStr to the unique id
	// copy the eventreq
	// find the old event object
	// show us as the object in the fRefTable
	//		we take the OSRef from the old context, point it at our context
	//
	//TODO - this whole operation causes a race condition for Event posting
	//  way up the chain we need to disable event posting
	// or copy the posted events afer this op completes
	
	fromContext.fFileDesc = kInvalidFileDesc;
	
	fWatchEventCalled = fromContext.fWatchEventCalled; 
	fUniqueID = fromContext.fUniqueID;
	fUniqueIDStr.Set((char*)&fUniqueID, sizeof(fUniqueID)),
	
	::memcpy( &fEventReq, &fromContext.fEventReq, sizeof( struct eventreq  ) );

	fRef.Set( fUniqueIDStr, this );
	fEventThread->fRefTable.Swap(&fRef);
	fEventThread->fRefTable.UnRegister(&fromContext.fRef);
}

void EventContext::InitNonBlocking(int inFileDesc, void* /*inPrivate*/)
{
	// inPrivate argument to be used by the Akamai file system
	
	//
	// Put the file descriptor in non-blocking mode
	fFileDesc = inFileDesc;

	int flag = ::fcntl(fFileDesc, F_GETFL, 0);
	int err = ::fcntl(fFileDesc, F_SETFL, flag | O_NONBLOCK);
	AssertV(err == 0, OSThread::GetErrno());
}

void EventContext::RequestEvent(int theMask)
{
#if DEBUG
	fModwatched = true;
#endif

	//
	// The first time this function gets called, we're supposed to
	// call watchevent. Each subsequent time, call modwatch. That's
	// the way the MacOS X event queue works.
	
	if (fWatchEventCalled)
	{
		fEventReq.er_eventbits = theMask;
#if MACOSXEVENTQUEUE
		if (modwatch(&fEventReq, theMask) != 0)
#else
		if (select_modwatch(&fEventReq, theMask) != 0)
#endif	
			AssertV(false, OSThread::GetErrno());
	}
	else
	{
		//allocate a Unique ID for this socket, and add it to the ref table
		fUniqueID = atomic_add(&sUniqueID, 1);

		fRef.Set(fUniqueIDStr, this);
		fEventThread->fRefTable.Register(&fRef);
			
		//fill out the eventreq data structure
		::memset( &fEventReq, '\0', sizeof(fEventReq));
		fEventReq.er_type = EV_FD;
		fEventReq.er_handle = fFileDesc;
		fEventReq.er_eventbits = theMask;
		fEventReq.er_data = (void*)fUniqueID;

		fWatchEventCalled = true;
#if MACOSXEVENTQUEUE
		if (watchevent(&fEventReq, theMask) != 0)
#else
		if (select_watchevent(&fEventReq, theMask) != 0)
#endif	
			//this should never fail, but if it does, cleanup.
			AssertV(false, OSThread::GetErrno());
			
	}
}

void EventThread::Entry()
{
	struct eventreq theCurrentEvent;
	::memset( &theCurrentEvent, '\0', sizeof(theCurrentEvent) );
	
	while (true)
	{
		int theErrno = EINTR;
		while (theErrno == EINTR)
		{
#if MACOSXEVENTQUEUE
			int theReturnValue = waitevent(&theCurrentEvent, NULL);
#else
			int theReturnValue = select_waitevent(&theCurrentEvent, NULL);
#endif	
			//Sort of a hack. In the POSIX version of the server, waitevent can return
			//an actual POSIX errorcode.
			if (theReturnValue >= 0)
				theErrno = theReturnValue;
			else
				theErrno = OSThread::GetErrno();
		}
		
		AssertV(theErrno == 0, theErrno);
		
		//ok, there's data waiting on this socket. Send a wakeup.
		Assert(theCurrentEvent.er_data != NULL);
		if (theCurrentEvent.er_data != NULL)
		{
			//The cookie in this event is an ObjectID. Resolve that objectID into
			//a pointer.
			StrPtrLen idStr((char*)&theCurrentEvent.er_data, sizeof(theCurrentEvent.er_data));
			OSRef* ref = fRefTable.Resolve(&idStr);
			if (ref != NULL)
			{
				EventContext* theContext = (EventContext*)ref->GetObject();
#if DEBUG
			
				theContext->fModwatched = false;
#endif
				theContext->fEventBits = theCurrentEvent.er_eventbits;
				theContext->ProcessEvent(theCurrentEvent.er_eventbits);
				fRefTable.Release(ref);
				
				
			}
		}

#if EVENT_CONTEXT_DEBUG
		SInt64	yieldStart = OS::Milliseconds();
#endif

		this->Yield();

#if EVENT_CONTEXT_DEBUG
		SInt64	yieldDur = OS::Milliseconds() - yieldStart;
		static SInt64	numZeroYields;
		
		if ( yieldDur > 1 )
		{
			printf( "EventThread time in OSTHread::Yield %i, numZeroYields %i\n", (long)yieldDur, (long)numZeroYields );
			numZeroYields = 0;
		}
		else
			numZeroYields++;
#endif
	}
}
