/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		OS.h

	Contains:  	OS utility functions

	$Log: OS.cpp,v $
	Revision 1.1.1.1  2000/01/17 23:35:57  wsanchez
	Import of StreamingServer-2.0 (Apple)
	
	Revision 1.7  1999/12/23 23:20:56  murata
	fixed little endian 64 bit byte swapping so constants are defined as longs to prevent compiler bug with >32 constant.
	fixed microseconds on non-MacOSx to return correct value.
	
	Revision 1.6  1999/12/21 01:43:31  lecroy
	*** empty log message ***
	
	Revision 1.5  1999/10/30 07:14:59  serenyi
	Fixed build problems for linux
	
	Revision 1.4  1999/10/27 17:53:45  serenyi
	Integrated Matthias's MacOS / GUSI changes
	
	Revision 1.3  1999/10/23 03:19:34  murata
	Changes to allow debugging codewarrior compiles on MacOS for PlaylistBroadcaster
	
	Revision 1.2  1999/09/24 19:59:12  murata
	Added new time function for conversion to Unix 1970 time from server boot time(milliseconds)
	
	Revision 1.1.1.1  1999/09/22 23:01:18  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:37  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:05:30  ds
	Created
		
	
*/

#include <stdlib.h>
#include <string.h>

#ifdef __MACOS__
#include <time.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/errno.h>

#include "OS.h"
#include "OSThread.h"
#include "MyAssert.h"

#if __MacOSX__
#include "timestamp.h"//for fast milliseconds routine
#endif

double 	OS::sDivisor = 0;
double OS::sMicroDivisor = 0;
SInt64  OS::s1970Msec = 0;
SInt64  OS::sMsecSince1970 = 0;
SInt64  OS::sMsecSince1900 = 0;
SInt64  OS::sInitialMsec = 0;

void OS::Initialize()
{
#if __MacOSX__
	//Setup divisor for milliseconds
	double divisorNumerator = 0;
	double divisorDenominator = 0;
	struct timescale theTimescale;
	
	(void)utimescale(&theTimescale);
	
	divisorNumerator = (double)theTimescale.tsc_numerator;
	divisorDenominator = (double)theTimescale.tsc_denominator;//we want milliseconds, not micro
	sMicroDivisor = divisorNumerator / divisorDenominator;
	divisorDenominator *= 1000;//we want milliseconds, not micro
	sDivisor = divisorNumerator / divisorDenominator;
#endif

	//setup t0 value for msec since 1900
	struct timeval t;
	struct timezone tz;
	int theErr = ::gettimeofday(&t, &tz);
	Assert(theErr == 0);

	//t.tv_sec is number of seconds since Jan 1, 1970. Convert to seconds since 1900	
	s1970Msec = 24 * 60 * 60;
	s1970Msec *= (70 * 365) + 17;
	
	sMsecSince1970 = ((SInt64)t.tv_sec * 1000) + ((SInt64)t.tv_usec / 1000);
	sMsecSince1900 = s1970Msec + sMsecSince1970;
	
	sInitialMsec = Milliseconds();
}

SInt64 OS::Milliseconds()
{
#if __MacOSX__
	//DMS - here is that superfast way of getting milliseconds. Note that this is time
	//relative to when the machine booted.
	UInt64 theCurTime = timestamp();
	Assert(sDivisor > 0);
	
	double theDoubleTime = (double)theCurTime;
	theDoubleTime *= sDivisor;
	return ((SInt64)theDoubleTime) - sInitialMsec;
#else
	struct timeval t;
	struct timezone tz;
	int theErr = ::gettimeofday(&t, &tz);
	Assert(theErr == 0);

	SInt64 curTime;
	curTime = t.tv_sec;
	curTime *= 1000;				// sec -> msec
	curTime += t.tv_usec / 1000;	// usec -> msec

	return curTime - sInitialMsec;
#endif
}

SInt64 OS::Microseconds()
{
#if __MacOSX__
	UInt64 theCurTime = timestamp();	
	double theDoubleTime = (double)theCurTime;
	theDoubleTime *= sMicroDivisor;
	return (SInt64)theDoubleTime - (sInitialMsec * 1000);
#else
	struct timeval t;
	struct timezone tz;
	int theErr = ::gettimeofday(&t, &tz);
	Assert(theErr == 0);

	SInt64 curTime;
	curTime = t.tv_sec;
	curTime *= 1000000;		// sec -> usec
	curTime += t.tv_usec;

	return curTime - (sInitialMsec * 1000);
#endif
}

SInt64	OS::HostToNetworkSInt64(SInt64 hostOrdered)
{
#if BIGENDIAN
	return hostOrdered;
#else
	return (SInt64) (  (UInt64)  (hostOrdered << 56) | (UInt64)  (((UInt64) 0x00ff0000 << 32) & (hostOrdered << 40))
		| (UInt64)  ( ((UInt64)  0x0000ff00 << 32) & (hostOrdered << 24)) | (UInt64)  (((UInt64)  0x000000ff << 32) & (hostOrdered << 8))
		| (UInt64)  ( ((UInt64)  0x00ff0000 << 8) & (hostOrdered >> 8)) | (UInt64)     ((UInt64)  0x00ff0000 & (hostOrdered >> 24))
		| (UInt64)  (  (UInt64)  0x0000ff00 & (hostOrdered >> 40)) | (UInt64)  ((UInt64)  0x00ff & (hostOrdered >> 56)) );
#endif
}


OS_Error OS::MakeDir(char *inPath, mode_t mode)
{
	struct stat theStatBuffer;
	if (::stat(inPath, &theStatBuffer) == -1)
	{
		//this directory doesn't exist, so let's try to create it
		if (::mkdir(inPath, mode) == -1)
			return (OS_Error)OSThread::GetErrno();
	}
	else if (!S_ISDIR(theStatBuffer.st_mode))
		return EEXIST;//there is a file at this point in the path!

	//directory exists
	return OS_NoErr;
}

OS_Error OS::RecursiveMakeDir(char *inPath, mode_t mode)
{
	Assert(inPath != NULL);
	
	//iterate through the path, replacing '/' with '\0' as we go
	char *thePathTraverser = inPath;
	
	//skip over the first / in the path.
	if (*thePathTraverser == '/')
		thePathTraverser++;
		
	while (*thePathTraverser != '\0')
	{
		if (*thePathTraverser == '/')
		{
			//we've found a filename divider. Now that we have a complete
			//filename, see if this partial path exists.
			
			//make the partial path into a C string
			*thePathTraverser = '\0';
			OS_Error theErr = MakeDir(inPath, mode);
			//there is a directory here. Just continue in our traversal
			*thePathTraverser = '/';

			if (theErr != OS_NoErr)
				return theErr;
		}
		thePathTraverser++;
	}
	
	//need to create the last directory in the path
	return MakeDir(inPath, mode);
}

