/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		OSCond.cpp

	Contains:	Implementation of OSCond class
	
	$Log: OSCond.cpp,v $
	Revision 1.3.2.1  2000/04/19 00:30:22  murata
	Removed assert following call to pthread_timed_wait. The error returned should be ignored.
	Bug #:2453341
	Submitted by:
	Reviewed by:
	
	Revision 1.3  2000/03/22 19:05:46  mythili
	*** empty log message ***
	
	Revision 1.2  2000/03/22 18:56:20  mythili
	Modified TimedWait to pass correct timeout values to pthread_cond_timedwait
	
	Revision 1.1.1.1  2000/01/17 23:35:57  wsanchez
	Import of StreamingServer-2.0 (Apple)
	
	Revision 1.6  1999/12/21 01:43:32  lecroy
	*** empty log message ***
	
	Revision 1.5  1999/12/15 23:33:25  serenyi
	Any pthread platform has mutex recursion implemented manually
	
	Revision 1.4  1999/11/22 01:28:05  serenyi
	Implemented recursive mutexes for platforms (IRIX, Solaris) that don't support mutex recursion internally
	
	Revision 1.3  1999/11/08 22:53:18  serenyi
	Updated code to reflect latest version of QTSS API
	
	Revision 1.2  1999/10/27 17:53:45  serenyi
	Integrated Matthias's MacOS / GUSI changes
	
	Revision 1.1.1.1  1999/09/22 23:01:18  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:38  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:05:31  ds
	Created
	

*/

#include "OSCond.h"
#include "OSMutex.h"
#include "OSThread.h"
#include "MyAssert.h"

#if __PTHREADS__
#include <sys/time.h>
#endif


OSCond::OSCond()
{
#if !__PTHREADS__
	fCondition = mycondition_alloc();
	Assert(fCondition != NULL);
#else
	int ret = pthread_cond_init(&fCondition, NULL);
	Assert(ret == 0);
#endif
}

OSCond::~OSCond()
{
#if !__PTHREADS__
	Assert(fCondition != NULL);
	mycondition_free(fCondition);
#else
	pthread_cond_destroy(&fCondition);
#endif
}

#if __PTHREADS__
void OSCond::TimedWait(OSMutex* inMutex, SInt32 inTimeoutInMilSecs)
{
	struct timespec ts;
	struct timeval tv;
	struct timezone tz;
	int	sec, usec;
	
	//These platforms do refcounting manually, and wait will release the mutex,
	// so we need to update the counts here
	Assert(inMutex->fHolder == pthread_self());
	inMutex->fHolderCount--;
	Assert(inMutex->fHolderCount == 0);
	inMutex->fHolder = 0;

	if (inTimeoutInMilSecs == 0)
		(void)pthread_cond_wait(&fCondition, &inMutex->fMutex);
	else
	{
		gettimeofday(&tv, &tz);
		sec = inTimeoutInMilSecs / 1000;
		inTimeoutInMilSecs = inTimeoutInMilSecs - (sec * 1000);
		Assert(inTimeoutInMilSecs < 1000);
		usec = inTimeoutInMilSecs * 1000;
		Assert(tv.tv_usec < 1000000);
		ts.tv_sec = tv.tv_sec + sec;
		ts.tv_nsec = (tv.tv_usec + usec) * 1000;
		Assert(ts.tv_nsec < 2000000000);
		if(ts.tv_nsec > 999999999)
		{
		     ts.tv_sec++;
		     ts.tv_nsec -= 1000000000;
		}
		int err = pthread_cond_timedwait(&fCondition, &inMutex->fMutex, &ts);
	}

	//Reverse of above
	Assert(inMutex->fHolder == 0);
	Assert(inMutex->fHolderCount == 0);
	inMutex->fHolderCount++;
	inMutex->fHolder = pthread_self();
}
#endif
