/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		osfile.cpp

	Contains:	simple file abstraction
					
	$Log: OSFileSource.cpp,v $
	Revision 1.3  2000/02/11 01:07:25  serenyi
	Server now supports cache control header and sends 304 Not Modified responses in appropriate situations
	
	Revision 1.2  2000/01/27 02:09:05  rtucker
	Merge changes from v81
	
	Revision 1.7  2000/01/19 19:38:21  serenyi
	Moved to new NEW macro
	
	Revision 1.6  1999/12/21 01:43:32  lecroy
	*** empty log message ***
	
	Revision 1.5  1999/11/08 22:53:19  serenyi
	Updated code to reflect latest version of QTSS API
	
	Revision 1.4  1999/10/28 20:59:45  serenyi
	Use OSThread::GetErrno
	
	Revision 1.3  1999/10/27 17:53:45  serenyi
	Integrated Matthias's MacOS / GUSI changes
	
	Revision 1.2  1999/10/23 03:19:34  murata
	Changes to allow debugging codewarrior compiles on MacOS for PlaylistBroadcaster
	
	Revision 1.1.1.1  1999/09/22 23:01:18  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:38  rtucker
	initial checkin from rrv14
	
	
	Revision 1.1  1999/06/02 18:14:50  serenyi
	Akamaizable version of OSFi
	
	Revision 1.1  1999/05/27 00:21:52  serenyi
	
	Revision 1.2  1999/02/19 23:06:12  ds
	Created
	
	
*/

#include <stdio.h>
#include <string.h>


#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/errno.h>

#include "OSFileSource.h"
#include "OSMemory.h"
#include "OSThread.h"

#if __MACOS__
#include "OSArrayObjectDeleter.h"
#endif

void OSFileSource::ReadEntireFile(char* inPath, StrPtrLen* outData)
{
	// Construct an OSFileSource object
	OSFileSource theFile(inPath);
	if (theFile.GetLength() == 0)
		return;
	
	// Allocate memory for the file data
	outData->Ptr = NEW char[theFile.GetLength() + 1];
	outData->Len = theFile.GetLength();
	
	// Read the data
	UInt32 recvLen = 0;
	OS_Error theErr = theFile.Read(outData->Ptr, outData->Len, &recvLen);
	Assert(theErr == OS_NoErr);
	Assert(recvLen == outData->Len);
}

void OSFileSource::Set(const char *inPath)
{
	Close();

#if __MACOS__
	OSCharArrayDeleter tempPath(NEW char[::strlen(inPath) + 2]);
	char* pathStart = &tempPath.GetObject()[0];
	
	// If inPath is relative, append a : 
	if (inPath[0] != '/')
	{
		*pathStart = ':';
		pathStart++;
	}
	// If inPath is absolute, skip past the first path marker
	else
		inPath++;

	strcpy(pathStart, inPath);
	
	// Convert all /'s to :'s
	while (*pathStart != '\0')
	{
		if (*pathStart == '/')
			*pathStart = ':';
		pathStart++;
	}

	// Use our new path, not the original one
	inPath = &tempPath.GetObject()[0];
	
#endif

#if __MacOSX__
	fFile = open(inPath, O_RDONLY | O_NO_MFS);
#else
	fFile = open(inPath, O_RDONLY);
#endif

	if (fFile != -1)
	{
		struct stat buf;
		if (::fstat(fFile, &buf) >= 0)
		{
			fLength = buf.st_size;
			fModDate = buf.st_mtime;
		}
		else
			Close();
	}	
}

void OSFileSource::Advise(UInt64 advisePos, UInt32 adviseAmt)
{
#if __MacOSX__
	//if the caller wants us to, issue an advise. Advise for the file area
	//immediately following the current read area
	struct radvisory radv;
	memset(&radv, 0, sizeof(radv));
	radv.ra_offset = (off_t)advisePos;
	radv.ra_count = (int)adviseAmt;
	::fcntl(fFile, F_RDADVISE, &radv);
#endif
}

OS_Error	OSFileSource::Read(void* inBuffer, UInt32 inLength, UInt32* outRcvLen)
{
	//optimization: avoid the extra kernel call. Only do an lseek if the file pointer really
	//doesn't match where we want to read from
	//if (fPosition != fReadPos)
		if (lseek(fFile, fPosition, SEEK_SET) == -1)
			return OSThread::GetErrno();
		
	int rcvLen = ::read(fFile, (char*)inBuffer, inLength);
	if (rcvLen == -1)
		return OSThread::GetErrno();

	if (outRcvLen != NULL)
		*outRcvLen = rcvLen;

	fPosition += rcvLen;
	fReadPos = fPosition;
	return OS_NoErr;
}

OS_Error	OSFileSource::Read(UInt64 inPosition, void* inBuffer, UInt32 inLength, UInt32* outRcvLen)
{
    this->Seek(inPosition);
    return this->Read(inBuffer,inLength,outRcvLen);
}


void	OSFileSource::Close()
{
	if ((fFile != -1) && (fShouldClose))
		::close(fFile);
	fFile = -1;
	fModDate = 0;
	fLength = 0;
	fPosition = 0;
	fReadPos = 0;
}

