/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		OSMutex.cpp

	Contains:	Platform - independent mutex header. The implementation of this object
				is platform - specific. Each platform must define an independent
				QTSSMutex.h & QTSSMutex.cpp file.
				
				This file is for Mac OS X Server only

	$Log: OSMutex.cpp,v $
	Revision 1.2  2000/01/27 02:09:05  rtucker
	Merge changes from v81
	
	Revision 1.10  2000/01/19 19:38:21  serenyi
	Moved to new NEW macro
	
	Revision 1.9  1999/12/21 01:43:34  lecroy
	*** empty log message ***
	
	Revision 1.8  1999/12/15 23:33:25  serenyi
	Any pthread platform has mutex recursion implemented manually
	
	Revision 1.7  1999/11/22 01:28:04  serenyi
	Implemented recursive mutexes for platforms (IRIX, Solaris) that don't support mutex recursion internally
	
	Revision 1.6  1999/11/15 01:23:17  serenyi
	Added in 4-char debugging codes, merged in solaris changes
	
	Revision 1.5  1999/11/08 23:03:09  serenyi
	Fixed macOS compilation problem
	
	Revision 1.4  1999/11/08 02:27:49  lecroy
	long long printf macro & sgi mods
	
	Revision 1.3  1999/10/30 07:14:59  serenyi
	Fixed build problems for linux
	
	Revision 1.2  1999/10/27 17:53:45  serenyi
	Integrated Matthias's MacOS / GUSI changes
	
	Revision 1.1.1.1  1999/09/22 23:01:19  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:38  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:06:13  ds
	Created
		

*/

#include "OSMutex.h"
#include <stdlib.h>
#include <string.h>

// Private globals
#if __PTHREADS__
static pthread_mutexattr_t	*sMutexAttr=NULL;
static pthread_once_t sMutexAttrInit = PTHREAD_ONCE_INIT;
static void MutexAttrInit();
#endif

OSMutex::OSMutex()
{
#if !__PTHREADS__
	fMutex = mymutex_alloc();
#else
	(void)pthread_once(&sMutexAttrInit, MutexAttrInit);
	(void)pthread_mutex_init(&fMutex, sMutexAttr);
	
	fHolder = 0;
	fHolderCount = 0;
#endif
}

#if __PTHREADS__
void MutexAttrInit()
{
	sMutexAttr = (pthread_mutexattr_t*)malloc(sizeof(pthread_mutexattr_t));
	::memset(sMutexAttr, 0, sizeof(pthread_mutexattr_t));
	pthread_mutexattr_init(sMutexAttr);
}
#endif

OSMutex::~OSMutex()
{
#if !__PTHREADS__
	mymutex_free(fMutex);
#else
	pthread_mutex_destroy(&fMutex);
#endif
}

#if __PTHREADS__
void		OSMutex::RecursiveLock()
{
	// We already have this mutex. Just refcount and return
	if (pthread_self() == fHolder)
	{
		fHolderCount++;
		return;
	}
	(void)pthread_mutex_lock(&fMutex);
	Assert(fHolder == 0);
	fHolder = pthread_self();
	fHolderCount++;
	Assert(fHolderCount == 1);
}

void		OSMutex::RecursiveUnlock()
{
	if (pthread_self() != fHolder)
		return;
		
	Assert(fHolderCount > 0);
	fHolderCount--;
	if (fHolderCount == 0)
	{
		fHolder = 0;
		pthread_mutex_unlock(&fMutex);
	}
}

Bool16		OSMutex::RecursiveTryLock()
{
	// We already have this mutex. Just refcount and return
	if (pthread_self() == fHolder)
	{
		fHolderCount++;
		return true;
	}

	int theErr = pthread_mutex_trylock(&fMutex);
	if (theErr != 0)
	{
		Assert(theErr == EBUSY);
		return false;
	}
	Assert(fHolder == 0);
	fHolder = pthread_self();
	fHolderCount++;
	Assert(fHolderCount == 1);
	return true;
}
#endif //__PTHREADS__