/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		OSMutex.h

	Contains:	Platform - independent mutex header. The implementation of this object
				is platform - specific. Each platform must define an independent
				OSMutex.h & OSMutex.cpp file.
				
				This file is for Mac OS X Server only

	$Log: OSMutex.h,v $
	Revision 1.2  2000/01/27 02:09:05  rtucker
	Merge changes from v81
	
	Revision 1.11  2000/01/19 19:38:22  serenyi
	Moved to new NEW macro
	
	Revision 1.10  1999/12/21 01:43:34  lecroy
	*** empty log message ***
	
	Revision 1.9  1999/12/15 23:33:25  serenyi
	Any pthread platform has mutex recursion implemented manually
	
	Revision 1.8  1999/11/22 01:28:04  serenyi
	Implemented recursive mutexes for platforms (IRIX, Solaris) that don't support mutex recursion internally
	
	Revision 1.7  1999/11/15 01:23:18  serenyi
	Added in 4-char debugging codes, merged in solaris changes
	
	Revision 1.6  1999/11/08 02:27:49  lecroy
	long long printf macro & sgi mods
	
	Revision 1.5  1999/11/02 04:01:51  serenyi
	Added a necessary include
	
	Revision 1.4  1999/10/30 07:14:59  serenyi
	Fixed build problems for linux
	
	Revision 1.3  1999/10/27 03:28:19  lecroy
	merged in changes from Akamai
	
	Revision 1.2  1999/10/07 08:55:01  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.1.1.1  1999/09/22 23:01:19  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:38  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:06:14  ds
	Created
	

*/

#ifndef _OSMUTEX_H_
#define _OSMUTEX_H_

#include <stdlib.h>
#include <sys/errno.h>

#if !__PTHREADS__
#include "mymutex.h"
#else
#include <pthread.h>
#endif

#include "OSHeaders.h"
#include "OSThread.h"
#include "MyAssert.h"

class OSCond;

class OSMutex
{
	public:

		OSMutex();
		~OSMutex();

		inline void Lock();
		inline void Unlock();
		
		// Returns true on successful grab of the lock, false on failure
		inline Bool16 TryLock();

	private:

#if !__PTHREADS__
		mymutex_t fMutex;
#else
		pthread_mutex_t fMutex;
		// These two platforms don't implement pthreads recursive mutexes, so
		// we have to do it manually
		pthread_t 	fHolder;
		UInt32		fHolderCount;
		
		void		RecursiveLock();
		void		RecursiveUnlock();
		Bool16		RecursiveTryLock();
#endif

		friend class OSCond;
};

class	OSMutexLocker
{
	public:

		OSMutexLocker(OSMutex *inMutexP) : fMutex(inMutexP) { if (fMutex != NULL) fMutex->Lock(); }
		~OSMutexLocker() {	if (fMutex != NULL) fMutex->Unlock(); }
		
		void Lock() 		{ if (fMutex != NULL) fMutex->Lock(); }
		void Unlock() 		{ if (fMutex != NULL) fMutex->Unlock(); }
		
	private:

		OSMutex*	fMutex;
};

void OSMutex::Lock()
{
#if !__PTHREADS__
	mymutex_lock(fMutex);
#else
	this->RecursiveLock();
#endif //!__PTHREADS__
}

void OSMutex::Unlock()
{
#if !__PTHREADS__
	mymutex_unlock(fMutex);
#else
	this->RecursiveUnlock();
#endif //!__PTHREADS__
}

Bool16 OSMutex::TryLock()
{
#if !__PTHREADS__
	return (Bool16)mymutex_try_lock(fMutex);
#else
	return this->RecursiveTryLock();
#endif //!__PTHREADS__
}

#endif //_OSMUTEX_H_