/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		OSThread.cpp

	Contains:	Thread abstraction implementation

	$Log: OSThread.cpp,v $
	Revision 1.2  2000/03/09 02:08:43  serenyi
	solaris build mods
	
	Revision 1.1.1.1  2000/01/17 23:35:58  wsanchez
	Import of StreamingServer-2.0 (Apple)
	
	Revision 1.7  2000/01/10 21:59:35  rtucker
	add thr_yield for solaris
	
	Revision 1.6  1999/12/23 18:45:08  rtucker
	mods to support platforms with non-preemptive threading packages
	
	Revision 1.5  1999/12/21 01:43:35  lecroy
	*** empty log message ***
	
	Revision 1.4  1999/10/30 07:14:59  serenyi
	Fixed build problems for linux
	
	Revision 1.3  1999/10/27 17:53:46  serenyi
	Integrated Matthias's MacOS / GUSI changes
	
	Revision 1.2  1999/10/07 08:54:58  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.1.1.1  1999/09/22 23:01:20  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:38  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:06:15  ds
	Created
		
	
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if __PTHREADS__
#include <pthread.h>
#if USE_THR_YIELD
#include <thread.h>
#endif
#else
#include <mach/mach.h>
#include <mach/cthreads.h>
#endif

#include "OSThread.h"
#include "Exception.h"
#include "MyAssert.h"

//
// OSThread.cp
//
void*	OSThread::sMainThreadData = NULL;

#if __PTHREADS__
pthread_key_t OSThread::gMainKey = 0;
pthread_once_t OSThread::gMakeMainKey = PTHREAD_ONCE_INIT;
#endif

OSThread::OSThread(char *threadName)
{
    int strlt = strlen(threadName);
	strlt = (strlt > 31) ? 31 : strlt;
	strncpy(fName, threadName, strlt);
	fName[strlt] = 0;	
   
	fStopRequested = false;
	fStarted = false;
	fRunning = false;
	fCancelThrown = false;
	fDetached = false;
	fJoined =false;
	fBlockStopRequest = false;
	fDoDetach = false;
	fThreadData = NULL;
	
#if __PTHREADS__
	pthread_once(&OSThread::gMakeMainKey, MakeMainKey);
#endif

#if DEBUG
	fLocksHeld = 0;
	memset(fLockTagArray, 0, sizeof(fLockTagArray));
#endif
}

OSThread::~OSThread()
{
	fDoDetach = false;//we must JOIN with the thread!
	//send the thread a stop exception
	this->SendStopRequest();
		
	fDoDetach = false;
	if (!fJoined && !fDetached)
		Join();
}

void OSThread::Start()
{
#if __PTHREADS__
	int err = pthread_create((pthread_t*)&fThreadID, NULL, _Entry, (void*)this);
	Assert(err == 0);
#else
	fThreadID = (UInt32)cthread_fork((cthread_fn_t)_Entry, (any_t)this);
#endif
}

void OSThread::Join()
{
	// What we're trying to do is allow the thread we want to delete to complete
	// running. So we wait for it to stop.
	Assert(!fJoined && !fDetached);
	fJoined = true;
#if __PTHREADS__
	void *retVal;
	pthread_join(pthread_self(), &retVal);
#else
	cthread_join((cthread_t)fThreadID);
#endif
}

void OSThread::Entry()
{
	Assert(false);
}

void OSThread::Yield()
{
	// on platforms who's threading is not pre-emptive yield 
	// to another thread
#if THREADING_IS_COOPERATIVE
	#if __PTHREADS__
		#if USE_THR_YIELD
			thr_yield();
		#else
			sched_yield();
		#endif
	#endif
#endif
	
}


void OSThread::Detach()
{
	Assert(!fDetached && !fJoined);
	fDetached = true;
}

void OSThread::SendStopRequest()
{
	fStopRequested = true;
	if (fDoDetach)
		Detach();
}

void OSThread::CheckForStopRequest()
{
	if (!fBlockStopRequest && fStopRequested && !fCancelThrown)
		ThrowStopRequest();
}


Bool16 OSThread::WillThrowStopRequest()
{
	return !fBlockStopRequest && fStopRequested && !fCancelThrown;
}


void OSThread::ThrowStopRequest()
{
	if (fStopRequested && !fCancelThrown) {
		fCancelThrown = true;
		Throw_(Cancel_E);
	}
}

void OSThread::CallEntry(OSThread* thread) 	// static method
{
	thread->fStarted = true;
	thread->fRunning = true;

	try
	{
		thread->Entry();
	} 
	catch(...)
	{
		Assert(thread->IsStopRequested());
	}

	thread->fRunning = false;
	
	if (thread->fDetached) {
		Assert(!thread->fJoined);
#if __PTHREADS__
		pthread_detach((pthread_t)thread->fThreadID);
#else
		cthread_detach((cthread_t)thread->fThreadID);
#endif
		delete thread;
	}
}

#if __PTHREADS__
void OSThread::MakeMainKey()
{
	pthread_key_create(&OSThread::gMainKey, NULL);
}
#endif

void* OSThread::_Entry(void *inThread)  //static
{
	OSThread* theThread = (OSThread*)inThread;
#if __PTHREADS__
	theThread->fThreadID = (UInt32)pthread_self();
	pthread_setspecific(OSThread::gMainKey, theThread);
#else
	theThread->fThreadID = (UInt32)cthread_self();
	cthread_set_data(cthread_self(), (any_t)theThread);
#endif
	OSThread::CallEntry(theThread);
	return NULL;
}

OSThread*	OSThread::GetCurrent()
{
#if __PTHREADS__
	return (OSThread *)pthread_getspecific(OSThread::gMainKey);
#else
	return (OSThread*)cthread_data(cthread_self());
#endif
}

#if DEBUG
void OSThread::IncrementLocksHeld(UInt32 tag)
{
	if ((fLocksHeld * 5) < (kTagArraySize - 5))
		//put this tag into the tag array
		::memcpy(&fLockTagArray[fLocksHeld * 5], &tag, sizeof(tag));
	fLocksHeld++;
}


void OSThread::DecrementLocksHeld(UInt32 tag)
{
	fLocksHeld--;
	//find a matching tag to remove out of the array
	for (UInt32 lockIndex = 0; lockIndex < kTagArraySize; lockIndex+=5)
	{
		if (::memcmp(&fLockTagArray[lockIndex], &tag, sizeof(tag)) == 0)
		{
			::memmove(&fLockTagArray[lockIndex], &fLockTagArray[lockIndex + 5], kTagArraySize - (lockIndex + 5));
			lockIndex = kTagArraySize; 
		}
	}
}

void OSThread::PrintLockTagArray()
{
	for (UInt32 lockIndex = 0; ((lockIndex < kTagArraySize) && (lockIndex < fLocksHeld)); lockIndex++)
		printf("Lock %ld: %s\n", lockIndex, &fLockTagArray[lockIndex * 5]);
}
#endif
