/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		OSThread.h

	Contains:	A thread abstraction

	$Log: OSThread.h,v $
	Revision 1.2  2000/02/04 00:46:25  serenyi
	Added date & expires headers to server response, server now parses if-modified-since dates
	
	Revision 1.1.1.1  2000/01/17 23:35:58  wsanchez
	Import of StreamingServer-2.0 (Apple)
	
	Revision 1.7  1999/12/23 18:45:08  rtucker
	mods to support platforms with non-preemptive threading packages
	
	Revision 1.6  1999/12/21 01:43:35  lecroy
	*** empty log message ***
	
	Revision 1.5  1999/11/15 01:23:18  serenyi
	Added in 4-char debugging codes, merged in solaris changes
	
	Revision 1.4  1999/10/28 21:00:10  serenyi
	Modified #includes
	
	Revision 1.3  1999/10/23 03:19:34  murata
	Changes to allow debugging codewarrior compiles on MacOS for PlaylistBroadcaster
	
	Revision 1.2  1999/10/07 08:55:01  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.1.1.1  1999/09/22 23:01:20  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:38  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:06:15  ds
	Created
		

*/

// OSThread.h
#ifndef __OSTHREAD__
#define __OSTHREAD__

#if __PTHREADS__
#if __solaris__
	#include <errno.h>
#else
	#include <sys/errno.h>
#endif
	#include <pthread.h>
#else
	#include <mach/mach.h>
	#include <mach/cthreads.h>
#endif

#include "OSHeaders.h"
#include "DateTranslator.h"

class OSThread
{

public:
								OSThread(char* threadName);
	virtual 					~OSThread();
	virtual 	void 			Entry();
				void 			Start();
				Bool16 			IsRunning()			{ return fRunning; }
				Bool16 			IsStopRequested() 	{ return fStopRequested; }
				void 			SendStopRequest();
				char* 			GetThreadName() 	{ return fName;}
				void 			Join();
				void 			Detach();
				static void		Yield();
				
				//When making a blocking system call, you can frame it
				//with these static function calls to support stop exceptions
				void			MakingBlockingCall() 	{ this->CheckForStopRequest(); }
				void			BlockingCallReturning() { this->CheckForStopRequest(); }
				void*			GetThreadData()			{ return fThreadData; }
				void			SetThreadData(void* inThreadData) { fThreadData = inThreadData; }
				
				// As a convienence to higher levels, each thread has its own date buffer
				DateBuffer*		GetDateBuffer()			{ return &fDateBuffer; }
				
				static void*	GetMainThreadData()		{ return sMainThreadData; }
				static void		SetMainThreadData(void* inData) { sMainThreadData = inData; }

#if DEBUG
				//for tracking how many locks are owned by a thread
				UInt32			GetNumLocksHeld() { return fLocksHeld; }
				void			IncrementLocksHeld(UInt32 tag);
				void			DecrementLocksHeld(UInt32 tag);
				void			PrintLockTagArray();
#endif

#if __PTHREADS__ || __MACOS__
	static  int			GetErrno() { return errno; }
#else
	static	int			GetErrno() { return cthread_errno();	}	
#endif

	
	static	OSThread*	GetCurrent();
	
private:

	void 			ThrowStopRequest();
	void 			CheckForStopRequest();	
	void			SuspendThread();
	void			ResumeThread();
	Bool16 		WillThrowStopRequest();

#if DEBUG
	static const UInt32 kTagArraySize = 50;
	UInt32	fLocksHeld;
	char	fLockTagArray[kTagArraySize];
#endif

#if __PTHREADS__
	static void MakeMainKey();
	static pthread_key_t	gMainKey;
	static pthread_once_t gMakeMainKey;
#endif

	Bool16 fStopRequested:1;
	Bool16 fStarted:1;
	Bool16 fRunning:1;
	Bool16 fDetached:1;
	Bool16 fJoined:1;
	Bool16 fCancelThrown:1;
	Bool16 fBlockStopRequest:1;
	Bool16 fDoDetach:1;

	char			fName[32];
	UInt32 			fThreadID;
	void*			fThreadData;
	DateBuffer		fDateBuffer;
	
	static void*	sMainThreadData;
	static void		CallEntry(OSThread* thread);		
	static void* 	_Entry(void* inThread);
};

#endif

