/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		Socket.cpp

	Contains:	implements Socket class
					
	$Log: Socket.cpp,v $
	Revision 1.1.1.1  2000/01/17 23:35:58  wsanchez
	Import of StreamingServer-2.0 (Apple)
	
	Revision 1.9  1999/12/21 01:43:35  lecroy
	*** empty log message ***
	
	Revision 1.8  1999/12/14 22:43:32  rtucker
	removed printf
	
	Revision 1.7  1999/12/14 19:38:46  rtucker
	made Defaults for Socket SO_RCVBUF 40K for DGRAM, 16L for STREAM
	
	Revision 1.6  1999/11/15 01:23:18  serenyi
	Added in 4-char debugging codes, merged in solaris changes
	
	Revision 1.5  1999/11/08 22:53:19  serenyi
	Updated code to reflect latest version of QTSS API
	
	Revision 1.4  1999/10/27 17:53:46  serenyi
	Integrated Matthias's MacOS / GUSI changes
	
	Revision 1.3  1999/10/27 03:28:19  lecroy
	merged in changes from Akamai
	
	Revision 1.2  1999/10/19 17:33:23  serenyi
	Added new prefs for the purposes of scaling TCP send buffers, added the ability to set these send buffers
	
	Revision 1.1.1.1  1999/09/22 23:01:20  lecroy
	importing
	
	Revision 1.3  1999/09/17 23:13:41  serenyi
	
	only read if the socket is connected
	
	Revision 1.2  1999/09/15 17:02:54  serenyi
	Reduced EPIPE errors
	
	Revision 1.1.1.1  1999/09/13 16:16:39  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:13:18  ds
	Created
	
	
*/

#include <string.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/errno.h>
#include <netinet/in.h>
#include <sys/uio.h>
#include <unistd.h>

#include "Socket.h"
#include "SocketUtils.h"
#include "OSMemory.h"

EventThread* Socket::sEventThread = NULL;

Socket::Socket(Task *notifytask, UInt32 inSocketType)
: 	EventContext(sEventThread),
	fState(inSocketType),
	fLocalAddrStrPtr(NULL),
	fLocalDNSStrPtr(NULL),
	fPortStr(fPortBuffer, kPortBufSizeInBytes)
{
	fLocalAddr.sin_addr.s_addr = 0;
	fLocalAddr.sin_port = 0;
	
	fDestAddr.sin_addr.s_addr = 0;
	fDestAddr.sin_port = 0;
	
	this->SetTask(notifytask);
}

OS_Error Socket::Open(int theType)
{
	Assert(fFileDesc == EventContext::kInvalidFileDesc);
	fFileDesc = ::socket(PF_INET, theType, 0);
	if (fFileDesc == EventContext::kInvalidFileDesc)
		return (OS_Error)OSThread::GetErrno();
			
	int 	socketBufSize;
	
	// use default buffer size of 16K for STREAM sockets
	// 40K for RAW and DATAGRAM sockets
	
	if ( theType == SOCK_STREAM ) 
		socketBufSize = 16*1024;
	else
		socketBufSize = 40*1024;
		
	//
	// Setup this socket's event context
	if (fState & kNonBlockingSocketType)
	{	
		this->InitNonBlocking(fFileDesc, NULL);	
	}

	int err = ::setsockopt(fFileDesc, SOL_SOCKET, SO_RCVBUF, (char*)&socketBufSize, sizeof(int));
	AssertV(err == 0, OSThread::GetErrno());
	
	return OS_NoErr;
}

void Socket::ReuseAddr()
{
	int one = 1;
	int err = ::setsockopt(fFileDesc, SOL_SOCKET, SO_REUSEADDR, (char*)&one, sizeof(int));
	Assert(err == 0);	
}

void	Socket::SetSocketBufSize(UInt32 inNewSize)
{
	int bufSize = inNewSize;
	int err = ::setsockopt(fFileDesc, SOL_SOCKET, SO_SNDBUF, (char*)&bufSize, sizeof(int));
	AssertV(err == 0, OSThread::GetErrno());
}


OS_Error Socket::Bind(UInt32 addr, UInt16 port)
{
	::memset(&fLocalAddr, 0, sizeof(fLocalAddr));
	fLocalAddr.sin_family = AF_INET;
	fLocalAddr.sin_port = htons(port);
	fLocalAddr.sin_addr.s_addr = htonl(addr);
	
	int err = ::bind(fFileDesc, (sockaddr *)&fLocalAddr, sizeof(fLocalAddr));
	
	if (err == -1)
	{
		fLocalAddr.sin_port = 0;
		fLocalAddr.sin_addr.s_addr = 0;
		return (OS_Error)OSThread::GetErrno();
	}
	fState |= kBound;
	return OS_NoErr;
}

StrPtrLen*	Socket::GetLocalAddrStr()
{
	//Use the array of IP addr strings to locate the string formatted version
	//of this IP address.
	if (fLocalAddrStrPtr == NULL)
	{
		for (UInt32 x = 0; x < SocketUtils::GetNumIPAddrs(); x++)
		{
			if (SocketUtils::GetIPAddr(x) == ntohl(fLocalAddr.sin_addr.s_addr))
			{
				fLocalAddrStrPtr = SocketUtils::GetIPAddrStr(x);
				break;
			}
		}
	}
	Assert(fLocalAddrStrPtr != NULL);
	return fLocalAddrStrPtr;
}

StrPtrLen*	Socket::GetLocalDNSStr()
{
	//Do the same thing as the above function, but for DNS names
	Assert(fLocalAddr.sin_addr.s_addr != INADDR_ANY);
	if (fLocalDNSStrPtr == NULL)
	{
		for (UInt32 x = 0; x < SocketUtils::GetNumIPAddrs(); x++)
		{
			if (SocketUtils::GetIPAddr(x) == ntohl(fLocalAddr.sin_addr.s_addr))
			{
				fLocalDNSStrPtr = SocketUtils::GetDNSNameStr(x);
				break;
			}
		}
	}

	//if we weren't able to get this DNS name, make the DNS name the same as the IP addr str.
	if (fLocalDNSStrPtr == NULL)
		fLocalDNSStrPtr = this->GetLocalAddrStr();

	Assert(fLocalDNSStrPtr != NULL);
	return fLocalDNSStrPtr;
}

StrPtrLen*	Socket::GetLocalPortStr()
{
	if (fPortStr.Len == kPortBufSizeInBytes)
	{
		int temp = ntohs(fLocalAddr.sin_port);
		::sprintf(fPortBuffer, "%d", temp);
		fPortStr.Len = ::strlen(fPortBuffer);
	}
	return &fPortStr;
}

OS_Error Socket::Send(const char* inData, const UInt32 inLength, UInt32* outLengthSent)
{
	Assert(inData != NULL);
	
	if (!(fState & kConnected))
		return (OS_Error)ENOTCONN;
		
	//int err = ::send(fFileDesc, inData, inLength, 0);//flags??
	int err;
    do {
       err = ::send(fFileDesc, inData, inLength, 0);//flags??
    } while((err == -1) && (OSThread::GetErrno() == EINTR));
	if (err == -1)
	{
		//Are there any errors that can happen if the client is connected?
		//Yes... EAGAIN. Means the socket is now flow-controleld
		int theErr = OSThread::GetErrno();
#if __solaris__
		// Solaris returns this error instead of EAGAIN
		if (theErr == ENOENT)
			theErr = EAGAIN;
#endif
		if ((theErr != EAGAIN) && (this->IsConnected()))
			fState ^= kConnected;//turn off connected state flag
		return (OS_Error)theErr;
	}
	
	*outLengthSent = err;
	return OS_NoErr;
}

OS_Error Socket::WriteV(const struct iovec* iov, const UInt32 numIOvecs, UInt32* outLenSent)
{
	Assert(iov != NULL);

	if (!(fState & kConnected))
		return (OS_Error)ENOTCONN;
		
	//int err = ::writev(fFileDesc, iov, numIOvecs);
	int err;
    do {
       err = ::writev(fFileDesc, iov, numIOvecs);//flags??
    } while((err == -1) && (OSThread::GetErrno() == EINTR));
	if (err == -1)
	{
		// Are there any errors that can happen if the client is connected?
		// Yes... EAGAIN. Means the socket is now flow-controleld
		int theErr = OSThread::GetErrno();
#if __solaris__
		// Solaris returns this error instead of EAGAIN
		if (theErr == ENOENT)
			theErr = EAGAIN;
#endif
		if ((theErr != EAGAIN) && (this->IsConnected()))
			fState ^= kConnected;//turn off connected state flag
		return (OS_Error)theErr;
	}

	*outLenSent = (UInt32)err;
	return OS_NoErr;
}

OS_Error Socket::Read(void *buffer, const UInt32 length, UInt32 *outRecvLenP)
{
	Assert(outRecvLenP != NULL);
	Assert(buffer != NULL);

	if (!(fState & kConnected))
		return (OS_Error)ENOTCONN;
			
	//int theRecvLen = ::recv(fFileDesc, buffer, length, 0);//flags??
	int theRecvLen;
    do {
       theRecvLen = ::recv(fFileDesc, buffer, length, 0);//flags??
    } while((theRecvLen == -1) && (OSThread::GetErrno() == EINTR));
	if (theRecvLen == -1)
	{
		// Are there any errors that can happen if the client is connected?
		// Yes... EAGAIN. Means the socket is now flow-controleld
		int theErr = OSThread::GetErrno();
#if __solaris__
		// Solaris returns this error instead of EAGAIN
		if (theErr == ENOENT)
			theErr = EAGAIN;
#endif
		if ((theErr != EAGAIN) && (this->IsConnected()))
			fState ^= kConnected;//turn off connected state flag
		return (OS_Error)theErr;
	}
	//if we get 0 bytes back from read, that means the client has disconnected.
	//Note that and return the proper error to the caller
	else if (theRecvLen == 0)
	{
		fState ^= kConnected;
		return (OS_Error)ENOTCONN;
	}
	Assert(theRecvLen > 0);
	*outRecvLenP = (UInt32)theRecvLen;
	return OS_NoErr;
}