/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		TCPListenerSocket.cpp

	Contains:	implements TCPListenerSocket class
					
	$Log: TCPListenerSocket.cpp,v $
	Revision 1.3  2000/03/09 22:52:17  serenyi
	Made work on freebsd
	
	Revision 1.2  2000/01/27 02:09:05  rtucker
	Merge changes from v81
	
	Revision 1.9  2000/01/19 19:38:22  serenyi
	Moved to new NEW macro
	
	Revision 1.8  1999/12/21 01:43:37  lecroy
	*** empty log message ***
	
	Revision 1.7  1999/11/15 01:23:18  serenyi
	Added in 4-char debugging codes, merged in solaris changes
	
	Revision 1.6  1999/11/08 22:53:19  serenyi
	Updated code to reflect latest version of QTSS API
	
	Revision 1.5  1999/10/27 17:53:46  serenyi
	Integrated Matthias's MacOS / GUSI changes
	
	Revision 1.4  1999/10/19 21:14:08  rtucker
	tcp enhancements
	
	Revision 1.3  1999/10/19 17:33:23  serenyi
	Added new prefs for the purposes of scaling TCP send buffers, added the ability to set these send buffers
	
	Revision 1.2  1999/10/14 23:18:34  rtucker
	changes to report server ip back to client over HTTP
	
	Revision 1.1.1.1  1999/09/22 23:01:22  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:39  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:13:21  ds
	Created
	
	
*/

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <unistd.h>
#include <sys/errno.h>

#include "TCPListenerSocket.h"
#include "Task.h"

OS_Error TCPListenerSocket::Listen(UInt32 queueLength)
{
	if (fFileDesc == EventContext::kInvalidFileDesc)
		return EBADF;
		
	int err = ::listen(fFileDesc, queueLength);
	if (err != 0)
		return (OS_Error)OSThread::GetErrno();
	return OS_NoErr;
}

OS_Error TCPListenerSocket::Initialize(UInt32 addr, UInt16 port)
{
	OS_Error err = this->TCPSocket::Open();
	if (err != 0)
		return err;
	
	this->RequestEvent(EV_RE);
	this->ReuseAddr();
	err = this->Bind(addr, port);
	
	if (err == 0)
	{
		// We never need to receive very much data, so just in case we get backed up
		// in reading, force the clients to flow control early
		int theRcvBufSize = 1024;
		err = ::setsockopt(fFileDesc, SOL_SOCKET, SO_RCVBUF, (char*)&theRcvBufSize, sizeof(int));
		AssertV(err == 0, OSThread::GetErrno());
	}
	
	if (err == 0)
		err = this->Listen(kListenQueueLength);
	return err;
}

void TCPListenerSocket::ProcessEvent(int /*eventBits*/)
{
	//we are executing on the same thread as every other
	//socket, so whatever you do here has to be fast.
	
	struct sockaddr_in addr;
#if __MacOSX__ || __FreeBSD__
	int size = sizeof(addr);
#else
	socklen_t size = sizeof(addr);
#endif
	Task* theTask = NULL;
	TCPSocket* theSocket = NULL;
	
	//fSocket data member of TCPSocket.
	while (true)
	{
		int osSocket = accept(fFileDesc, (struct sockaddr*)&addr, &size);
		if (osSocket == -1)
		{
			//take a look at what this error is.
			int acceptError = OSThread::GetErrno();
			
			//If it's EAGAIN, there's nothing on the listen queue right now,
			//so modwatch and return
			if (acceptError == EAGAIN)
				break;
#if __solaris__
			// For some reason solaris returns this error from accept.
			if (acceptError == ENOENT)
				break;
#endif
			//if this error gets returned, we're out of file desciptors, so let's
			//idle the listener until the condition clears up.
			else if (acceptError == EMFILE)
			{
				fOutOfDescriptors = true;
				this->SetIdleTimer(kTimeBetweenAcceptsInMsec);
				return;//don't modwatch, because we're out of file descriptors anyway.
			}
			else
			{
#if DEBUG
				printf("TCP Listener socket got %d err from accept\n",acceptError);
#endif
				continue;
			}
		}
		if ((theTask = this->GetSessionTask(&theSocket)) == NULL)
			//this should be a disconnect. do an ioctl call?
			close(osSocket);
		else
		{	
			Assert(osSocket != EventContext::kInvalidFileDesc);
			
			//set options on the socket
			//we are a server, always disable nagle algorithm
			int one = 1;
			int err = ::setsockopt(osSocket, IPPROTO_TCP, TCP_NODELAY, (char*)&one, sizeof(int));
			AssertV(err == 0, OSThread::GetErrno());

			err = ::setsockopt(osSocket, SOL_SOCKET, SO_KEEPALIVE, (char*)&one, sizeof(int));
			AssertV(err == 0, OSThread::GetErrno());

			int	sndBufSize = 8L * 1024L;
			err = ::setsockopt(osSocket, SOL_SOCKET, SO_SNDBUF, (char*)&sndBufSize, sizeof(int));
			AssertV(err == 0, OSThread::GetErrno());

			int	rcvBufSize = 1L * 1024L;
			err = ::setsockopt(osSocket, SOL_SOCKET, SO_RCVBUF, (char*)&rcvBufSize, sizeof(int));
			AssertV(err == 0, OSThread::GetErrno());

			//setup the socket. When there is data on the socket,
			//theTask will get an kReadEvent event
			theSocket->Set(osSocket, &addr, this);
			theSocket->InitNonBlocking(osSocket, NULL);
			theSocket->SetTask(theTask);
			theSocket->RequestEvent(EV_RE);
		}
	}
	
	fOutOfDescriptors = false;
	//after every accept we should modwatch to make sure we will continue
	//to get listen events
	this->RequestEvent(EV_RE);
}
SInt64 TCPListenerSocket::Run()
{
	//This function will get called when we have run out of file descriptors.
	//All we need to do is check the listen queue to see if the situation has
	//cleared up.
	(void)this->GetEvents();
	this->ProcessEvent(Task::kReadEvent);
	return 0;
}
