/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		UDPSocketPool.h

	Contains:	Object that creates & maintains UDP socket pairs in a pool.

	$Log: UDPSocketPool.h,v $
	Revision 1.3  2000/03/14 23:49:00  serenyi
	Changed port range to be 6970
	
	Revision 1.2  2000/01/27 02:09:06  rtucker
	Merge changes from v81
	
	Revision 1.3  2000/01/19 19:38:23  serenyi
	Moved to new NEW macro
	
	Revision 1.2  1999/12/21 01:43:38  lecroy
	*** empty log message ***
	
	Revision 1.1.1.1  1999/09/22 23:01:15  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:40  rtucker
	initial checkin from rrv14
	
	
	Revision 1.1  1999/02/19 23:15:02  ds
	Created
	
*/

#ifndef __UDPSOCKETPOOL_H__
#define __UDPSOCKETPOOL_H__

#include "UDPDemuxer.h"
#include "UDPSocket.h"
#include "OSMutex.h"
#include "OSQueue.h"

class UDPSocketPair;

class UDPSocketPool
{
	public:
	
		UDPSocketPool() : fMutex() {}
		virtual ~UDPSocketPool() {}
		
		//Skanky access to member data
		OSMutex*	GetMutex() 			{ return &fMutex; }
		OSQueue*	GetSocketQueue()	{ return &fUDPQueue; }
		
		//Gets a UDP socket out of the pool. 
		//inIPAddr = IP address you'd like this pair to be bound to.
		//inPort = port you'd like this pair to be bound to, or 0 if you don't care
		//inSrcIPAddr = srcIP address of incoming packets for the demuxer.
		//inSrcPort = src port of incoming packets for the demuxer.
		//This may return NULL if no pair is available that meets the criteria.
		UDPSocketPair*	GetUDPSocketPair(UInt32 inIPAddr, UInt16 inPort,
											UInt32 inSrcIPAddr, UInt16 inSrcPort);
		
		//When done using a UDP socket pair retrieved via GetUDPSocketPair, you must
		//call this function. Doing so tells the pool which UDP sockets are in use,
		//keeping the number of UDP sockets allocated at a minimum.
		void ReleaseUDPSocketPair(UDPSocketPair* inPair);

		UDPSocketPair*	CreateUDPSocketPair(UInt32 inAddr, UInt16 inPort);
		
	protected:
	
		//Because UDPSocket is a base class, and this pool class is intended to be
		//a general purpose class for all types of UDP sockets (reflector, standard),
		//there must be a virtual fuction for actually constructing the derived UDP sockets
		virtual UDPSocketPair*	ConstructUDPSocketPair() = 0;
		virtual void			DestructUDPSocketPair(UDPSocketPair* inPair) = 0;
	
	private:
	
		enum
		{
			kLowestUDPPort = 6970,	//UInt16
			kHighestUDPPort = 65535 //UInt16
		};
	
		OSQueue fUDPQueue;
		OSMutex fMutex;
};

class UDPSocketPair
{
	public:
		
		UDPSocketPair(UDPSocket* inSocketA, UDPSocket* inSocketB)
			: fSocketA(inSocketA), fSocketB(inSocketB), fRefCount(0), fElem(this) {}
		~UDPSocketPair() {}
	
		UDPSocket*	GetSocketA() { return fSocketA; }
		UDPSocket*	GetSocketB() { return fSocketB; }
		
	private:
	
		UDPSocket* 	fSocketA;
		UDPSocket* 	fSocketB;
		UInt32 		fRefCount;
		OSQueueElem fElem;
		
		friend class UDPSocketPool;
};
#endif // __UDPSOCKETPOOL_H__

