/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		RTCPPacket.cpp

	Contains:  	RTCPReceiverPacket de-packetizing classes

	$Log: RTCPPacket.cpp,v $
	Revision 1.1.1.1  2000/01/17 23:36:03  wsanchez
	Import of StreamingServer-2.0 (Apple)
	
	Revision 1.4  1999/12/21 01:44:01  lecroy
	*** empty log message ***
	
	Revision 1.3  1999/12/04 04:20:54  murata
	Remove check and false result from  parse in RTCPReceiver packet if report count is 0.
	Fix to bug 2412622.
	
	Revision 1.2  1999/10/07 08:54:40  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.1.1.1  1999/09/22 23:01:48  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:51  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:05:30  cl
	Created
		
	
*/


#include "RTCPPacket.h"

//returns true if successful, false otherwise
Bool16 RTCPPacket::ParsePacket(UInt8* inPacketBuffer, UInt32 inPacketLen)
{
	if (inPacketLen < kRTCPPacketSizeInBytes)
		return false;
	fReceiverPacketBuffer = inPacketBuffer;

	//the length of this packet can be no less than the advertised length (which is
	//in 32-bit words, so we must multiply) plus the size of the header (4 bytes)
	if (inPacketLen < ((this->GetPacketLength() * 4) + kRTCPHeaderSizeInBytes))
		return false;
	
	//do some basic validation on the packet
	if (this->GetVersion() != kSupportedRTCPVersion)
		return false;
		
	return true;
}

#ifdef DEBUG_RTCP_PACKETS
void RTCPReceiverPacket::Dump()//Override
{
	RTCPPacket::Dump();
	
	for (int i = 0;i<this->GetReportCount(); i++)
	{
		printf( "	[%d] rptSrcID==%lu, fracLost==%d, totLost==%d, highSeq#==%lu, jit==%lu, lastSRTime==%lu, lastSRDelay==%lu \n",
	                         i,
	                         this->GetReportSourceID(i),
	                         this->GetFractionLostPackets(i),
	                         this->GetTotalLostPackets(i),
	                         this->GetHighestSeqNumReceived(i),
	                         this->GetJitter(i),
	                         this->GetLastSenderReportTime(i),
	                         this->GetLastSenderReportDelay(i) );
	}


}
#endif


Bool16 RTCPReceiverPacket::ParseReceiverReport(UInt8* inPacketBuffer, UInt32 inPacketLength)
{
	Bool16 ok = this->ParsePacket(inPacketBuffer, inPacketLength);
	if (!ok)
		return false;
	
	fRTCPReceiverReportArray = inPacketBuffer + kRTCPPacketSizeInBytes;
	
	//this is the maximum number of reports there could possibly be
	int theNumReports = (inPacketLength - kRTCPPacketSizeInBytes) / kReportBlockOffsetSizeInBytes;

	//if the number of receiver reports is greater than the theoretical limit, return an error.
	if (this->GetReportCount() > theNumReports)
		return false;
		
	return true;
}

UInt32 RTCPReceiverPacket::GetCumulativeFractionLostPackets()
{
	float avgFractionLost = 0;
	for (short i = 0; i < this->GetReportCount(); i++)
	{
		avgFractionLost += this->GetFractionLostPackets(i);
		avgFractionLost /= (i+1);
	}
	
	return (UInt32)avgFractionLost;
}


UInt32 RTCPReceiverPacket::GetCumulativeJitter()
{
	float avgJitter = 0;
	for (short i = 0; i < this->GetReportCount(); i++)
	{
		avgJitter += this->GetJitter(i);
		avgJitter /= (i + 1);
	}
	
	return (UInt32)avgJitter;
}


UInt32 RTCPReceiverPacket::GetCumulativeTotalLostPackets()
{
	UInt32 totalLostPackets = 0;
	for (short i = 0; i < this->GetReportCount(); i++)
	{
		totalLostPackets += this->GetTotalLostPackets(i);
	}
	
	return totalLostPackets;
}



#ifdef DEBUG_RTCP_PACKETS
void RTCPPacket::Dump()
{
	
	printf("\n--\n");
	
	printf( "vers==%d, pad==%d, ReportCount==%d, type==%d, length==%d, sourceID==%ld\n",
			 this->GetVersion(),
             (int)this->GetHasPadding(),
             this->GetReportCount(),
             (int)this->GetPacketType(),
             (int)this->GetPacketLength(),
             this->GetPacketSourceID() );
}
#endif


