/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Portions Copyright (c) 1999 Apple Computer, Inc.  All Rights
 * Reserved.  This file contains Original Code and/or Modifications of
 * Original Code as defined in and that are subject to the Apple Public
 * Source License Version 1.1 (the "License").  You may not use this file
 * except in compliance with the License.  Please obtain a copy of the
 * License at http://www.apple.com/publicsource and read it before using
 * this file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON- INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		RTSPClient.cpp

	Contains:	.  
					
	
*/

#include "RTSPClient.h"
#include "StringParser.h"
#include "OSMemory.h"

#include <sys/errno.h>

static char* sEmptyString = "";

RTSPClient::RTSPClient(UInt32 inSocketType)
:	fSocket(NULL, inSocketType),
	fHostAddr(0),
	fHostPort(0),
	fCSeq(1),
	fStatus(0),
	fSessionID(sEmptyString),
	fServerPort(0),
	fContentLength(0),
	fRecvContentBuffer(NULL),
	fSendBufferLen(0),
	fLengthSent(0),
	fContentRecvLen(0),
	fHeaderRecvLen(0),
	fTransactionStarted(false),
	fReceiveInProgress(false),
	fReceivedResponse(false),
	fEventMask(0)
{}

RTSPClient::~RTSPClient()
{
	delete [] fRecvContentBuffer;
	delete [] fURL.Ptr;
	if (fSessionID.Ptr != sEmptyString)
		delete [] fSessionID.Ptr;
}

void RTSPClient::Set(UInt32 hostAddr, UInt16 hostPort, const StrPtrLen& inURL)
{
	delete [] fURL.Ptr;
	fURL.Ptr = new char[inURL.Len + 1];
	fURL.Len = inURL.Len;
	::memcpy(fURL.Ptr, inURL.Ptr, inURL.Len);
	fURL.Ptr[fURL.Len] = '\0';

	fHostAddr = hostAddr;
	fHostPort = hostPort;	
}

OS_Error RTSPClient::SendDescribe()
{
	if (!fTransactionStarted)
		::sprintf(fSendBuffer, "DESCRIBE %s RTSP/1.0\r\nCSeq: %lu\r\nAccept: application.sdp\r\nUser-agent: QTSS/2.0\r\n\r\n", fURL.Ptr, fCSeq);
	
	return this->DoTransaction();
}

OS_Error RTSPClient::SendSetup(UInt32 inTrackID, UInt16 inClientPort)
{
	if (!fTransactionStarted)
		::sprintf(fSendBuffer, "SETUP %s/trackID=%lu RTSP/1.0\r\nCSeq: %lu\r\n%sTransport: RTP/AVP;unicast;client_port=%u-%u\r\nUser-agent: QTSS/2.0\r\n\r\n", fURL.Ptr, inTrackID, fCSeq, fSessionID.Ptr, inClientPort, inClientPort + 1);
	
	return this->DoTransaction();
}

OS_Error RTSPClient::SendPlay()
{
	if (!fTransactionStarted)
		::sprintf(fSendBuffer, "PLAY %s RTSP/1.0\r\nCSeq: %lu\r\n%sRange: npt=0.0\r\nUser-agent: QTSS/2.0\r\n\r\n", fURL.Ptr, fCSeq, fSessionID.Ptr);
	
	return this->DoTransaction();
}

OS_Error RTSPClient::SendTeardown()
{
	if (!fTransactionStarted)
		::sprintf(fSendBuffer, "TEARDOWN %s RTSP/1.0\r\nCSeq: %lu\r\n%sUser-agent: QTSS/2.0\r\n\r\n", fURL.Ptr, fCSeq, fSessionID.Ptr);
	
	return this->DoTransaction();
}

OS_Error RTSPClient::DoTransaction()
{
	OS_Error theErr = OS_NoErr;
	fEventMask = 0;
	
	if (!fTransactionStarted)
	{
		fCSeq++;
		fSendBufferLen = ::strlen(fSendBuffer);
		fLengthSent = 0;
		fTransactionStarted = true;
	}
	
	if (!fReceiveInProgress)
	{
		theErr = this->CheckConnection();
		if (theErr != OS_NoErr)
			return theErr;
		theErr = this->SendRequest();
	}
	if (fReceiveInProgress)
	{
		Assert(theErr == OS_NoErr);
		theErr = this->ReceiveResponse();
	}
	return theErr;
}

OS_Error RTSPClient::CheckConnection()
{
	OS_Error theErr = OS_NoErr;
	if (!fSocket.IsBound())
	{
		theErr = fSocket.Open();
		if (theErr == OS_NoErr)
			theErr = fSocket.Bind(0, 0);

		if (theErr != OS_NoErr)
			return theErr;
	}
	if (!fSocket.IsConnected())
	{
		theErr = fSocket.Connect(fHostAddr, fHostPort);
		if (theErr == EINPROGRESS)
		{
			fEventMask = EV_RE | EV_WR;
			return theErr;
		}
	}
	return theErr;
}

OS_Error RTSPClient::SendRequest()
{
	while (fLengthSent < fSendBufferLen)
	{
		UInt32 theLengthSent = 0;
		OS_Error theErr = fSocket.Send(&fSendBuffer[fLengthSent], fSendBufferLen - fLengthSent, &theLengthSent);
		fLengthSent += theLengthSent;
		
		if (theErr != OS_NoErr)
		{
			fEventMask = EV_WR;// wait for the socket to become writeable
			return theErr;
		}
	}
	
	// Done sending request, we're moving onto receiving the response
	fReceiveInProgress = true;
	return OS_NoErr;
}

OS_Error RTSPClient::ReceiveResponse()
{
	OS_Error theErr = OS_NoErr;
	
	while (!fReceivedResponse)
	{
		UInt32 theRecvLen = 0;
		theErr = fSocket.Read(&fRecvHeaderBuffer[fHeaderRecvLen], kReqBufSize - fHeaderRecvLen, &theRecvLen);
		if (theErr != OS_NoErr)
		{
			fEventMask = EV_RE;// wait for the socket to become readable
			return theErr;
		}
		fHeaderRecvLen += theRecvLen;
		
		// Check to see if we've gotten a complete header
		fRecvHeaderBuffer[fHeaderRecvLen] = '\0';
		
		char* theResponseData = ::strstr(fRecvHeaderBuffer, "\r\n\r\n");
		if (theResponseData != NULL)
		{
			// skip past the \r\n\r\n
			theResponseData += 4;
			
			// We've got a new response
			fReceivedResponse = true;
			
			// Zero out fields that will change with every RTSP response
			fServerPort = 0;
			fStatus = 0;
			fContentLength = 0;
		
			// Parse the response.
			StrPtrLen theData(fRecvHeaderBuffer, (theResponseData - (&fRecvHeaderBuffer[0])));
			StringParser theParser(&theData);
			
			theParser.ConsumeLength(NULL, 9); //skip past RTSP/1.0
			fStatus = theParser.ConsumeInteger(NULL);
			
			while (theParser.GetDataRemaining() > 0)
			{
				static StrPtrLen sSessionHeader("Session");
				static StrPtrLen sContentLenHeader("Content-length");
				static StrPtrLen sTransportHeader("Transport");
				
				StrPtrLen theKey;
				theParser.GetThruEOL(&theKey);
				
				if (theKey.NumEqualIgnoreCase(sSessionHeader.Ptr, sSessionHeader.Len))
				{
					if (fSessionID.Len == 0)
					{
						// Copy the session ID and store it.
						// First figure out how big the session ID is. We copy
						// everything up until the first ';' returned from the server
						UInt32 keyLen = 0;
						while (theKey.Ptr[keyLen] != ';')
							keyLen++;
						
						// Append an EOL so we can stick this thing transparently into the SETUP request
						
						fSessionID.Ptr = NEW char[keyLen + 3];
						fSessionID.Len = keyLen + 2;
						::memcpy(fSessionID.Ptr, theKey.Ptr, keyLen);
						::memcpy(fSessionID.Ptr + keyLen, "\r\n", 2);//Append a EOL
						fSessionID.Ptr[keyLen + 2] = '\0';
					}
				}
				else if (theKey.NumEqualIgnoreCase(sContentLenHeader.Ptr, sContentLenHeader.Len))
				{
					StringParser theCLengthParser(&theKey);
					theCLengthParser.ConsumeUntil(NULL, StringParser::sDigitMask);
					fContentLength = theCLengthParser.ConsumeInteger(NULL);
					
					delete [] fRecvContentBuffer;
					fRecvContentBuffer = NEW char[fContentLength + 1];
					
					// Figure out how much of the content body we've already received
					// in the header buffer
					fContentRecvLen = fHeaderRecvLen - (theResponseData - &fRecvHeaderBuffer[0]);

					// Immediately copy the bit of the content body that we've already
					// read off of the socket.
					::memcpy(fRecvContentBuffer, theResponseData, fContentRecvLen);
					
				}
				else if (theKey.NumEqualIgnoreCase(sTransportHeader.Ptr, sTransportHeader.Len))
				{
					StringParser theTransportParser(&theKey);
					StrPtrLen theSubHeader;

					while (theTransportParser.GetDataRemaining() > 0)
					{
						static StrPtrLen sServerPort("server_port");

						theTransportParser.GetThru(&theSubHeader, ';');
						if (theSubHeader.NumEqualIgnoreCase(sServerPort.Ptr, sServerPort.Len))
						{
							StringParser thePortParser(&theSubHeader);
							thePortParser.ConsumeUntil(NULL, StringParser::sDigitMask);
							fServerPort = thePortParser.ConsumeInteger(NULL);
						}
					}
				}
			}
		}
		else if (fHeaderRecvLen == kReqBufSize)
			return ENOBUFS; // This response is too big for us to handle!
	}
	
	while (fContentLength > fContentRecvLen)
	{
		UInt32 theContentRecvLen = 0;
		theErr = fSocket.Read(&fRecvContentBuffer[fContentRecvLen], fContentLength - fContentRecvLen, &theContentRecvLen);
		if (theErr != OS_NoErr)
		{
			fEventMask = EV_RE;
			return theErr;
		}
		fContentRecvLen += theContentRecvLen;		
	}
	
	
	// We're all done, reset all our state information.
	fContentRecvLen = 0;
	fHeaderRecvLen = 0;
	fReceivedResponse = false;
	fReceiveInProgress = false;
	fTransactionStarted = false;
	
	return OS_NoErr;
}

#define _RTSPCLIENTTESTING_ 0

#include "SocketUtils.h"
#include "OS.h"

#if _RTSPCLIENTTESTING_
int main(int argc, char * argv[]) 
{
	OS::Initialize();
	Socket::Initialize();
	SocketUtils::Initialize();

	UInt32 ipAddr = (UInt32)ntohl(::inet_addr("17.221.41.111"));
	UInt16 port = 554;
	StrPtrLen theURL("rtsp://17.221.41.111/mystery.mov");

	RTSPClient theClient(Socket::kNonBlockingSocketType);
	theClient.Set(ipAddr, port, theURL);
	
	OS_Error theErr = EINVAL;
	while (theErr != OS_NoErr)
	{
		theErr = theClient.SendDescribe();
		sleep(1);
	}
	Assert(theClient.GetStatus() == 200);
	theErr = EINVAL;
	while (theErr != OS_NoErr)
	{
		theErr = theClient.SendSetup(3, 6790);
		sleep(1);
	}
	printf("Server port: %d\n", theClient.GetServerPort());
	Assert(theClient.GetStatus() == 200);
	theErr = EINVAL;
	while (theErr != OS_NoErr)
	{
		theErr = theClient.SendSetup(4, 6792);
		sleep(1);
	}
	printf("Server port: %d\n", theClient.GetServerPort());
	Assert(theClient.GetStatus() == 200);
	theErr = EINVAL;
	while (theErr != OS_NoErr)
	{
		theErr = theClient.SendPlay();
		sleep(1);
	}
	Assert(theClient.GetStatus() == 200);
	theErr = EINVAL;
	while (theErr != OS_NoErr)
	{
		theErr = theClient.SendTeardown();
		sleep(1);
	}
	Assert(theClient.GetStatus() == 200);
}
#endif