/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		QTSSCallbacks.cpp

	Contains:	Implements QTSS Callback functions.
					
	$Log: QTSSCallbacks.cpp,v $
	Revision 1.3  2000/02/11 01:07:23  serenyi
	Server now supports cache control header and sends 304 Not Modified responses in appropriate situations
	
	Revision 1.2  2000/01/27 02:09:13  rtucker
	Merge changes from v81
	
	Revision 1.9  2000/01/26 21:32:15  serenyi
	Added new QTSS_RequestIdleTime callback, made it so the ClientSessionClosing role can register for async events
	
	Revision 1.8  2000/01/19 19:37:03  serenyi
	Moved to new NEW macro
	
	Revision 1.7  1999/12/21 01:44:03  lecroy
	*** empty log message ***
	
	Revision 1.6  1999/11/08 22:52:33  serenyi
	Updated code to reflect changes in QTSS API
	
	Revision 1.5  1999/10/27 17:55:02  serenyi
	Integrated Matthias's MacOS / GUSI changes
	
	Revision 1.4  1999/10/07 08:52:29  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.3  1999/09/30 18:45:17  serenyi
	Fixed #2392578 - Dictionary should not attempt to return attrID from AddAttribute (in QTSSDictionaryMap)
	
	Revision 1.2  1999/09/24 19:54:42  murata
	Added API call support QTSS_MilliSecsTo1970Secs
	
	Revision 1.1.1.1  1999/09/22 23:01:49  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:52  rtucker
	initial checkin from rrv14
	
	

	Created: Tue, Mar 2, 1999 @ 4:23 PM
*/

#include "QTSSCallbacks.h"
#include "QTSSModuleInterface.h"
#include "QTSSDictionary.h"
#include "QTSSStream.h"
#include "OSMemory.h"
#include "RTSPRequestInterface.h"
#include "RTPSession.h"
#include "OS.h"
#include "EventContext.h"
#include "Socket.h"

#include <sys/errno.h>

#pragma mark __MEMORY_ROUTINES__

void* 	QTSSCallbacks::QTSS_New(FourCharCode /*inMemoryIdentifier*/, UInt32 inSize)
{
	//
	// This callback is now deprecated because the server no longer uses FourCharCodes
	// for memory debugging. For clients that still use it, the default, non-debugging
	// version of New is used.
	
	//return OSMemory::New(inSize, inMemoryIdentifier, false);
	return OSMemory::New(inSize); 
}

void	QTSSCallbacks::QTSS_Delete(void* inMemory)
{
	OSMemory::Delete(inMemory);
}

void	QTSSCallbacks::QTSS_Milliseconds(SInt64* outMilliseconds)
{
	if (outMilliseconds != NULL)
		*outMilliseconds = OS::Milliseconds();
}

void	QTSSCallbacks::QTSS_ConvertToUnixTime(SInt64 *inQTSS_MilliSecondsPtr, time_t* outSecondsPtr)
{
	if ( (NULL != outSecondsPtr) && (NULL != inQTSS_MilliSecondsPtr) )
		*outSecondsPtr = OS::ConvertToUnixTime(*inQTSS_MilliSecondsPtr);
}


#pragma mark __STARTUP_ROUTINES__

QTSS_Error 	QTSSCallbacks::QTSS_AddRole(QTSS_Role inRole)
{
	QTSS_ModuleState* theState = (QTSS_ModuleState*)OSThread::GetMainThreadData();
	if (OSThread::GetCurrent() != NULL)
		theState = (QTSS_ModuleState*)OSThread::GetCurrent()->GetThreadData();
		
	// Roles can only be added before modules have had their Initialize role invoked.
	if ((theState == NULL) ||  (theState->curRole != QTSS_Register_Role))
		return QTSS_OutOfState;
		
	// There can only be one QTSS_RTSPRequest processing module
	if ((inRole == QTSS_RTSPRequest_Role) && (QTSSModule::HasRTSPRequestModule()))
		return QTSS_RequestFailed;
	
	return theState->curModule->AddRole(inRole);
}


#pragma mark __DICTIONARY_ROUTINES__

QTSS_Error	QTSSCallbacks::QTSS_AddAttribute(QTSS_ObjectType inType, const char* inName, QTSS_AttrFunctionPtr inFunctionPtr)
{
	QTSS_ModuleState* theState = (QTSS_ModuleState*)OSThread::GetMainThreadData();
	if (OSThread::GetCurrent() != NULL)
		theState = (QTSS_ModuleState*)OSThread::GetCurrent()->GetThreadData();

	// Attributes can only be added before modules have had their Initialize role invoked.
	if ((theState == NULL) || (theState->curRole != QTSS_Register_Role))
		return QTSS_OutOfState;

	UInt32 theDictionaryIndex = QTSSModuleInterface::GetDictionaryIndex(inType);
	if (theDictionaryIndex == QTSSModuleInterface::kIllegalDictionary)
		return QTSS_BadArgument;
		
	QTSSDictionaryMap* theMap = QTSSModuleInterface::GetDictionaryMap(theDictionaryIndex);
	return theMap->AddAttribute(inName, inFunctionPtr, true);
}

QTSS_Error	QTSSCallbacks::QTSS_IDForAttr(QTSS_ObjectType inType, const char* inTag, QTSS_AttributeID* outID)
{
	UInt32 theDictionaryIndex = QTSSModuleInterface::GetDictionaryIndex(inType);
	if (theDictionaryIndex == QTSSModuleInterface::kIllegalDictionary)
		return QTSS_BadArgument;
		
	return QTSSModuleInterface::GetDictionaryMap(theDictionaryIndex)->GetAttrID(inTag, outID);
}

QTSS_Error 	QTSSCallbacks::QTSS_GetValuePtr (QTSS_Object inDictionary, QTSS_AttributeID inID, UInt32 inIndex, void** outBuffer, UInt32* outLen)
{
	if (inDictionary == NULL)
		return QTSS_BadArgument;
	return ((QTSSDictionary*)inDictionary)->GetValue(inID, inIndex, outBuffer, outLen);
}


QTSS_Error 	QTSSCallbacks::QTSS_GetValue (QTSS_Object inDictionary, QTSS_AttributeID inID, UInt32 inIndex, void* ioBuffer, UInt32* ioLen)
{
	if (inDictionary == NULL)
		return QTSS_BadArgument;
	return ((QTSSDictionary*)inDictionary)->GetValue(inID, inIndex, ioBuffer, ioLen);
}

QTSS_Error 	QTSSCallbacks::QTSS_SetValue (QTSS_Object inDictionary, QTSS_AttributeID inID, UInt32 inIndex, const void* inBuffer,  UInt32 inLen)
{
	if (inDictionary == NULL)
		return QTSS_BadArgument;
	return ((QTSSDictionary*)inDictionary)->SetValue(inID, inIndex, inBuffer, inLen);
}


#pragma mark __STREAM_ROUTINES__

QTSS_Error 	QTSSCallbacks::QTSS_Write(QTSS_StreamRef inStream, void* inBuffer, UInt32 inLen, UInt32* outLenWritten, UInt32 inFlags)
{
	if (inStream == NULL)
		return QTSS_BadArgument;
	QTSS_Error theErr = ((QTSSStream*)inStream)->Write(inBuffer, inLen, outLenWritten, inFlags);
	
	// Server internally propogates POSIX errorcodes such as EAGAIN and ENOTCONN up to this
	// level. The API guarentees that no POSIX errors get returned, so we have QTSS_Errors
	// to replace them. So we have to replace them here.
	if (theErr == EAGAIN)
		return QTSS_WouldBlock;
	else if (theErr > 0)
		return QTSS_NotConnected;
	else
		return theErr;
}

QTSS_Error 	QTSSCallbacks::QTSS_WriteV(QTSS_StreamRef inStream, iovec* inVec, UInt32 inNumVectors, UInt32 inTotalLength, UInt32* outLenWritten)
{
	if (inStream == NULL)
		return QTSS_BadArgument;
	QTSS_Error theErr = ((QTSSStream*)inStream)->WriteV(inVec, inNumVectors, inTotalLength, outLenWritten);

	// Server internally propogates POSIX errorcodes such as EAGAIN and ENOTCONN up to this
	// level. The API guarentees that no POSIX errors get returned, so we have QTSS_Errors
	// to replace them. So we have to replace them here.
	if (theErr == EAGAIN)
		return QTSS_WouldBlock;
	else if (theErr > 0)
		return QTSS_NotConnected;
	else
		return theErr;
}

QTSS_Error 	QTSSCallbacks::QTSS_Flush(QTSS_StreamRef inStream)
{
	if (inStream == NULL)
		return QTSS_BadArgument;
	QTSS_Error theErr = ((QTSSStream*)inStream)->Flush();

	// Server internally propogates POSIX errorcodes such as EAGAIN and ENOTCONN up to this
	// level. The API guarentees that no POSIX errors get returned, so we have QTSS_Errors
	// to replace them. So we have to replace them here.
	if (theErr == EAGAIN)
		return QTSS_WouldBlock;
	else if (theErr > 0)
		return QTSS_NotConnected;
	else
		return theErr;
}


#pragma mark __SERVICE_ROUTINES__

QTSS_Error	QTSSCallbacks::QTSS_AddService(const char* inServiceName, QTSS_ServiceFunctionPtr inFunctionPtr)
{
	QTSS_ModuleState* theState = (QTSS_ModuleState*)OSThread::GetMainThreadData();
	if (OSThread::GetCurrent() != NULL)
		theState = (QTSS_ModuleState*)OSThread::GetCurrent()->GetThreadData();

	// This may happen if this callback is occurring on module-created thread
	if (theState == NULL)
		return QTSS_OutOfState;
		
	// Roles can only be added before modules have had their Initialize role invoked.
	if (theState->curRole != QTSS_Register_Role)
		return QTSS_OutOfState;

	return QTSSModuleInterface::GetDictionaryMap(QTSSModuleInterface::kServiceDictIndex)->
				AddAttribute(inServiceName, (QTSS_AttrFunctionPtr)inFunctionPtr, true);
}

QTSS_Error	QTSSCallbacks::QTSS_IDForService(const char* inTag, QTSS_ServiceID* outID)
{
	return QTSSModuleInterface::GetDictionaryMap(QTSSModuleInterface::kServiceDictIndex)->
				GetAttrID(inTag, outID);
}

QTSS_Error	QTSSCallbacks::QTSS_DoService(QTSS_ServiceID inID, QTSS_ServiceFunctionArgsPtr inArgs)
{
	// Make sure that the service ID is in fact valid
	
	QTSSDictionaryMap* theMap = QTSSModuleInterface::GetDictionaryMap(QTSSModuleInterface::kServiceDictIndex);
	if (!theMap->IsValidID(inID))
		return QTSS_IllegalService;
	
	// Get the service function	
	QTSS_ServiceFunctionPtr theFunction = (QTSS_ServiceFunctionPtr)theMap->GetAttrFunction(inID);

	// Invoke it, return the result.	
	return (theFunction)(inArgs);
}

#pragma mark __RTSP_ROUTINES__

QTSS_Error QTSSCallbacks::QTSS_SendRTSPHeaders(QTSS_RTSPRequestObject inRef)
{
	if (inRef == NULL)
		return QTSS_BadArgument;
		
	((RTSPRequestInterface*)inRef)->SendHeader();
	return QTSS_NoErr;
}

QTSS_Error QTSSCallbacks::QTSS_AppendRTSPHeader(QTSS_RTSPRequestObject inRef,
														QTSS_RTSPHeader inHeader,
														char* inValue,
														UInt32 inValueLen)
{
	if ((inRef == NULL) || (inValue == NULL))
		return QTSS_BadArgument;
	if (inHeader >= qtssNumHeaders)
		return QTSS_BadArgument;
	
	StrPtrLen theValue(inValue, inValueLen);
	((RTSPRequestInterface*)inRef)->AppendHeader(inHeader, &theValue);
	return QTSS_NoErr;
}

QTSS_Error QTSSCallbacks::QTSS_SendStandardRTSPResponse(QTSS_RTSPRequestObject inRTSPRequest,
															QTSS_Object inRTPInfo,
															UInt32 inFlags)
{
	if ((inRTSPRequest == NULL) || (inRTPInfo == NULL))
		return QTSS_BadArgument;
		
	switch (((RTSPRequestInterface*)inRTSPRequest)->GetMethod())
	{
		case qtssDescribeMethod:
			((RTPSession*)inRTPInfo)->SendDescribeResponse((RTSPRequestInterface*)inRTSPRequest);
			return QTSS_NoErr;
		case qtssSetupMethod:
		{
			// Because QTSS_SendStandardRTSPResponse supports sending a proper 304 Not Modified on a SETUP,
			// but a caller typically won't be adding a stream for a 304 response, we have the policy of
			// making the caller pass in the QTSS_ClientSessionObject instead. That means we need to do
			// different things here depending...
			if (((RTSPRequestInterface*)inRTSPRequest)->GetStatus() == qtssRedirectNotModified)
				(void)((RTPSession*)inRTPInfo)->DoSessionSetupResponse((RTSPRequestInterface*)inRTSPRequest);
			else
				((RTPStream*)inRTPInfo)->SendSetupResponse((RTSPRequestInterface*)inRTSPRequest);
			return QTSS_NoErr;
		}
		case qtssPlayMethod:
			((RTPSession*)inRTPInfo)->SendPlayResponse((RTSPRequestInterface*)inRTSPRequest, inFlags);
			return QTSS_NoErr;
		case qtssPauseMethod:
			((RTPSession*)inRTPInfo)->SendPauseResponse((RTSPRequestInterface*)inRTSPRequest);
			return QTSS_NoErr;
		case qtssTeardownMethod:
			((RTPSession*)inRTPInfo)->SendTeardownResponse((RTSPRequestInterface*)inRTSPRequest);
			return QTSS_NoErr;
	}
	return QTSS_BadArgument;
}


#pragma mark __RTP_ROUTINES__


QTSS_Error	QTSSCallbacks::QTSS_AddRTPStream(QTSS_ClientSessionObject inClientSession, QTSS_RTSPRequestObject inRTSPRequest, QTSS_RTPStreamObject* outStream, QTSS_AddStreamFlags inFlags)
{
	if ((inClientSession == NULL) || (inRTSPRequest == NULL) ||(outStream == NULL))
		return QTSS_BadArgument;
	return ((RTPSession*)inClientSession)->AddStream((RTSPRequestInterface*)inRTSPRequest, (RTPStream**)outStream, inFlags);
}

QTSS_Error	QTSSCallbacks::QTSS_Play(QTSS_ClientSessionObject inClientSession, QTSS_RTSPRequestObject inRTSPRequest, QTSS_PlayFlags inPlayFlags)
{
	if (inClientSession == NULL)
		return QTSS_BadArgument;
	return ((RTPSession*)inClientSession)->Play((RTSPRequestInterface*)inRTSPRequest, inPlayFlags);
}

QTSS_Error	QTSSCallbacks::QTSS_Pause(QTSS_ClientSessionObject inClientSession)
{
	if (inClientSession == NULL)
		return QTSS_BadArgument;
	((RTPSession*)inClientSession)->Pause();
	return QTSS_NoErr;
}

QTSS_Error	QTSSCallbacks::QTSS_Teardown(QTSS_ClientSessionObject inClientSession)
{
	if (inClientSession == NULL)
		return QTSS_BadArgument;
	
	((RTPSession*)inClientSession)->Teardown();
	return QTSS_NoErr;
}

#pragma mark __ASYNC_ROUTINES__

QTSS_Error	QTSSCallbacks::QTSS_CreateEventContext(int inFileDesc, QTSS_EventContextRef* outContext, void* inPrivate)
{
	EventContext* theEventContext = NEW EventContext(Socket::GetEventThread());
	theEventContext->InitNonBlocking(inFileDesc, inPrivate);
	*outContext = (QTSS_EventContextRef)theEventContext;
	return QTSS_NoErr;
}

QTSS_Error	QTSSCallbacks::QTSS_DestroyEventContext(QTSS_EventContextRef inContext)
{
	EventContext* theContext = (EventContext*)inContext;
	theContext->Cleanup();
	delete theContext;
	return QTSS_NoErr;
}

QTSS_Error	QTSSCallbacks::QTSS_RequestEvent(QTSS_EventContextRef inContext, int inEventMask)
{
	QTSS_ModuleState* theState = (QTSS_ModuleState*)OSThread::GetMainThreadData();
	if (OSThread::GetCurrent() != NULL)
		theState = (QTSS_ModuleState*)OSThread::GetCurrent()->GetThreadData();
		
	// This may happen if this callback is occurring on module-created thread
	if (theState == NULL)
		return QTSS_RequestFailed;
		
	if (theState->curTask == NULL)
		return QTSS_OutOfState;
		
	theState->eventRequested = true;
	EventContext* theContext = (EventContext*)inContext;
	theContext->SetTask(theState->curTask);
	theContext->RequestEvent(inEventMask);
	return QTSS_NoErr;
}
QTSS_Error	QTSSCallbacks::QTSS_SetIdleTimer(SInt64 inMsecToWait)
{
	QTSS_ModuleState* theState = (QTSS_ModuleState*)OSThread::GetMainThreadData();
	if (OSThread::GetCurrent() != NULL)
		theState = (QTSS_ModuleState*)OSThread::GetCurrent()->GetThreadData();
		
	// This may happen if this callback is occurring on module-created thread
	if (theState == NULL)
		return QTSS_RequestFailed;
		
	if (theState->curTask == NULL)
		return QTSS_OutOfState;
		
	theState->eventRequested = true;
	theState->idleTime = inMsecToWait;
	return QTSS_NoErr;
}