/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		QTSSDictionary.cpp

	Contains:	Implementation of object defined in QTSSDictionary.h
					
	$Log: QTSSDictionary.cpp,v $
	Revision 1.2  2000/01/27 02:09:13  rtucker
	Merge changes from v81
	
	Revision 1.12  2000/01/19 19:37:04  serenyi
	Moved to new NEW macro
	
	Revision 1.11  1999/12/21 01:44:03  lecroy
	*** empty log message ***
	
	Revision 1.10  1999/11/15 01:20:48  serenyi
	Added 4-char codes for mem debugging, merged in solaris diffs
	
	Revision 1.9  1999/11/08 22:52:34  serenyi
	Updated code to reflect changes in QTSS API
	
	Revision 1.8  1999/11/03 23:33:33  serenyi
	Got rid of expiration, added new errorcode
	
	Revision 1.7  1999/10/27 17:55:02  serenyi
	Integrated Matthias's MacOS / GUSI changes
	
	Revision 1.6  1999/10/07 08:52:30  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.5  1999/10/06 17:26:08  rtucker
	never mind, index is unsigned.  users of Setvalue shoudl be sure to
	pass unsigned vals to avoid <0 indexes being converted to large
	unsigned values.
	
	Revision 1.4  1999/10/06 17:17:37  rtucker
	added index checking to SetValue
	
	Revision 1.3  1999/09/30 18:45:18  serenyi
	Fixed #2392578 - Dictionary should not attempt to return attrID from AddAttribute (in QTSSDictionaryMap)
	
	Revision 1.2  1999/09/30 00:29:23  serenyi
	Improved API validation code
	
	Revision 1.1.1.1  1999/09/22 23:01:50  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:53  rtucker
	initial checkin from rrv14
	
	

	Created: Tue, Mar 2, 1999 @ 4:23 PM
*/


#include "QTSSDictionary.h"

#include <string.h>
#include <stdio.h>

#include "MyAssert.h"
#include "OSMemory.h"

#include <sys/errno.h>


#pragma mark __QTSS_DICTIONARY__

QTSSDictionary::QTSSDictionary(QTSSDictionaryMap* inMap, OSMutex* inMutex) 
: fNumAttributes(inMap->GetNumAttrs()), fMap(inMap), fMutexP(inMutex)
{
	fAttributes = NEW DictValueElement[fNumAttributes];
}


QTSS_Error QTSSDictionary::GetValue(QTSS_AttributeID inAttrID, UInt32 inIndex,
											void** outValueBuffer, UInt32* outValueLen,
											Bool16 isInternal)
{
	// This function gets called by modules, so do good errorchecking
	
	if ((inAttrID < 0) || ((UInt32)inAttrID >= fNumAttributes))
		return QTSS_BadArgument;
	if ((outValueBuffer == NULL) || (outValueLen == NULL))
		return QTSS_BadArgument;
	if ((!isInternal) && (!fMap->IsPreemptiveSafe(inAttrID)))
		return QTSS_NotPreemptiveSafe;
	// An iterated attribute cannot have a param retrieval function
	if ((inIndex > 0) && (fMap->GetAttrFunction(inAttrID) != NULL))
		return QTSS_BadIndex;
	// Check to make sure the index parameter is legal
	if (inIndex > fAttributes[inAttrID].fNumAttributes)
		return QTSS_BadIndex;
	
	// Retrieve the parameter
	char* theBuffer = fAttributes[inAttrID].fAttributeData.Ptr;
	*outValueLen = fAttributes[inAttrID].fAttributeData.Len;
	
	if ((*outValueLen == 0) && (fMap->GetAttrFunction(inAttrID) != NULL))
	{
		// If the parameter doesn't have a value assigned yet, and there is an attribute
		// retrieval function provided, invoke that function now.
		
		theBuffer = (char*)fMap->GetAttrFunction(inAttrID)(this, outValueLen);

		//If the param retrieval function didn't return an explicit value for this attribute,
		//refetch the parameter out of the array, in case the function modified it.
		
		if (*outValueLen == 0)
		{
			theBuffer = fAttributes[inAttrID].fAttributeData.Ptr;
			*outValueLen = fAttributes[inAttrID].fAttributeData.Len;
		}
		
	}
#if DEBUG
	else
		// Make sure we aren't outside the bounds of attribute memory
		Assert(fAttributes[inAttrID].fAllocatedLen >=
			(fAttributes[inAttrID].fAttributeData.Len * (fAttributes[inAttrID].fNumAttributes + 1)));
#endif

	// Return an error if there is no data for this attribute
	if (*outValueLen == 0)
		return QTSS_ValueNotFound;
		
	// Adjust for the iteration, if any
	theBuffer += fAttributes[inAttrID].fAttributeData.Len * inIndex;
	*outValueBuffer = theBuffer;
	
	return QTSS_NoErr;
}

QTSS_Error QTSSDictionary::GetValue(QTSS_AttributeID inAttrID, UInt32 inIndex,
											void* ioValueBuffer, UInt32* ioValueLen)
{
	void* tempValueBuffer;
	UInt32 tempValueLen = 0;

	if (ioValueLen == NULL)
		return QTSS_BadArgument;
	
	// If there is a mutex, lock it and get a pointer to the proper attribute
	OSMutexLocker locker(fMutexP);
	QTSS_Error theErr = this->GetValue(inAttrID, inIndex, &tempValueBuffer,
											&tempValueLen, true);

	if (theErr == QTSS_NoErr)
	{
		// If caller provided a buffer that's too small for this attribute, report that error
		if (tempValueLen > *ioValueLen)
			theErr = QTSS_NotEnoughSpace;
			
		// Only copy out the attribute if the buffer is big enough
		if ((ioValueBuffer != NULL) && (theErr == QTSS_NoErr))
			::memcpy(ioValueBuffer, tempValueBuffer, tempValueLen);
			
		// Always set the ioValueLen to be the actual length of the attribute.
		*ioValueLen = tempValueLen;
	}
	return theErr;
}

QTSS_Error QTSSDictionary::SetValue(QTSS_AttributeID inAttrID, UInt32 inIndex,
													const void* inBuffer,  UInt32 inLen,
													Bool16 obeyReadOnly)
{
	if ((inAttrID < 0) || ((UInt32)inAttrID >= fNumAttributes))
		return QTSS_BadArgument;
	if ((obeyReadOnly) && (!fMap->IsWriteable(inAttrID)))
		return QTSS_ReadOnly;
	if ((fMap->GetAttrFunction(inAttrID) != NULL) && (inIndex > 0))
		return QTSS_BadIndex;
	if ((inBuffer == NULL) && (inLen > 0))
		return QTSS_BadArgument;
		
	// If there is a mutex, make this action atomic.
	OSMutexLocker locker(fMutexP);
	
	// If this attribute is iterated, this new value
	// must be the same size as all the others.
	if ((fAttributes[inAttrID].fNumAttributes > 0) &&
		(inLen != fAttributes[inAttrID].fAttributeData.Len))
		return QTSS_BadArgument;
		
	if ((inLen * (inIndex + 1)) > fAttributes[inAttrID].fAllocatedLen)
	{
		// We need to reallocate this buffer.
		UInt32 theLen = 2 * (inLen * (inIndex + 1));// Allocate 2wice as much as we need
		char* theNewBuffer = NEW char[theLen];
		// Copy out the old attribute data
		::memcpy(theNewBuffer, fAttributes[inAttrID].fAttributeData.Ptr,
										fAttributes[inAttrID].fAllocatedLen);
		
		// Now get rid of the old stuff. Delete the buffer
		// if it was already allocated internally
		if (fAttributes[inAttrID].fAllocatedInternally)
			delete [] fAttributes[inAttrID].fAttributeData.Ptr;
		
		// Finally, update this attribute structure with all the new values.
		fAttributes[inAttrID].fAttributeData.Ptr = theNewBuffer;
		fAttributes[inAttrID].fAllocatedLen = theLen;
		fAttributes[inAttrID].fAllocatedInternally = true;
	}
		
	// At this point, we should always have enough space to write what we want
	Assert(fAttributes[inAttrID].fAllocatedLen >= (inLen * (inIndex + 1)));
	
	// Set the number of attributes to be proper
	if (inIndex > fAttributes[inAttrID].fNumAttributes)
		fAttributes[inAttrID].fNumAttributes = inIndex;
		
	// Copy the new data to the right place in our data buffer
	::memcpy(fAttributes[inAttrID].fAttributeData.Ptr + (inLen * inIndex), inBuffer, inLen);
	fAttributes[inAttrID].fAttributeData.Len = inLen;
	return QTSS_NoErr;
}

UInt32	QTSSDictionary::GetNumValues(QTSS_AttributeID inAttrID)
{
	// fNumAttributes is offset from 0, we need to convert this to a
	// normal count of how many values there are.
	
	if (fAttributes[inAttrID].fAttributeData.Len == 0)
		return 0;
	else
		return fAttributes[inAttrID].fNumAttributes + 1;
}

void	QTSSDictionary::SetVal(	QTSS_AttributeID inAttrID,
									void* inValueBuffer,
									UInt32 inBufferLen)
{ 
	Assert(inAttrID >= 0);
	Assert((UInt32)inAttrID < fNumAttributes);
	fAttributes[inAttrID].fAttributeData.Ptr = (char*)inValueBuffer;
	fAttributes[inAttrID].fAttributeData.Len = inBufferLen;
	fAttributes[inAttrID].fAllocatedLen = inBufferLen;
	
	// This function doesn't set fNumAttributes & fAllocatedInternally.
}

void	QTSSDictionary::SetEmptyVal(QTSS_AttributeID inAttrID, void* inBuf, UInt32 inBufLen)
{
	Assert(inAttrID >= 0);
	Assert((UInt32)inAttrID < fNumAttributes);
	fAttributes[inAttrID].fAttributeData.Ptr = (char*)inBuf;
	fAttributes[inAttrID].fAllocatedLen = inBufLen;

}


#pragma mark __QTSS_DICTIONARY_MAP__

QTSSDictionaryMap::QTSSDictionaryMap(UInt32 inNumReservedAttrs)
: 	fNextAvailableID(inNumReservedAttrs), fAttrArraySize(inNumReservedAttrs)
{
	if (fAttrArraySize < kMinArraySize)
		fAttrArraySize = kMinArraySize;
	fAttrArray = NEW QTSSDictionaryMap::AttrInfo[fAttrArraySize];
}

QTSS_Error QTSSDictionaryMap::AddAttribute(	const char* inAttrName,
												QTSS_AttrFunctionPtr inFuncPtr,
												Bool16 isPreemptiveSafe)
{
	if (::strlen(inAttrName) > QTSS_MAX_ATTRIBUTE_NAME_SIZE)
		return QTSS_BadArgument;
	
	if (fAttrArraySize == fNextAvailableID)
	{
		// If there currently isn't an attribute array, or if the current array
		// is full, allocate a new array and copy all the old stuff over to the new array.
		
		UInt32 theNewArraySize = fAttrArraySize * 2;
		if (theNewArraySize == 0)
			theNewArraySize = kMinArraySize;
		
		QTSSDictionaryMap::AttrInfo* theNewArray = NEW QTSSDictionaryMap::AttrInfo[theNewArraySize];
		if (fAttrArray != NULL)
		{
			::memcpy(theNewArray, fAttrArray, sizeof(QTSSDictionaryMap::AttrInfo) * fAttrArraySize);
			delete [] fAttrArray;
		}
		fAttrArray = theNewArray;
		fAttrArraySize = theNewArraySize;
	}
	
	QTSS_AttributeID theID = fNextAvailableID;
	fNextAvailableID++;

	// Copy the information into the first available element
	// Currently, all attributes added in this fashion are always writeable
	this->SetAttribute(theID, inAttrName, inFuncPtr, isPreemptiveSafe, true);	
	return QTSS_NoErr;
}

void QTSSDictionaryMap::SetAttribute(	QTSS_AttributeID inID,
										const char* inAttrName,
										QTSS_AttrFunctionPtr inFuncPtr,
										Bool16 isPreemptiveSafe,
										Bool16 isWriteable)
{
	Assert(inID >= 0);
	Assert(::strlen(inAttrName) < QTSS_MAX_ATTRIBUTE_NAME_SIZE);

	//Copy the information into the first available element
	
	::strcpy(&fAttrArray[inID].fAttrName[0], inAttrName);
	fAttrArray[inID].fFuncPtr = inFuncPtr;
	fAttrArray[inID].fIsPreemptiveSafe = isPreemptiveSafe;	
	fAttrArray[inID].fIsWriteable = isWriteable;	
}


QTSS_Error QTSSDictionaryMap::GetAttrID(const char* inAttrName, QTSS_AttributeID* outID)
{
	if (outID == NULL)
		return QTSS_BadArgument;

	for (UInt32 count = 0; count < fNextAvailableID; count++)
	{
		if (::strcmp(&fAttrArray[count].fAttrName[0], inAttrName) == 0)
		{
			*outID = count;
			return QTSS_NoErr;
		}	
	}
	return QTSS_BadArgument;
}

#if __DICTIONARY_TESTING__
static void* TestAttrFunc(QTSS_Object inServer, UInt32* outLen);
static void* TestAttrFunc2(QTSS_Object inServer, UInt32* outLen);

void* TestAttrFunc(QTSS_Object inServer, UInt32* outLen)
{
	return NULL;
}
void* TestAttrFunc2(QTSS_Object inServer, UInt32* outLen)
{
	return NULL;
}

void QTSSDictionary::Test()
{
	QTSSDictionaryMap theMap(0);
	Assert(theMap.GetNumAttrs() == 0);
	UInt32 id1;
	QTSS_Error theErr = theMap.AddAttribute("foo", NULL, true, &id1);
	Assert(theErr == QTSS_NoErr);
	UInt32 id2;
	theErr = theMap.AddAttribute("foo2", TestAttrFunc, false, &id2);
	Assert(theErr == QTSS_NoErr);
	UInt32 id3;
	theErr = theMap.AddAttribute("foo3", TestAttrFunc2, false, &id3);
	Assert(theErr == QTSS_NoErr);
	UInt32 id4;
	theErr = theMap.AddAttribute("foo4", TestAttrFunc2, false, &id4);
	Assert(theErr == QTSS_NoErr);
	UInt32 id5;
	theErr = theMap.AddAttribute("foo4dfdsfdfaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa", TestAttrFunc2, false, &id5);
	Assert(theErr == QTSS_BadArgument);
	theErr = theMap.AddAttribute("foo4dfdsfdfaaaaaaaaaaaaaa", TestAttrFunc2, false, NULL);
	Assert(theErr == QTSS_BadArgument);
	UInt32 id6;
	theErr = theMap.AddAttribute("foo6", NULL, false, &id6);
	Assert(theErr == QTSS_NoErr);	
	theMap.SetAttribute(id4, "foo7", NULL, false, false);
	Assert(theMap.GetNumAttrs() == 5);
	
	QTSSDictionary theDict(&theMap);
	char boofer[100];
	UInt32 len = 0;
	
	theErr = theDict.SetValue(id2, 2, "poopie", 6);
	Assert(theErr == QTSS_BadIndex);
	theErr = theDict.SetValue(id2, 0, "poopie", 6);
	Assert(theErr == QTSS_NoErr);
	theErr = theDict.SetValue(id2, 0, "heinosity", 9);
	Assert(theErr == QTSS_NoErr);
	theErr = theDict.SetValue(id2, 0, "poopieheinosity", 15);
	Assert(theErr == QTSS_NoErr);
	
	theErr = theDict.GetValue(342, 3, (void**)NULL, (UInt32*)NULL);
	Assert(theErr == QTSS_BadArgument);
	theErr = theDict.GetValue(-2, 0, boofer, &len);
	Assert(theErr == QTSS_BadArgument);
	theErr = theDict.GetValue(-1, 0, boofer, &len);
	Assert(theErr == QTSS_BadArgument);

	theErr = theDict.GetValue(id1, 0, boofer, &len);
	Assert(theErr == QTSS_BadIndex);
	theErr = theDict.SetValue(id1, 2, "foop", 4);
	Assert(theErr == QTSS_NoErr);
	theErr = theDict.SetValue(id1, 1, "foopie", 6);
	Assert(theErr == QTSS_BadArgument);
	theErr = theDict.SetValue(id1, 0, "foopie", 6);
	Assert(theErr == QTSS_BadArgument);
	theErr = theDict.SetValue(id4, 0, "foopie", 6);
	Assert(theErr == EPERM);
	theErr = theDict.SetValue(id1, 0, "doop", 4);
	Assert(theErr == QTSS_NoErr);
	
	UInt32 theLen = 0;
	char* theBuffer = NULL;
	theErr = theDict.GetValue(id1, 3, (void**)&theBuffer, (UInt32*)&theLen);
	Assert(theErr == QTSS_BadIndex);
	theErr = theDict.GetValue(id1, 1, (void**)&theBuffer, (UInt32*)&theLen);
	Assert(theErr == QTSS_NoErr);
	theErr = theDict.GetValue(id1, 0, (void**)&theBuffer, (UInt32*)&theLen);
	Assert(theErr == QTSS_NoErr);
	Assert(theLen == 4);
	Assert(::strncmp(theBuffer, "doop", 4) == 0);
	theErr = theDict.GetValue(id1, 2, (void**)&theBuffer, (UInt32*)&theLen);
	Assert(theErr == QTSS_NoErr);
	Assert(theLen == 4);
	Assert(::strncmp(theBuffer, "foop", 4) == 0);
	
	char theBuf[10];
	theLen = 3;
	theErr = theDict.GetValue(id1, 2, (void*)theBuf, (UInt32*)&theLen);
	Assert(theErr == E2BIG);
	Assert(theLen == 4);
	theErr = theDict.GetValue(id1, 2, (void*)NULL, (UInt32*)&theLen);
	Assert(theErr == QTSS_NoErr);
	Assert(theLen == 4);
	theErr = theDict.GetValue(id1, 2, (void*)theBuf, (UInt32*)NULL);
	Assert(theErr == QTSS_BadArgument);
	theLen = 6;
	theErr = theDict.GetValue(id1, 2, theBuf, &theLen);
	Assert(theErr == QTSS_NoErr);
	Assert(theLen == 4);
	Assert(::strncmp(theBuf, "foop", 4) == 0);
	theErr = theDict.GetValue(id1, 0, theBuf, &theLen);
	Assert(theErr == QTSS_NoErr);
	Assert(theLen == 4);
	Assert(::strncmp(theBuf, "doop", 4) == 0);

	theErr = theDict.GetValue(id2, 1, theBuf, &theLen);
	Assert(theErr == QTSS_BadIndex);

	theLen = 15;
	theErr = theDict.GetValue(id2, 0, theBuf, &theLen);
	Assert(theErr == QTSS_NoErr);
	Assert(theLen == 15);
	Assert(::strncmp(theBuf, "poopieheinosity", 15) == 0);

	theErr = theDict.GetValue(id2, 0, (void**)&theBuffer, (UInt32*)&theLen);
	Assert(theErr == EPERM);
}
#endif
