/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		QTSSDictionary.h

	Contains:	Definitions of two classes: QTSSDictionary and QTSSDictionaryMap.
				Collectively, these classes implement the "dictionary" APIs in QTSS
				API. A QTSSDictionary corresponds to a QTSS_Object,
				a QTSSDictionaryMap corresponds to a QTSS_ObjectType.
					
	$Log: QTSSDictionary.h,v $
	Revision 1.1.1.1  2000/01/17 23:36:03  wsanchez
	Import of StreamingServer-2.0 (Apple)
	
	Revision 1.6  1999/12/21 01:44:03  lecroy
	*** empty log message ***
	
	Revision 1.5  1999/11/08 22:52:34  serenyi
	Updated code to reflect changes in QTSS API
	
	Revision 1.4  1999/10/07 08:52:43  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.3  1999/09/30 18:45:18  serenyi
	Fixed #2392578 - Dictionary should not attempt to return attrID from AddAttribute (in QTSSDictionaryMap)
	
	Revision 1.2  1999/09/30 00:29:24  serenyi
	Improved API validation code
	
	Revision 1.1.1.1  1999/09/22 23:01:50  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:53  rtucker
	initial checkin from rrv14
	
	

	Created: Tue, Mar 2, 1999 @ 4:23 PM
*/



#ifndef _QTSSDICTIONARY_H_
#define _QTSSDICTIONARY_H_

#include <stdlib.h>

#include "QTSS.h"
#include "OSHeaders.h"
#include "OSMutex.h"
#include "StrPtrLen.h"
#include "MyAssert.h"
#include "QTSSStream.h"

class QTSSDictionaryMap;

#define __DICTIONARY_TESTING__ 0


class QTSSDictionary : public QTSSStream
{
	public:
	
		//
		// CONSTRUCTOR / DESTRUCTOR
		
		QTSSDictionary(QTSSDictionaryMap* inMap, OSMutex* inMutex = NULL);
		virtual ~QTSSDictionary() { delete [] fAttributes; }
		
		//
		// QTSS API CALLS
		
		// This version of GetAttribute copies the element into a buffer provided by the caller
		// Returns: 	QTSS_BadArgument, QTSS_NotPreemptiveSafe (if attribute is not preemptive safe),
		//				QTSS_BadIndex (if inIndex is bad)
		QTSS_Error GetValue(QTSS_AttributeID inAttrID, UInt32 inIndex, void* ioValueBuffer, UInt32* ioValueLen);

		//This version of GetAttribute returns a pointer to the internal buffer for the attribute.
		//Only usable if the attribute is preemptive safe.
		//
		// Returns: 	Same as above, but also QTSS_NotEnoughSpace, if value is too big for buffer.
		QTSS_Error GetValue(QTSS_AttributeID inAttrID, UInt32 inIndex, void** outValueBuffer, UInt32* outValueLen)
						{ return GetValue(inAttrID, inIndex, outValueBuffer, outValueLen, false); }
		
		// Returns:		QTSS_BadArgument, QTSS_ReadOnly (if attribute is read only),
		//				QTSS_BadIndex (attempt to set indexed parameter with param retrieval)
		QTSS_Error SetValue(QTSS_AttributeID inAttrID, UInt32 inIndex,
							const void* inBuffer,  UInt32 inLen, Bool16 obeyReadOnly = true);
		
		//
		// ACCESSORS
		
		QTSSDictionaryMap*	GetDictionaryMap() { return fMap; }
		
		// Returns the number of values associated with a given attribute
		UInt32				GetNumValues(QTSS_AttributeID inAttrID);
		
		// Meant only for internal server use. Does no error checking,
		// doesn't invoke the param retrieval function.
		StrPtrLen*	GetValue(QTSS_AttributeID inAttrID) 
					{
					 	Assert(inAttrID >= 0); Assert(inAttrID < (SInt32)fNumAttributes);
						return &fAttributes[inAttrID].fAttributeData;
					}

		//
		// MODIFIERS

		// These functions are meant to be used by the server when it is setting up the
		// dictionary attributes. They do no error checking.

		// They don't set fNumAttributes & fAllocatedInternally.
		void	SetVal(QTSS_AttributeID inAttrID, void* inValueBuffer, UInt32 inBufferLen);
		void	SetVal(QTSS_AttributeID inAttrID, StrPtrLen* inNewValue)
					{ this->SetVal(inAttrID, inNewValue->Ptr, inNewValue->Len); }

		// Call this if you want to assign empty storage to an attribute
		void	SetEmptyVal(QTSS_AttributeID inAttrID, void* inBuf, UInt32 inBufLen);
		
#if __DICTIONARY_TESTING__
		static void	Test(); // API test for these objects
#endif

	private:
	
		// Utility routine used by the two external flavors of GetValue
		QTSS_Error GetValue(QTSS_AttributeID inAttrID, UInt32 inIndex,
											void** outValueBuffer, UInt32* outValueLen,
											Bool16 isInternal);

		struct DictValueElement
		{
			// This stores all necessary information for each attribute value.
			
			DictValueElement() : 	fAllocatedLen(0), fNumAttributes(0),
									fAllocatedInternally(false) {}
			~DictValueElement() { if (fAllocatedInternally) delete [] fAttributeData.Ptr; }
			
			StrPtrLen	fAttributeData;	// The data
			UInt32		fAllocatedLen;	// How much space do we have allocated?
			UInt32		fNumAttributes;	// If this is an iterated attribute, how many?
			Bool16		fAllocatedInternally; //Should we delete this memory?
		};
		
		UInt32				fNumAttributes;
		DictValueElement*	fAttributes;
		QTSSDictionaryMap* 	fMap;
		OSMutex*			fMutexP;
};

class QTSSDictionaryMap
{
	public:
	
		// There must be one of these for each dictionary type. This object
		// maps Attribute names to attribute IDs for a given dictionary type.
		// It also stores all related attribute information, such as whether
		// the attribute is read only or read write.

		struct AttrInfo
		{
			// This is all the relevent information for each dictionary
			// attribute.
			
			char 					fAttrName[QTSS_MAX_ATTRIBUTE_NAME_SIZE + 1];
			QTSS_AttrFunctionPtr 	fFuncPtr;
			Bool16 					fIsPreemptiveSafe;
			Bool16					fIsWriteable;
		};
		
		//
		// CONSTRUCTOR / DESTRUCTOR
		
		QTSSDictionaryMap(UInt32 inNumReservedAttrs);

		// Doesn't free up any internally allocated memory!
		~QTSSDictionaryMap(){}


		//
		// QTSS API CALLS
				
		// Both functions either return QTSS_BadArgument or QTSS_NoErr
		QTSS_Error 		GetAttrID(const char* inAttrName, QTSS_AttributeID* outID);
		QTSS_Error 		AddAttribute(const char* inAttrName, QTSS_AttrFunctionPtr inFuncPtr,
														Bool16 isPreemptiveSafe);

		// ACCESSORS
		
		// These functions do no error checking. Be careful.

		UInt32			GetNumAttrs() 	{ return fNextAvailableID; }
		
		Bool16					IsPreemptiveSafe(QTSS_AttributeID inAttrID) 
			{ Assert(inAttrID < (QTSS_AttributeID)fNextAvailableID); return fAttrArray[inAttrID].fIsPreemptiveSafe; }

		Bool16					IsWriteable(QTSS_AttributeID inAttrID) 
			{ Assert(inAttrID < (QTSS_AttributeID)fNextAvailableID); return fAttrArray[inAttrID].fIsWriteable; }

		QTSS_AttrFunctionPtr	GetAttrFunction(QTSS_AttributeID inAttrID)
			{ Assert(inAttrID < (QTSS_AttributeID)fNextAvailableID); return fAttrArray[inAttrID].fFuncPtr; }
									
		Bool16					IsValidID(QTSS_AttributeID inAttrID) { return ((inAttrID < (SInt32)fNextAvailableID) && (inAttrID >= 0)); }

		// These functions allow you to iterate through the entire map.
		// inIndex may be a value from 0 -> GetNumAttrs()
		char*	GetAttrName(UInt32 inIndex)
			{ Assert(inIndex < fNextAvailableID); return &fAttrArray[inIndex].fAttrName[0]; }
		UInt32	GetAttrID(UInt32 inIndex)
			{ Assert(inIndex < fNextAvailableID); return inIndex; }
		

		// MODIFIERS
		
		// Sets this attribute ID to have this information
		
		void		SetAttribute(QTSS_AttributeID inID, const char* inAttrName, QTSS_AttrFunctionPtr inFuncPtr,
										Bool16 isPreemptiveSafe, Bool16 isWriteable);
		
	private:
	
		enum
		{
			kMinArraySize = 20
		};

		UInt32 							fNextAvailableID;
		UInt32 							fAttrArraySize;
		AttrInfo*					 	fAttrArray;
};

#endif
