/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		QTSSErrorLogModule.cpp

	Contains:	Implementation of object defined in .h file.

	$Log: QTSSErrorLogModule.cpp,v $
	Revision 1.2  2000/01/27 02:09:13  rtucker
	Merge changes from v81
	
	Revision 1.5  2000/01/19 19:37:04  serenyi
	Moved to new NEW macro
	
	Revision 1.4  1999/12/21 01:44:03  lecroy
	*** empty log message ***
	
	Revision 1.3  1999/12/17 02:07:51  murata
	Use "Darwin Streaming" on non MacOSX systems
	
	Revision 1.2  1999/10/07 08:52:30  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.1.1.1  1999/09/22 23:01:50  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:53  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:08:28  ds
	Created
	

*/

#include <string.h>
#include "QTSSErrorLogModule.h"
#include "QTSSMessages.h"
#include "QTSSRollingLog.h"
#include "QTSServerInterface.h"
#include "OSMemory.h"

// QTSSERRORLOG CLASS DEFINITION

class QTSSErrorLog : public QTSSRollingLog
{
	public:
	
		QTSSErrorLog() : QTSSRollingLog() {}
		virtual ~QTSSErrorLog() {}
	
		virtual char* GetLogName() 	{ return QTSServerInterface::GetServer()->GetPrefs()->GetErrorLogName();}
		
		virtual char* GetLogDir() 	{ return QTSServerInterface::GetServer()->GetPrefs()->GetErrorLogDir();}
		
		virtual UInt32 GetRollIntervalInDays()	{ return QTSServerInterface::GetServer()->GetPrefs()->GetErrorRollIntervalInDays();}
		
		virtual UInt32 GetMaxLogBytes()	{ return QTSServerInterface::GetServer()->GetPrefs()->GetMaxErrorLogBytes();}
	
};

// STATIC DATA

static OSMutex*				sLogMutex = NULL;//Log module isn't reentrant
static QTSSErrorLog*		sErrorLog = NULL;

// STATIC FUNCTIONS

// The dispatch function for this module
static QTSS_Error 	QTSSErrorLogModuleDispatch(QTSS_Role inRole, QTSS_RoleParamPtr inParamBlock);

// A service routine allowing other modules to roll the log
static QTSS_Error 	RollErrorLog(QTSS_ServiceFunctionArgsPtr inArgs);

static QTSS_Error 		Register();
static QTSS_Error		Shutdown();
		
static QTSS_Error		LogError(QTSS_RoleParamPtr inParamBlock);
static void					CheckErrorLogState();


// FUNCTION IMPLEMENTATIONS

QTSS_Error QTSSErrorLogModule_Main(void* inPrivateArgs)
{
	return _stublibrary_main(inPrivateArgs, QTSSErrorLogModuleDispatch);
}


QTSS_Error QTSSErrorLogModuleDispatch(QTSS_Role inRole, QTSS_RoleParamPtr inParamBlock)
{
	switch (inRole)
	{
		case QTSS_Register_Role:
			return Register();
		case QTSS_ErrorLog_Role:
			return LogError(inParamBlock);
		case QTSS_Shutdown_Role:
			return Shutdown();
	}
	return QTSS_NoErr;
}


// ROLE METHODS

QTSS_Error Register()
{
	sLogMutex = new OSMutex();
	
	// Do role & service setup
	
	(void)QTSS_AddRole(QTSS_ErrorLog_Role);
	(void)QTSS_AddRole(QTSS_Shutdown_Role);
	
	(void)QTSS_AddService("RollErrorLog", &RollErrorLog);
	
	// Unlike most modules, all initialization for this module happens in
	// the register role. This is so that this error log can begin logging
	// errors ASAP.
	
	CheckErrorLogState();
		
	//format a date for the startup time
	char theDateBuffer[QTSSRollingLog::kMaxDateBufferSizeInBytes];
	Bool16 result = QTSSRollingLog::FormatDate(theDateBuffer);
	
	char tempBuffer[1024];
	if (result)
#ifndef __MacOSX__
		::sprintf(tempBuffer, "#Darwin Streaming Startup STARTUP %s\n", theDateBuffer);
#else
		::sprintf(tempBuffer, "#QuickTime Streaming Startup STARTUP %s\n", theDateBuffer);
#endif

	// log startup message to error log as well.
	if ((result) && (sErrorLog != NULL))
		sErrorLog->WriteToLog(tempBuffer, !kAllowLogToRoll);
	
	return QTSS_NoErr;
}

QTSS_Error Shutdown()
{
	//log shutdown message
	//format a date for the shutdown time
	char theDateBuffer[QTSSRollingLog::kMaxDateBufferSizeInBytes];
	Bool16 result = QTSSRollingLog::FormatDate(theDateBuffer);
	
	char tempBuffer[1024];
	if (result)
#ifndef __MacOSX__
		::sprintf(tempBuffer, "#Darwin Streaming Shutdown SHUTDOWN %s\n", theDateBuffer);
#else
		::sprintf(tempBuffer, "#QuickTime Streaming Shutdown SHUTDOWN %s\n", theDateBuffer);
#endif

	if ((result) && (sErrorLog != NULL))
		sErrorLog->WriteToLog(tempBuffer, !kAllowLogToRoll);
		
	return QTSS_NoErr;
}


QTSS_Error LogError(QTSS_RoleParamPtr inParamBlock)
{
	Assert(NULL != inParamBlock->errorParams.inBuffer);
	if (inParamBlock->errorParams.inBuffer == NULL)
		return QTSS_NoErr;
		
	QTSSPrefs* thePrefs = QTSServerInterface::GetServer()->GetPrefs();
		
	OSMutexLocker locker(sLogMutex);
	if (thePrefs->GetErrorLogVerbosity() >= inParamBlock->errorParams.inVerbosity)
	{
		//The error logger is the bottleneck for any and all messages printed by the server.
		//For debugging purposes, these messages can be printed to stdout as well.
		if (thePrefs->IsScreenLoggingEnabled())
			::printf("%s\n", inParamBlock->errorParams.inBuffer);
		
		CheckErrorLogState();
		
		if (sErrorLog == NULL)
			return QTSS_NoErr;
			
		//timestamp the error
		char theDateBuffer[QTSSRollingLog::kMaxDateBufferSizeInBytes];
		Bool16 result = QTSSRollingLog::FormatDate(theDateBuffer);
		//for now, just ignore the error.
		if (!result)
			theDateBuffer[0] = '\0';

		Assert(strlen(theDateBuffer) + ::strlen(inParamBlock->errorParams.inBuffer) < 1024);
		char tempBuffer[1024];
		::sprintf(tempBuffer, "%s: %s\n", theDateBuffer, inParamBlock->errorParams.inBuffer);
		
		sErrorLog->WriteToLog(tempBuffer, kAllowLogToRoll);
	}
	return QTSS_NoErr;
}


void CheckErrorLogState()
{
	//this function makes sure the logging state is in synch with the preferences.
	//extern variable declared in QTSSPreferences.h

	QTSSPrefs* thePrefs = QTSServerInterface::GetServer()->GetPrefs();

	//check error log.
	if ((NULL == sErrorLog) && (thePrefs->IsErrorLogEnabled()))
	{
		sErrorLog = NEW QTSSErrorLog();
		sErrorLog->EnableLog();
	}

	if ((NULL != sErrorLog) && (!thePrefs->IsErrorLogEnabled()))
	{
		delete sErrorLog;
		sErrorLog = NULL;
	}
}

// SERVICE ROUTINES

QTSS_Error RollErrorLog(QTSS_ServiceFunctionArgsPtr /*inArgs*/)
{
	OSMutexLocker locker(sLogMutex);
	if (sErrorLog != NULL)
		sErrorLog->RollLog();
	return QTSS_NoErr;
}



