/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		QTSSModule.cpp

	Contains:	Implements object defined in QTSSModule.h
					
	$Log: QTSSModule.cpp,v $
	Revision 1.3  2000/03/07 22:01:09  serenyi
	extern "C"d module entrypoints, fixed QT file bugs
	
	Revision 1.2  2000/01/27 02:09:13  rtucker
	Merge changes from v81
	
	Revision 1.6  2000/01/19 19:37:04  serenyi
	Moved to new NEW macro
	
	Revision 1.5  1999/12/21 01:44:04  lecroy
	*** empty log message ***
	
	Revision 1.4  1999/11/08 22:52:34  serenyi
	Updated code to reflect changes in QTSS API
	
	Revision 1.3  1999/10/27 17:55:02  serenyi
	Integrated Matthias's MacOS / GUSI changes
	
	Revision 1.2  1999/10/07 08:52:31  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.1.1.1  1999/09/22 23:01:51  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:53  rtucker
	initial checkin from rrv14
	
	

	Created: Tue, Mar 2, 1999 @ 4:23 PM
*/

#include <sys/errno.h>

#include "QTSSModule.h"
#include "OSArrayObjectDeleter.h"
#include "OSMemory.h"
#include "StringParser.h"
#include "Socket.h"

Bool16	QTSSModule::sHasRTSPRequestModule = false;

QTSSModule::QTSSModule(char* inName, char* inPath)
: 	fQueueElem(this),
	fPath(NULL),
	fFragment(NULL),
	fDispatchFunc(NULL)
{
	if ((inPath != NULL) && (inPath[0] != '\0'))
	{
		// Create a code fragment if this module is being loaded from disk
		
		fFragment = NEW OSCodeFragment(inPath);
		fPath = NEW char[::strlen(inPath) + 2];
		::strcpy(fPath, inPath);
	}
	
	// If there is a name, copy it into the module object's internal buffer
	if (inName != NULL)
		this->SetName(inName);
				
	::memset(fRoleArray, 0, sizeof(fRoleArray));
}

QTSS_Error 	QTSSModule::Initialize(QTSS_CallbacksPtr inCallbacks, QTSS_MainEntryPointPtr inEntrypoint)
{
	QTSS_Error theErr = QTSS_NoErr;
	
	// Load fragment from disk if necessary
	
	if ((fFragment != NULL) && (inEntrypoint == NULL))
		theErr = this->LoadFromDisk(&inEntrypoint);
	if (theErr != QTSS_NoErr)
		return theErr;
		
	// At this point, we must have an entrypoint
	if (inEntrypoint == NULL)
		return QTSS_NotAModule;
		
	// Invoke the private initialization routine
	QTSS_PrivateArgs thePrivateArgs;
	thePrivateArgs.inServerAPIVersion = QTSS_API_VERSION;
	thePrivateArgs.inCallbacks = inCallbacks;
	thePrivateArgs.outStubLibraryVersion = 0;
	thePrivateArgs.outDispatchFunction = NULL;
	
	theErr = (inEntrypoint)(&thePrivateArgs);
	if (theErr != QTSS_NoErr)
		return theErr;
		
	if (thePrivateArgs.outStubLibraryVersion != thePrivateArgs.inServerAPIVersion)
		return QTSS_WrongVersion;
	
	// Set the dispatch function so we'll be able to invoke this module later on
	
	fDispatchFunc = thePrivateArgs.outDispatchFunction;
	return QTSS_NoErr;
}

QTSS_Error QTSSModule::LoadFromDisk(QTSS_MainEntryPointPtr* outEntrypoint)
{
	static StrPtrLen sMainEntrypointName("_Main");
	
	Assert(outEntrypoint != NULL);
	
	// Modules only need to be initialized if they reside on disk. 
	if (fFragment == NULL)
		return QTSS_NoErr;
	
	if (!fFragment->IsValid())
		return QTSS_NotAModule;
		
	// fPath is actually a path. Extract the file name.
	
	StrPtrLen theFileName(fPath);
	StringParser thePathParser(&theFileName);
	
	while (thePathParser.GetThru(&theFileName, '/'))
		;
	Assert(theFileName.Len > 0);
	Assert(theFileName.Ptr != NULL);

	// At this point, theFileName points to the file name. Make this the module name.
	this->SetName(theFileName.Ptr);
	
	// 
	// The main entrypoint symbol name is the file name plus that _Main__ string up there.
	OSCharArrayDeleter theSymbolName(NEW char[theFileName.Len + sMainEntrypointName.Len + 2]);
	::strcpy(theSymbolName, theFileName.Ptr);
	
	::strcat(theSymbolName, sMainEntrypointName.Ptr);
	*outEntrypoint = (QTSS_MainEntryPointPtr)fFragment->GetSymbol(theSymbolName.GetObject());
	return QTSS_NoErr;
}

void		QTSSModule::SetName(char* inName)
{
	fModuleName.Ptr = NEW char[::strlen(inName) + 1];
	::strcpy(fModuleName.Ptr, inName);
	fModuleName.Len = ::strlen(fModuleName.Ptr);
}

QTSS_Error	QTSSModule::AddRole(QTSS_Role inRole)
{
	switch (inRole)
	{
		// Map actual QTSS Role names to our private enum values. Turn on the proper one
		// in the role array
		case QTSS_Initialize_Role:			fRoleArray[kInitializeRole] = true;			break;
		case QTSS_Shutdown_Role:			fRoleArray[kShutdownRole] = true;			break;
		case QTSS_RTSPFilter_Role:			fRoleArray[kRTSPFilterRole] = true;			break;
		case QTSS_RTSPRoute_Role:			fRoleArray[kRTSPRouteRole] = true;			break;
		case QTSS_RTSPAuthorize_Role:		fRoleArray[kRTSPAuthRole] = true;			break;
		case QTSS_RTSPPreProcessor_Role:	fRoleArray[kRTSPPreProcessorRole] = true;	break;
		case QTSS_RTSPRequest_Role:			fRoleArray[kRTSPRequestRole] = true;		break;
		case QTSS_RTSPPostProcessor_Role:	fRoleArray[kRTSPPostProcessorRole] = true;	break;
		case QTSS_RTSPSessionClosing_Role:	fRoleArray[kRTSPSessionClosingRole] = true;	break;
		case QTSS_RTPSendPackets_Role:		fRoleArray[kRTPSendPacketsRole] = true;	 	break;
		case QTSS_ClientSessionClosing_Role:fRoleArray[kClientSessionClosingRole] = true;break;
		case QTSS_RTCPProcess_Role:			fRoleArray[kRTCPProcessRole] = true;	 	break;
		case QTSS_ErrorLog_Role:			fRoleArray[kErrorLogRole] = true;			break;
		case QTSS_RereadPrefs_Role:			fRoleArray[kRereadPrefsRole] = true;		break;
		default:
			return QTSS_BadArgument;
	}
	
	if (inRole == QTSS_RTSPRequest_Role)
		sHasRTSPRequestModule = true;
			
	return QTSS_NoErr;
}

