/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		QTSSModuleInterface.cpp

	Contains:	Implements object defined in QTSSModuleInterface.h
					
	$Log: QTSSModuleInterface.cpp,v $
	Revision 1.1.1.1  2000/01/17 23:36:03  wsanchez
	Import of StreamingServer-2.0 (Apple)
	
	Revision 1.6  1999/12/21 01:44:04  lecroy
	*** empty log message ***
	
	Revision 1.5  1999/11/08 22:52:34  serenyi
	Updated code to reflect changes in QTSS API
	
	Revision 1.4  1999/10/18 20:18:57  serenyi
	Changed "dictType" to "objectType"
	
	Revision 1.3  1999/10/07 08:52:31  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.2  1999/09/30 00:29:51  serenyi
	Added assert logging
	
	Revision 1.1.1.1  1999/09/22 23:01:51  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:53  rtucker
	initial checkin from rrv14
	
	

	Created: Tue, Mar 2, 1999 @ 4:23 PM
*/


#include "QTSSModuleInterface.h"
#include "QTSSRTSPProtocol.h"
#include "OSMemory.h"
#include "OS.h"

// STATIC DATA

OSQueue 				QTSSModuleInterface::sModuleQueue;
			
QTSSDictionaryMap*		QTSSModuleInterface::sDictionaryMaps[kNumDictionaries];
QTSSModule**			QTSSModuleInterface::sModuleArray[QTSSModule::kNumRoles];
UInt32					QTSSModuleInterface::sNumModulesInRole[QTSSModule::kNumRoles];
QTSSErrorLogStream		QTSSModuleInterface::sErrorLogStream;
QTSSDictionary*			QTSSModuleInterface::sServerDict = NULL;

QTSS_Error	QTSSErrorLogStream::Write(void* inBuffer, UInt32 inLen, UInt32* outLenWritten, UInt32 inFlags)
{
	// For the error log stream, the flags are considered to be the verbosity
	// of the error.
	if (inFlags >= qtssIllegalVerbosity)
		inFlags = qtssMessageVerbosity;
		
 	QTSSModuleInterface::LogError(inFlags, (char*)inBuffer);
 	if (outLenWritten != NULL)
 		*outLenWritten = inLen;
 		
	return QTSS_NoErr;
}

void QTSSErrorLogStream::LogAssert(char* inMessage)
{
 	QTSSModuleInterface::LogError(qtssAssertVerbosity, inMessage);
}


void QTSSModuleInterface::Initialize()
{
	for (UInt32 y = 0; y < QTSSModule::kNumRoles; y++)
	{
		sModuleArray[y] = NULL;
		sNumModulesInRole[y] = 0;
	}

	sDictionaryMaps[kServerDictIndex] 		= new QTSSDictionaryMap(qtssSvrNumParams);	//kServerDictIndex
	sDictionaryMaps[kPrefsDictIndex] 		= new QTSSDictionaryMap(qtssPrefsNumParams);	//kPrefsDictIndex
	sDictionaryMaps[kTextMessagesDictIndex] = new QTSSDictionaryMap(qtssMsgNumParams);	//kTextMessagesDictIndex
	sDictionaryMaps[kServiceDictIndex] 		= new QTSSDictionaryMap(0);					//kServiceDictIndex
	sDictionaryMaps[kRTPStreamDictIndex] 	= new QTSSDictionaryMap(qtssRTPStrNumParams);//kRTPStreamDictIndex
	sDictionaryMaps[kClientSessionDictIndex]= new QTSSDictionaryMap(qtssCliSesNumParams);//kRTPSessionDictIndex
	sDictionaryMaps[kRTSPSessionDictIndex] 	= new QTSSDictionaryMap(qtssRTSPSesNumParams);//kRTSPSessionDictIndex
	sDictionaryMaps[kRTSPRequestDictIndex] 	= new QTSSDictionaryMap(qtssRTSPReqNumParams);//kRTSPRequestDictIndex
	sDictionaryMaps[kRTSPHeaderDictIndex] 	= new QTSSDictionaryMap(qtssNumHeaders);		//kRTSPHeaderDictIndex
}

UInt32	QTSSModuleInterface::GetDictionaryIndex(QTSS_ObjectType inType)
{
	 switch (inType)
	 {
	 	case qtssRTPStreamObjectType:		return kRTPStreamDictIndex;
	 	case qtssClientSessionObjectType:	return kClientSessionDictIndex;
	 	case qtssRTSPSessionObjectType:		return kRTSPSessionDictIndex;
	 	case qtssRTSPRequestObjectType:		return kRTSPRequestDictIndex;
	 	case qtssRTSPHeaderObjectType:		return kRTSPHeaderDictIndex;
	 	case qtssServerObjectType:			return kServerDictIndex;
	 	case qtssPrefsObjectType:			return kPrefsDictIndex;
	 	case qtssTextMessagesObjectType:	return kTextMessagesDictIndex;
	 	default:							return kIllegalDictionary;
	 }
	 return kIllegalDictionary;
}

void	QTSSModuleInterface::LogError(	QTSS_ErrorVerbosity inVerbosity,
										char* inBuffer)
{
	QTSS_RoleParams theParams;
	theParams.errorParams.inVerbosity = inVerbosity;
	theParams.errorParams.inBuffer = inBuffer;

	for (UInt32 x = 0; x < QTSSModuleInterface::GetNumModulesInRole(QTSSModule::kErrorLogRole); x++)
		(void)QTSSModuleInterface::GetModule(QTSSModule::kErrorLogRole, x)->CallDispatch(QTSS_ErrorLog_Role, &theParams);

	// If this is a fatal error, set the proper attribute in the RTSPServer dictionary
	if ((inVerbosity == qtssFatalVerbosity) && (sServerDict != NULL))
	{
		QTSS_ServerState theState = qtssFatalErrorState;
		(void)sServerDict->SetValue(qtssSvrState, 0, &theState, sizeof(theState));
	}
}
