/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		QTSSModuleInterface.h

	Contains:	Stores all global data directly related to QTSS Module API stuff.
					
	$Log: QTSSModuleInterface.h,v $
	Revision 1.1.1.1  2000/01/17 23:36:03  wsanchez
	Import of StreamingServer-2.0 (Apple)
	
	Revision 1.5  1999/12/21 01:44:05  lecroy
	*** empty log message ***
	
	Revision 1.4  1999/11/08 22:52:35  serenyi
	Updated code to reflect changes in QTSS API
	
	Revision 1.3  1999/10/07 08:52:44  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.2  1999/09/30 00:29:51  serenyi
	Added assert logging
	
	Revision 1.1.1.1  1999/09/22 23:01:51  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:53  rtucker
	initial checkin from rrv14
	
	

	Created: Tue, Mar 2, 1999 @ 4:23 PM
*/

#ifndef __QTSSMODULEINTERFACE_H__
#define __QTSSMODULEINTERFACE_H__

#include "QTSS.h"
#include "QTSSDictionary.h"
#include "QTSSStream.h"
#include "QTSSModule.h"


class QTSSModuleInterface;

// This object also functions as our assert logger
class QTSSErrorLogStream : public QTSSStream, public AssertLogger
{
	public:
	
		// This QTSSStream is used by modules to write to the error log
	
		QTSSErrorLogStream() {}
		virtual ~QTSSErrorLogStream() {}
		
		virtual QTSS_Error	Write(void* inBuffer, UInt32 inLen, UInt32* outLenWritten, UInt32 inFlags);
		virtual void		LogAssert(char* inMessage);
};


class QTSSModuleInterface
{
	public:
	
		//
		// INITIALIZE
		//
		// This creates dictionary maps
		static void Initialize();
		
		//
		// DICTIONARY MAPS
		
		// All dictionary maps are stored in this object, and are accessable
		// through these routines
		
		// This enum allows all QTSSDictionaryMaps to be stored in an array	
		enum
		{
			kServerDictIndex 				= 0,
			kPrefsDictIndex 				= 1,
			kTextMessagesDictIndex 			= 2,
			kServiceDictIndex 				= 3,
			
			kRTPStreamDictIndex 			= 4,
			kClientSessionDictIndex	 		= 5,
			kRTSPSessionDictIndex 			= 6,
			kRTSPRequestDictIndex 			= 7,
			kRTSPHeaderDictIndex 			= 8,

			kNumDictionaries				= 9,
			kIllegalDictionary				= kNumDictionaries
		};
		
		// This function converts a QTSS_ObjectType to an index
		static UInt32					GetDictionaryIndex(QTSS_ObjectType inType);

		// Using one of the above predefined indexes, this returns the corresponding map
		static QTSSDictionaryMap*		GetDictionaryMap(UInt32 inIndex)
			{ Assert(inIndex < kNumDictionaries); return sDictionaryMaps[inIndex]; }
		
		//
		// MODULE STORAGE
		
		// All module objects are stored here, and are accessable through
		// these routines.
		
		// Returns the number of modules that act in a given role
		static UInt32		GetNumModulesInRole(QTSSModule::RoleIndex inRole)
								{ Assert(inRole < QTSSModule::kNumRoles); return sNumModulesInRole[inRole]; }
		
		// Allows the caller to iterate over all modules that act in a given role			
		static QTSSModule*	GetModule(QTSSModule::RoleIndex inRole, UInt32 inIndex)
								{ 	Assert(inRole < QTSSModule::kNumRoles);
									Assert(inIndex < sNumModulesInRole[inRole]);
									return sModuleArray[inRole][inIndex];
								}
		
		// All modules are placed on a queue, accessable here.	
		static OSQueue*		GetModuleQueue() { return &sModuleQueue; }
		
		//
		// ERROR LOGGING
		
		// Invokes the error logging modules with some data
		static void		LogError(	QTSS_ErrorVerbosity inVerbosity, char* inBuffer);
		
		// Returns the error log stream
		static QTSS_StreamRef GetErrorLogStream() { return &sErrorLogStream; }

	protected:
	
		// GLOBAL MODULE DATA
		// This stuff is set up by the derived (QTSServer) class
		
		static OSQueue 					sModuleQueue;
			
		static QTSSDictionaryMap*		sDictionaryMaps[kNumDictionaries];
		
		static QTSSModule**				sModuleArray[QTSSModule::kNumRoles];
		static UInt32					sNumModulesInRole[QTSSModule::kNumRoles];
		static QTSSErrorLogStream		sErrorLogStream;
		
		static QTSSDictionary*			sServerDict;
};





#endif //__QTSSMODULEINTERFACE_H__
