/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		QTSSPrefs.cpp

	Contains:	Implements class defined in QTSSPrefs.h.

	Change History (most recent first):
	
*/

#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "QTSSPrefs.h"
#include "MyAssert.h"
#include "OSMemory.h"
#include "QTSSModuleInterface.h"
 
#pragma mark __STATIC_DATA__

//These variables are used to store the default settings once
//we start altering the defaults by reading out of the file.
//It doesn't really matter what they are set to right now.
UInt32 	QTSSPrefs::sDefaultRTSPTimeout = 0;
UInt32	QTSSPrefs::sDefaultRealRTSPTimeoutInSecs = 180;
UInt32 	QTSSPrefs::sDefaultRTPTimeout = 60;
SInt32 	QTSSPrefs::sDefaultMaximumConnections = 1000;
SInt32	QTSSPrefs::sDefaultMaxBandwidth = 100000;
char*  	QTSSPrefs::sDefaultMoviesFolder = "/Local/Library/QuickTimeStreaming/Movies";
UInt32	QTSSPrefs::sDefaultIPAddress = INADDR_ANY;
char*	QTSSPrefs::sDefaultIPAddressStr = "0.0.0.0";
Bool16	QTSSPrefs::sDefaultBreakOnAssert = true;
Bool16 	QTSSPrefs::sDefaultAutoRestart = true;
UInt32	QTSSPrefs::sDefaultTBUpdateTime = 1;
UInt32	QTSSPrefs::sDefaultABUpdateTime = 60;
UInt32	QTSSPrefs::sDefaultSafePlayDuration = 600;
char*	QTSSPrefs::sDefaultModuleFolder = "/usr/sbin/StreamingServerModules";

char*	QTSSPrefs::sDefaultErrorLogName = "Error";
char*	QTSSPrefs::sDefaultErrorLogDir = "/Local/Library/QuickTimeStreaming/Logs/";
UInt32	QTSSPrefs::sErrorRollIntervalInDays = 0;
UInt32	QTSSPrefs::sErrorLogBytes = 256000;
UInt32	QTSSPrefs::sErrorLogVerbosity = 2;
Bool16	QTSSPrefs::sScreenLoggingEnabled = true;
Bool16	QTSSPrefs::sErrorLogEnabled = true;

SInt32	QTSSPrefs::sDefaultTCPMinThinDelayToleranceInMSec = 50;
SInt32	QTSSPrefs::sDefaultTCPMaxThinDelayToleranceInMSec = 300;
SInt32	QTSSPrefs::sDefaultTCPVideoDelayToleranceInMSec = 500;
SInt32	QTSSPrefs::sDefaultTCPAudioDelayToleranceInMSec = 2500;
Bool16  QTSSPrefs::sDefaultDoReportHTTPConnectionAddress = false;

UInt32	QTSSPrefs::sDefaultMinTCPBufferSizeInBytes 	= 8192;
UInt32	QTSSPrefs::sDefaultMaxTCPBufferSizeInBytes 	= 32768;
Float32	QTSSPrefs::sDefaultTCPSecondsToBuffer		= .33;

Bool16	QTSSPrefs::sDefaultAuthorizationEnabled 	= true;
char*	QTSSPrefs::sDefaultAuthorizationRealm		= "QT Streaming Server";

char*	QTSSPrefs::sDefaultPidFileName		= "/var/run/dss2.pid";
char*	QTSSPrefs::sDefaultRunUserName		= "";
char*	QTSSPrefs::sDefaultRunGroupName		= "";

char* QTSSPrefs::sPrefsKeyStrings[] = 
{	
	//qtssPrefsRTSPTimeoutParam
	"rtsp_timeout",
	//qtssPrefsRealRTSPTimeoutParam
	"real_rtsp_timeout",
	//qtssPrefsRTPTimeoutParam,
	"rtp_timeout",
	//qtssPrefsMaximumConnectionsParam
	"maximum_connections",
	//qtssPrefsMaximumBandwidthParam
	"maximum_bandwidth",
	//qtssPrefsMovieFolderParam
	"movie_folder",
	//qtssPrefsRTSPIPAddrParam
	"bind_ip_addr",
	//qtssPrefsBreakOnAssertParam
	"break_on_assert",
	//qtssPrefsAutoRestartParam,
	"auto_restart",
	//qtssPrefsTotalBytesUpdateParam
	"total_bytes_update",
	//qtssPrefsAvgBandwidthUpdateParam
	"average_bandwidth_update",
	//qtssPrefsSafePlayDurationParam
	"safe_play_duration",
	//qtssPrefsModuleFolderParam
	"module_folder",
	//qtssPrefsErrorLogNameParam
	"error_logfile_name",
	//qtssPrefsErrorLogDirParam
	"error_logfile_dir",
	//qtssPrefsErrorRollIntervalParam
	"error_logfile_interval",
	//qtssPrefsMaxErrorLogSizeParam
	"error_logfile_size",
	//qtssPrefsErrorLogVerbosityParam
	"error_logfile_verbosity",
	//qtssPrefsScreenLoggingParam
	"screen_logging",
	//qtssPrefsErrorLogEnabledParam
	"error_logging",

	//qtssPrefsTCPMinThinDelayToleranceInMSec 
	"tcp_min_thin_delay_tolerance",
	// qtssPrefsTCPMaxThinDelayToleranceInMSec  
	"tcp_max_thin_delay_tolerance",
	// qtssPrefsTCPVideoDelayToleranceInMSec
	"tcp_max_video_delay_tolerance",
	//qtssPrefsTCPAudioDelayToleranceInMSec 
	"tcp_max_audio_delay_tolerance",
	
	//qtssPrefsMinTCPBufferSizeInBytes
	"min_tcp_buffer_size",
	//qtssPrefsMaxTCPBufferSizeInBytes
	"max_tcp_buffer_size",
	//qtssPrefsTCPSecondsToBuffer
	"tcp_seconds_to_buffer",
	
	// qtssPrefsDoReportHTTPConnectionAddress
	"do_report_http_connection_ip_address",
	
	//qtssPrefsDefaultAuthorizationRealm 
	"default_authorization_realm",
	
	//qtssPrefsPidFileName
	"pid_file_name",
	//qtssPrefsRunUserName
	"run_user_name",	
	//qtssPrefsRunGroupName
	"run_group_name"
};

char* 	QTSSPrefs::sEnabledString = "enabled";

#pragma mark __RTSP_PREFS__

QTSSPrefs::QTSSPrefs(PrefsSource* inPrefsSource)
:	QTSSDictionary(QTSSModuleInterface::GetDictionaryMap(QTSSModuleInterface::kPrefsDictIndex), &fPrefsMutex),
 	fPrefsSource(inPrefsSource),
	fRTSPTimeoutInSecs(0),
	fRTSPTimeoutString(fRTSPTimeoutBuf, 0),
	fRealRTSPTimeoutInSecs(0),
	fRTPTimeoutInSecs(0),
	fMaximumConnections(0),
	fMaxBandwidthInKBits(0),
	fIPAddress(0),
	fBreakOnAssert(false),
	fAutoRestart(false),
	fTBUpdateTimeInSecs(0),
	fABUpdateTimeInSecs(0),
	fSafePlayDurationInSecs(0),
	fErrorRollIntervalInDays(0),
	fErrorLogBytes(0),
	fErrorLogVerbosity(3),
	fScreenLoggingEnabled(true),
	fErrorLogEnabled(false),
	fTCPMinThinDelayToleranceInMSec(50),
	fTCPMaxThinDelayToleranceInMSec(300),
	fTCPVideoDelayToleranceInMSec(500),
	fTCPAudioDelayToleranceInMSec(2500),
	fMinTCPBufferSizeInBytes(0),
	fMaxTCPBufferSizeInBytes(0),
	fTCPSecondsToBuffer(0),
	fDoReportHTTPConnectionAddress(false),
	fPrefsMutex()
{
	SetupAttributes();
	RereadPreferences();
}

void QTSSPrefs::Initialize()
{
	QTSSDictionaryMap* theMap = QTSSModuleInterface::GetDictionaryMap(QTSSModuleInterface::kPrefsDictIndex);
	Assert(theMap != NULL);
	
	for (UInt32 x = 0; x < qtssPrefsNumParams; x++)
		theMap->SetAttribute(x, sPrefsKeyStrings[x], NULL, false, false);
}


void QTSSPrefs::SetupAttributes()
{
	this->SetVal(qtssPrefsRTSPTimeout, 		&fRTSPTimeoutInSecs, 		sizeof(fRTSPTimeoutInSecs));
	this->SetVal(qtssPrefsRealRTSPTimeout, 	&fRealRTSPTimeoutInSecs, 	sizeof(fRealRTSPTimeoutInSecs));
	this->SetVal(qtssPrefsRTPTimeout, 		&fRTPTimeoutInSecs, 		sizeof(fRTPTimeoutInSecs));
	this->SetVal(qtssPrefsMaximumConnections,&fMaximumConnections, 		sizeof(fMaximumConnections));
	this->SetVal(qtssPrefsMaximumBandwidth, &fMaxBandwidthInKBits, 		sizeof(fMaxBandwidthInKBits));
	this->SetVal(qtssPrefsBreakOnAssert, 	&fBreakOnAssert, 			sizeof(fBreakOnAssert));
	this->SetVal(qtssPrefsAutoRestart, 		&fAutoRestart, 				sizeof(fAutoRestart));
	this->SetVal(qtssPrefsTotalBytesUpdate, &fTBUpdateTimeInSecs, 		sizeof(fTBUpdateTimeInSecs));
	this->SetVal(qtssPrefsAvgBandwidthUpdate,&fABUpdateTimeInSecs, 		sizeof(fABUpdateTimeInSecs));
	this->SetVal(qtssPrefsSafePlayDuration, &fSafePlayDurationInSecs, 	sizeof(fSafePlayDurationInSecs));

	this->SetVal(qtssPrefsErrorRollInterval, &fErrorRollIntervalInDays, sizeof(fErrorRollIntervalInDays));
	this->SetVal(qtssPrefsMaxErrorLogSize, 	&fErrorLogBytes, 			sizeof(fErrorLogBytes));
	this->SetVal(qtssPrefsErrorLogVerbosity, &fErrorLogVerbosity, 		sizeof(fErrorLogVerbosity));
	this->SetVal(qtssPrefsScreenLogging, 	&fScreenLoggingEnabled, 	sizeof(fScreenLoggingEnabled));
	this->SetVal(qtssPrefsErrorLogEnabled, 	&fErrorLogEnabled, 			sizeof(fErrorLogEnabled));


	this->SetVal(qtssPrefsTCPMinThinDelayToleranceInMSec, 	&fTCPMinThinDelayToleranceInMSec, sizeof(fTCPMinThinDelayToleranceInMSec));
	this->SetVal(qtssPrefsTCPMaxThinDelayToleranceInMSec, 	&fTCPMaxThinDelayToleranceInMSec, sizeof(fTCPMaxThinDelayToleranceInMSec));
	this->SetVal(qtssPrefsTCPVideoDelayToleranceInMSec, 	&fTCPVideoDelayToleranceInMSec,   sizeof(fTCPVideoDelayToleranceInMSec));
	this->SetVal(qtssPrefsTCPAudioDelayToleranceInMSec, 	&fTCPAudioDelayToleranceInMSec,   sizeof(fTCPAudioDelayToleranceInMSec));

	this->SetVal(qtssPrefsMinTCPBufferSizeInBytes, 	&fMinTCPBufferSizeInBytes,	sizeof(fMinTCPBufferSizeInBytes));
	this->SetVal(qtssPrefsMaxTCPBufferSizeInBytes, 	&fMaxTCPBufferSizeInBytes,	sizeof(fMaxTCPBufferSizeInBytes));
	this->SetVal(qtssPrefsTCPSecondsToBuffer, 	&fTCPSecondsToBuffer,   		sizeof(fTCPSecondsToBuffer));

	this->SetVal(qtssPrefsDoReportHTTPConnectionAddress, 	&fDoReportHTTPConnectionAddress,   sizeof(fDoReportHTTPConnectionAddress));

}



void QTSSPrefs::RereadPreferences()
{
	static const UInt32 kMaxPrefValueSize = 1024;
	char thePrefValue[kMaxPrefValueSize];
	
	OSMutexLocker locker(&fPrefsMutex);
	
	// Use the names of the attributes in the attribute map as the key values for
	// finding preferences in the config file.
	
	for (UInt32 x = 0; x < this->GetDictionaryMap()->GetNumAttrs(); x++)
	{
		thePrefValue[0] = '\0';
		(void)fPrefsSource->GetValue(this->GetDictionaryMap()->GetAttrName(x), &thePrefValue[0]);
		
		// Add this preference into the dictionary.
		
		switch(this->GetDictionaryMap()->GetAttrID(x))
		{
			case qtssPrefsRTSPTimeout:
			{
				this->ProcessUIntPref(thePrefValue, &fRTSPTimeoutInSecs, sDefaultRTSPTimeout);
				if (fRTSPTimeoutString.Len == 0)
				{
					::sprintf(fRTSPTimeoutString.Ptr, "%lu", fRTSPTimeoutInSecs);
					fRTSPTimeoutString.Len = ::strlen(fRTSPTimeoutString.Ptr);
				}
				break;
			}
			
			case qtssPrefsRealRTSPTimeout:
				this->ProcessUIntPref(thePrefValue, &fRealRTSPTimeoutInSecs, sDefaultRealRTSPTimeoutInSecs);
				break;
			case qtssPrefsRTPTimeout:
				this->ProcessUIntPref(thePrefValue, &fRTPTimeoutInSecs, sDefaultRTPTimeout);
				break;
			case qtssPrefsMaximumConnections:
				this->ProcessSIntPref(thePrefValue, &fMaximumConnections, sDefaultMaximumConnections);
				break;
			case qtssPrefsMaximumBandwidth:
				this->ProcessSIntPref(thePrefValue, &fMaxBandwidthInKBits, sDefaultMaxBandwidth);
				break;
			case qtssPrefsMovieFolder:
				this->ProcessStringPref(qtssPrefsMovieFolder, thePrefValue, sDefaultMoviesFolder);
				break;
			case qtssPrefsRTSPIPAddr:
			{
				fIPAddress = sDefaultIPAddress;
#ifndef __solaris__
				if (thePrefValue[0] != '\0')
				{
					struct in_addr theAddress;
					if (::inet_aton(thePrefValue, &theAddress) != 0)
	                	fIPAddress = ntohl(theAddress.s_addr);
	            }
				this->ProcessStringPref(qtssPrefsRTSPIPAddr, thePrefValue, sDefaultIPAddressStr);
#endif
	            break;
			}
			
			case qtssPrefsBreakOnAssert:
				this->ProcessBooleanPref(thePrefValue, &fBreakOnAssert, sDefaultBreakOnAssert);
				break;
			case qtssPrefsAutoRestart:
				this->ProcessBooleanPref(thePrefValue, &fAutoRestart, sDefaultAutoRestart);
				break;
			case qtssPrefsTotalBytesUpdate:
				this->ProcessUIntPref(thePrefValue, &fTBUpdateTimeInSecs, sDefaultTBUpdateTime);
				break;
			case qtssPrefsAvgBandwidthUpdate:
				this->ProcessUIntPref(thePrefValue, &fABUpdateTimeInSecs, sDefaultABUpdateTime);
				break;
			case qtssPrefsSafePlayDuration:
				this->ProcessUIntPref(thePrefValue, &fSafePlayDurationInSecs, sDefaultSafePlayDuration);
				break;
			case qtssPrefsModuleFolder:
				this->ProcessStringPref(qtssPrefsModuleFolder, thePrefValue, sDefaultModuleFolder);
				break;
			case qtssPrefsErrorLogName:
				this->ProcessStringPref(qtssPrefsErrorLogName, thePrefValue, sDefaultErrorLogName);
				break;
			case qtssPrefsErrorLogDir:
				this->ProcessStringPref(qtssPrefsErrorLogDir, thePrefValue, sDefaultErrorLogDir);
				break;
			case qtssPrefsErrorRollInterval:
				this->ProcessUIntPref(thePrefValue, &fErrorRollIntervalInDays, sErrorRollIntervalInDays);
				break;
			case qtssPrefsMaxErrorLogSize:
				this->ProcessUIntPref(thePrefValue, &fErrorLogBytes, sErrorLogBytes);
				break;
			case qtssPrefsErrorLogVerbosity:
				this->ProcessUIntPref(thePrefValue, &fErrorLogVerbosity, sErrorLogVerbosity);
				break;
			case qtssPrefsScreenLogging:
				this->ProcessBooleanPref(thePrefValue, &fScreenLoggingEnabled, sScreenLoggingEnabled);
				break;
			case qtssPrefsErrorLogEnabled:
				this->ProcessBooleanPref(thePrefValue, &fErrorLogEnabled, sErrorLogEnabled);
				break;
			case qtssPrefsTCPMinThinDelayToleranceInMSec:
				this->ProcessSIntPref(thePrefValue, &fTCPMinThinDelayToleranceInMSec, sDefaultTCPMinThinDelayToleranceInMSec);
				break;
			case qtssPrefsTCPMaxThinDelayToleranceInMSec:
				this->ProcessSIntPref(thePrefValue, &fTCPMaxThinDelayToleranceInMSec, sDefaultTCPMaxThinDelayToleranceInMSec);
				break;
			case qtssPrefsTCPVideoDelayToleranceInMSec:
				this->ProcessSIntPref(thePrefValue, &fTCPVideoDelayToleranceInMSec, sDefaultTCPVideoDelayToleranceInMSec);
				break;
			case qtssPrefsTCPAudioDelayToleranceInMSec:
				this->ProcessSIntPref(thePrefValue, &fTCPAudioDelayToleranceInMSec, sDefaultTCPAudioDelayToleranceInMSec);
				break;				

			case qtssPrefsMinTCPBufferSizeInBytes:
				this->ProcessUIntPref(thePrefValue, &fMinTCPBufferSizeInBytes, sDefaultMinTCPBufferSizeInBytes);
				break;				
			case qtssPrefsMaxTCPBufferSizeInBytes:
				this->ProcessUIntPref(thePrefValue, &fMaxTCPBufferSizeInBytes, sDefaultMaxTCPBufferSizeInBytes);
				break;				
			case qtssPrefsTCPSecondsToBuffer:
				this->ProcessFloat32Pref(thePrefValue, &fTCPSecondsToBuffer, sDefaultTCPSecondsToBuffer);
				break;				

			case qtssPrefsDoReportHTTPConnectionAddress:
				this->ProcessBooleanPref(thePrefValue, &fDoReportHTTPConnectionAddress, sDefaultDoReportHTTPConnectionAddress);
				break;

			case qtssPrefsDefaultAuthorizationRealm:
				this->ProcessStringPref(qtssPrefsDefaultAuthorizationRealm, thePrefValue, sDefaultAuthorizationRealm);
				break;

			case qtssPrefsPidFileName:
				this->ProcessStringPref(qtssPrefsPidFileName, thePrefValue, sDefaultPidFileName);
				break;

			case qtssPrefsRunUserName:
				this->ProcessStringPref(qtssPrefsRunUserName, thePrefValue, sDefaultRunUserName);
				break;

			case qtssPrefsRunGroupName:
				this->ProcessStringPref(qtssPrefsRunGroupName, thePrefValue, sDefaultRunGroupName);
				break;

			default:
				this->ProcessStringPref(this->GetDictionaryMap()->GetAttrID(x), thePrefValue, NULL);
		}
	}
}

void QTSSPrefs::ProcessBooleanPref(char* inBuffer, Bool16* inPref, Bool16 inDefault)
{
	if (inBuffer[0] == '\0')
		*inPref = inDefault;
	else if (::strcmp(inBuffer, sEnabledString) == 0)
		*inPref = true;
	else
		*inPref = false;
}

void QTSSPrefs::ProcessUIntPref(char* inBuffer, UInt32* inPref, UInt32 inDefault)
{
	if (inBuffer[0] == '\0')
		*inPref = inDefault;
	else
		::sscanf(inBuffer, "%ld", inPref);
}

void QTSSPrefs::ProcessSIntPref(char* inBuffer, SInt32* inPref, SInt32 inDefault)
{
	if (inBuffer[0] == '\0')
		*inPref = inDefault;
	else
		::sscanf(inBuffer, "%ld", inPref);
}

void QTSSPrefs::ProcessFloat32Pref(char* inBuffer, Float32* inPref, Float32 inDefault)
{
	if (inBuffer[0] == '\0')
		*inPref = inDefault;
	else
		::sscanf(inBuffer, "%f", inPref);
}

char*	QTSSPrefs::GetMovieFolder(char* inBuffer, UInt32* ioLen)
{
	OSMutexLocker locker(&fPrefsMutex);

	// Get the movie folder attribute
	StrPtrLen* theMovieFolder = this->GetValue(qtssPrefsMovieFolder);

	// If the movie folder path fits inside the provided buffer, copy it there
	if (theMovieFolder->Len < *ioLen)
		::memcpy(inBuffer, theMovieFolder->Ptr, theMovieFolder->Len);
	else
	{
		// Otherwise, allocate a buffer to store the path
		inBuffer = NEW char[theMovieFolder->Len + 2];
		::memcpy(inBuffer, theMovieFolder->Ptr, theMovieFolder->Len);
	}
	*ioLen = theMovieFolder->Len;
	return inBuffer;
}

void	QTSSPrefs::ProcessStringPref(UInt32 inID, char* inNewValue, char* inDefaultValue)
{
	// If there is no new value, try to set this to the default value.
	if ((inNewValue == NULL) || (inNewValue[0] == '\0'))
		inNewValue = inDefaultValue;
		
	// If there already is an attribute here, we KNOW that we have allocated
	// it with a previous call to RereadPreferences, because these prefs
	// are read-only, so delete it.
	
	StrPtrLen* theAttribute = this->GetValue(inID);
	if (theAttribute->Ptr != NULL)
		delete [] theAttribute->Ptr;
		
	// Clear out the attribute
	
	this->SetVal(inID, NULL, 0);
	Assert(theAttribute->Ptr == NULL);
	
	// If there is a new attribute, allocate some new memory for 
	// the new attribute, and copy the data into the newly allocated buffer
	if (inNewValue != NULL)
	{
		char* attrBuffer = NEW char[::strlen(inNewValue) + 2];
		::strcpy(attrBuffer, inNewValue);
		this->SetVal(inID, attrBuffer, ::strlen(attrBuffer));
	}
}

// SERVICE ROUTINES

