/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		QTSServer.cpp

	Contains:	Implements object defined in QTSServer.h
	
	$Log: QTSServer.cpp,v $
	Revision 1.3  2000/01/28 08:31:15  serenyi
	Fixed startup sequence so RTCPs get processed
	
	Revision 1.2  2000/01/27 02:09:13  rtucker
	Merge changes from v81
	
	Revision 1.19  2000/01/26 21:32:15  serenyi
	Added new QTSS_RequestIdleTime callback, made it so the ClientSessionClosing role can register for async events
	
	Revision 1.18  2000/01/26 18:33:26  serenyi
	#2426075 Integrated changes from Kostya Vasilyev
	
	Revision 1.17  2000/01/21 23:52:55  rtucker
	
	logging changes
	
	Revision 1.16  2000/01/19 19:37:05  serenyi
	Moved to new NEW macro
	
	Revision 1.15  1999/12/21 01:44:05  lecroy
	*** empty log message ***
	
	Revision 1.14  1999/11/17 20:40:28  serenyi
	Server still starts up if any one of the listener ports couldn't be inited
	
	Revision 1.13  1999/11/09 12:41:31  murata
	Turn off access module testing.
	
	Revision 1.12  1999/11/08 22:52:35  serenyi
	Updated code to reflect changes in QTSS API
	
	Revision 1.11  1999/10/29 23:26:51  murata
	Added #define to add the QTAccessModule for debugging as a static linked file
	
	Revision 1.10  1999/10/28 01:48:37  rtucker
	fixed admin->reread prefs bug
	
	Revision 1.9  1999/10/27 17:55:03  serenyi
	Integrated Matthias's MacOS / GUSI changes
	
	Revision 1.8  1999/10/21 22:07:25  lecroy
	made access & fileprivs modules dynamic, added AFTER_BUILD scripts to strip them and move them around, renamed QTSSPasswd to qtpasswd, probably some other stuff too...
	
	Revision 1.7  1999/10/19 22:13:31  lecroy
	took out both auth modules
	
	Revision 1.6  1999/10/14 23:18:43  rtucker
	changes to report server ip back to client over HTTP
	
	Revision 1.5  1999/10/09 01:09:43  lecroy
	Added QTSSAccessModule
	
	Revision 1.4  1999/10/07 08:52:31  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.3  1999/09/30 00:30:11  serenyi
	Added assert logging
	
	Revision 1.2  1999/09/24 19:57:02  murata
	Added support for new API call QTSS_MilliSecsTo1970Secs.
	
	Revision 1.1.1.1  1999/09/22 23:01:50  lecroy
	importing
	
	Revision 1.3  1999/09/15 17:04:38  serenyi
	Changed RereadPrefsService to be safely callable at any time
	
	Revision 1.2  1999/09/14 00:46:28  murata
	fix bugs: RTSPSession.cpp now not returning error if not HTTP stream
	
	Revision 1.3  1999/02/26 05:14:57  cl
	Added HTTP stats (RTSPWebStatsModule)
	
	Revision 1.2  1999/02/19 23:08:35  ds
	Created
	

*/

#define QTACCESS_DEBUGGING 0

#include <sys/types.h>
#include <dirent.h>
#include <sys/errno.h>
#if !__MACOS__
#include <unistd.h>
#include <grp.h>
#include <pwd.h>
#endif

#include "QTSServer.h"

#include "OSMemory.h"
#include "OSArrayObjectDeleter.h"
#include "SocketUtils.h"
#include "TCPListenerSocket.h"
#include "Task.h"

#include "QTSS_Private.h"
#include "QTSSCallbacks.h"
#include "QTSSModuleUtils.h"

//Compile time modules
#include "QTSSErrorLogModule.h"
#include "QTSSFileModule.h"
#if __MacOSX__
#include "QTSSvrControlModule.h"
#endif
#include "QTSSAccessLogModule.h"
#include "QTSSFlowControlModule.h"
#include "QTSSReflectorModule.h"
#include "QTSSSplitterModule.h"
#include "QTSSRelayModule.h"
#include "QTSSWebStatsModule.h"
#if MEMORY_DEBUGGING
#include "QTSSWebDebugModule.h"
#endif

#if QTACCESS_DEBUGGING
	#include "QTSSAccessModule.h"
#endif

#include "RTSPRequestInterface.h"
#include "RTSPSessionInterface.h"
#include "RTPSessionInterface.h"
#include "RTSPSession.h"
#include "RTPStream.h"
#include "RTCPTask.h"

// CLASS DEFINITIONS

class RTSPListenerSocket : public TCPListenerSocket
{
	public:
	
		RTSPListenerSocket() {}
		virtual ~RTSPListenerSocket() {}
		
		//sole job of this object is to implement this function
		virtual Task*	GetSessionTask(TCPSocket** outSocket);
};

class RTPSocketPool : public UDPSocketPool
{
	public:
	
		// Pool of UDP sockets for use by the RTP server
		
		RTPSocketPool() {}
		~RTPSocketPool() {}
		
		virtual UDPSocketPair*	ConstructUDPSocketPair();
		virtual void			DestructUDPSocketPair(UDPSocketPair* inPair);
};



char* 			QTSServer::sPortPrefString = "rtsp_port";
QTSS_Callbacks	QTSServer::sCallbacks;

QTSServer::~QTSServer()
{
	//
	// Grab the prefs mutex. This is to make sure we can't reread prefs
	// WHILE shutting down, which would cause some weirdness for QTSS API
	// (some modules could get QTSS_RereadPrefs_Role after QTSS_Shutdown, which would be bad)
	OSMutexLocker locker(this->GetPrefs()->GetMutex());

	QTSS_ModuleState theModuleState;
	theModuleState.curRole = QTSS_Shutdown_Role;
	theModuleState.curTask = NULL;
	OSThread::SetMainThreadData(&theModuleState);

	for (UInt32 x = 0; x < QTSSModuleInterface::GetNumModulesInRole(QTSSModule::kShutdownRole); x++)
		(void)QTSSModuleInterface::GetModule(QTSSModule::kShutdownRole, x)->CallDispatch(QTSS_Shutdown_Role, NULL);

	OSThread::SetMainThreadData(NULL);
}

Bool16 QTSServer::Initialize(PrefsSource* inPrefsSource, PrefsSource* inMessagesSource, UInt16 inPortOverride)
{
	static const UInt32 kRTPSessionMapSize = 577;
	
	sServerDict = this;
	this->InitCallbacks();

	//
	// DICTIONARY INITIALIZATION
	
	QTSSPrefs::Initialize();
	QTSSMessages::Initialize();
	RTSPRequestInterface::Initialize();
	RTSPSessionInterface::Initialize();
	RTPSessionInterface::Initialize();
	RTPStream::Initialize();
	RTSPSession::Initialize();
	
	//
	// STUB SERVER INITIALIZATION
	//
	// Construct stub versions of the prefs and messages dictionaries. We need
	// both of these to initialize the server, but they have to be stubs because
	// their QTSSDictionaryMaps will presumably be modified when modules get loaded.
	
	fSrvrPrefs = new QTSSPrefs(inPrefsSource);
	fSrvrMessages = new QTSSMessages(inMessagesSource);
	QTSSModuleUtils::Initialize(fSrvrMessages, fSrvrPrefs, QTSSModuleInterface::GetErrorLogStream());

	//
	// SETUP ASSERT BEHAVIOR
	//
	// Depending on the server preference, we will either break when we hit an
	// assert, or log the assert to the error log
	if (!fSrvrPrefs->ShouldServerBreakOnAssert())
		SetAssertLogger(&sErrorLogStream);// the error log stream is our assert logger
		
	//
	// CREATE GLOBAL OBJECTS
	fSocketPool = new RTPSocketPool();
	fRTPMap = new OSRefTable(kRTPSessionMapSize);
	
	//
	// Load ERROR LOG module only. This is good in case there is a startup error.
	
	QTSSModule* theLoggingModule = new QTSSModule("QTSSErrorLogModule");
	(void)theLoggingModule->Initialize(&sCallbacks, &QTSSErrorLogModule_Main);
	(void)AddModule(theLoggingModule);
	this->BuildModuleRoleArrays();

	//
	// DEFAULT IP ADDRESS & DNS NAME
	if (!this->GetDefaultIPAddr())
		return false;
		
	//
	// BEGIN LISTENING
	UInt32	numListeners;
	if ( !this->CreateListeners(fSrvrPrefs, inPortOverride, &numListeners) )
		QTSSModuleUtils::LogError(qtssWarningVerbosity, qtssMsgSomePortsFailed, 0);
	else if ( numListeners == 0 )
		QTSSModuleUtils::LogError(qtssWarningVerbosity, qtssMsgNoPortsSucceeded, 0);
	// Server still starts up even if it couldn't listen on any port.
		
	//
	// SETUP UDP SOCKETS
	if (!this->SetupUDPSockets())
		return false;

	//
	// SWITCH TO RUN USER AND GROUP ID
	if (!this->SwitchPersonality())
		return false;

	//
	// LOAD AND INITIALIZE ALL MODULES
		
	LoadModules(fSrvrPrefs);
	LoadCompiledInModules();
	this->BuildModuleRoleArrays();

	// Finish setting up modules. Create our final prefs & messages objects,
	// register all global dictionaries, and invoke the modules in their Init roles.
	fSrvrPrefs = new QTSSPrefs(inPrefsSource);
	fSrvrMessages = new QTSSMessages(inMessagesSource);
	QTSSModuleUtils::Initialize(fSrvrMessages, fSrvrPrefs, QTSSModuleInterface::GetErrorLogStream());

	//
	// ADD REREAD PREFERENCES SERVICE
	(void)QTSSModuleInterface::GetDictionaryMap(QTSSModuleInterface::kServiceDictIndex)->
				AddAttribute(QTSS_REREAD_PREFS_SERVICE, (QTSS_AttrFunctionPtr)QTSServer::RereadPrefsService, true);

	//
	// INVOKE INITIALIZE ROLE
	this->DoInitRole();
	
	if (fServerState != qtssFatalErrorState)
		fServerState = qtssRunningState; // Server is done starting up!
		
	return true;
}

void QTSServer::StartTasks()
{
	fRTCPTask = new RTCPTask();
	fStatsTask = new RTPStatsUpdaterTask();
}

Bool16 QTSServer::GetDefaultIPAddr()
{
	//check to make sure there is an available ip interface
	if (SocketUtils::GetNumIPAddrs() == 0)
	{
		QTSSModuleUtils::LogError(qtssFatalVerbosity, qtssMsgNotConfiguredForIP, 0);
		return false;
	}

	//find out what our default IP addr is & dns name
	if (fSrvrPrefs->GetRTSPIPAddress() == INADDR_ANY)
		fDefaultIPAddr = SocketUtils::GetIPAddr(0);
	else
		fDefaultIPAddr = fSrvrPrefs->GetRTSPIPAddress();
		
	for (UInt32 ipAddrIter = 0; ipAddrIter < SocketUtils::GetNumIPAddrs(); ipAddrIter++)
	{
		if (SocketUtils::GetIPAddr(ipAddrIter) == fDefaultIPAddr)
		{
			this->SetVal(qtssSvrDefaultDNSName, SocketUtils::GetDNSNameStr(ipAddrIter));
			Assert(this->GetValue(qtssSvrDefaultDNSName)->Ptr != NULL);
			break;
		}
	}
	if (this->GetValue(qtssSvrDefaultDNSName)->Ptr == NULL)
	{
		//If we've gotten here, what has probably happened is the IP address (explicitly
		//entered as a preference) doesn't exist
		QTSSModuleUtils::LogError(qtssFatalVerbosity, qtssMsgDefaultRTSPAddrUnavail, 0);
		return false;	
	}
	return true;
}				


Bool16 QTSServer::CreateListeners(QTSSPrefs* inPrefs, UInt16 inPortOverride, UInt32 *numListeners)
{
	// return true if some listeners were desired and created.
	// set numListeners to the number succcessully opened
	
	QTSS_Error err = QTSS_NoErr;
	
	//If there is a port override in effect, ignore any settings, and use that port
	if (inPortOverride != 0)
	{
		//Because of a bug in the GCC MacOS X compiler, we can't create an actual array of objects
		fListeners = NEW TCPListenerSocket*[2];
		fListeners[0] = NEW RTSPListenerSocket();

		err = this->SetValue(qtssSvrRTSPPorts, 0, &inPortOverride, sizeof(inPortOverride), false);
		Assert(err == QTSS_NoErr);
		
		err = fListeners[0]->Initialize(inPrefs->GetRTSPIPAddress(), inPortOverride);
		fNumListeners = 1;
		
		if ( err == QTSS_NoErr )
			*numListeners = 1;
			
		return (err == QTSS_NoErr);
	}

	static const UInt32 kMaxPrefValueSize = 1024;
	char thePrefValue[kMaxPrefValueSize];

	//Figure out how many listeners we need (based on the number of ports we are listening on)

	//loop through all the ports. The first time, just to find out how many,
	//the second time, to actually read the values into the array
	Bool16 ok = true;
	for (fNumListeners = 0; ok; fNumListeners++)
	{
		thePrefValue[0] = '\0';
		// function return 0 if it couldn't find the preference
		ok = (Bool16)inPrefs->GetPrefsSource()->
				GetValueByIndex(sPortPrefString, fNumListeners, &thePrefValue[0]);
	}
	fNumListeners--;

	//if there aren't any ports entered, server won't listen on anything
	if (fNumListeners == 0)
	{	*numListeners = 0;
		return true;
	}
	
	//Now that we know how many listeners we need to create, go ahead and create them
	//Because of a bug in the GCC MacOS X compiler, we can't create an actual array of objects
	fListeners = NEW TCPListenerSocket*[fNumListeners + 2];
	
	UInt32 numOpened = 0;
	
	//Open all the listeners on all the proper ports
	for (UInt32 theIndex = 0; theIndex < fNumListeners; theIndex++)
	{
		fListeners[theIndex] = NEW RTSPListenerSocket();

		thePrefValue[0] = '\0';
		ok = (Bool16)inPrefs->GetPrefsSource()->
				GetValueByIndex(sPortPrefString, theIndex, &thePrefValue[0]);
		Assert(ok);
		
		if (thePrefValue[0] != '\0')
		{
			UInt16 thePort = (UInt16)::strtol(thePrefValue, NULL, 10);
			err = fListeners[theIndex]->Initialize(inPrefs->GetRTSPIPAddress(), thePort);
			if (err == EADDRINUSE)
				QTSSModuleUtils::LogError(qtssWarningVerbosity, qtssListenPortInUse, 0, thePrefValue);
			else if (err == EACCES)
				QTSSModuleUtils::LogError(qtssWarningVerbosity, qtssListenPortAccessDenied, 0, thePrefValue);
			else if (err != QTSS_NoErr)
				QTSSModuleUtils::LogError(qtssWarningVerbosity, qtssListenPortError, 0, thePrefValue);
			else
				// Add this port to the dictionary
				(void)this->SetValue(qtssSvrRTSPPorts, numOpened++, &thePort, sizeof(thePort), false);
		}
	}
	
	// Return true if we opened 1, false otherwise
	*numListeners = numOpened;
	
	return (numOpened > 0);
}

Bool16	QTSServer::SetupUDPSockets()
{
	//function finds all IP addresses on this machine, and binds 1 RTP / RTCP
	//socket pair to a port pair on each address.
	
	UInt32 theNumAllocatedPairs = 0;
	for (UInt32 theNumPairs = 0; theNumPairs < SocketUtils::GetNumIPAddrs(); theNumPairs++)
	{
		if (fSocketPool->CreateUDPSocketPair(SocketUtils::GetIPAddr(theNumPairs), 0) != NULL)
			theNumAllocatedPairs++;
	}
	//only return an error if we couldn't allocate ANY pairs of sockets
	if (theNumAllocatedPairs == 0)
		return false;
	return true;
}

Bool16	QTSServer::SwitchPersonality()
{
#if !__MACOS__
	OSCharArrayDeleter runGroupName(fSrvrPrefs->GetRunGroupName());
	if (::strlen(runGroupName.GetObject()) > 0)
	{
		struct group* gr = ::getgrnam(runGroupName.GetObject());
		if (gr == NULL || ::setgid(gr->gr_gid) == -1)
		{
			QTSSModuleUtils::LogError(qtssFatalVerbosity, qtssMsgCannotSetRunGroup, 0,
					runGroupName.GetObject(), strerror(OSThread::GetErrno()));
			return false;
		}
	}
	
	OSCharArrayDeleter runUserName(fSrvrPrefs->GetRunUserName());
	if (::strlen(runUserName.GetObject()) > 0)
	{
		struct passwd* pw = ::getpwnam(runUserName.GetObject());
		if (pw == NULL || ::setuid(pw->pw_uid) == -1)
		{
			QTSSModuleUtils::LogError(qtssFatalVerbosity, qtssMsgCannotSetRunUser, 0,
					runUserName.GetObject(), strerror(OSThread::GetErrno()));
			return false;
		}
	}
#endif	
	return true;
}

#pragma mark __MODULE_LOADING_ROUTINES__

void	QTSServer::LoadCompiledInModules()
{
	// MODULE DEVELOPERS SHOULD ADD THE FOLLOWING THREE LINES OF CODE TO THIS
	// FUNCTION IF THEIR MODULE IS BEING COMPILED INTO THE SERVER.
	//
	// QTSSModule* myModule = new QTSSModule("__MODULE_NAME__");
	// (void)myModule->Initialize(&sCallbacks, &__MODULE_MAIN_ROUTINE__);
	// (void)AddModule(myModule);
	//
	// The following modules are all compiled into the server. 
	
	QTSSModule* theFileModule = new QTSSModule("QTSSFileModule");
	(void)theFileModule->Initialize(&sCallbacks, &QTSSFileModule_Main);
	(void)AddModule(theFileModule);

	QTSSModule* theReflectorModule = new QTSSModule("QTSSReflectorModule");
	(void)theReflectorModule->Initialize(&sCallbacks, &QTSSReflectorModule_Main);
	(void)AddModule(theReflectorModule);

	QTSSModule* theSplitterModule = new QTSSModule("QTSSSplitterModule");
	(void)theSplitterModule->Initialize(&sCallbacks, &QTSSSplitterModule_Main);
	(void)AddModule(theSplitterModule);

	QTSSModule* theRelayModule = new QTSSModule("QTSSRelayModule");
	(void)theRelayModule->Initialize(&sCallbacks, &QTSSRelayModule_Main);
	(void)AddModule(theRelayModule);

#if __MacOSX__
	QTSSModule* theSCmodule = new QTSSModule("QTSSSvrControlModule");
	(void)theSCmodule->Initialize(&sCallbacks, &QTSSvrControlModule_Main);
	(void)AddModule(theSCmodule);
#endif

	QTSSModule* theAccessLog = new QTSSModule("QTSSAccessLogModule");
	(void)theAccessLog->Initialize(&sCallbacks, &QTSSAccessLogModule_Main);
	(void)AddModule(theAccessLog);

	QTSSModule* theFlowControl = new QTSSModule("QTSSFlowControlModule");
	(void)theFlowControl->Initialize(&sCallbacks, &QTSSFlowControlModule_Main);
	(void)AddModule(theFlowControl);

	QTSSModule* theWebStats = new QTSSModule("QTSSWebStatsModule");
	(void)theWebStats->Initialize(&sCallbacks, &QTSSWebStatsModule_Main);
	(void)AddModule(theWebStats);

#if MEMORY_DEBUGGING
	QTSSModule* theWebDebug = new QTSSModule("QTSSWebDebugModule");
	(void)theWebDebug->Initialize(&sCallbacks, &QTSSWebDebugModule_Main);
	(void)AddModule(theWebDebug);
#endif

#if QTACCESS_DEBUGGING
	QTSSModule* theQTACESSmodule = new QTSSModule("QTSSAccessModule");
	(void)theQTACESSmodule->Initialize(&sCallbacks, &QTSSAccessModule_Main);
	(void)AddModule(theQTACESSmodule);
#endif

}



void	QTSServer::InitCallbacks()
{
	sCallbacks.addr[kNewCallback] = 				(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_New;
	sCallbacks.addr[kDeleteCallback] = 				(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_Delete;
	sCallbacks.addr[kMillisecondsCallback] = 		(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_Milliseconds;
	sCallbacks.addr[kConvertToUnixTimeCallback] = 	(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_ConvertToUnixTime;

	sCallbacks.addr[kAddRoleCallback] = 			(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_AddRole;
	sCallbacks.addr[kAddAttributeCallback] = 		(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_AddAttribute;
	sCallbacks.addr[kIDForTagCallback] = 			(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_IDForAttr;
	sCallbacks.addr[kGetAttributePtrByIDCallback] = (QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_GetValuePtr;
	sCallbacks.addr[kGetAttributeByIDCallback] = 	(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_GetValue;
	sCallbacks.addr[kSetAttributeByIDCallback] = 	(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_SetValue;

	sCallbacks.addr[kWriteCallback] = 				(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_Write;
	sCallbacks.addr[kWriteVCallback] = 				(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_WriteV;

	sCallbacks.addr[kAddServiceCallback] = 			(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_AddService;
	sCallbacks.addr[kIDForServiceCallback] = 		(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_IDForService;
	sCallbacks.addr[kDoServiceCallback] = 			(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_DoService;

	sCallbacks.addr[kSendRTSPHeadersCallback] = 	(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_SendRTSPHeaders;
	sCallbacks.addr[kAppendRTSPHeadersCallback] = 	(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_AppendRTSPHeader;
	sCallbacks.addr[kSendStandardRTSPCallback] = 	(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_SendStandardRTSPResponse;

	sCallbacks.addr[kAddRTPStreamCallback] = 		(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_AddRTPStream;
	sCallbacks.addr[kPlayCallback] = 				(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_Play;
	sCallbacks.addr[kPauseCallback] = 				(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_Pause;
	sCallbacks.addr[kTeardownCallback] = 			(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_Teardown;

	sCallbacks.addr[kCreateEventContextCallback] = 	(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_CreateEventContext;
	sCallbacks.addr[kDestroyEventContextCallback] = (QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_DestroyEventContext;
	sCallbacks.addr[kRequestEventCallback] = 		(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_RequestEvent;
	sCallbacks.addr[kSetIdleTimerCallback] = 		(QTSS_CallbackProcPtr)QTSSCallbacks::QTSS_SetIdleTimer;
}

#if 0
char* QTSServer::GetModuleDirectory(QTSSPrefs* inPrefs)
{
	char* theModuleDirName = inPrefs->GetModuleDirectory();
	Assert(theModuleDirName != NULL);
	
	UInt32 dirSize = 256;
	char* dirPath = NULL;
	
	while (true)
	{
		// We don't know how big the path to the current working directory is,
		// so loop until we pass in a buffer that's big enough.
		
		char* theCurrentDir = new char[dirSize];
		dirPath = ::getcwd(theCurrentDir, dirSize);
		
		// Check to make sure this buffer can accomodate the moduleDirName
		// being appended.
		if ((dirPath != NULL) && (::strlen(dirPath) < (dirSize - (::strlen(theModuleDirName) + 2))))
			break;
			
		// If this buffer is not big enough, free it up and allocate 2wice
		delete [] theCurrentDir;
		dirSize *= 2;
	}
	
	// Append the module dir name.
	::strcat(dirPath, "/");
	::strcat(dirPath, theModuleDirName);
	return dirPath;
}
#endif

void QTSServer::LoadModules(QTSSPrefs* inPrefs)
{
	// Fetch the name of the module directory and open it.
	
	OSCharArrayDeleter theModDirName(inPrefs->GetModuleDirectory());
	Assert(theModDirName != NULL);
		
	DIR* theDir = ::opendir(theModDirName.GetObject());
	if (theDir == NULL)
	{
		QTSSModuleUtils::LogError(qtssWarningVerbosity, qtssMsgNoModuleFolder, 0);	
		return;
	}
	
	while (true)
	{
		// Iterate over each file in the directory, attempting to construct
		// a module object from that file.
		
		struct dirent* theFile = ::readdir(theDir);
		if (theFile == NULL)
			break;
		
		// Ignore these silly directory names
		
		if (::strcmp(theFile->d_name, ".") == 0)
			continue;
		if (::strcmp(theFile->d_name, "..") == 0)
			continue;
		
		{
			//
			// Construct a full path to this module
			UInt32 totPathLen = ::strlen(theModDirName.GetObject()) + ::strlen(theFile->d_name);
			OSCharArrayDeleter theModPath(NEW char[totPathLen + 4]);
			::strcpy(theModPath.GetObject(), theModDirName.GetObject());
			::strcat(theModPath.GetObject(), "/");
			::strcat(theModPath.GetObject(), theFile->d_name);
			
			//
			// Construct a QTSSModule object, and attempt to initialize the module
			QTSSModule* theNewModule = NEW QTSSModule(theFile->d_name, theModPath.GetObject());
			QTSS_Error theErr = theNewModule->Initialize(&sCallbacks);
			
			if (theErr != QTSS_NoErr)
			{
				QTSSModuleUtils::LogError(qtssWarningVerbosity, qtssMsgBadModule, theErr,
												theNewModule->GetName()->Ptr);
				delete theNewModule;
			}
			//
			// If the module was successfully initialized, add it to our module queue
			else if (!this->AddModule(theNewModule))
			{
				QTSSModuleUtils::LogError(qtssWarningVerbosity, qtssMsgRegFailed, theErr,
												theNewModule->GetName()->Ptr);
				delete theNewModule;
			}
		}
	}		
}

Bool16 QTSServer::AddModule(QTSSModule* inModule)
{
	Assert(inModule->IsInitialized());
	
	// Prepare to invoke the module's Register role. Setup the Register param block
	QTSS_ModuleState theModuleState;

	theModuleState.curModule = inModule;
	theModuleState.curRole = QTSS_Register_Role;
	theModuleState.curTask = NULL;
	OSThread::SetMainThreadData(&theModuleState);
	
	// Currently we do nothing with the module name
	QTSS_RoleParams theRegParams;
	
	// If the module returns an error from the QTSS_Register role, don't put it anywhere
	if (inModule->CallDispatch(QTSS_Register_Role, &theRegParams) != QTSS_NoErr)
		return false;
		
	OSThread::SetMainThreadData(NULL);
		
	sModuleQueue.EnQueue(inModule->GetQueueElem());
	return true;
}
 
void QTSServer::DoInitRole()
{
	QTSS_RoleParams theInitParams;
	theInitParams.initParams.inServer = 		this;
	theInitParams.initParams.inPrefs = 			fSrvrPrefs;
	theInitParams.initParams.inMessages = 		fSrvrMessages;
	theInitParams.initParams.inErrorLogStream = &sErrorLogStream;
		
	QTSS_ModuleState theModuleState;
	theModuleState.curRole = QTSS_Initialize_Role;
	theModuleState.curTask = NULL;
	OSThread::SetMainThreadData(&theModuleState);

	for (UInt32 x = 0; x < QTSSModuleInterface::GetNumModulesInRole(QTSSModule::kInitializeRole); x++)
	{
		QTSSModule* theModule = QTSSModuleInterface::GetModule(QTSSModule::kInitializeRole, x);
		QTSS_Error theErr = theModule->CallDispatch(QTSS_Initialize_Role, &theInitParams);

		if (theErr != QTSS_NoErr)
		{
			// If the module reports an error when initializing itself,
			// delete the module and pretend it was never there.
			QTSSModuleUtils::LogError(qtssWarningVerbosity, qtssMsgInitFailed, theErr,
											theModule->GetName()->Ptr);
											
			sModuleQueue.Remove(theModule->GetQueueElem());
			delete theModule;
		}
	}

	OSThread::SetMainThreadData(NULL);
}

void QTSServer::BuildModuleRoleArrays()
{
	OSQueueIter theIter(&sModuleQueue);
	QTSSModule* theModule = NULL;
		
	// Make sure these variables are cleaned out in case they've already been inited.
	
	DestroyModuleRoleArrays();

	// Loop through all the roles of all the modules, recording the number of
	// modules in each role, and also recording which modules are doing what.
	
	for (UInt32 x = 0; x < QTSSModule::kNumRoles; x++)
	{
		sNumModulesInRole[x] = 0;
		for (theIter.Reset(); !theIter.IsDone(); theIter.Next())
		{
			theModule = (QTSSModule*)theIter.GetCurrent()->GetEnclosingObject();
			if (theModule->RunsInRole(x))
				sNumModulesInRole[x] += 1;
		}
	
		if (sNumModulesInRole[x] > 0)
		{
			UInt32 moduleIndex = 0;
			sModuleArray[x] = new QTSSModule*[sNumModulesInRole[x] + 1];
			for (theIter.Reset(); !theIter.IsDone(); theIter.Next())
			{
				theModule = (QTSSModule*)theIter.GetCurrent()->GetEnclosingObject();
				if (theModule->RunsInRole(x))
				{
					sModuleArray[x][moduleIndex] = theModule;
					moduleIndex++;
				}
			}
		}
	}
}

void QTSServer::DestroyModuleRoleArrays()
{
	for (UInt32 x = 0; x < QTSSModule::kNumRoles; x++)
	{
		sNumModulesInRole[x] = 0;
		if (sModuleArray[x] != NULL)
			delete [] sModuleArray[x];
		sModuleArray[x] = NULL;	
	}
}

#pragma mark __RTSP_LISTENER_SOCKET__

Task*	RTSPListenerSocket::GetSessionTask(TCPSocket** outSocket)
{
	Assert(outSocket != NULL);
	
	// when the server is behing a round robin DNS, the client needs to knwo the IP address ot the server
	// so that it can direct the "POST" half of the connection to the same machine when tunnelling RTSP thru HTTP
	Bool16	doReportHTTPConnectionAddress = QTSServerInterface::GetServer()->GetPrefs()->GetDoReportHTTPConnectionAddress();
	
	RTSPSession* theTask = NEW RTSPSession(doReportHTTPConnectionAddress);
	*outSocket = theTask->GetSocket();	// out socket is not attached to a unix socket yet.

	return theTask;
}


#pragma mark __RTP_SOCKET_POOL__


UDPSocketPair*	RTPSocketPool::ConstructUDPSocketPair()
{
	Task* theTask = ((QTSServer*)QTSServerInterface::GetServer())->fRTCPTask;
	
	//construct a pair of UDP sockets, the lower one for RTP data (outgoing only, no demuxer
	//necessary), and one for RTCP data (incoming, so definitely need a demuxer).
	//These are nonblocking sockets that DON'T receive events (we are going to poll for data)
	return NEW
		UDPSocketPair(	NEW UDPSocket(theTask, Socket::kNonBlockingSocketType),
						NEW UDPSocket(theTask, UDPSocket::kWantsDemuxer | Socket::kNonBlockingSocketType));
}

void RTPSocketPool::DestructUDPSocketPair(UDPSocketPair* inPair)
{
	delete inPair->GetSocketA();
	delete inPair->GetSocketB();
	delete inPair;
}

#pragma mark __REREAD_PREFS_SERVICE__

QTSS_Error QTSServer::RereadPrefsService(QTSS_ServiceFunctionArgsPtr /*inArgs*/)
{
	//
	// This function can only be called safely when the server is completely running.
	// Ensuring this is a bit complicated because of preemption. Here's how it's done...
	
	QTSServerInterface* theServer = QTSServerInterface::GetServer();
	
	// This is to make sure this function isn't being called before the server is
	// completely started up.
	if ((theServer == NULL) || (theServer->GetServerState() != qtssRunningState))
		return QTSS_OutOfState;

	// Because the server must have started up, and because this object always stays
	// around (until the process dies), we can now safely get this object.
	QTSSPrefs* thePrefs = theServer->GetPrefs();
	
	// Grab the prefs mutex. We want to make sure that calls to RereadPrefsService
	// are serialized. This also prevents the server from shutting down while in
	// this function, because the QTSServer destructor grabs this mutex as well.
	OSMutexLocker locker(thePrefs->GetMutex());
	
	// Finally, check the server state again. The state may have changed
	// to qtssShuttingDownState or qtssFatalErrorState in this time, though
	// at this point we have the prefs mutex, so we are guarenteed that the
	// server can't actually shut down anymore
	if (theServer->GetServerState() != qtssRunningState)
		return QTSS_OutOfState;
	
	// Ok, we're ready to reread preferences now.
	
	//
	// Reread preferences
	thePrefs->RereadPreferences();

	//
	// Now that we are done rereading the prefs, invoke all modules in the RereadPrefs
	// role so they can update their internal prefs caches.
	for (UInt32 x = 0; x < QTSSModuleInterface::GetNumModulesInRole(QTSSModule::kRereadPrefsRole); x++)
	{
		QTSSModule* theModule = QTSSModuleInterface::GetModule(QTSSModule::kRereadPrefsRole, x);
		(void)theModule->CallDispatch(QTSS_RereadPrefs_Role, NULL);
	}
	return QTSS_NoErr;
}


