/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		QTSServer.h

	Contains:	This object is responsible for bringing up & shutting down
				the server. It also loads & initializes all modules.
	
	$Log: QTSServer.h,v $
	Revision 1.3  2000/01/28 08:31:15  serenyi
	Fixed startup sequence so RTCPs get processed
	
	Revision 1.2  2000/01/27 02:09:13  rtucker
	Merge changes from v81
	
	Revision 1.5  2000/01/26 18:33:26  serenyi
	#2426075 Integrated changes from Kostya Vasilyev
	
	Revision 1.4  2000/01/21 23:52:56  rtucker
	
	logging changes
	
	Revision 1.3  1999/12/21 01:44:05  lecroy
	*** empty log message ***
	
	Revision 1.2  1999/10/07 08:52:44  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.1.1.1  1999/09/22 23:01:50  lecroy
	importing
	
	Revision 1.2  1999/09/15 17:04:39  serenyi
	Changed RereadPrefsService to be safely callable at any time
	
	Revision 1.1.1.1  1999/09/13 16:16:53  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:08:35  ds
	Created
	

*/

#ifndef __QTSSERVER_H__
#define __QTSSERVER_H__

#include "QTSServerInterface.h"
#include "QTSSModuleInterface.h"

class RTCPTask;
class RTSPListenerSocket;
class RTPSocketPool;

class QTSServer : public QTSServerInterface, public QTSSModuleInterface
{
	public:

		QTSServer() {}
		virtual ~QTSServer();

		//
		// Initialize
		//
		// This function starts the server. If it returns true, the server has
		// started up sucessfully. If it returns false, a fatal error occurred
		// while attempting to start the server.
		//
		// This function *must* be called before the server creates any threads,
		// because one of its actions is to change the server to the right UID / GID.
		// Threads will only inherit these if they are created afterwards.
		Bool16 Initialize(PrefsSource* inPrefsSource, PrefsSource* inMessagesSource,
							UInt16 inPortOverride);
							
		
		//
		// StartTasks
		//
		// The server has certain global tasks that it runs for things like stats
		// updating and RTCP processing. This function must be called to start those
		// going, and it must be called after Initialize				
		void StartTasks();


		//
		// RereadPrefsService
		//
		// This service is registered by the server (calling "RereadPreferences").
		// It rereads the preferences. Anyone can call this to reread the preferences,
		// and it may be called safely at any time, though it will fail with a
		// QTSS_OutOfState if the server isn't in the qtssRunningState.
		
		static QTSS_Error RereadPrefsService(QTSS_ServiceFunctionArgsPtr inArgs);

	private:
	
		//
		// GLOBAL TASKS
		RTCPTask* 			fRTCPTask;
		RTPStatsUpdaterTask*fStatsTask;

		static char*		sPortPrefString;
		
		//
		// Module loading & unloading routines
		
		static QTSS_Callbacks	sCallbacks;
		
		// Sets up QTSS API callback routines
		void					InitCallbacks();
		
		// Loads compiled-in modules
		void					LoadCompiledInModules();

		// Loads modules from disk
		void 					LoadModules(QTSSPrefs* inPrefs);
		
		// Adds a module to the module array
		Bool16				 	AddModule(QTSSModule* inModule);
		
		// Call module init roles
		void 					DoInitRole();
		
		// Build & destroy the optimized role / module arrays for invoking modules
		void 					BuildModuleRoleArrays();
		void 					DestroyModuleRoleArrays();
		
		//Returns false if server couldn't listen on one or more of the ports, true otherwise
		Bool16 					CreateListeners(QTSSPrefs* inPrefs, UInt16 inPortOverride, UInt32 *numListeners);

		Bool16 					GetDefaultIPAddr();
		Bool16					SetupUDPSockets();

		Bool16					SwitchPersonality();
		
#ifndef __MACOS__
		static pid_t 			sMainPid;
#endif
				
		friend class RTPSocketPool;
};
#endif // __QTSSERVER_H__


