/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		QTSServerInterface.h

	Contains:	This object defines an interface for getting and setting server-wide
				attributes, and storing global server resources.
				
				There can be only one of these objects per process, so there
				is a static accessor.
					
	$Log: QTSServerInterface.h,v $
	Revision 1.1.1.1  2000/01/17 23:36:03  wsanchez
	Import of StreamingServer-2.0 (Apple)
	
	Revision 1.6  1999/12/21 01:44:06  lecroy
	*** empty log message ***
	
	Revision 1.5  1999/12/02 19:26:27  serenyi
	Got rid of "RoadRunner", added a platform string to the Server: header
	
	Revision 1.4  1999/10/14 17:59:55  lecroy
	Added qtssAPIVersion attribute
	
	Revision 1.3  1999/10/07 08:52:44  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.2  1999/09/30 00:30:39  serenyi
	Added support for HTTP session attribute
	
	Revision 1.1.1.1  1999/09/22 23:01:51  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:53  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:08:37  ds
	Created
	

*/


#ifndef __QTSSERVERINTERFACE_H__
#define __QTSSERVERINTERFACE_H__

#include "QTSS.h"
#include "QTSSDictionary.h"
#include "QTSSPrefs.h"
#include "QTSSMessages.h"
#include "atomic.h"

#include "OSMutex.h"
#include "Task.h"
#include "TCPListenerSocket.h"

class OSRefTable;
class UDPSocketPool;
class QTSSPrefs;
class QTSSMessages;
class RTPStatsUpdaterTask;
class RTPSessionInterface;

class QTSServerInterface : public QTSSDictionary
{
	public:
	
		//Initialize must be called right off the bat to initialize dictionary resources
		static void		Initialize();

		//
		// CONSTRUCTOR / DESTRUCTOR
		
		QTSServerInterface();
		virtual ~QTSServerInterface() {}
		
		//
		//
		// STATISTICS MANIPULATION
		// These functions are how the server keeps its statistics current
		
		void				AlterCurrentRTSPSessionCount(SInt32 inDifference)
			{ OSMutexLocker locker(&fMutex); fNumRTSPSessions += inDifference; }
		void				AlterCurrentRTSPHTTPSessionCount(SInt32 inDifference)
			{ OSMutexLocker locker(&fMutex); fNumRTSPHTTPSessions += inDifference; }
		void				SwapFromRTSPToHTTP()
			{ OSMutexLocker locker(&fMutex); fNumRTSPSessions--; fNumRTSPHTTPSessions++; }
			
		//total rtp bytes sent by the server
		void			IncrementTotalRTPBytes(UInt32 bytes)
										{ (void)atomic_add(&fPeriodicRTPBytes, bytes); }
		//total rtp packets sent by the server
		void			IncrementTotalPackets()
										{ (void)atomic_add(&fPeriodicRTPPackets, 1); }
										
		// Also increments current RTP session count
		void			IncrementTotalRTPSessions()
			{ OSMutexLocker locker(&fMutex); fNumRTPSessions++; fTotalRTPSessions++; }
		void			AlterCurrentRTPSessionCount(SInt32 inDifference)
			{ OSMutexLocker locker(&fMutex); fNumRTPSessions += inDifference; }

		
		//
		// ACCESSORS
		
		QTSS_ServerState	GetServerState() 		{ return fServerState; }
		UInt32				GetNumRTPSessions()		{ return fNumRTPSessions; }
		UInt32				GetTotalRTPSessions()	{ return fTotalRTPSessions; }
		UInt32				GetCurBandwidthInBits()	{ return fCurrentRTPBandwidthInBits; }
		UInt32				GetAvgBandwidthInBits()	{ return fAvgRTPBandwidthInBits; }
		UInt32				GetRTPPacketsPerSec()	{ return fRTPPacketsPerSecond; }
		UInt64				GetTotalRTPBytes()		{ return fTotalRTPBytes; }
		
		//
		//
		// GLOBAL OBJECTS REPOSITORY
		// This object is in fact global, so there is an accessor for it as well.
		
		static QTSServerInterface*	GetServer()			{ return sServer; }
		
		//Allows you to map RTP session IDs (strings) to actual RTP session objects
		OSRefTable*			GetRTPSessionMap() 			{ return fRTPMap; }
	
		//Server provides a statically created & bound UDPSocket / Demuxer pair
		//for each IP address setup to serve RTP. You access those pairs through
		//this function. This returns a pair pre-bound to the IPAddr specified.
		UDPSocketPool*		GetSocketPool() 			{ return fSocketPool; }

		QTSSPrefs*			GetPrefs()					{ return fSrvrPrefs; }
		QTSSMessages*		GetMessages()				{ return fSrvrMessages; }
		
		//
		//
		// SERVER NAME & VERSION
		
		static StrPtrLen&	GetServerName()				{ return sServerNameStr; }
		static StrPtrLen&	GetServerVersion() 			{ return sServerVersionStr; }
		static StrPtrLen&	GetServerPlatform() 		{ return sServerPlatformStr; }
		static StrPtrLen&	GetServerBuildDate() 		{ return sServerBuildDateStr; }
		static StrPtrLen&	GetServerHeader()			{ return sServerHeaderPtr; }
		
		
	protected:

		// Setup by the derived RTSPServer object

		//Sockets are allocated global to the server, and arbitrated through this pool here.
		//RTCP data is processed completely within the following task.
		UDPSocketPool*				fSocketPool;
		
		// All RTP sessions are put into this map
		OSRefTable*					fRTPMap;
		
		QTSSPrefs*					fSrvrPrefs;
		QTSSMessages*				fSrvrMessages;

		QTSS_ServerState			fServerState;
		UInt32						fDefaultIPAddr;
		
		// Array of pointers to TCPListenerSockets.
		TCPListenerSocket** 		fListeners;
		UInt32						fNumListeners; // Number of elements in the array
		
	private:
	
		enum
		{
			kMaxServerHeaderLen = 100
		};

		static UInt32		sServerAPIVersion;
		static StrPtrLen	sServerNameStr;
		static StrPtrLen	sServerVersionStr;
		static StrPtrLen	sServerPlatformStr;
		static StrPtrLen	sServerBuildDateStr;
		static char			sServerHeader[kMaxServerHeaderLen];
		static StrPtrLen	sServerHeaderPtr;

		OSMutex				fMutex;
		UInt32				fNumRTSPSessions;
		UInt32				fNumRTSPHTTPSessions;
		UInt32				fNumRTPSessions;

		//stores the total number of connections since startup.
		UInt32				fTotalRTPSessions;
		//stores the total number of bytes served since startup
		UInt64				fTotalRTPBytes;
		//total number of rtp packets sent since startup
		UInt64				fTotalRTPPackets;

		//because there is no 64 bit atomic add (for obvious reasons), we efficiently
		//implement total byte counting by atomic adding to this variable, then every
		//once in awhile updating the sTotalBytes.
		unsigned int		fPeriodicRTPBytes;
		unsigned int		fPeriodicRTPPackets;
		
		//stores the current served bandwidth in BITS per second
		UInt32				fCurrentRTPBandwidthInBits;
		UInt32				fAvgRTPBandwidthInBits;
		UInt32				fRTPPacketsPerSecond;
		
		// stores # of UDP sockets in the server currently (gets updated lazily via.
		// param retrieval function)
		UInt32				fTotalUDPSockets;
		
		// are we out of descriptors?
		Bool16				fIsOutOfDescriptors;
		
		// Param retrieval functions
		static void* GetTotalUDPSockets(QTSS_Object inServer, UInt32* outLen);
		static void* IsOutOfDescriptors(QTSS_Object inServer, UInt32* outLen);

		static QTSServerInterface*	sServer;
		static QTSSDictionaryMap::AttrInfo	sAttributes[];
		
		friend class RTPStatsUpdaterTask;
};

class RTPStatsUpdaterTask : public Task
{
	public:
	
		// This class runs periodically to compute current totals & averages
		RTPStatsUpdaterTask();
		virtual ~RTPStatsUpdaterTask() {}
	
	private:
	
		virtual SInt64 Run();
		RTPSessionInterface* GetNewestSession(OSRefTable* inRTPSessionMap);

		
		SInt64 fLastBandwidthTime;
		SInt64 fLastBandwidthAvg;
		SInt64 fLastBytesSent;
};

#endif // __QTSSERVERINTERFACE_H__

