/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		RTPStream.cpp

	Contains:	Implementation of RTPStream class. 
					
	$Log: RTPStream.cpp,v $
	Revision 1.2  2000/02/11 01:07:23  serenyi
	Server now supports cache control header and sends 304 Not Modified responses in appropriate situations
	
	Revision 1.1.1.1  2000/01/17 23:36:03  wsanchez
	Import of StreamingServer-2.0 (Apple)
	
	Revision 1.21  1999/12/21 01:44:08  lecroy
	*** empty log message ***
	
	Revision 1.20  1999/12/15 23:36:04  serenyi
	Must do a trylock when in processpacket to avoid a deadlock
	
	Revision 1.19  1999/12/13 22:29:04  serenyi
	Part of fixing 2415738. Must grab RTSPSession mutex before doing an interleaved write.
	
	Revision 1.18  1999/12/09 22:01:58  rtucker
	 only send ;timeout param in Session: when 'rtsp_timeout' > 0
	
	Revision 1.17  1999/12/08 23:25:45  rtucker
	OPTIONS now tickles RTP Stream
	we now report timeout value in first SETUP response
	
	Revision 1.16  1999/11/08 22:52:36  serenyi
	Updated code to reflect changes in QTSS API
	
	Revision 1.15  1999/10/28 20:57:49  serenyi
	Added #if DEBUG in one place
	
	Revision 1.14  1999/10/27 17:55:03  serenyi
	Integrated Matthias's MacOS / GUSI changes
	
	Revision 1.13  1999/10/27 03:28:21  lecroy
	merged in changes from Akamai
	
	Revision 1.12  1999/10/19 21:14:17  rtucker
	tcp enhancements
	
	Revision 1.11  1999/10/15 00:41:22  serenyi
	Don't update stats if we get EAGAIN!
	
	Revision 1.10  1999/10/14 01:17:45  rtucker
	updates for thinning
	
	Revision 1.9  1999/10/11 18:41:26  serenyi
	Added time units to qtssRTPStrBufferDelay param
	
	Revision 1.8  1999/10/08 01:12:51  rtucker
	added support for TCP thining and thinnig prefs
	
	Revision 1.7  1999/10/07 08:52:32  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.6  1999/10/06 00:26:11  murata
	Make the qtssRTPStrBufferDelay read/write so the file module can modify the value.
	
	Revision 1.5  1999/10/05 23:34:04  murata
	added qtssRTPStrBufferDelay suppor
	
	Revision 1.4  1999/10/05 01:21:46  serenyi
	Fixed #2392594. InterleavedWrite not properly checking for RTSPSession not being available before doing a write
	
	Revision 1.3  1999/09/29 18:31:24  rtucker
	turned RTPStream debugging off
	
	Revision 1.2  1999/09/24 21:18:24  rtucker
	added code for debugging TCP flow control
	
	Revision 1.1.1.1  1999/09/22 23:01:52  lecroy
	importing
	
	Revision 1.5  1999/09/17 22:59:06  serenyi
	Only update timeouts if valid rtcp is found
	
	Revision 1.3  1999/09/17 02:47:49  serenyi
	Added check for fRTSPSession being NULL, because it can go away sometimes
	
	Revision 1.2  1999/09/15 17:05:03  serenyi
	Wrapped code in mutexes
	
	Revision 1.1.1.1  1999/09/13 16:16:54  rtucker
	initial checkin from rrv14
	
	
	Revision 1.4  1999/02/24 19:27:20  ds
	RTPStream now uses new UDPSocketPool interface for getting its UDP sockets
	
	Revision 1.2  1999/02/19 23:08:27  ds
	Created
	

*/

/*

	9.21.99 rtucker - CVS is down, so I might forget this by the time tis back up
	
	
	added RefreshTimer calls to InterleavedWrite so that we keep the RTPSession 
	and RTSPSession alive during HTTP connections when we don't get RTCP receiver 
	reports.
	
	put code back into InterleavedWrite so that it reports the correct # of bytes
	written back up the chain
	
	
*/

#include <stdlib.h>

#include "RTPStream.h"

#include "QTSSModuleUtils.h"
#include "QTSServerInterface.h"
#include "OS.h"

#include "RTCPPacket.h"
#include "RTCPAPPPacket.h"
#include "RTCPSRPacket.h"
#include "SocketUtils.h"
#include <sys/errno.h>

#if DEBUG
#define RTP_TCP_STREAM_DEBUG 0
#else
#define RTP_TCP_STREAM_DEBUG 0
#endif

QTSSDictionaryMap::AttrInfo	RTPStream::sAttributes[] = 
{
	/* 0  */ { "qtssRTPStrTrackID", 				NULL, 	true,	true   },
	/* 1  */ { "qtssRTPStrSSRC", 					NULL, 	true,	false  }, 
	/* 2  */ { "qtssRTPStrPayloadName", 			NULL, 	true,	true   },
	/* 3  */ { "qtssRTPStrPayloadType", 			NULL, 	true,	true   },
	/* 4  */ { "qtssRTPStrFirstSeqNumber",			NULL, 	true,	true   },
	/* 5  */ { "qtssRTPStrFirstTimestamp", 			NULL, 	true,	true   },
	/* 6  */ { "qtssRTPStrTimescale", 				NULL, 	true,	true   },
	/* 7  */ { "qtssRTPStrQualityLevel", 			NULL, 	true,	true   },
	/* 8  */ { "qtssRTPStrNumQualityLevels", 		NULL, 	true,	true   },
	/* 9  */ { "qtssRTPStrBufferDelayInSecs", 		NULL, 	true,	true   },
	
	/* 10 */ { "qtssRTPStrFractionLostPackets", 	NULL, 	true,	false  },
	/* 11 */ { "qtssRTPStrTotalLostPackets", 		NULL, 	true,	false  },
	/* 12 */ { "qtssRTPStrJitter", 					NULL, 	true,	false  },
	/* 13 */ { "qtssRTPStrRecvBitRate", 			NULL, 	true,	false  },
	/* 14 */ { "qtssRTPStrAvgLateMilliseconds", 	NULL, 	true,	false  },
	/* 15 */ { "qtssRTPStrPercentPacketsLost", 		NULL, 	true,	false  },
	/* 16 */ { "qtssRTPStrAvgBufDelayInMsec", 		NULL, 	true,	false  },
	/* 17 */ { "qtssRTPStrGettingBetter", 			NULL, 	true,	false  },
	/* 18 */ { "qtssRTPStrGettingWorse", 			NULL, 	true,	false  },
	/* 19 */ { "qtssRTPStrNumEyes", 				NULL, 	true,	false  },
	/* 20 */ { "qtssRTPStrNumEyesActive", 			NULL, 	true,	false  },
	/* 21 */ { "qtssRTPStrNumEyesPaused", 			NULL, 	true,	false  },
	/* 22 */ { "qtssRTPStrTotPacketsRecv", 			NULL, 	true,	false  },
	/* 23 */ { "qtssRTPStrTotPacketsDropped", 		NULL, 	true,	false  },
	/* 24 */ { "qtssRTPStrTotPacketsLost", 			NULL, 	true,	false  },
	/* 25 */ { "qtssRTPStrClientBufFill", 			NULL, 	true,	false  },
	/* 26 */ { "qtssRTPStrFrameRate", 				NULL, 	true,	false  },
	/* 27 */ { "qtssRTPStrExpFrameRate", 			NULL, 	true,	false  },
	/* 28 */ { "qtssRTPStrAudioDryCount", 			NULL, 	true,	false  },
	/* 29 */ { "qtssRTPStrIsTCP", 					NULL, 	true,	false  },
	/* 30 */ { "qtssRTPStrStreamRef", 				NULL, 	true,	false  },
	/* 31 */ { "qtssRTPStrCurrentPacketDelay", 		NULL, 	true,	true  }

};

StrPtrLen	RTPStream::sChannelNums[] =
{
	StrPtrLen("0"),
	StrPtrLen("1"),
	StrPtrLen("2"),
	StrPtrLen("3"),
	StrPtrLen("4"),
	StrPtrLen("5"),
	StrPtrLen("6"),
	StrPtrLen("7"),
	StrPtrLen("8"),
	StrPtrLen("9")
};

QTSS_ModuleState RTPStream::sRTCPProcessModuleState = { NULL, 0, NULL, false };

void	RTPStream::Initialize()
{
	for (int x = 0; x < qtssRTPStrNumParams; x++)
		QTSSModuleInterface::GetDictionaryMap(QTSSModuleInterface::kRTPStreamDictIndex)->
			SetAttribute(x, sAttributes[x].fAttrName, sAttributes[x].fFuncPtr,
				sAttributes[x].fIsPreemptiveSafe, sAttributes[x].fIsWriteable);
}

RTPStream::RTPStream(UInt32 inSSRC, RTPSessionInterface* inSession)
: 	QTSSDictionary(QTSSModuleInterface::GetDictionaryMap(QTSSModuleInterface::kRTPStreamDictIndex), NULL),
	fSockets(NULL),
	fSession(inSession),
	fRemoteAddr(0),
	fRemoteRTPPort(0),
	fRemoteRTCPPort(0),
	fLastSenderReportTime(0),
	fPacketCount(0),
	fByteCount(0),
	fTrackID(0),
	fSsrc(inSSRC),
	fSsrcStringPtr(fSsrcString, 0),
	fPayloadType(qtssUnknownPayloadType),
	fFirstSeqNumber(0),
	fFirstTimeStamp(0),
	fTimescale(0),
	fStreamURLPtr(fStreamURL, 0),
	fQualityLevel(0),
	fNumQualityLevels(0),
	fBufferDelay(3.0),
	
	fFractionLostPackets(0),
	fTotalLostPackets(0),
	fJitter(0),
	fReceiverBitRate(0),
	fAvgLateMsec(0),
	fPercentPacketsLost(0),
	fAvgBufDelayMsec(0),
	fIsGettingBetter(0),
	fIsGettingWorse(0),
	fNumEyes(0),
	fNumEyesActive(0),
	fNumEyesPaused(0),
	fTotalPacketsRecv(0),
	fTotalPacketsDropped(0),
	fTotalPacketsLost(0),
	fClientBufferFill(0),
	fFrameRate(0),
	fExpectedFrameRate(0),
	fAudioDryCount(0),
	fIsTCP(false),
	fCurrentPacketDelay(0),
	fStreamRef(this),
	fRTPChannel(0),
	fRTCPChannel(0)
{
#if DEBUG
	fNumPacketsDroppedOnTCPFlowControl = 0;
#endif
	//format the ssrc as a string
	::sprintf(fSsrcString, "%ld", fSsrc);
	fSsrcStringPtr.Len = ::strlen(fSsrcString);
	Assert(fSsrcStringPtr.Len < kMaxSsrcSizeInBytes);

	// SETUP DICTIONARY ATTRIBUTES
	
	this->SetVal(qtssRTPStrTrackID, 			&fTrackID, 				sizeof(fTrackID));
	this->SetVal(qtssRTPStrSSRC, 				&fSsrc, 				sizeof(fSsrc));
	this->SetEmptyVal(qtssRTPStrPayloadName, 	&fPayloadNameBuf, 		kDefaultPayloadBufSize);
	this->SetVal(qtssRTPStrPayloadType, 		&fPayloadType, 			sizeof(fPayloadType));
	this->SetVal(qtssRTPStrFirstSeqNumber, 		&fFirstSeqNumber, 		sizeof(fFirstSeqNumber));
	this->SetVal(qtssRTPStrFirstTimestamp, 		&fFirstTimeStamp, 		sizeof(fFirstTimeStamp));
	this->SetVal(qtssRTPStrTimescale, 			&fTimescale, 			sizeof(fTimescale));
	this->SetVal(qtssRTPStrQualityLevel, 		&fQualityLevel, 		sizeof(fQualityLevel));
	this->SetVal(qtssRTPStrNumQualityLevels, 	&fNumQualityLevels, 	sizeof(fNumQualityLevels));
	this->SetVal(qtssRTPStrBufferDelayInSecs, 	&fBufferDelay, 			sizeof(fBufferDelay));
	this->SetVal(qtssRTPStrFractionLostPackets,	&fFractionLostPackets, 	sizeof(fFractionLostPackets));
	this->SetVal(qtssRTPStrTotalLostPackets, 	&fTotalLostPackets, 	sizeof(fTotalLostPackets));
	this->SetVal(qtssRTPStrJitter, 				&fJitter, 				sizeof(fJitter));
	this->SetVal(qtssRTPStrRecvBitRate, 		&fReceiverBitRate, 		sizeof(fReceiverBitRate));
	this->SetVal(qtssRTPStrAvgLateMilliseconds, &fAvgLateMsec, 			sizeof(fAvgLateMsec));
	this->SetVal(qtssRTPStrPercentPacketsLost, 	&fPercentPacketsLost, 	sizeof(fPercentPacketsLost));
	this->SetVal(qtssRTPStrAvgBufDelayInMsec, 	&fAvgBufDelayMsec, 		sizeof(fAvgBufDelayMsec));
	this->SetVal(qtssRTPStrGettingBetter, 		&fIsGettingBetter, 		sizeof(fIsGettingBetter));
	this->SetVal(qtssRTPStrGettingWorse, 		&fIsGettingWorse, 		sizeof(fIsGettingWorse));
	this->SetVal(qtssRTPStrNumEyes, 			&fNumEyes, 				sizeof(fNumEyes));
	this->SetVal(qtssRTPStrNumEyesActive, 		&fNumEyesActive, 		sizeof(fNumEyesActive));
	this->SetVal(qtssRTPStrNumEyesPaused, 		&fNumEyesPaused, 		sizeof(fNumEyesPaused));
	this->SetVal(qtssRTPStrTotPacketsRecv, 		&fTotalPacketsRecv, 	sizeof(fTotalPacketsRecv));
	this->SetVal(qtssRTPStrTotPacketsDropped, 	&fTotalPacketsDropped, 	sizeof(fTotalPacketsDropped));
	this->SetVal(qtssRTPStrTotPacketsLost, 		&fTotalPacketsLost, 	sizeof(fTotalPacketsLost));
	this->SetVal(qtssRTPStrClientBufFill, 		&fClientBufferFill, 	sizeof(fClientBufferFill));
	this->SetVal(qtssRTPStrFrameRate, 			&fFrameRate, 			sizeof(fFrameRate));
	this->SetVal(qtssRTPStrExpFrameRate, 		&fExpectedFrameRate, 	sizeof(fExpectedFrameRate));
	this->SetVal(qtssRTPStrAudioDryCount, 		&fAudioDryCount, 		sizeof(fAudioDryCount));
	this->SetVal(qtssRTPStrIsTCP, 				&fIsTCP, 				sizeof(fIsTCP));
	this->SetVal(qtssRTPStrCurrentPacketDelay,	&fCurrentPacketDelay, 	sizeof(fCurrentPacketDelay));
	this->SetVal(qtssRTPStrStreamRef, 			&fStreamRef, 			sizeof(fStreamRef));
	
	
}

RTPStream::~RTPStream()
{
	QTSS_Error err = QTSS_NoErr;
	if (fSockets != NULL)
	{
		// If there is an UDP socket pair associated with this stream, make sure to free it up
		Assert(fSockets->GetSocketB()->GetDemuxer() != NULL);
		fSockets->GetSocketB()->GetDemuxer()->
			UnregisterTask(fRemoteAddr, fRemoteRTCPPort, this);
		Assert(err == QTSS_NoErr);
	
		QTSServerInterface::GetServer()->GetSocketPool()->ReleaseUDPSocketPair(fSockets);
	}

#if RTP_TCP_STREAM_DEBUG
	if ( fIsTCP )
		::printf( "DEBUG: ~RTPStream %li sends got EAGAIN'd.\n", (long)fNumPacketsDroppedOnTCPFlowControl );
#endif
}

QTSS_Error RTPStream::Setup(RTSPRequestInterface* request, QTSS_AddStreamFlags inFlags)
{
	//Get the URL for this track
	fStreamURLPtr.Len = kMaxStreamURLSizeInBytes;
	if (request->GetValue(qtssRTSPReqFileName, 0, fStreamURLPtr.Ptr, &fStreamURLPtr.Len) != QTSS_NoErr)
		return QTSSModuleUtils::SendErrorResponse(request, qtssClientBadRequest, qtssMsgFileNameTooLong);
	fStreamURL[fStreamURLPtr.Len] = '\0';//just in case someone wants to use string routines
	
	// Check to see if this RTP stream should be sent over TCP.
	if (request->IsTCP())
	{
		fIsTCP = true;
		
		request->GetSession()->InterleaveSetup();
		
		// If it is, get 2 channel numbers from the RTSP session.
		fRTPChannel = request->GetSession()->GetChannelNumber();
		fRTCPChannel = request->GetSession()->GetChannelNumber();
		
		// If we are interleaving, this is all we need to do to setup.
		return QTSS_NoErr;
	}
	
	//Get and store the remote addresses provided by the client. The remote addr is the
	//same as the RTSP client's IP address, unless an alternate was specified in the
	//transport header.
	fRemoteAddr = request->GetSession()->GetSocket()->GetRemoteAddr();
	if (request->GetDestAddr() != INADDR_ANY)
	{
		// Sending data to other addresses could be used in malicious ways, therefore
		// it is up to the module as to whether this sort of request might be allowed
		if (!(inFlags & qtssASFlagsAllowDestination))
			return QTSSModuleUtils::SendErrorResponse(request, qtssClientBadRequest, qtssMsgAltDestNotAllowed);
		fRemoteAddr = request->GetDestAddr();
	}
	fRemoteRTPPort = request->GetClientPortA();
	fRemoteRTCPPort = request->GetClientPortB();

	if ((fRemoteRTPPort == 0) || (fRemoteRTCPPort == 0))
		return QTSSModuleUtils::SendErrorResponse(request, qtssClientBadRequest, qtssMsgNoClientPortInTransport);		
	
	//make sure that the client is advertising an even-numbered RTP port,
	//and that the RTCP port is actually one greater than the RTP port
	if ((fRemoteRTPPort & 1) != 0)
		return QTSSModuleUtils::SendErrorResponse(request, qtssClientBadRequest, qtssMsgRTPPortMustBeEven);		
		
	if (fRemoteRTCPPort != (fRemoteRTPPort + 1))
		return QTSSModuleUtils::SendErrorResponse(request, qtssClientBadRequest, qtssMsgRTCPPortMustBeOneBigger);		
	
	// Find the right source address for this stream. If it isn't specified in the
	// RTSP request, assume it is the same interface as for the RTSP request.
	UInt32 sourceAddr = request->GetSession()->GetSocket()->GetLocalAddr();
	if ((request->GetSourceAddr() != INADDR_ANY) && (SocketUtils::IsLocalIPAddr(request->GetSourceAddr())))
		sourceAddr = request->GetSourceAddr();

	// If the destination address is multicast, we need to setup multicast socket options
	// on the sockets. Because these options may be different for each stream, we need
	// a dedicated set of sockets
	if (SocketUtils::IsMulticastIPAddr(fRemoteAddr))
	{
		fSockets = QTSServerInterface::GetServer()->GetSocketPool()->CreateUDPSocketPair(sourceAddr, 0);
		
		if (fSockets != NULL)
		{
			//Set options on both sockets. Not really sure why we need to specify an
			//outgoing interface, because these sockets are already bound to an interface!
			QTSS_Error err = fSockets->GetSocketA()->SetTtl(request->GetTtl());
			if (err == QTSS_NoErr)
				err = fSockets->GetSocketB()->SetTtl(request->GetTtl());
			if (err == QTSS_NoErr)
				err = fSockets->GetSocketA()->SetMulticastInterface(fSockets->GetSocketA()->GetLocalAddr());
			if (err == QTSS_NoErr)
				err = fSockets->GetSocketB()->SetMulticastInterface(fSockets->GetSocketB()->GetLocalAddr());
			if (err != QTSS_NoErr)
				return QTSSModuleUtils::SendErrorResponse(request, qtssServerInternal, qtssMsgCantSetupMulticast);		
		}
	}
	else
		fSockets = QTSServerInterface::GetServer()->GetSocketPool()->GetUDPSocketPair(sourceAddr, 0, fRemoteAddr, 
																						fRemoteRTCPPort);

	if (fSockets == NULL)
		return QTSSModuleUtils::SendErrorResponse(request, qtssServerInternal, qtssMsgOutOfPorts);		
		
	//finally, register with the demuxer to get RTCP packets from the proper address
	Assert(fSockets->GetSocketB()->GetDemuxer() != NULL);
	QTSS_Error err = fSockets->GetSocketB()->GetDemuxer()->RegisterTask(fRemoteAddr, fRemoteRTCPPort, this);
	//errors should only be returned if there is a routing problem, there should be none
	Assert(err == QTSS_NoErr);
	return QTSS_NoErr;
}

void RTPStream::SendSetupResponse( RTSPRequestInterface* inRequest )
{
	if (fSession->DoSessionSetupResponse(inRequest) != QTSS_NoErr)
		return;
		
	inRequest->AppendDateAndExpires();
	this->AppendTransport(inRequest);
	inRequest->SendHeader();
}

void RTPStream::AppendTransport(RTSPRequestInterface* request)
{
	// We are either going to append the RTP / RTCP port numbers (UDP),
	// or the channel numbers (TCP, interleaved)
	if (!fIsTCP)
	{
		// Append UDP socket port numbers.
		UDPSocket* theRTPSocket = fSockets->GetSocketA();
		UDPSocket* theRTCPSocket = fSockets->GetSocketB();
		request->AppendTransportHeader(theRTPSocket->GetLocalPortStr(), theRTCPSocket->GetLocalPortStr(), NULL, NULL);
	}
	else if (fRTCPChannel < kNumPrebuiltChNums)
		// We keep a certain number of channel number strings prebuilt, so most of the time
		// we won't have to call sprintf
		request->AppendTransportHeader(NULL, NULL, &sChannelNums[fRTPChannel],  &sChannelNums[fRTCPChannel]);
	else
	{
		// If these channel numbers fall outside prebuilt range, we will have to call sprintf.
		char rtpChannelBuf[10];
		char rtcpChannelBuf[10];
		::sprintf(rtpChannelBuf, "%d", fRTPChannel);		
		::sprintf(rtcpChannelBuf, "%d", fRTCPChannel);
		
		StrPtrLen rtpChannel(rtpChannelBuf);
		StrPtrLen rtcpChannel(rtcpChannelBuf);

		request->AppendTransportHeader(NULL, NULL, &rtpChannel, &rtcpChannel);
	}
}

void	RTPStream::AppendRTPInfo(QTSS_RTSPHeader inHeader, RTSPRequestInterface* request, UInt32 inFlags)
{
	//format strings for the various numbers we need to send back to the client
	char rtpTimeBuf[20];
	StrPtrLen rtpTimeBufPtr;
	if (inFlags & qtssPlayRespWriteTrackInfo)
	{
		::sprintf(rtpTimeBuf, "%ld", fFirstTimeStamp);
		rtpTimeBufPtr.Set(rtpTimeBuf, ::strlen(rtpTimeBuf));
		Assert(rtpTimeBufPtr.Len < 20);
	}	
	
	char seqNumberBuf[20];
	StrPtrLen seqNumberBufPtr;
	if (inFlags & qtssPlayRespWriteTrackInfo)
	{
		::sprintf(seqNumberBuf, "%d", fFirstSeqNumber);
		seqNumberBufPtr.Set(seqNumberBuf, ::strlen(seqNumberBuf));
		Assert(seqNumberBufPtr.Len < 20);
	}
	
	//only advertise our ssrc to the client if we are in fact sending RTCP
	//sender reports. Otherwise, this ssrc is invalid!
	UInt32 theSsrcStringLen = fSsrcStringPtr.Len;
	if (inFlags & qtssPlayRespWriteTrackInfo)
		fSsrcStringPtr.Len = 0;
	
	request->AppendRTPInfoHeader(inHeader, &fStreamURLPtr, &seqNumberBufPtr,
									&fSsrcStringPtr, &rtpTimeBufPtr);
	fSsrcStringPtr.Len = theSsrcStringLen;
}

/*********************************
/
/	InterleavedWrite
/
/	Write the given RTP packet out on the RTSP channel in interleaved format.
/	update quality levels and statistics
/	on success refresh the RTP session timeout to keep it alive
/
*/

QTSS_Error	RTPStream::InterleavedWrite(void* inBuffer, UInt32 inLen, UInt32* outLenWritten, unsigned char channel)
{
	
	// Check to see if fSession is NULL. This may happen if we are about to go away.
	// (Look at RTPSession::Teardown)
	// return EAGAIN because it is known to be well tolerated.
	Assert(fSession != NULL);
	if (!fSession->GetSessionMutex()->TryLock())
		return EAGAIN;
		
	if (fSession->GetRTSPSession() == NULL)
	{
		fSession->GetSessionMutex()->Unlock();
		return EAGAIN;
	}
		
	QTSS_Error err = fSession->GetRTSPSession()->InterleavedWrite( inBuffer, inLen, outLenWritten, channel);
	
	// Make sure to unlock the mutex
	fSession->GetSessionMutex()->Unlock();

#if DEBUG
	if ( err == EAGAIN )
	{
		fNumPacketsDroppedOnTCPFlowControl++;
	}
#endif		

	// reset the timeouts when the connection is still alive
	// wehn transmitting over HTTP, we're not going to get
	// RTCPs that would normally Refresh the session time.
	if ( err == QTSS_NoErr )
		fSession->RefreshTimeout(); // RTSP session gets refreshed internally in WriteV

			
	// set quality level for thinning based on time delay
	// in delivering packets.
	//+rt NOTE -- we need to change quality numbers to a global range, then
	// let the modules map their actual capabilities of thinning
	// against that.  ( for both TCP and UDP delivery )
	if ( fNumQualityLevels >  2 )
	{
		QTSServerInterface* theServer = QTSServerInterface::GetServer();
		
		if ( fCurrentPacketDelay < theServer->GetPrefs()->GetTCPMinThinDelayToleranceInMSec() )
		{	
			fQualityLevel = 0;
		}
		else if ( fCurrentPacketDelay < theServer->GetPrefs()->GetTCPMaxThinDelayToleranceInMSec() )
		{
			fQualityLevel = 2 > fNumQualityLevels ? fNumQualityLevels : 2;
			fQualityLevel /= 2;
		}
		else 
			fQualityLevel =  2 > fNumQualityLevels ? fNumQualityLevels : 2;
	}
	
	#if RTP_TCP_STREAM_DEBUG
	//printf( "DEBUG: RTPStream fCurrentPacketDelay %li, fQualityLevel %i\n", (long)fCurrentPacketDelay, (int)fQualityLevel );
	#endif

	return err;
	
}


QTSS_Error	RTPStream::Write(void* inBuffer, UInt32 inLen, UInt32* outLenWritten, UInt32 inFlags)
{
	QTSS_Error err = QTSS_NoErr;
	
	if (inFlags == qtssWriteFlagsIsRTCP)
	{
		if ( fIsTCP )	// write out in interleave format on the RTSP TCP channel
		{	
			err = this->InterleavedWrite( inBuffer, inLen, outLenWritten, fRTCPChannel );
		}
		else if ( inLen > 0 )
			(void)fSockets->GetSocketB()->SendTo(fRemoteAddr, fRemoteRTCPPort, inBuffer, inLen);
	}
	else if (inFlags == qtssWriteFlagsIsRTP)
	{
		if ( fIsTCP )	// write out in interleave format on the RTSP TCP channel
		{	
			err = this->InterleavedWrite( inBuffer, inLen, outLenWritten, fRTPChannel );		
		}
		else if ( inLen > 0 )
			(void)fSockets->GetSocketA()->SendTo(fRemoteAddr, fRemoteRTPPort, inBuffer, inLen);

		if ( err == QTSS_NoErr && inLen > 0 )
		{
			// Update statistics if we were actually able to send the data (don't
			// update if the socket is flow controlled or some such thing)
			
			fSession->UpdatePacketsSent(1);
			fSession->UpdateBytesSent(inLen);
			QTSServerInterface::GetServer()->IncrementTotalRTPBytes(inLen);
			QTSServerInterface::GetServer()->IncrementTotalPackets();
			
			//stream statistics
			fPacketCount++;
			fByteCount += inLen;
		
			// Send an RTCP sender report if it's time. Again, we only want to send an
			// RTCP if the RTP packet was sent sucessfully
			SInt64 theTime = OS::Milliseconds();

			//send an RTCP packet if we are supposed to and it's time
			if ((fSession->GetPlayFlags() & qtssPlayFlagsSendRTCP) &&
				(theTime > (fLastSenderReportTime + (kSenderReportIntervalInSecs * 1000))))
			{
				fLastSenderReportTime = theTime;
				
				RTCPSRPacket* theSR = fSession->GetSRPacket();
				theSR->SetSSRC(fSsrc);
				theSR->SetNTPTimestamp(fSession->GetNTPPlayTime() +
										OS::ConvertMsecToFixed64Sec(theTime - fSession->GetPlayTime()));
				UInt32* timeStampP = (UInt32*)inBuffer;
				theSR->SetRTPTimestamp(ntohl(timeStampP[1]));// timestamp is ALREADY network ordered!
				theSR->SetPacketCount(fPacketCount);
				theSR->SetByteCount(fByteCount);
				
				
				if ( fIsTCP )	// write out in interleave format on the RTSP TCP channel
				{	UInt32	wasWritten;
				
					(void)this->InterleavedWrite( theSR->GetSRPacket(), theSR->GetSRPacketLen(), &wasWritten, fRTCPChannel );
				}
				else
					(void)fSockets->GetSocketB()->SendTo(fRemoteAddr, fRemoteRTCPPort,
													theSR->GetSRPacket(), theSR->GetSRPacketLen());
			}
		}
	}
	else
		return QTSS_BadArgument;//qtssWriteFlagsIsRTCP or qtssWriteFlagsIsRTP wasn't specified

	if (outLenWritten != NULL)
		*outLenWritten = inLen;
	return err;
}


void RTPStream::ProcessPacket(StrPtrLen* inPacket)
{
	StrPtrLen currentPtr(*inPacket);
	
	while ( currentPtr.Len > 0 )
	{
		/*
			Due to the variable-type nature of RTCP packets, this is a bit unusual...
			We initially treat the packet as a generic RTCPPacket in order to determine its'
			actual packet type.  Once that is figgered out, we treat it as its' actual packet type
		*/
		RTCPPacket rtcpPacket;
		if (!rtcpPacket.ParsePacket((UInt8*)currentPtr.Ptr, currentPtr.Len))
			return;//abort if we discover a malformed RTCP packet

		//if this is the first packet, verify that it is an RR and the padding bit is 0 (see RTP RFC)
		if (currentPtr.Ptr == inPacket->Ptr)
			if ((rtcpPacket.GetPacketType() != RTCPReceiverPacket::kPacketType) ||
			 	(rtcpPacket.GetHasPadding() == true))
			 	return;
		
		switch (rtcpPacket.GetPacketType())
		{
			case RTCPReceiverPacket::kPacketType:
			{
				RTCPReceiverPacket receiverPacket;
				if (!receiverPacket.ParseReceiverReport((UInt8*)currentPtr.Ptr, currentPtr.Len))
					return;//abort if we discover a malformed receiver report

				fFractionLostPackets = receiverPacket.GetCumulativeFractionLostPackets();
				fTotalLostPackets = receiverPacket.GetCumulativeTotalLostPackets();
				fJitter = receiverPacket.GetCumulativeJitter();
			}
			break;
			
			case RTCPCompressedQTSSPacket::kPacketType:
			{
				RTCPCompressedQTSSPacket compressedQTSSPacket;
				if (!compressedQTSSPacket.ParseCompressedQTSSPacket((UInt8*)currentPtr.Ptr, currentPtr.Len))
					return;//abort if we discover a malformed app packet

				fReceiverBitRate = 		compressedQTSSPacket.GetReceiverBitRate();
				fAvgLateMsec = 			compressedQTSSPacket.GetAverageLateMilliseconds();
				fPercentPacketsLost = 	compressedQTSSPacket.GetPercentPacketsLost();
				fAvgBufDelayMsec = 		compressedQTSSPacket.GetAverageBufferDelayMilliseconds();
				fIsGettingBetter = (UInt16)compressedQTSSPacket.GetIsGettingBetter();
				fIsGettingWorse = (UInt16)compressedQTSSPacket.GetIsGettingWorse();
				fNumEyes = 				compressedQTSSPacket.GetNumEyes();
				fNumEyesActive = 		compressedQTSSPacket.GetNumEyesActive();
				fNumEyesPaused = 		compressedQTSSPacket.GetNumEyesPaused();
				fTotalPacketsRecv = 	compressedQTSSPacket.GetTotalPacketReceived();
				fTotalPacketsDropped = 	compressedQTSSPacket.GetTotalPacketsDropped();
				fTotalPacketsLost = 	compressedQTSSPacket.GetTotalPacketsLost();
				fClientBufferFill = 	compressedQTSSPacket.GetClientBufferFill();
				fFrameRate = 			compressedQTSSPacket.GetFrameRate();
				fExpectedFrameRate = 	compressedQTSSPacket.GetExpectedFrameRate();
				fAudioDryCount = 		compressedQTSSPacket.GetAudioDryCount();
			}
			break;
			
			case SourceDescriptionPacket::kPacketType:
			{
#ifdef DEBUG_RTCP_PACKETS
				SourceDescriptionPacket sedsPacket;
				if (!sedsPacket.Parse(currentPtr.Ptr, currentPtr.Len))
					return;//abort if we discover a malformed app packet

				sedsPacket.Dump();
#endif
			}
			break;
			
			//default:
			//	WarnV(false, "Unknown RTCP Packet Type");
			break;
		
		}
		
		
		currentPtr.Ptr += (rtcpPacket.GetPacketLength() * 4 ) + 4;
		currentPtr.Len -= (rtcpPacket.GetPacketLength() * 4 ) + 4;
	}

	// Invoke the RTCP modules, allowing them to process this packet
	QTSS_RoleParams theParams;
	theParams.rtcpProcessParams.inRTPStream = this;
	theParams.rtcpProcessParams.inClientSession = fSession;
	theParams.rtcpProcessParams.inRTCPPacketData = inPacket->Ptr;
	theParams.rtcpProcessParams.inRTCPPacketDataLen = inPacket->Len;
	
	// We don't allow async events from this role, so just set an empty module state.
	OSThread::GetCurrent()->SetThreadData(&sRTCPProcessModuleState);
	
	// Modules are guarenteed atomic access to the session. Also, the RTSP Session accessed
	// below could go away at any time. So we need to lock the RTP session mutex.
	// *BUT*, when this function is called the caller already has the UDP socket pool &
	// UDP Demuxer mutexes. Blocking on grabbing this mutex could cause a deadlock.
	// So, dump this RTCP packet if we can't get the mutex.
	if (!fSession->GetSessionMutex()->TryLock())
		return;

	//no matter what happens (whether or not this is a valid packet) reset the timeouts
	fSession->RefreshTimeout();
	if (fSession->GetRTSPSession() != NULL)
		fSession->GetRTSPSession()->RefreshTimeout();
	
	// Invoke RTCP processing modules
	for (UInt32 x = 0; x < QTSSModuleInterface::GetNumModulesInRole(QTSSModule::kRTCPProcessRole); x++)
		(void)QTSSModuleInterface::GetModule(QTSSModule::kRTCPProcessRole, x)->CallDispatch(QTSS_RTCPProcess_Role, &theParams);

	fSession->GetSessionMutex()->Unlock();
}
