/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		RTPStream.h

	Contains:	Represents a single client stream (audio, video, etc).
				Control API is similar to overall session API.
				
				Contains all stream-specific data & resources, used by Session when it
				wants to send out or receive data for this stream
				
				This is also the class that implements the RTP stream dictionary
				for QTSS API.
				
	$Log: RTPStream.h,v $
	Revision 1.3  2000/02/11 01:07:23  serenyi
	Server now supports cache control header and sends 304 Not Modified responses in appropriate situations
	
	Revision 1.2  2000/01/27 02:09:14  rtucker
	Merge changes from v81
	
	Revision 1.7  2000/01/26 21:32:16  serenyi
	Added new QTSS_RequestIdleTime callback, made it so the ClientSessionClosing role can register for async events
	
	Revision 1.6  1999/12/21 01:44:08  lecroy
	*** empty log message ***
	
	Revision 1.5  1999/11/02 00:14:04  murata
	Changed bool to Bool16 like it is supposed to be.
	
	Revision 1.4  1999/10/08 01:12:51  rtucker
	added support for TCP thining and thinnig prefs
	
	
	
	
	VS:
	
	Revision 1.3  1999/10/07 08:52:45  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.2  1999/10/05 23:00:14  murata
	added bufferdelay float32 for logging and api
	
	Revision 1.1.1.1  1999/09/22 23:01:53  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:54  rtucker
	initial checkin from rrv14
	
	
	Revision 1.4  1999/02/24 19:28:06  ds
	RTPStream now uses new UDPSocketPool interface for getting its UDP sockets
	
	Revision 1.2  1999/02/19 23:08:27  ds
	Created
	

*/

#ifndef __RTPSTREAM_H__
#define __RTPSTREAM_H__

#include "QTSS.h"
#include "QTSSDictionary.h"

#include "UDPDemuxer.h"
#include "UDPSocketPool.h"

#include "RTSPRequestInterface.h"
#include "RTPSessionInterface.h"

class RTPStream : public QTSSDictionary, public UDPDemuxerTask
{
	public:
		
		// Initializes dictionary resources
		static void	Initialize();

		//
		// CONSTRUCTOR / DESTRUCTOR
		
		RTPStream(UInt32 inSSRC, RTPSessionInterface* inSession);
		virtual ~RTPStream();
		
		//
		//ACCESS FUNCTIONS
		
		UInt32		GetSSRC()					{ return fSsrc; }
		
		// Setup uses the info in the RTSPRequestInterface to associate
		// all the necessary resources, ports, sockets, etc, etc, with this
		// stream.
		QTSS_Error Setup(RTSPRequestInterface* request, QTSS_AddStreamFlags inFlags);
		
		// Write sends RTP data to the client. Caller must specify
		// either qtssWriteFlagsIsRTP or qtssWriteFlagsIsRTCP
		virtual QTSS_Error	Write(void* inBuffer, UInt32 inLen,
										UInt32* outLenWritten, UInt32 inFlags);
		
		//UTILITY FUNCTIONS:
		//These are not necessary to call and do not manipulate the state of the
		//stream. They may, however, be useful services exported by the server
		
		// Formats a standard setup response.
		void			SendSetupResponse(RTSPRequestInterface* request);

		//Formats a transport header for this stream. 
		void			AppendTransport(RTSPRequestInterface* request);
		
		//Formats a RTP-Info header for this stream.
		//Isn't useful unless you've already called Play()
		void			AppendRTPInfo(QTSS_RTSPHeader inHeader,
										RTSPRequestInterface* request, UInt32 inFlags);

		//When we get a new RTCP packet, we can directly invoke the RTP session and tell it
		//to process the packet right now!
		void ProcessPacket(StrPtrLen* inPacket);

	private:
	
		enum
		{
			kMaxSsrcSizeInBytes 		= 12,
			kMaxStreamURLSizeInBytes 	= 32,
			kDefaultPayloadBufSize		= 32,
			kSenderReportIntervalInSecs = 7,
			kNumPrebuiltChNums			= 10
		};
	
		//either pointers to the statically allocated sockets (maintained by the server)
		//or fresh ones (only fresh in extreme special cases)
		UDPSocketPair* 			fSockets;
		RTPSessionInterface*	fSession;

		//who am i sending to?
		UInt32		fRemoteAddr;
		UInt16		fRemoteRTPPort;
		UInt16		fRemoteRTCPPort;
		
		//RTCP stuff 
		SInt64		fLastSenderReportTime;
		UInt32		fPacketCount;
		UInt32		fByteCount;
		
		// DICTIONARY ATTRIBUTES
		
		//Module assigns a streamID to this object
		UInt32		fTrackID;
		
		//low-level RTP stuff 
		UInt32		fSsrc;
		char		fSsrcString[kMaxSsrcSizeInBytes];
		StrPtrLen	fSsrcStringPtr;
		
		//Payload name and codec type.
		char				fPayloadNameBuf[kDefaultPayloadBufSize];
		QTSS_RTPPayloadType	fPayloadType;

		//Media information.
		UInt16		fFirstSeqNumber;//used in sending the play response
		UInt32		fFirstTimeStamp;//RTP time
		UInt32		fTimescale;
		
		//what is the URL for this stream?
		char		fStreamURL[kMaxStreamURLSizeInBytes];
		StrPtrLen	fStreamURLPtr;
		
		UInt32		fQualityLevel;
		UInt32		fNumQualityLevels;
		
		Float32 	fBufferDelay; // from the sdp
		
		// RTCP data
		UInt32		fFractionLostPackets;
		UInt32		fTotalLostPackets;
		UInt32		fJitter;
		UInt32		fReceiverBitRate;
		UInt16		fAvgLateMsec;
		UInt16		fPercentPacketsLost;
		UInt16		fAvgBufDelayMsec;
		UInt16		fIsGettingBetter;
		UInt16		fIsGettingWorse;
		UInt32		fNumEyes;
		UInt32		fNumEyesActive;
		UInt32		fNumEyesPaused;
		UInt32		fTotalPacketsRecv;
		UInt16		fTotalPacketsDropped;
		UInt16		fTotalPacketsLost;
		UInt16		fClientBufferFill;
		UInt16		fFrameRate;
		UInt16		fExpectedFrameRate;
		UInt16		fAudioDryCount;
		
		Bool16		fIsTCP;
		SInt32		fCurrentPacketDelay;  // dictionary: qtssRTPStrCurrentPacketDelay
		
		// Pointer to the stream ref (this is just a this pointer)
		QTSS_StreamRef	fStreamRef;
		

#if DEBUG
		UInt32		fNumPacketsDroppedOnTCPFlowControl;
#endif
		
		// If we are interleaving RTP data over the TCP connection,
		// these are channel numbers to use for RTP & RTCP
		unsigned char	fRTPChannel;
		unsigned char	fRTCPChannel;
		
		// acutally write the data out that way
		QTSS_Error	InterleavedWrite(void* inBuffer, UInt32 inLen, UInt32* outLenWritten, unsigned char channel );

		static QTSSDictionaryMap::AttrInfo	sAttributes[];
		static StrPtrLen					sChannelNums[];
		static QTSS_ModuleState				sRTCPProcessModuleState;
};
#endif // __RTPSTREAM_H__
