/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		RTSPRequest.cpp

	Contains:	Implementation of RTSPRequest class.

	$Log: RTSPRequest.cpp,v $
	Revision 1.3  2000/03/07 23:08:39  murata
	changes to support improved logging and logging module bug fixes. Changes include
	new API attributes in the client session object and error reporting in QTFileLib.
	Bug #:2440344, 2429487,2412973, 2412969.
	Submitted by:
	Reviewed by:
	
	Revision 1.2  2000/02/04 00:46:25  serenyi
	Added date & expires headers to server response, server now parses if-modified-since dates
	
	Revision 1.1.1.1  2000/01/17 23:36:03  wsanchez
	Import of StreamingServer-2.0 (Apple)
	
	Revision 1.8  2000/01/11 18:32:58  rtucker
	strenghened ParseURI error handling of DecodeURL failures
	
	Revision 1.7  1999/12/21 01:44:08  lecroy
	*** empty log message ***
	
	Revision 1.6  1999/11/09 12:02:14  murata
	fix a memory leak. The preference GetAuthorizationRealm call allocates memory.
	
	Revision 1.5  1999/11/03 18:05:45  serenyi
	If illegal / unknown RTSP header is found, dictionary would get corrupted. This was fixed by this checkin
	
	Revision 1.4  1999/10/27 17:55:03  serenyi
	Integrated Matthias's MacOS / GUSI changes
	
	Revision 1.3  1999/10/27 03:26:02  murata
	Changes are for optimizing authentication code in the request.
	
	Revision 1.2  1999/10/07 08:52:33  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.1.1.1  1999/09/22 23:01:53  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:54  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:08:31  ds
	Created
	
	
*/


#include "RTSPRequest.h"
#include "RTSPProtocol.h"
#include "QTSServerInterface.h"

#include "RTSPSessionInterface.h"
#include "StringParser.h"
#include "StringTranslator.h"
#include "OS.h"
#include "QTSS.h"
#include "QTSSModuleUtils.h"
#include "base64.h"
#include "OSArrayObjectDeleter.h"
#include "DateTranslator.h"

#include <arpa/inet.h>
#include <sys/errno.h>

UInt8
RTSPRequest::sURLStopConditions[] =
{
	0, 0, 0, 0, 0, 0, 0, 0, 0, 1, //0-9      //'\t' is a stop condition
	1, 0, 0, 1, 0, 0, 0, 0, 0, 0, //10-19    //'\r' & '\n' are stop conditions
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //20-29
	0, 0, 1, 0, 0, 0, 0, 0, 0, 0, //30-39    //' '
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //40-49
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //50-59
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //60-69
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //70-79
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //80-89
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //90-99
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //100-109
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //110-119
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //120-129
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //130-139
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //140-149
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //150-159
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //160-169
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //170-179
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //180-189
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //190-199
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //200-209
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //210-219
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //220-229
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //230-239
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, //240-249
	0, 0, 0, 0, 0, 0 			 //250-255
};

static StrPtrLen	sDefaultRealm("QT Streaming Server");

//Parses the request
QTSS_Error RTSPRequest::Parse(StrPtrLen *stream)
{
	Assert(stream != NULL);
	StringParser parser(stream);

	//parse status line.
	QTSS_Error error = ParseFirstLine(parser);

	//handle any errors that come up	
	if (error != QTSS_NoErr)
		return error;
		
	error = this->ParseHeaders(parser);
	if (error != QTSS_NoErr)
		return error;
	
	//Response headers should set themselves up to reflect what's in the request headers
	fResponseKeepAlive = fRequestKeepAlive;
	
	//Make sure that there was some path that was extracted from this request. If not, there is no way
	//we can process the request, so generate an error
	if (this->GetValue(qtssRTSPReqFilePath)->Len == 0)
		return QTSSModuleUtils::SendErrorResponse(this, qtssClientBadRequest, qtssMsgNoURLInRequest);
	
	return QTSS_NoErr;
}

//returns: StatusLineTooLong, SyntaxError, BadMethod
QTSS_Error RTSPRequest::ParseFirstLine(StringParser &parser)
{	
	//first get the method
	StrPtrLen theParsedData;
	parser.ConsumeWord(&theParsedData);
	this->SetVal(qtssRTSPReqMethodStr, theParsedData.Ptr, theParsedData.Len);
	
	
	//THIS WORKS UNDER THE ASSUMPTION THAT:
	//valid HTTP/1.1 headers are: GET, HEAD, POST, PUT, OPTIONS, DELETE, TRACE
	fMethod = RTSPProtocol::GetMethod(theParsedData);
	if (fMethod == qtssIllegalMethod)
		return QTSSModuleUtils::SendErrorResponse(this, qtssClientBadRequest, qtssMsgBadRTSPMethod, &theParsedData);
	
	//no longer assume this is a space... instead, just consume whitespace
	parser.ConsumeWhitespace();

	//now parse the uri
	QTSS_Error err = ParseURI(parser);
	if (err != QTSS_NoErr)
		return err;

	//no longer assume this is a space... instead, just consume whitespace
	parser.ConsumeWhitespace();

	//if there is a version, consume the version string
	StrPtrLen versionStr;
	parser.ConsumeUntil(&versionStr, StringParser::sEOLMask);
	
	//check the version
	if (versionStr.Len > 0)
		fVersion = RTSPProtocol::GetVersion(versionStr);

	//go past the end of line
	if (!parser.ExpectEOL())
		return QTSSModuleUtils::SendErrorResponse(this, qtssClientBadRequest, qtssMsgNoRTSPVersion);
	return QTSS_NoErr;
}

//returns: SyntaxError if there was an error in the uri. Or InternalServerError
QTSS_Error RTSPRequest::ParseURI(StringParser &parser)
{
	//read in the complete URL, set it to be the qtssAbsoluteURLParam
	StrPtrLen theAbsURL;
	parser.ConsumeUntil(&theAbsURL, sURLStopConditions);
	this->SetVal(qtssRTSPReqAbsoluteURL, &theAbsURL);
	
	StringParser urlParser(&theAbsURL);
		
	//we always should have a slash before the uri.
	//If not, that indicates this is a full URI
	if (*theAbsURL.Ptr != '/')
	{
		//if it is a full URL, store the host name off in a separate parameter
		StrPtrLen theRTSPString;
		urlParser.ConsumeLength(&theRTSPString, 7); //consume "rtsp://"
		//assign the host field here to the proper QTSS param
		StrPtrLen theHost;
		urlParser.ConsumeUntil(&theHost, '/');
		fHeaderDictionary.SetVal(qtssHostHeader, &theHost);
	}
	
	//whatever is in this position in the URl must be the relative URL. Store that
	//in the qtssURLParam.
	this->SetVal(qtssRTSPReqURI, urlParser.GetCurrentPosition(), urlParser.GetDataReceivedLen() - urlParser.GetDataParsedLen());
	
	//path strings are statically allocated. Therefore, if they are longer than
	//this length we won't be able to handle the request.
	StrPtrLen* theURLParam = this->GetValue(qtssRTSPReqURI);
	if (theURLParam->Len > RTSPRequestInterface::kMaxFilePathSizeInBytes)
		return QTSSModuleUtils::SendErrorResponse(this, qtssClientBadRequest, qtssMsgURLTooLong, theURLParam);

	//decode the URL, put the result in the separate buffer for the file path,
	//set the file path StrPtrLen to the proper value
	SInt32 theBytesWritten = StringTranslator::DecodeURL(theURLParam->Ptr, theURLParam->Len,
												fFilePath, RTSPRequestInterface::kMaxFilePathSizeInBytes);
	//if negative, an error occurred, reported as an QTSS_Error
	//we also need to leave room for a terminator.
	if ((theBytesWritten < 0) || (theBytesWritten == RTSPRequestInterface::kMaxFilePathSizeInBytes))
	{
		return QTSSModuleUtils::SendErrorResponse(this, qtssClientBadRequest, qtssMsgURLInBadFormat, theURLParam);
	}

	//setup the proper QTSS param
	fFilePath[theBytesWritten] = '\0';
	this->SetVal(qtssRTSPReqFilePath, fFilePath, theBytesWritten);
	return QTSS_NoErr;
}


//throws eHTTPNoMoreData and eHTTPOutOfBuffer
QTSS_Error RTSPRequest::ParseHeaders(StringParser& parser)
{
	StrPtrLen theKeyWord;
	Bool16 isStreamOK;
	
	//Repeat until we get a \r\n\r\n, which signals the end of the headers
	
	while ((parser.PeekFast() != '\r') && (parser.PeekFast() != '\n')) 
	{
		//First get the header identifier
		
		isStreamOK = parser.GetThru(&theKeyWord, ':');
		if (!isStreamOK)
			return QTSSModuleUtils::SendErrorResponse(this, qtssClientBadRequest, qtssMsgNoColonAfterHeader);
							
		if (parser.PeekFast() == ' ') {	// handle space, if any
			isStreamOK = parser.Expect(' ');
			Assert(isStreamOK);
		}
		
		//Look up the proper header enumeration based on the header string.
		//Use the enumeration to look up the dictionary ID of this header,
		//and set that dictionary attribute to be whatever is in the body of the header
		
		UInt32 theHeader = RTSPProtocol::GetRequestHeader(theKeyWord);

		StrPtrLen theHeaderVal;
		isStreamOK = parser.GetThruEOL(&theHeaderVal);

		// If this is an unknown header, ignore it. Otherwise, set the proper
		// dictionary attribute
		if (theHeader != qtssIllegalHeader)
		{
			Assert(theHeader < qtssNumHeaders);
			fHeaderDictionary.SetVal(theHeader, &theHeaderVal);
		}
		if (!isStreamOK)
			return QTSSModuleUtils::SendErrorResponse(this, qtssClientBadRequest, qtssMsgNoEOLAfterHeader);
		
		//some headers require some special processing. If this code begins
		//to get out of control, we made need to come up with a function pointer table
		switch (theHeader)
		{
			case qtssTransportHeader: 			ParseTransportHeader(); break;
			case qtssRangeHeader:				ParseRangeHeader(); 	break;
			case qtssIfModifiedSinceHeader:		ParseIfModSinceHeader();break;
			default:	break;
		}
	}
	isStreamOK = parser.ExpectEOL();
	Assert(isStreamOK);
	return QTSS_NoErr;
}

void RTSPRequest::ParseTransportHeader()
{
	StringParser theTransportParser(fHeaderDictionary.GetValue(qtssTransportHeader));
	
	//transport header from client: Transport: RTP/AVP;unicast;client_port=5000-5001\r\n
	//								Transport: RTP/AVP;multicast;ttl=15;destination=229.41.244.93;client_port=5000-5002\r\n
	//								Transport: RTP/AVP/TCP;unicast\r\n
	
	StrPtrLen theTransportSubHeader;
	(void)theTransportParser.GetThru(&theTransportSubHeader, ';');
	
	while (theTransportSubHeader.Len > 0)
	{
		// Extract the relevent information from the relevent subheader.
		// So far we care about 3 sub-headers
		
		switch (*theTransportSubHeader.Ptr)
		{
			case 'R':	// RTP/AVP/??? Is this TCP or UDP?
			{
				if ((theTransportSubHeader.Len == 11) && (theTransportSubHeader.Ptr[8] == 'T'))
					fIsTCP = true;
				break;
			}
			case 'c':	//client_port sub-header
			{
				this->ParseClientPortSubHeader(&theTransportSubHeader);
				break;
			}
			case 'd':	//destination sub-header
			{
				static StrPtrLen sDestinationSubHeader("destination");
				
				//Parse the header, extract the destination address
				this->ParseAddrSubHeader(&theTransportSubHeader, &sDestinationSubHeader, &fDestinationAddr);
				break;
			}
			case 's':	//source sub-header
			{
				//Same as above code
				static StrPtrLen sSourceSubHeader("source");
				this->ParseAddrSubHeader(&theTransportSubHeader, &sSourceSubHeader, &fSourceAddr);
				break;
			}
			case 't':	//time-to-live sub-header
			{
				this->ParseTimeToLiveSubHeader(&theTransportSubHeader);
				break;
			}
		}
		
		// Move onto the next parameter
		(void)theTransportParser.GetThru(&theTransportSubHeader, ';');
	}
}

void  RTSPRequest::ParseRangeHeader()
{
	StringParser theRangeParser(fHeaderDictionary.GetValue(qtssRangeHeader));

	// Setup the start and stop time dictionary attributes
	this->SetVal(qtssRTSPReqStartTime, &fStartTime, sizeof(fStartTime));
	this->SetVal(qtssRTSPReqStopTime, &fStopTime, sizeof(fStopTime));

	theRangeParser.ConsumeLength(NULL, 4);//consume "npt="
	fStartTime = (Float64)theRangeParser.ConsumeFloat();
	//see if there is a stop time as well.
	if (theRangeParser.GetDataRemaining() > 1)
	{
		theRangeParser.ConsumeLength(NULL, 1);
		fStopTime = (Float64)theRangeParser.ConsumeFloat();
	}
}

void  RTSPRequest::ParseIfModSinceHeader()
{
	fIfModSinceDate = DateTranslator::ParseDate(fHeaderDictionary.GetValue(qtssIfModifiedSinceHeader));

	// Only set the param if this is a legal date
	if (fIfModSinceDate != 0)
		this->SetVal(qtssRTSPReqIfModSinceDate, &fIfModSinceDate, sizeof(fIfModSinceDate));
}

void RTSPRequest::ParseAddrSubHeader(StrPtrLen* inSubHeader, StrPtrLen* inHeaderName, UInt32* outAddr)
{
	StringParser theSubHeaderParser(inSubHeader);

	// First make sure this is the proper subheader
	if (!inSubHeader->Equal(*inHeaderName))
		return;
		
	//Find the IP address
	theSubHeaderParser.ConsumeUntil(NULL, StringParser::sDigitMask);

	//Set the addr string param.
	StrPtrLen theAddr(theSubHeaderParser.GetCurrentPosition(), theSubHeaderParser.GetDataRemaining());
	
	//Convert the string to a UInt32 IP address
	char theTerminator = theAddr.Ptr[theAddr.Len];
	theAddr.Ptr[theAddr.Len] = '\0';
	
	*outAddr = ntohl(::inet_addr(theAddr.Ptr));
	
	theAddr.Ptr[theAddr.Len] = theTerminator;
}

void RTSPRequest::ParseClientPortSubHeader(StrPtrLen* inClientPortSubHeader)
{
	static StrPtrLen sClientPortSubHeader("client_port=");

	StringParser theSubHeaderParser(inClientPortSubHeader);

	// Skip over to the first port
	StrPtrLen theFirstBit;
	theSubHeaderParser.ConsumeUntil(&theFirstBit, StringParser::sDigitMask);
	
	// Make sure this is the client port subheader
	if (!theFirstBit.Equal(sClientPortSubHeader))
		return;

	// Store the two client ports as integers
	fClientPortA = (UInt16)theSubHeaderParser.ConsumeInteger(NULL);
	theSubHeaderParser.ConsumeLength(NULL, 1);
	fClientPortB = (UInt16)theSubHeaderParser.ConsumeInteger(NULL);
}

void RTSPRequest::ParseTimeToLiveSubHeader(StrPtrLen* inTimeToLiveSubHeader)
{
	static StrPtrLen sTimeToLiveSubHeader("ttl=");

	StringParser theSubHeaderParser(inTimeToLiveSubHeader);

	// Skip over to the first part
	StrPtrLen theFirstBit;
	theSubHeaderParser.ConsumeUntil(&theFirstBit, StringParser::sDigitMask);
	// Make sure this is the ttl subheader
	if (!theFirstBit.Equal(sTimeToLiveSubHeader))
		return;
	
	// Parse out the time to live...
	fTtl = (UInt16)theSubHeaderParser.ConsumeInteger(NULL);
}

QTSS_Error RTSPRequest::ParseAuthNameAndPassword(void)
 {	
	char decodedLine[kAuthNameAndPasswordBuffSize] = {};
	char codedLine[kAuthNameAndPasswordBuffSize] = {};
	StrPtrLen	authLine;
	
	QTSS_Error	theErr = QTSS_NoErr;
	
	do //once
	{	QTSSDictionary *theRTSPHeaders = this->GetHeaderDictionary();
		theErr = theRTSPHeaders->GetValue(qtssAuthorizationHeader, 0,(void **) &authLine.Ptr, &authLine.Len);
		//PrintfStrPtrLen(  &authLine );
		
		if (QTSS_NoErr != theErr) 
		{	//printf("error GetValue (qtssAuthorizationHeader) \n");
			break;
		}
		
		Assert( authLine.Len < kAuthNameAndPasswordBuffSize );
		if (authLine.Len >= kAuthNameAndPasswordBuffSize) authLine.Len = kAuthNameAndPasswordBuffSize -1;
		memcpy (codedLine,authLine.Ptr,authLine.Len);
		authLine.Set(codedLine, authLine.Len);
		
		if (0 == authLine.Len) break;
		
		{
			StrPtrLen	authWord("");
			StringParser parsedAuthLine(&authLine);
			
			parsedAuthLine.ConsumeWord(&authWord);
			if (!authWord.Equal(StrPtrLen("Basic",strlen("Basic") ) ) ) break;
			
			parsedAuthLine.ConsumeWhitespace();
			parsedAuthLine.ConsumeWord(&authWord);
			if (0 == authWord.Len ) break;		
			if (authWord.Len > 1024) break;
			
			(void) Base64decode(decodedLine, authWord.Ptr);
			StrPtrLen	nameAndPassword(decodedLine);
			StrPtrLen	name("");
			StrPtrLen	password("");
			StrPtrLen	dump("");
			StringParser parsedNameAndPassword(&nameAndPassword);
			
			//PrintfStrPtrLen(  &nameAndPassword );

			parsedNameAndPassword.ConsumeUntil(&name,':');			
			parsedNameAndPassword.ConsumeLength(&dump, 1);
			parsedNameAndPassword.GetThruEOL(&password);
			
			//PrintfStrPtrLen(  &password );
			
			if (name.Len > 0) 
			{	
				theErr = this->SetValue(qtssRTSPReqUserName,0,  name.Ptr , name.Len, false);
			}
			else
			{	//printf("name.len is bad");
			}

			if (QTSS_NoErr != theErr) 
			{	//printf("SetValue  qtssRTSPReqUserName has err = %ld \n",theErr);
				break;
			}
			
			//printf("RTSPRequest saving dictionary password =%s len = %ld \n",password.Ptr, password.Len);

			
			if (password.Len > 0) theErr = this->SetValue(qtssRTSPReqUserPassword,0,  password.Ptr , password.Len, false);
			if (QTSS_NoErr != theErr) 
			{	//printf("SetValue  qtssRTSPReqUserPassword has err = %ld \n",theErr);
				break;
			}
		}
	} while (false);
	
	return theErr;
};

QTSS_Error RTSPRequest::SetupAuthLocalPath(void)
{
	QTSS_Error theErr = QTSS_NoErr;

	do //once
	{
		
	 	QTSS_AttributeID theID = qtssRTSPReqFilePath;

		if (qtssSetupMethod == fMethod)
		{	//printf("a setup request\n");
			theID = qtssRTSPReqFilePathTrunc;	
		}
		
	 	UInt32 outLen = 0;
		OSCharArrayDeleter fullPath(NULL);
		fullPath.SetObject(QTSSModuleUtils::GetFullPath(this, theID, &outLen, NULL)); // GetFullPath allocates mem
		
		char *path = fullPath.GetObject();			
		if ( path == NULL)
		{	Assert(0);
			break;
		}
					
		(void)this->SetValue(qtssRTSPReqLocalPath, 0, path,outLen, false);
	
		
	} while (false); 
	
	return theErr;
}

QTSS_Error RTSPRequest::SendChallenge(void)
{	
	QTSS_Error theErr = QTSS_NoErr;
	char *prefRealmPtr = NULL;
	
	do 
	{	
		char realmBuff[kRealmBuffSize] = "Basic realm=\"";
		StrPtrLen challenge(realmBuff);
		StrPtrLen whichRealm;
		
		// Get the module's realm
		StrPtrLen moduleRealm;
		theErr = this->GetValue(qtssRTSPReqURLRealm, 0,  (void **) &moduleRealm.Ptr, &moduleRealm.Len);
		if ( (QTSS_NoErr == theErr) && (moduleRealm.Len > 0) )
		{
			whichRealm = moduleRealm;	
		}
		else
		{
			theErr = QTSS_NoErr;
			// Get the default realm from the config file or use the static default if config realm is not found
			QTSServerInterface* theServer = QTSServerInterface::GetServer();
			prefRealmPtr = theServer->GetPrefs()->GetAuthorizationRealm(); // allocates memory
			Assert(prefRealmPtr != NULL);
			if (prefRealmPtr != NULL)
			{	whichRealm.Set(prefRealmPtr, strlen(prefRealmPtr));
			}	
			else
			{
				whichRealm = sDefaultRealm;
			}
		}
		
		int realmLen = whichRealm.Len + challenge.Len + 2; // add 2 based on double quote char + end of string 0x00
		if (realmLen > kRealmBuffSize) // The realm is too big so use the default realm
		{	Assert(0);
			whichRealm = sDefaultRealm;
		}
		memcpy(&challenge.Ptr[challenge.Len],whichRealm.Ptr,whichRealm.Len);
		int newLen = challenge.Len + whichRealm.Len;
		
		challenge.Ptr[newLen] = '"'; // add the terminating "" this was accounted for with the size check above
		challenge.Ptr[newLen + 1] = 0;// add the 0 terminator this was accounted for with the size check above
		challenge.Len = newLen +1; // set the real size of the string excluding the 0.
		
		#if (0)
		{  // test code
			char test[256] = {};
			memcpy(test,sDefaultRealm.Ptr,sDefaultRealm.Len);
			printf("the static realm =%s \n",test);

			memcpy(test,prefRealm.Ptr,prefRealm.Len);
			printf("the Pref realm =%s \n",test);

			memcpy(test,moduleRealm.Ptr,moduleRealm.Len);
			printf("the moduleRealm  =%s \n",test);
		
			memcpy(test,whichRealm.Ptr,whichRealm.Len);
			printf("the challenge realm  =%s \n",test);
			
			memcpy(test,challenge.Ptr,challenge.Len);
			printf("the challenge string  =%s len = %ld\n",test, challenge.Len);
		}
		#endif

		fStatus = qtssClientUnAuthorized;
		this->SetResponseKeepAlive(true);
		this->AppendHeader(qtssWWWAuthenticateHeader, &challenge);
		this->SendHeader();

	
	} while (false);
	
	if (prefRealmPtr != NULL)
	{	
		delete[] prefRealmPtr;	
	}
	
	return theErr;
}
