/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		RTSPRequest.h

	Contains:	This class encapsulates a single RTSP request. It stores the meta data
				associated with a request, and provides an interface (through its base
				class) for modules to access request information
				
				It divides the request into a series of states.
				
				State 1: At first, when the object is constructed or right after 
						 its Reset function is called, it is in an uninitialized state.
				State 2: Parse() parses an RTSP header. Once this function returns,
						 most of the request-related query functions have been setup.
						 (unless an error occurs)
				State 3: GetHandler() uses the request information to create the
						 proper request Handler object for this request. After that,
						 it is the Handler object's responsibilty to process the
						 request and send a response to the client.
					
	$Log: RTSPRequest.h,v $
	Revision 1.3  2000/03/07 23:08:39  murata
	changes to support improved logging and logging module bug fixes. Changes include
	new API attributes in the client session object and error reporting in QTFileLib.
	Bug #:2440344, 2429487,2412973, 2412969.
	Submitted by:
	Reviewed by:
	
	Revision 1.2  2000/02/04 00:46:24  serenyi
	Added date & expires headers to server response, server now parses if-modified-since dates
	
	Revision 1.1.1.1  2000/01/17 23:36:03  wsanchez
	Import of StreamingServer-2.0 (Apple)
	
	Revision 1.3  1999/12/21 01:44:09  lecroy
	*** empty log message ***
	
	Revision 1.2  1999/10/27 03:26:02  murata
	Changes are for optimizing authentication code in the request.
	
	Revision 1.1.1.1  1999/09/22 23:01:53  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:54  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:08:32  ds
	Created
	

*/

#ifndef __RTSPREQUEST_H__
#define __RTSPREQUEST_H__

#include "RTSPRequestInterface.h"
#include "RTSPRequestStream.h"
#include "RTSPResponseStream.h"
#include "RTSPSessionInterface.h"
#include "StringParser.h"
#include "QTSSRTSPProtocol.h"


//HTTPRequest class definition
class RTSPRequest : public RTSPRequestInterface
{
public:

	//CONSTRUCTOR / DESTRUCTOR
	//these do very little. Just initialize / delete some member data.
	//
	//Arguments:		session: the session this request is on (massive cyclical dependency)
	RTSPRequest(RTSPSessionInterface* session)
		: RTSPRequestInterface(session) {}
	virtual ~RTSPRequest() {}
	
	//Parses the request. Returns an error handler if there was an error encountered
	//in parsing.
	QTSS_Error Parse(StrPtrLen *stream);
	QTSS_Error ParseAuthNameAndPassword(void);
	QTSS_Error SetupAuthLocalPath(void);
	QTSS_Error SendChallenge(void);
	
private:

	//PARSING
	enum { kRealmBuffSize = 512, kAuthNameAndPasswordBuffSize = 512 };

	//Parsing the URI line (first line of request
	QTSS_Error ParseFirstLine(StringParser &parser);
	
	//Utility functions called by above
	QTSS_Error ParseURI(StringParser &parser);

	//Parsing the rest of the headers
	//This assumes that the parser is at the beginning of the headers. It will parse
	//the headers, fill out the data & HTTPParameters object.
	//
	//Returns:		A handler object signifying that a fatal syntax error has occurred
	QTSS_Error ParseHeaders(StringParser& parser);
	
	//Functions to parse the contents of particuarly complicated headers (as a convienence
	//for modules)
	void  	ParseRangeHeader();
	void 	ParseTransportHeader();
	void  	ParseIfModSinceHeader();
	void 	ParseAddrSubHeader(StrPtrLen* inSubHeader, StrPtrLen* inHeaderName, UInt32* outAddr);
	void 	ParseClientPortSubHeader(StrPtrLen* inClientPortSubHeader);
	void 	ParseTimeToLiveSubHeader(StrPtrLen* inTimeToLiveSubHeader);


	static UInt8	sURLStopConditions[];
};
#endif // __RTSPREQUEST_H__

