/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		RTSPRequestInterface.cp

	Contains:	Implementation of class defined in RTSPRequestInterface.h
	

	$Log: RTSPRequestInterface.cpp,v $
	Revision 1.5  2000/03/14 23:48:32  serenyi
	Made writev kAlwaysBuffer again
	
	Revision 1.4  2000/03/09 02:15:37  serenyi
	changed QTSS_WriteV to not buffer from always buffer
	
	Revision 1.3  2000/02/11 01:07:24  serenyi
	Server now supports cache control header and sends 304 Not Modified responses in appropriate situations
	
	Revision 1.2  2000/02/04 00:46:24  serenyi
	Added date & expires headers to server response, server now parses if-modified-since dates
	
	Revision 1.1.1.1  2000/01/17 23:36:03  wsanchez
	Import of StreamingServer-2.0 (Apple)
	
	Revision 1.8  1999/12/21 01:44:09  lecroy
	*** empty log message ***
	
	Revision 1.7  1999/12/08 23:25:46  rtucker
	OPTIONS now tickles RTP Stream
	we now report timeout value in first SETUP response
	
	Revision 1.6  1999/10/27 17:55:03  serenyi
	Integrated Matthias's MacOS / GUSI changes
	
	Revision 1.5  1999/10/27 03:26:02  murata
	Changes are for optimizing authentication code in the request.
	
	Revision 1.4  1999/10/16 00:27:45  murata
	use new GetPref Calls.
	
	Revision 1.3  1999/10/07 08:52:33  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.2  1999/09/30 00:31:40  serenyi
	Use core server generation of Server: header
	
	Revision 1.1.1.1  1999/09/22 23:01:53  lecroy
	importing
	
	Revision 1.1.1.1  1999/09/13 16:16:54  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:08:32  ds
	Created
	

*/


//INCLUDES:
#include <sys/types.h>
#include <sys/uio.h>

#include "RTSPRequestInterface.h"
#include "RTSPSessionInterface.h"
#include "RTSPRequestStream.h"

#include "StringParser.h"
#include "OSMemory.h"
#include "OSThread.h"
#include "DateTranslator.h"

#include "QTSSPrefs.h"
#include "QTSServerInterface.h"

char 		RTSPRequestInterface::sPremadeHeader[kStaticHeaderSizeInBytes];
StrPtrLen	RTSPRequestInterface::sPremadeHeaderPtr(sPremadeHeader, kStaticHeaderSizeInBytes);
StrPtrLen	RTSPRequestInterface::sColonSpace(": ", 2);

QTSSDictionaryMap::AttrInfo	RTSPRequestInterface::sAttributes[] =
{
	/* 0 */ { "qtssRTSPReqFullRequest",			NULL,					true,	false },
	/* 1 */ { "qtssRTSPReqMethodStr",			NULL,					true,	false },
	/* 2 */ { "qtssRTSPReqFilePath",			NULL,					true,	false },
	/* 3 */ { "qtssRTSPReqURI",					NULL,					true,	false },
	/* 4 */ { "qtssRTSPReqFilePathTrunc",		GetTruncatedPath,		true,	false },
	/* 5 */ { "qtssRTSPReqFileName",			GetFileName,			true,	false },
	/* 6 */ { "qtssRTSPReqFileDigit",			GetFileDigit,			true,	false },
	/* 7 */ { "qtssRTSPReqAbsoluteURL",			NULL,					true,	false },
	/* 8 */ { "qtssRTSPReqTruncAbsoluteURL",	GetTruncatedPath,		true,	false },
	/* 9 */ { "qtssRTSPReqMethod",				NULL,					true,	false },
	/* 10 */ { "qtssRTSPReqStatusCode",			NULL,					true,	true  },
	/* 11 */ { "qtssRTSPReqStartTime",			NULL,					true,	false },
	/* 12 */ { "qtssRTSPReqStopTime",			NULL,					true,	false },
	/* 13 */ { "qtssRTSPReqRespKeepAlive",		NULL,					true,	true  },
	/* 14 */ { "qtssRTSPReqRootDir",			NULL,					true,	true  },
	/* 15 */ { "qtssRTSPReqRealStatusCode",		GetRealStatusCode,		true,	false },
	/* 16 */ { "qtssRTSPReqStreamRef",			NULL,					true,	false },
	
	/* 17 */ { "qtssRTSPReqUserName",			NULL,					true,	false },
	/* 18 */ { "qtssRTSPReqUserPassword",		NULL,					true,	false },
	/* 19 */ { "qtssRTSPReqUserAllowed",		NULL,					true,	true  },
	/* 20 */ { "qtssRTSPReqURLRealm",			NULL,					true,	true  },
	/* 21 */ { "qtssRTSPReqLocalPath",			NULL,					true,	true  },
	/* 22 */ { "qtssRTSPReqIfModSinceDate",		NULL,					true,	false }
};

void  RTSPRequestInterface::Initialize()
{
	//make a partially complete header
	StringFormatter headerFormatter(sPremadeHeaderPtr.Ptr, kStaticHeaderSizeInBytes);
	PutStatusLine(&headerFormatter, qtssSuccessOK, RTSPProtocol::k10Version);
	headerFormatter.Put(QTSServerInterface::GetServerHeader());
	headerFormatter.PutEOL();
	headerFormatter.Put(RTSPProtocol::GetHeaderString(qtssCSeqHeader));
	headerFormatter.Put(sColonSpace);
	sPremadeHeaderPtr.Len = headerFormatter.GetCurrentOffset();
	Assert(sPremadeHeaderPtr.Len < kStaticHeaderSizeInBytes);
	
	//Setup all the dictionary stuff
	for (UInt32 x = 0; x < qtssRTSPReqNumParams; x++)
		QTSSModuleInterface::GetDictionaryMap(QTSSModuleInterface::kRTSPRequestDictIndex)->
			SetAttribute(x, sAttributes[x].fAttrName, sAttributes[x].fFuncPtr,
							sAttributes[x].fIsPreemptiveSafe, sAttributes[x].fIsWriteable);
	
	QTSSDictionaryMap* theHeaderMap = QTSSModuleInterface::GetDictionaryMap(QTSSModuleInterface::kRTSPHeaderDictIndex);
	for (UInt32 y = 0; y < qtssNumHeaders; y++)
		theHeaderMap->SetAttribute(y, RTSPProtocol::GetHeaderString(y).Ptr, NULL, true, false);
}

//CONSTRUCTOR / DESTRUCTOR: very simple stuff
RTSPRequestInterface::RTSPRequestInterface(RTSPSessionInterface *session)
:	QTSSDictionary(QTSSModuleInterface::GetDictionaryMap(QTSSModuleInterface::kRTSPRequestDictIndex)),
	//fMethod(0), //parameter need not be set
	fStatus(qtssSuccessOK), //parameter need not be set
	fRealStatusCode(0),
	fRequestKeepAlive(true),
	//fResponseKeepAlive(true), //parameter need not be set
	fVersion(RTSPProtocol::k10Version),
	fStartTime(-1),
	fStopTime(-1),
	fClientPortA(0),
	fClientPortB(0),
	fTtl(0),
	fDestinationAddr(0),
	fSourceAddr(0),
	fIsTCP(false),
	fIfModSinceDate(0),
	fStreamRef(this),
	fMovieFolderPtr(&fMovieFolderPath[0]),
	fHeaderDictionary(QTSSModuleInterface::GetDictionaryMap(QTSSModuleInterface::kRTSPHeaderDictIndex)),
	fAllowed(true),
	fSession(session),
	fOutputStream(session->GetOutputStream()),
	fStandardHeadersWritten(false)
{
	//Setup QTSS parameters that can be setup now. These are typically the parameters that are actually
	//pointers to binary variable values. Because these variables are just member variables of this object,
	//we can properly initialize their pointers right off the bat.
	RTSPRequestStream* input = session->GetInputStream();
	this->SetVal(qtssRTSPReqFullRequest, input->GetRequestBuffer()->Ptr, input->GetRequestBuffer()->Len);
	this->SetVal(qtssRTSPReqMethod, &fMethod, sizeof(fMethod));
	this->SetVal(qtssRTSPReqStatusCode, &fStatus, sizeof(fStatus));
	this->SetVal(qtssRTSPReqRespKeepAlive, &fResponseKeepAlive, sizeof(fResponseKeepAlive));
	this->SetVal(qtssRTSPReqStreamRef, &fStreamRef, sizeof(fStreamRef));
	
	// Get the default root directory from QTSSPrefs, and store that in the proper parameter
	// Note that the GetMovieFolderPath function may allocate memory, so we check for that
	// in this object's destructor and free that memory if necessary.
	UInt32 pathLen = kMovieFolderBufSizeInBytes;
	fMovieFolderPtr = QTSServerInterface::GetServer()->GetPrefs()->GetMovieFolder(fMovieFolderPtr, &pathLen);
	this->SetVal(qtssRTSPReqRootDir, fMovieFolderPtr, pathLen);
	
	//There are actually other attributes that point to member variables that we COULD setup now, but they are attributes that
	//typically aren't set for every request, so we lazy initialize those when we parse the request

	this->SetVal(qtssRTSPReqUserAllowed, &fAllowed, sizeof(fAllowed));

}

void RTSPRequestInterface::AppendHeader(QTSS_RTSPHeader inHeader, StrPtrLen* inValue)
{
	if (!fStandardHeadersWritten)
		this->WriteStandardHeaders();
		
	fOutputStream->Put(RTSPProtocol::GetHeaderString(inHeader));
	fOutputStream->Put(sColonSpace);
	fOutputStream->Put(*inValue);
	fOutputStream->PutEOL();
}

void RTSPRequestInterface::PutStatusLine(StringFormatter* putStream, QTSS_RTSPStatusCode status,
										RTSPProtocol::RTSPVersion version)
{
	putStream->Put(RTSPProtocol::GetVersionString(version));
	putStream->PutSpace();
	putStream->Put(RTSPProtocol::GetStatusCodeAsString(status));
	putStream->PutSpace();
	putStream->Put(RTSPProtocol::GetStatusCodeString(status));
	putStream->PutEOL();	
}

void RTSPRequestInterface::AppendDateAndExpires()
{
	if (!fStandardHeadersWritten)
		this->WriteStandardHeaders();

	Assert(OSThread::GetCurrent() != NULL);
	DateBuffer* theDateBuffer = OSThread::GetCurrent()->GetDateBuffer();
	theDateBuffer->InexactUpdate(); // Update the date buffer to the current date & time
	StrPtrLen theDate(theDateBuffer->GetDateBuffer(), DateBuffer::kDateBufferLen);
	
	// Append dates, and have this response expire immediately
	this->AppendHeader(qtssDateHeader, &theDate);
	this->AppendHeader(qtssExpiresHeader, &theDate);
}


void RTSPRequestInterface::AppendSessionHeaderWithTimeout( StrPtrLen* inSessionID, StrPtrLen* inTimeout )
{

	// Append a session header if there wasn't one already
	if ( GetHeaderDictionary()->GetValue(qtssSessionHeader)->Len == 0)
	{	
		if (!fStandardHeadersWritten)
			this->WriteStandardHeaders();

		static StrPtrLen	sTimeoutString(";timeout=");

		// Just write out the session header and session ID
		fOutputStream->Put( RTSPProtocol::GetHeaderString(qtssSessionHeader ) );
		fOutputStream->Put(sColonSpace);
		fOutputStream->Put( *inSessionID );
		
		
		if ( inTimeout != NULL )
		{
			fOutputStream->Put( sTimeoutString );
			fOutputStream->Put( *inTimeout );
		}
		
		
		fOutputStream->PutEOL();
	}
}

void RTSPRequestInterface::AppendTransportHeader(StrPtrLen* serverPortA,
													StrPtrLen* serverPortB,
													StrPtrLen* channelA,
													StrPtrLen* channelB)
{
	static StrPtrLen	sServerPortString(";server_port=");
	static StrPtrLen	sInterleavedString(";interleaved=");

	if (!fStandardHeadersWritten)
		this->WriteStandardHeaders();

	// Just write out the same transport header the client sent to us.
	fOutputStream->Put(RTSPProtocol::GetHeaderString(qtssTransportHeader));
	fOutputStream->Put(sColonSpace);
	fOutputStream->Put(*fHeaderDictionary.GetValue(qtssTransportHeader));
	
	// Append the server ports, if provided.
	if (serverPortA != NULL)
	{
		fOutputStream->Put(sServerPortString);
		fOutputStream->Put(*serverPortA);
		fOutputStream->PutChar('-');
		fOutputStream->Put(*serverPortB);
	}
	
	// Append channel #'s, if provided
	if (channelA != NULL)
	{
		fOutputStream->Put(sInterleavedString);
		fOutputStream->Put(*channelA);
		fOutputStream->PutChar('-');
		fOutputStream->Put(*channelB);
	}

	fOutputStream->PutEOL();
}

void RTSPRequestInterface::AppendContentBaseHeader(StrPtrLen* theURL)
{
	if (!fStandardHeadersWritten)
		this->WriteStandardHeaders();

	fOutputStream->Put(RTSPProtocol::GetHeaderString(qtssContentBaseHeader));
	fOutputStream->Put(sColonSpace);
	fOutputStream->Put(*theURL);
	fOutputStream->PutChar('/');
	fOutputStream->PutEOL();
}

void RTSPRequestInterface::AppendRTPInfoHeader(QTSS_RTSPHeader inHeader,
												StrPtrLen* url, StrPtrLen* seqNumber,
												StrPtrLen* ssrc, StrPtrLen* rtpTime)
{
	static StrPtrLen sURL("url=", 4);
	static StrPtrLen sSeq(";seq=", 5);
	static StrPtrLen sSsrc(";ssrc=", 6);
	static StrPtrLen sRTPTime(";rtptime=", 9);

	if (!fStandardHeadersWritten)
		this->WriteStandardHeaders();

	fOutputStream->Put(RTSPProtocol::GetHeaderString(inHeader));
	if (inHeader != qtssSameAsLastHeader)
		fOutputStream->Put(sColonSpace);
		
	//Only append the various bits of RTP information if they actually have been
	//providied
	if ((url != NULL) && (url->Len > 0))
	{
		fOutputStream->Put(sURL);
		fOutputStream->Put(*url);
	}
	if ((seqNumber != NULL) && (seqNumber->Len > 0))
	{
		fOutputStream->Put(sSeq);
		fOutputStream->Put(*seqNumber);
	}
	if ((ssrc != NULL) && (ssrc->Len > 0))
	{
		fOutputStream->Put(sSsrc);
		fOutputStream->Put(*ssrc);
	}
	if ((rtpTime != NULL) && (rtpTime->Len > 0))
	{
		fOutputStream->Put(sRTPTime);
		fOutputStream->Put(*rtpTime);
	}
	fOutputStream->PutEOL();
}


void RTSPRequestInterface::WriteStandardHeaders()
{
	static StrPtrLen	sCloseString("Close", 5);

	Assert(sPremadeHeader != NULL);
	fStandardHeadersWritten = true; //must be done here to prevent recursive calls
	
	//if this is a "200 OK" response (most HTTP responses), we have some special
	//optmizations here
	if (fStatus == qtssSuccessOK)
	{
		fOutputStream->Put(sPremadeHeaderPtr);
		StrPtrLen* cSeq = fHeaderDictionary.GetValue(qtssCSeqHeader);
		Assert(cSeq != NULL);
		if (cSeq->Len > 1)
			fOutputStream->Put(*cSeq);
		else if (cSeq->Len == 1)
			fOutputStream->PutChar(*cSeq->Ptr);
		fOutputStream->PutEOL();
	}
	else
	{
		//other status codes just get built on the fly
		PutStatusLine(fOutputStream, fStatus, RTSPProtocol::k10Version);
		fOutputStream->Put(QTSServerInterface::GetServerHeader());
		fOutputStream->PutEOL();
		AppendHeader(qtssCSeqHeader, fHeaderDictionary.GetValue(qtssCSeqHeader));
	}

	//append sessionID header
	StrPtrLen* incomingID = fHeaderDictionary.GetValue(qtssSessionHeader);
	if ((incomingID != NULL) && (incomingID->Len > 0))
		AppendHeader(qtssSessionHeader, incomingID);

	//follows the HTTP/1.1 convention: if server wants to close the connection, it
	//tags the response with the Connection: close header
	if (!fResponseKeepAlive)
		AppendHeader(qtssConnectionHeader, &sCloseString);
}

void RTSPRequestInterface::SendHeader()
{
	if (!fStandardHeadersWritten)
		this->WriteStandardHeaders();
	fOutputStream->PutEOL();
}

QTSS_Error
RTSPRequestInterface::Write(void* inBuffer, UInt32 inLength, UInt32* outLenWritten, UInt32 /*inFlags*/)
{
	//now just write whatever remains into the output buffer
	fOutputStream->Put((char*)inBuffer, inLength);
	
	if (outLenWritten != NULL)
		*outLenWritten = inLength;
		
	return QTSS_NoErr;
}

QTSS_Error
RTSPRequestInterface::WriteV(iovec* inVec, UInt32 inNumVectors, UInt32 inTotalLength, UInt32* outLenWritten)
{
	(void)fOutputStream->WriteV(inVec, inNumVectors, inTotalLength, NULL,
												RTSPResponseStream::kAlwaysBuffer);
	if (outLenWritten != NULL)
		*outLenWritten = inTotalLength;
	return QTSS_NoErr;	
}

#pragma mark __PARAM_RETRIEVAL_FUNCTIONS__

//param retrieval functions described in .h file
void* RTSPRequestInterface::GetTruncatedPath(QTSS_Object inRequest, UInt32* /*outLen*/)
{
	//We only want this param retrieval function to be invoked once, so set the
	//qtssRTSPReqFilePathTruncParam to be a valid pointer
	
	//This function works for 2 attributes, because they are almost the same: qtssRTSPReqFilePathTruncParam, qtssRTSPReqTruncAbsoluteURLParam
	
	RTSPRequestInterface* theRequest = (RTSPRequestInterface*)inRequest;
	theRequest->SetVal(qtssRTSPReqFilePathTrunc, theRequest->GetValue(qtssRTSPReqFilePath));
	theRequest->SetVal(qtssRTSPReqTruncAbsoluteURL, theRequest->GetValue(qtssRTSPReqAbsoluteURL));

	//Adjust the length to truncate off the last file in the path
	
	StrPtrLen* theTruncPathParam = theRequest->GetValue(qtssRTSPReqFilePathTrunc);
	StrPtrLen* theAbsTruncPathParam = theRequest->GetValue(qtssRTSPReqTruncAbsoluteURL);
	theTruncPathParam->Len--;
	theAbsTruncPathParam->Len--;
	while (theTruncPathParam->Ptr[theTruncPathParam->Len] != '/')
	{
		theTruncPathParam->Len--;
		theAbsTruncPathParam->Len--;
	}

	return NULL;
}

void* RTSPRequestInterface::GetFileName(QTSS_Object inRequest, UInt32* /*outLen*/)
{
	//We only want this param retrieval function to be invoked once, so set the
	//qtssRTSPReqFilePathTruncParam to be a valid pointer
	
	RTSPRequestInterface* theRequest = (RTSPRequestInterface*)inRequest;
	theRequest->SetVal(qtssRTSPReqFileName, theRequest->GetValue(qtssRTSPReqFilePath));

	StrPtrLen* theFileNameParam = theRequest->GetValue(qtssRTSPReqFileName);

	//paranoid check
	if (theFileNameParam->Len == 0)
		return theFileNameParam;
		
	//walk back in the file name until we hit a /
	SInt32 x = theFileNameParam->Len - 1;
	for (; x > 0; x--)
		if (theFileNameParam->Ptr[x] == '/')
			break;
	//once we do, make the tempPtr point to the next character after the slash,
	//and adjust the length accordingly
	if (theFileNameParam->Ptr[x] == '/')
	{
		theFileNameParam->Ptr = (&theFileNameParam->Ptr[x]) + 1;
		theFileNameParam->Len -= (x + 1);
	}
	
	return NULL;		
}


void* RTSPRequestInterface::GetFileDigit(QTSS_Object inRequest, UInt32* /*outLen*/)
{
	//We only want this param retrieval function to be invoked once, so set the
	//qtssRTSPReqFileDigitParam to be a valid pointer
	
	RTSPRequestInterface* theRequest = (RTSPRequestInterface*)inRequest;
	theRequest->SetVal(qtssRTSPReqFileDigit, theRequest->GetValue(qtssRTSPReqFilePath));

	StrPtrLen* theFileDigit = theRequest->GetValue(qtssRTSPReqFileDigit);

	UInt32	theFilePathLen = theRequest->GetValue(qtssRTSPReqFilePath)->Len;
	theFileDigit->Ptr += theFilePathLen - 1;
	theFileDigit->Len = 0;
	while ((StringParser::sDigitMask[*theFileDigit->Ptr] != 0) &&
			(theFileDigit->Len <= theFilePathLen))
	{
		theFileDigit->Ptr--;
		theFileDigit->Len++;
	}
	//termination condition means that we aren't actually on a digit right now.
	//Move pointer back onto the digit
	theFileDigit->Ptr++;
	
	return NULL;
}

void* RTSPRequestInterface::GetRealStatusCode(QTSS_Object inRequest, UInt32* outLen)
{
	// Set the fRealStatusCode variable based on the current fStatusCode.
	// Return it to the caller (this means this function will be invoked each
	// time the parameter is requested, which is what we want).
	RTSPRequestInterface* theReq = (RTSPRequestInterface*)inRequest;
	theReq->fRealStatusCode = RTSPProtocol::GetStatusCode(theReq->fStatus);
	*outLen = sizeof(UInt32);
	return &theReq->fRealStatusCode;
}
