/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		RTSPSessionInterface.cpp

	Contains:	Implementation of RTSPSessionInterface object.
	
	$Log: RTSPSessionInterface.cpp,v $
	Revision 1.3  2000/03/09 02:15:52  serenyi
	changed QTSS_WriteV to not buffer from always buffer
	
	Revision 1.2  2000/01/27 02:09:14  rtucker
	Merge changes from v81
	
	Revision 1.16  2000/01/19 19:37:06  serenyi
	Moved to new NEW macro
	
	Revision 1.15  1999/12/21 01:44:11  lecroy
	*** empty log message ***
	
	Revision 1.14  1999/12/06 21:18:39  rtucker
	changes for Solaris alignment resttrictions
	
	Revision 1.13  1999/11/15 01:20:49  serenyi
	Added 4-char codes for mem debugging, merged in solaris diffs
	
	Revision 1.12  1999/11/08 22:52:37  serenyi
	Updated code to reflect changes in QTSS API
	
	Revision 1.11  1999/10/27 17:55:04  serenyi
	Integrated Matthias's MacOS / GUSI changes
	
	Revision 1.10  1999/10/19 21:14:18  rtucker
	tcp enhancements
	
	Revision 1.9  1999/10/19 02:34:03  murata
	Added code to save off the request authorization params in the session for logging.
	
	Revision 1.8  1999/10/13 03:48:11  murata
	Support for Authorization: moved authenticate parsing into server (step 1 of moving
	authentication into server and leaving authorization in modules.
	
	Revision 1.7  1999/10/08 00:00:17  murata
	Setting the client ID as the IP address here.
	
	Revision 1.6  1999/10/07 08:52:34  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.5  1999/10/06 07:16:56  serenyi
	Fixed bugs in base64 decoding stuff, it now works with fragmented packets & pipelined requests
	
	Revision 1.4  1999/10/06 00:05:24  murata
	support for average bit rate and removal of conditional compile flag for the new logging fields
	
	Revision 1.3  1999/09/30 00:33:30  serenyi
	Bugfixes, added new attribute for session type
	
	Revision 1.2  1999/09/24 19:57:28  murata
	New logging field support.
	
	Revision 1.1.1.1  1999/09/22 23:01:54  lecroy
	importing
	
	Revision 1.2  1999/09/17 02:36:49  serenyi
	Fixed bug where if RTSP session is half closed (on http), server would consider the connection dead.
	
	Revision 1.1.1.1  1999/09/13 16:16:54  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:08:38  ds
	Created
	

*/

#include "atomic.h"

#include "RTSPSessionInterface.h"
#include "QTSServerInterface.h"
#include "OSMemory.h"

#include <sys/errno.h>


#if DEBUG
	#define RTSP_SESSION_INTERFACE_DEBUGGING 0
#else
	#define RTSP_SESSION_INTERFACE_DEBUGGING 0
#endif



unsigned int 			RTSPSessionInterface::sSessionIDCounter = kFirstRTSPSessionID;
Bool16					RTSPSessionInterface::sDoBase64Decoding = true;

QTSSDictionaryMap::AttrInfo	RTSPSessionInterface::sAttributes[] = 
{   /*fields:   fAttrName, fFuncPtr, fIsPreemptiveSafe, fIsWriteable */
	/* 0 */ { "qtssRTSPSesID", 				NULL, 			true,	false },
	/* 1 */ { "qtssRTSPSesLocalAddr", 		SetupParams, 	true,	false },
	/* 2 */ { "qtssRTSPSesLocalAddrStr",	SetupParams, 	true,	false },
	/* 3 */ { "qtssRTSPSesLocalDNS",		SetupParams, 	true,	false },
	/* 4 */ { "qtssRTSPSesRemoteAddr",		SetupParams, 	true,	false },
	/* 5 */ { "qtssRTSPSesRemoteAddrStr",	SetupParams, 	true,	false },
	/* 6 */ { "qtssRTSPSesEventCntxt",		NULL, 			true,	false },
	/* 7 */ { "qtssRTSPSesType",			NULL, 			true,	false },
	/* 8 */ { "qtssRTSPSesStreamRef",		NULL, 			true,	false },
	
	/* 9 */ { "qtssRTSPSesLastUserName",	NULL, 			true,	false  },
	/* 10 */{ "qtssRTSPSesLastUserPassword",NULL, 			true,	false  },
	/* 11 */{ "qtssRTSPSesLastURLRealm",	NULL, 			true,	false  }
};


void	RTSPSessionInterface::Initialize()
{
	for (UInt32 x = 0; x < qtssRTSPSesNumParams; x++)
		QTSSModuleInterface::GetDictionaryMap(QTSSModuleInterface::kRTSPSessionDictIndex)->
			SetAttribute(x, sAttributes[x].fAttrName, sAttributes[x].fFuncPtr, sAttributes[x].fIsPreemptiveSafe, sAttributes[x].fIsWriteable);
}


RTSPSessionInterface::RTSPSessionInterface() 
:	QTSSDictionary(QTSSModuleInterface::GetDictionaryMap(QTSSModuleInterface::kRTSPSessionDictIndex)),
	Task(), 
	fTimeoutTask(this, QTSServerInterface::GetServer()->GetPrefs()->GetRealRTSPTimeoutInSecs() * 1000),
	fInputStream(&fSocket),
	fOutputStream(&fSocket, &fTimeoutTask),
	fSessionMutex(),
	fTCPCoalesceBuffer(NULL),
	fNumInCoalesceBuffer(0),
	fSocket(this, Socket::kNonBlockingSocketType),
	fOutputSocketP(&fSocket),
	fInputSocketP(&fSocket),
	fSessionType(qtssRTSPSession),
	fLiveSession(true),
	fObjectHolders(0),
	fCurChannelNum(0),
	fStreamRef(this)
{
	fSessionID = (UInt32)atomic_add(&sSessionIDCounter, 1);
	this->SetVal(qtssRTSPSesID, &fSessionID, sizeof(fSessionID));
	this->SetVal(qtssRTSPSesEventCntxt, &fOutputSocketP, sizeof(fOutputSocketP));
	this->SetVal(qtssRTSPSesType, &fSessionType, sizeof(fSessionType));
	this->SetVal(qtssRTSPSesStreamRef, &fStreamRef, sizeof(fStreamRef));

	this->SetEmptyVal(qtssRTSPSesLastUserName, &fUserNameBuf[0], kMaxUserNameLen);
	this->SetEmptyVal(qtssRTSPSesLastUserPassword, &fUserPasswordBuf[0], kMaxUserPasswordLen);
	this->SetEmptyVal(qtssRTSPSesLastURLRealm, &fUserRealmBuf[0], kMaxUserRealmLen);
}


RTSPSessionInterface::~RTSPSessionInterface()
{
	// If the input socket is != output socket, the input socket was created dynamically
 	if (fInputSocketP != fOutputSocketP) 
 		delete fInputSocketP;
 	
 	delete [] fTCPCoalesceBuffer;

}

void RTSPSessionInterface::DecrementObjectHolderCount()
{
	this->Signal(Task::kReadEvent);//have the object wakeup in case it can go away.
	atomic_sub(&fObjectHolders, 1);
}

QTSS_Error RTSPSessionInterface::Write(void* inBuffer, UInt32 inLength,
											UInt32* outLenWritten, UInt32 /*inFlags*/)
{
	iovec theVec[2];
	theVec[1].iov_base = (char*)inBuffer;
	theVec[1].iov_len = inLength;
	return fOutputStream.WriteV(theVec, 2, inLength, outLenWritten, RTSPResponseStream::kDontBuffer);
}

QTSS_Error RTSPSessionInterface::WriteV(iovec* inVec, UInt32 inNumVectors, UInt32 inTotalLength, UInt32* outLenWritten)
{
	return fOutputStream.WriteV(inVec, inNumVectors, inTotalLength, outLenWritten, RTSPResponseStream::kDontBuffer);
}


void RTSPSessionInterface::InterleaveSetup()
{
	if ( !fTCPCoalesceBuffer )
	{
		fTCPCoalesceBuffer = new char[kTCPCoalesceBufferSize];

	}
}

/*********************************
/
/	InterleavedWrite
/
/	Write the given RTP packet out on the RTSP channel in interleaved format.
/
*/

QTSS_Error RTSPSessionInterface::InterleavedWrite(void* inBuffer, UInt32 inLen, UInt32* outLenWritten, unsigned char channel)
{

	if ( inLen == 0 && fNumInCoalesceBuffer == 0 )
	{	*outLenWritten = 0;
		return QTSS_NoErr;
	}
		
	// First attempt to grab the RTSPSession mutex. This is to prevent writing data to
	// the connection at the same time an RTSPRequest is being processed. We cannot
	// wait for this mutex to be freed (there would be a deadlock possibility), so
	// just try to grab it, and if we can't, then just report it as an EAGAIN
	if ( this->GetSessionMutex()->TryLock() == false )
	{
		return EAGAIN;
	}

	// DMS - this struct should be packed.
	//rt todo -- is this struct more portable (byte alignment could be a problem)?
	struct	RTPInterleaveHeader
	{
		unsigned char header;
		unsigned char channel;
		UInt16		len;
	};
	
	struct 	iovec 				iov[3];
	QTSS_Error					err = QTSS_NoErr;
	
	

	// flush rules
	if ( ( inLen > kTCPCoalesceDirectWriteSize || inLen == 0 ) && fNumInCoalesceBuffer > 0 
		|| ( inLen + fNumInCoalesceBuffer + kInteleaveHeaderSize > kTCPCoalesceBufferSize ) && fNumInCoalesceBuffer > 0
		)
	{
		UInt32		buffLenWritten;
		
		// skip iov[0], WriteV uses it
		iov[1].iov_base = fTCPCoalesceBuffer;
		iov[1].iov_len = fNumInCoalesceBuffer;
		
		err = this->GetOutputStream()->WriteV( iov, 2, fNumInCoalesceBuffer, &buffLenWritten, RTSPResponseStream::kAllOrNothing );

	#if RTSP_SESSION_INTERFACE_DEBUGGING 
		::printf("InterleavedWrite: flushing %li\n", fNumInCoalesceBuffer );
	#endif
		
		if ( err == QTSS_NoErr )
			fNumInCoalesceBuffer = 0;
	}
	
	
	
	if ( err == QTSS_NoErr )
	{
		
		if ( inLen > kTCPCoalesceDirectWriteSize )	
		{
			struct RTPInterleaveHeader	rih;
			
			// write direct to stream
			rih.header = '$';
			rih.channel = channel;
			rih.len = htons(inLen);
			
			iov[1].iov_base = (char*)&rih;
			iov[1].iov_len = sizeof(rih);
			
			iov[2].iov_base = (char*)inBuffer;
			iov[2].iov_len = inLen;

			err = this->GetOutputStream()->WriteV( iov, 3, inLen + sizeof(rih), outLenWritten, RTSPResponseStream::kAllOrNothing );

		#if RTSP_SESSION_INTERFACE_DEBUGGING 
			::printf("InterleavedWrite: bypass %li\n", inLen );
		#endif
			
		}
		else
		{
			// coalesce with other small writes
			
			fTCPCoalesceBuffer[fNumInCoalesceBuffer] = '$';
			fNumInCoalesceBuffer++;;
			
			fTCPCoalesceBuffer[fNumInCoalesceBuffer] = channel;
			fNumInCoalesceBuffer++;
			
			//*((short*)&fTCPCoalesceBuffer[fNumInCoalesceBuffer]) = htons(inLen);
			// if we ever turn TCPCoalesce back on, this should be optimized
			// for processors w/o alignment restrictions as above.
			
			SInt16	pcketLen = htons(inLen);
			::memcpy( &fTCPCoalesceBuffer[fNumInCoalesceBuffer], &pcketLen, 2 );
			fNumInCoalesceBuffer += 2;
			
			::memcpy( &fTCPCoalesceBuffer[fNumInCoalesceBuffer], inBuffer, inLen );
			fNumInCoalesceBuffer += inLen;
		
		#if RTSP_SESSION_INTERFACE_DEBUGGING 
			::printf("InterleavedWrite: coalesce %li, total bufff %li\n", inLen, fNumInCoalesceBuffer);
		#endif
		}
	}
	
	if ( err == QTSS_NoErr )
	{	
		/* 	if no error sure to correct outLenWritten, cuz WriteV above includes the interleave header count
		
			 GetOutputStream()->WriteV guarantees all or nothing for writes
			 if no error, then all was written.
		*/
		if ( outLenWritten != NULL )
			*outLenWritten = inLen;
	}

	this->GetSessionMutex()->Unlock();


	return err;
	
}

/*
	take the TCP socket away from a RTSP session that's
	waiting to be snarfed.
	
*/

void	RTSPSessionInterface::SnarfInputSocket( RTSPSessionInterface* fromRTSPSession )
{
	Assert( fromRTSPSession != NULL );
	Assert( fromRTSPSession->fOutputSocketP != NULL );
	
	// grab the unused, but already read fromsocket data
	// this should be the first RTSP request
	if (sDoBase64Decoding)
		fInputStream.IsBase64Encoded(true); // client sends all data base64 encoded
	fInputStream.SnarfRetreat( fromRTSPSession->fInputStream );

	if (fInputSocketP == fOutputSocketP)
		fInputSocketP = NEW TCPSocket( this, Socket::kNonBlockingSocketType );
	else
		fInputSocketP->Cleanup(); 	// if this is a socket replacing an old socket, we need
									// to make sure the file descriptor gets closed
	fInputSocketP->SnarfSocket( fromRTSPSession->fSocket );
	
	// fInputStream, meet your new input socket
	fInputStream.AttachToSocket( fInputSocketP );
}

#pragma mark __PARAM_RETRIEVAL_FUNCTIONS__

 void* RTSPSessionInterface::SetupParams(QTSS_Object inSession, UInt32* /*outLen*/)
 {
 	RTSPSessionInterface* theSession = (RTSPSessionInterface*)inSession;
 
 	theSession->fLocalAddr = theSession->fSocket.GetLocalAddr();
 	theSession->fRemoteAddr = theSession->fSocket.GetRemoteAddr();
 	
 	StrPtrLen* theLocalAddrStr = theSession->fSocket.GetLocalAddrStr();
 	StrPtrLen* theLocalDNSStr = theSession->fSocket.GetLocalDNSStr();
 	StrPtrLen* theRemoteAddrStr = theSession->fSocket.GetRemoteAddrStr();
 	
	theSession->SetVal(qtssRTSPSesLocalAddr, &theSession->fLocalAddr, sizeof(theSession->fLocalAddr));
	theSession->SetVal(qtssRTSPSesLocalAddrStr, theLocalAddrStr->Ptr, theLocalAddrStr->Len);
	theSession->SetVal(qtssRTSPSesLocalDNS, theLocalDNSStr->Ptr, theLocalDNSStr->Len);
	theSession->SetVal(qtssRTSPSesRemoteAddr, &theSession->fRemoteAddr, sizeof(theSession->fRemoteAddr));
	theSession->SetVal(qtssRTSPSesRemoteAddrStr, theRemoteAddrStr->Ptr, theRemoteAddrStr->Len);

	return NULL;
 }
