/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		RTSPSessionInterface.h

	Contains:	Presents an API for session-wide resources for modules to use.
				Implements the RTSP Session dictionary for QTSS API.
	
	$Log: RTSPSessionInterface.h,v $
	Revision 1.2  2000/03/07 23:08:39  murata
	changes to support improved logging and logging module bug fixes. Changes include
	new API attributes in the client session object and error reporting in QTFileLib.
	Bug #:2440344, 2429487,2412973, 2412969.
	Submitted by:
	Reviewed by:
	
	Revision 1.1.1.1  2000/01/17 23:36:04  wsanchez
	Import of StreamingServer-2.0 (Apple)
	
	Revision 1.11  1999/12/21 01:44:11  lecroy
	*** empty log message ***
	
	Revision 1.10  1999/12/03 22:04:11  rtucker
	changed kTCPCoalesceDirectWriteSize to 0, to eliminate problems due to
	coalescing very low bandwidth movies.
	
	Revision 1.9  1999/11/08 22:52:37  serenyi
	Updated code to reflect changes in QTSS API
	
	Revision 1.8  1999/10/19 21:14:18  rtucker
	tcp enhancements
	
	Revision 1.7  1999/10/19 02:34:03  murata
	Added code to save off the request authorization params in the session for logging.
	
	Revision 1.6  1999/10/13 03:48:11  murata
	Support for Authorization: moved authenticate parsing into server (step 1 of moving
	authentication into server and leaving authorization in modules.
	
	Revision 1.5  1999/10/07 08:52:46  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.4  1999/10/06 07:16:56  serenyi
	Fixed bugs in base64 decoding stuff, it now works with fragmented packets & pipelined requests
	
	Revision 1.3  1999/09/30 00:33:30  serenyi
	Bugfixes, added new attribute for session type
	
	Revision 1.2  1999/09/24 19:57:29  murata
	New logging field support.
	
	Revision 1.1.1.1  1999/09/22 23:01:49  lecroy
	importing
	
	Revision 1.3  1999/09/17 02:36:49  serenyi
	Fixed bug where if RTSP session is half closed (on http), server would consider the connection dead.
	
	Revision 1.2  1999/09/15 17:05:55  serenyi
	Updated stuff
	
	Revision 1.1.1.1  1999/09/13 16:16:55  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:08:39  ds
	Created
	

*/

#ifndef __RTSPSESSIONINTERFACE_H__
#define __RTSPSESSIONINTERFACE_H__

#include "RTSPRequestStream.h"
#include "RTSPResponseStream.h"
#include "Task.h"
#include "QTSS.h"
#include "QTSSDictionary.h"
#include "QTSSModuleInterface.h"
#include "atomic.h"

class RTSPSessionInterface : public QTSSDictionary, public Task
{
public:

	//Initialize must be called right off the bat to initialize dictionary resources
	static void		Initialize();
	static void		SetBase64Decoding(Bool16 newVal) { sDoBase64Decoding = newVal; }
	
	RTSPSessionInterface();
	virtual ~RTSPSessionInterface();
	
	//Is this session alive? If this returns false, clean up and begone as
	//fast as possible
	Bool16 IsLiveSession() 		{ return fSocket.IsConnected() && fLiveSession; }
	
	// Allows clients to refresh the timeout
	void RefreshTimeout()		{ fTimeoutTask.RefreshTimeout(); }
	
	// In order to facilitate sending out of band data on the RTSP connection,
	// other objects need to have direct pointer access to this object. But,
	// because this object is a task object it can go away at any time. If # of
	// object holders is > 0, the RTSPSession will NEVER go away. However,
	// the object managing the session should be aware that if IsLiveSession returns
	// false it may be wise to relinquish control of the session
	void IncrementObjectHolderCount() { (void)atomic_add(&fObjectHolders, 1); }
	void DecrementObjectHolderCount();
	
	// If RTP data is interleaved into the RTSP connection, we need to associate
	// a unique channel number with each rtp stream. This function allocates
	// a new channel number, unique wrt this RTSP session.
	unsigned char		GetChannelNumber() 	{ return fCurChannelNum++; }
	
	//Two main things are persistent through the course of a session, not
	//associated with any one request. The RequestStream (which can be used for
	//getting data from the client), and the socket. OOps, and the ResponseStream
	RTSPRequestStream*	GetInputStream() 	{ return &fInputStream; }
	RTSPResponseStream* GetOutputStream()	{ return &fOutputStream; }
	TCPSocket*			GetSocket() 		{ return &fSocket; }
	OSMutex*			GetSessionMutex()	{ return &fSessionMutex; }
	
	UInt32				GetSessionID()		{ return fSessionID; }
	
	// QTSS STREAM FUNCTIONS
	
	// Allows non-buffered writes to the client. These will flow control.
	
	// THE FIRST ENTRY OF THE IOVEC MUST BE BLANK!!!
	virtual QTSS_Error WriteV(iovec* inVec, UInt32 inNumVectors, UInt32 inTotalLength, UInt32* outLenWritten);
	virtual QTSS_Error Write(void* inBuffer, UInt32 inLength, UInt32* outLenWritten, UInt32 inFlags);

	// performs RTP over RTSP
	QTSS_Error 	InterleavedWrite(void* inBuffer, UInt32 inLen, UInt32* outLenWritten, unsigned char channel);
	void		InterleaveSetup();

	enum
	{
		kMaxUserNameLen 		= 32,
		kMaxUserPasswordLen 	= 32,
		kMaxUserRealmLen	 	= 64
	};

protected:
	enum
	{
		kFirstRTSPSessionID 	= 1,	//UInt32
	};

	//Each RTSP session has a unique number that identifies it.

	char				fUserNameBuf[kMaxUserNameLen];
	char				fUserPasswordBuf[kMaxUserPasswordLen];
	char				fUserRealmBuf[kMaxUserRealmLen];

	TimeoutTask			fTimeoutTask;//allows the session to be timed out
	
	RTSPRequestStream 	fInputStream;
	RTSPResponseStream 	fOutputStream;
	
	// Any RTP session sending interleaved data on this RTSP session must
	// be prevented from writing while an RTSP request is in progress
	OSMutex				fSessionMutex;
	
	// for coalescing small interleaved writes into a single TCP frame
	enum
	{
		  kTCPCoalesceBufferSize = 1450	//1450 is the max data space in an TCP segment over ent
		, kTCPCoalesceDirectWriteSize = 0 // if > this # bytes bypass coalescing and make a direct write
		, kInteleaveHeaderSize = 4  // '$ '+ 1 byte ch ID + 2 bytes length
	};
	char*		fTCPCoalesceBuffer;
	SInt32		fNumInCoalesceBuffer;


	//+rt  socket we get from "accept()"
	TCPSocket			fSocket;
	TCPSocket*			fOutputSocketP;
	TCPSocket*			fInputSocketP;	// <-- usually same as fSocketP, unless we're HTTP Proxying
	
	void		SnarfInputSocket( RTSPSessionInterface* fromRTSPSession );
	
	// What session type are we?
	QTSS_RTSPSessionType	fSessionType;
	
	Bool16				fLiveSession;
	unsigned int		fObjectHolders;
	unsigned char		fCurChannelNum;
	
	QTSS_StreamRef		fStreamRef;

	UInt32				fSessionID;
	UInt32				fLocalAddr;
	UInt32				fRemoteAddr;
	
	
	static unsigned int	sSessionIDCounter;
	static Bool16			sDoBase64Decoding;
	
	//Dictionary support
	
	// Param retrieval function
	static void* 		SetupParams(QTSS_Object inSession, UInt32* outLen);
	
	static QTSSDictionaryMap::AttrInfo	sAttributes[];
};
#endif // __RTSPSESSIONINTERFACE_H__

